/*
	Name: asttapiprovider.cpp
	Copyright: Under the Mozilla Public License Version 1.1 or later
	Author: ACTIVA
*/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Activa asttapiprovider.
 *
 * The Initial Developer of the Original Code is ACTIVA.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): none
 *
 * ***** END LICENSE BLOCK ***** */

#include "asttapiprovider.h"
#include "config.h"
#include "activatsp_helper.h"



// Microsoft recommended algorithm for calculating the permanent line ID
#define MAKEPERMLINEID(dwPermProviderID, dwDeviceID)  ((LOWORD(dwPermProviderID) << 16) | dwDeviceID)

// { dwDialPause, dwDialSpeed, dwDigitDuration, dwWaitForDialtone }
LINEDIALPARAMS      g_dpMin = { 100,  50, 100,  100 };
LINEDIALPARAMS      g_dpDef = { 250,  50, 250,  500 };
LINEDIALPARAMS      g_dpMax = { 1000, 50, 1000, 1000 };

// some feature dial commands
#define	FEATURE_LOGIN						"Login"         //old "*7361"
#define	FEATURE_LOGOUT						"Logout"        //old "*7360"
#define	FEATURE_NOTREADY					"NotReady"      //old "*7350"
#define	FEATURE_READY						"Ready"         //old "*7351"

#define	FEATURE_RECORDING_START				"StartRecording"
#define	FEATURE_RECORDING_STOP				"StopRecording"
#define	FEATURE_RECORDING_PAUSE				"PauseRecording"
#define	FEATURE_RECORDING_UNPAUSE			"UnpauseRecording"
#define	FEATURE_RECORDING_CHANGE_FILENAME	"ChangeRecordingFileName"

#define	FEATURE_DUMP						"Dump"			//old "*0000"

/**
 * Construcs a new TapiLine.
 * TapiLine models a tsp line
 */
TapiLine::TapiLine()
{
	m_xref = 0;			// will use xref as HDRVLINE
	m_htapiLine = 0;
	m_dwDeviceID = 0;
	m_deviceID = "";
	m_fnEventProc = NULL;
};

/**
*	Constructs a new TapiCall
*/
TapiCall::TapiCall() 
{
	m_pTapiLine = NULL;
	m_htapiCall = 0;
	m_callId = 0;
	m_address = "";				
	m_callerId  = "";			
	m_callerIdName  = "";		
	m_calledDevice  = "";		
	m_redirectingDevice  = "";	
	m_userData = "";
	m_eLineCallOrigin = linecallorigin_unknown;
	m_isPredictive = false;
	m_state = tapicallstate_unknown ;
	m_cause = ast_cause_normal;
}

/**
*	update TapiCall info. If a given new attribute is "" class attribute will not be changed.
*	\param  callerId
*	\param callerIdName
*	\param calledDevice
*	\param lastRedirectingDevice
*	\param state
*/
void	TapiCall::Update(std::string callerId, std::string callerIdName, std::string calledDevice, std::string lastRedirectingDevice, eTapiCall_state state)
{
	if (callerId!="") {
		m_callerId = callerId;
	}
	if (callerIdName!="") {
		m_callerIdName = callerIdName;
	}
	if (calledDevice!="") {
		m_calledDevice = calledDevice;
	}
	if (lastRedirectingDevice!="") {
		m_redirectingDevice = lastRedirectingDevice;
	}
	if (state !=tapicallstate_unknown) {
		m_state = state;
	}
}

/**
*	Convert a TapiCall to String
*	\return string with the information of the TapiCall
*/
std::string	TapiCall::DumpToString()
{
	std::string	str;
	str = "TapiLine=0x"+ToString(m_pTapiLine);
	str += " hdrivercall=0x" +ToString(this);
	str += " htapicall=0x" +ToString(m_htapiCall);
	str += " callID=" + ToString(m_callId);
	str += " address=" +m_address;
	str += " state=" +m_state;
	str += " CID=" +m_callerId;
	str += " CIDName=" +m_callerIdName;
	str += " lastRedir=" +m_redirectingDevice;
	str += " called=" +m_calledDevice;
	str += " userData="+m_userData;
	str += " isPredictive="+ToString(m_isPredictive);
	str += " cause="+ToString(m_cause);
	return str;
}

/**
*	Constructor
*/
ASTTapiProvider::ASTTapiProvider()
{
	m_nLockDebugCounter=0;
}

/**
*	Destructor
*/
ASTTapiProvider::~ASTTapiProvider()
{
	// relese memory and clean maps
	CallMapCleanup();

	DeleteCriticalSection(&m_csMapLock);
}

	
/**
 * factory method
 * creates a ASTTapiProvider and attach itself as a cstalistener
 *	\param logCallBack
 *	\param fnCompletionProc
 *	\param dwPermanentProviderID
 *	\param dwLineDeviceIDBase
 *	\param v_vOutDnVector
 *	\return NULL if cannot create
 */
ASTTapiProvider	*ASTTapiProvider::Create(ASTLogListener *logCallBack, ASYNC_COMPLETION fnCompletionProc, DWORD dwPermanentProviderID, DWORD dwLineDeviceIDBase, std::vector<std::string> v_vOutDnVector)
{
	ASTTapiProvider	*instance = new ASTTapiProvider();
	if (instance==NULL) {
		return NULL;
	}

	ASTCstaProviderInterface	*cstaListener = CreateASTCSTAProvider(instance, logCallBack);
	if (cstaListener==NULL) { // ups... rollback
		delete instance;
		return NULL;
	}

	instance->m_cstaProvider = cstaListener;
	instance->m_fnCompletionProc = fnCompletionProc;
	instance->m_dwPermanentProviderID = dwPermanentProviderID;
    instance->m_dwLineDeviceIDBase = dwLineDeviceIDBase;

	InitializeCriticalSection(&instance->m_csMapLock);
	InitializeCriticalSection(&instance->m_csLineOpenCloseLock);
	instance->m_log.SetOrigin("TAPIProvider");
	instance->m_log.SetLogger(logCallBack);
	std::string	v = getVersion();
	instance->m_log.LogInfo("__________________ ASTTapiProvider::Create __________________");
	instance->m_log.LogInfo("_  ActivaTSP.tsp v%8s                                  _", v.c_str());
	instance->m_log.LogInfo("_____________________________________________________________");

	instance->m_vOutDnVector=v_vOutDnVector;
	
	return instance;
}

/**
*	Connect to an Asterisk PBX
*	\param Host ip address of Asterisk PBX
*	\param port port of connection
*	\param user connect with specify user
*	\param passwd password
*	\param eAuthType authentification type 0-plaintext 1-md5
*	\return AC_RESULT
*/
AC_RESULT	ASTTapiProvider::Connect(std::string sHost, int nPort, std::string user, std::string pwd, AUTH_TYPE eAuthType)
{
	return m_cstaProvider->Connect(sHost, nPort, user, pwd, eAuthType);
}

/**
*	Disconnect from an Asterisk PBX
*	\return true if the disconnection was correctly
*/
bool	ASTTapiProvider::Disconnect()
{
	bool rc = m_cstaProvider->Disconnect();
	if (rc) {
		CallMapCleanup();
	}

	return rc;

}

/**
*	Open a Monitor over a device
*	\param device
*	\return Monitor reference
*/
MonitorXref	ASTTapiProvider::OpenMonitor(std::string deviceId)
{
	return m_cstaProvider->OpenMonitor(deviceId);
}

/**
*	Close a Monitor
*	\param Monitor reference
*	\param true if the monitor was closed correctly, otherwise false
*/
bool	ASTTapiProvider::CloseMonitor(MonitorXref xref)
{
	return m_cstaProvider->CloseMonitor(xref);
}

/**
*	Set do Not Disturb feature
*	\param device
*	\param bool
*	\return AC_RESULT
*/
AC_RESULT	ASTTapiProvider::SetDoNotDisturb(std::string deviceId, bool bDND_on, std::string dndContext)
{
	m_log.LogTrace("SetDoNotDisturb device=%s, state=%s", deviceId.c_str(), ToString(bDND_on));
	return m_cstaProvider->SetDoNotDisturb(deviceId, bDND_on, registryGetAgentContext("from-internal"));
}

/**
*	Add a new TapiLine to Map
*	\param xref
*	\param htapiLine
*	\param pLineInfo
*/
void		ASTTapiProvider::AddTapiLine(MonitorXref xref, HTAPILINE htapiLine, TapiLine *pLineInfo)
{
	m_log.LogTrace("AddTapiLine xref=%d htapiLine=0x%x pTapiLine=0x%x", xref, htapiLine, pLineInfo);

	if (htapiLine==NULL || pLineInfo==NULL) {
		m_log.LogWarn("  do nothing: htapiLine==NULL || pLineInfo==NULL");
		return ;
	}
	// protect map access with mutex
	WaitForMapLock();
		pLineInfo->m_xref = xref;
		pLineInfo->m_htapiLine = htapiLine;
		m_mapXref2TapiLine[xref] = pLineInfo ;
		m_mapHtapiLine2xref[htapiLine] = xref ;
	ReleaseMapLock();
}

/**
 * gets TapiLine associated to hdriverLIne.
 * opaque hdriverLine interpreted as xref
 *	\param xref
 *	\return pointer to TapiLine
 */
TapiLine*		ASTTapiProvider::GetTapiLine(MonitorXref xref)
{
	// protect map access with mutex
	WaitForMapLock();
	if (m_mapXref2TapiLine.find(xref) == m_mapXref2TapiLine.end()) {
		ReleaseMapLock();
		return NULL;
	}

	TapiLine	*pTapiLine = m_mapXref2TapiLine[xref];
	ReleaseMapLock();
	return pTapiLine;
}

/**
 *	gets TapiLine assiciated to htapiLine.
 *	\param htapiLine 
 *	\return pointer to TapiLine
 */
TapiLine*		ASTTapiProvider::GetTapiLine(HTAPILINE htapiLine)
{
	// protect map access with mutex
	WaitForMapLock();
	if (m_mapHtapiLine2xref.find(htapiLine) == m_mapHtapiLine2xref.end()) {	// not found
		ReleaseMapLock();
		return NULL;
	}

	MonitorXref	xref = m_mapHtapiLine2xref[htapiLine];
	TapiLine	*pTapiLine = m_mapXref2TapiLine[xref];
	ReleaseMapLock();	
	return pTapiLine;
}

/**
*	Remove TapiLine
*	\param hdriverLine
*	\return pointer to TapiLine
*/
TapiLine*		ASTTapiProvider::RemoveTapiLine(HDRVLINE hdriverLine)
{
	m_log.LogTrace("RemoveTapiLine hdrvLine=0x%x", hdriverLine);

	MonitorXref	xref = (MonitorXref)(void*) hdriverLine;

	// protect map access with mutex
	WaitForMapLock();
	if (m_mapXref2TapiLine.find(xref) == m_mapXref2TapiLine.end()) {
		ReleaseMapLock();
		m_log.LogTrace("  not here!");
		return NULL;
	}
	TapiLine	*pTapiLine = m_mapXref2TapiLine[xref];
	HTAPILINE	htapiLine = pTapiLine->m_htapiLine;
	m_mapXref2TapiLine.erase(xref);
	m_mapHtapiLine2xref.erase(htapiLine);
	m_log.LogTrace("  was %s", pTapiLine->m_deviceID.c_str());
	ReleaseMapLock();
	return pTapiLine;
}

/**
 *	Adds or Updates TapiCall into the map. 
 *  callid may be null (for originating calls for example)
 *	warning: caller is responsible for calling ReleaseMapLock()
 *	\param newTapiCall
 *	\param TapiLine
 *	\param htapiCall
 *	\param callid identifier of call
 *	\param address identify the device of call
 *	\param callerid who is calling
 *	\param calleridname
 *	\param calledDevice who has been called
 *	\param lastRedirection
 *	\param userData
 */
void		ASTTapiProvider::AddTapiCall(TapiCall *pNewTapiCall, TapiLine *pTapiLine, HTAPICALL htapiCall, DWORD callId, std::string address, std::string callerId, std::string callerIdName, std::string calledDevice, std::string lastRedirectionDevice, std::string userData, eTapiCall_state state)
{
	m_log.LogTrace("ASTTapiProvider::AddTapiCall pNewTapiCall=0x%x callId=%d, htapicall=0x%x, callerId=%s, calledId=%s userData=%s", pNewTapiCall, callId, htapiCall, callerId.c_str(), calledDevice.c_str(), userData.c_str());

	if (pNewTapiCall==NULL || pTapiLine==NULL || htapiCall==NULL) {
		m_log.LogWarn("  do nothing: htapiCall==NULL || pTapiLIne==NULL || pNewTapiCall==NULL");
		return ;
	}

	HDRVCALL	hdrvCall = (HDRVCALL) pNewTapiCall ;
	// protect map access with critical section
	WaitForMapLock();

	if (GetTapiCall(pNewTapiCall) != NULL ) { // if it was in the map, start over again...
		RemoveTapiCall(pNewTapiCall);		
	}
	
	pNewTapiCall->m_pTapiLine = pTapiLine;	// this is the associated line
	pNewTapiCall->m_htapiCall = htapiCall;
	pNewTapiCall->m_callId = callId;
	pNewTapiCall->m_address = address;
	pNewTapiCall->m_callerId = callerId;
	pNewTapiCall->m_callerIdName = callerIdName ;
	pNewTapiCall->m_calledDevice = calledDevice;
	pNewTapiCall->m_redirectingDevice = lastRedirectionDevice;
	pNewTapiCall->m_userData = userData;
	if (state !=tapicallstate_unknown) {
		pNewTapiCall->m_state = state;
	}
	pNewTapiCall->m_cause = ast_cause_normal;

	TapiCall	*pTC = GetTapiCall(callId, pTapiLine);
	if (!pTC) {	// not in map, so insert
		if (!GetTapiCall(pNewTapiCall)) {	// not even in map
			m_mapTapiCall[hdrvCall] = pNewTapiCall;			// add entry to main map
			m_mapHtapiCall2HdrvCall[htapiCall] = hdrvCall ;	// adjust helper maps
		}

		if (callId) {	// callid == 0 will not add to callid map 
			// add to Map_CallId2HdrvCallSet multimap
			if (m_mapCallId2HdrvCallSet.find(callId)==m_mapCallId2HdrvCallSet.end()) {	// none found
				m_log.LogTrace("  first TapiCall for this callid: [%d] => 0x%x", callId, hdrvCall);
				Set_hdrvCall *pv = new Set_hdrvCall();
				pv->insert(hdrvCall);
				m_mapCallId2HdrvCallSet[callId] = pv;
			} else {
				m_log.LogTrace("  add new TapiLine for this callid: [%d] => 0x%x", callId, hdrvCall);
				m_mapCallId2HdrvCallSet[callId]->insert(hdrvCall);
			}
		}
	}else{
		m_log.LogTrace(" already exist a tapiCall 0x%x with callId %d", pTC, callId);
	}

	ReleaseMapLock();

	DumpMaps();
}

/**
*	Gets TapiCall from map (or NULL if not found).
*	\param TapiCall
*	\return TapiCall
*/
TapiCall*		ASTTapiProvider::GetTapiCall(TapiCall* pTapiCall)
{
	HDRVCALL	hdrvCall = (HDRVCALL) pTapiCall ;

	// protect map access with mutex
	WaitForMapLock();
	if (m_mapTapiCall.find(hdrvCall) == m_mapTapiCall.end()) {	// not found
		ReleaseMapLock();
		return NULL;
	}

	TapiCall	*pRealTapiCall = m_mapTapiCall[hdrvCall];

	ReleaseMapLock();	
	return pRealTapiCall;
}


/**
 * Looks for a TapiCall in call-map, where given callId ad TapiLine
 *	\param callId identifier of call
 *	\param TapiLine
 *	\return TapiCall
 */
TapiCall *	ASTTapiProvider::GetTapiCall(DWORD callId, TapiLine *pTapiLine)
{
	// do not search for null callId match
	if (callId < 1 || pTapiLine == NULL)
		return NULL;

	// protect map access with mutex
	WaitForMapLock();
	if (m_mapCallId2HdrvCallSet.find(callId) == m_mapCallId2HdrvCallSet.end()) {
		ReleaseMapLock();
		return NULL;
	}

	Set_hdrvCall  *pSet_hdriver = m_mapCallId2HdrvCallSet[callId];
	for (Set_hdrvCall::iterator i = pSet_hdriver->begin(); i!=pSet_hdriver->end(); i++) {
		HDRVCALL	hdrvCall = *i ;
		TapiCall	*pTC = GetTapiCall(hdrvCall);
		if (pTC) {
			if (pTC->m_pTapiLine == pTapiLine) { // found
				ReleaseMapLock();
				return pTC;
			}
		}
	}

	ReleaseMapLock();
	return NULL;
}

/**
 * gets TapiCall assiciated to htapiCall.
 *	\param htapiCall
 *	\return TapiCall
 */
TapiCall*		ASTTapiProvider::GetTapiCall(HTAPICALL htapiCall)
{
	// protect map access with mutex
	WaitForMapLock();
	if (m_mapHtapiCall2HdrvCall.find(htapiCall) == m_mapHtapiCall2HdrvCall.end()) {	// not found
		ReleaseMapLock();
		return NULL;
	}

	HDRVCALL	hdrvCall = m_mapHtapiCall2HdrvCall[htapiCall];
	TapiCall	*pTapiCall = m_mapTapiCall[hdrvCall];

	ReleaseMapLock();	
	return pTapiCall;
}

/**
 * Remove TapiCall from call map. It does *not* delete pTapiCall object itself,
 * thus the caller must delete it.
 *	\param TapiCall
 *	\return TapiCall
 */
TapiCall*	ASTTapiProvider::RemoveTapiCall(TapiCall* pOpaqueTapiCall)
{
	HDRVCALL	hdrvCall = (HDRVCALL) pOpaqueTapiCall ;

	m_log.LogTrace("ASTTapiProvider::RemoveTapiCall hdrivercall=0x%x", hdrvCall);

	DumpMaps();

	// protect map access with mutex
	WaitForMapLock();
	if (m_mapTapiCall.find(hdrvCall) == m_mapTapiCall.end()) {
		ReleaseMapLock();
		m_log.LogError("  HDRVCALL Not FOUND !!!");
		return NULL;
	}

	TapiCall	*pRealTapiCall = m_mapTapiCall[hdrvCall];
	HTAPICALL	htapiCall = pRealTapiCall->m_htapiCall;
	DWORD		dwCallId = pRealTapiCall->m_callId;	

	// remove from map
	m_mapTapiCall.erase(hdrvCall);

	// remove from helper maps
	m_mapHtapiCall2HdrvCall.erase(htapiCall);
	if (dwCallId) {	// remember that pTapiCall->callId  may be NULL
		m_log.LogTrace("  call had callid=%d", dwCallId);
		if (m_mapCallId2HdrvCallSet.find(dwCallId) != m_mapCallId2HdrvCallSet.end()) {
			Set_hdrvCall *pSet_hdriver = m_mapCallId2HdrvCallSet[dwCallId];
			m_log.LogTrace("    ...with %d associated hdrvrcalls", pSet_hdriver->size());

			if (pSet_hdriver->find(hdrvCall) != pSet_hdriver->end()) {	// here it is
				m_log.LogTrace("  delete htcall 0x%x and hdrvCall 0x%x", htapiCall, hdrvCall);
				pSet_hdriver->erase(hdrvCall);
			}

			if (pSet_hdriver->size()==0) {
				m_log.LogTrace("  deleted the last htapicall and the set of callid %d", dwCallId);
				m_mapCallId2HdrvCallSet.erase(dwCallId);
				delete pSet_hdriver;
			}
		}
	}

	ReleaseMapLock();

	return pRealTapiCall;
}

/**
*
*/
bool		ASTTapiProvider::CompletionProc(DRV_REQUESTID dwRequestID, LONG lResult)
{
	if (m_fnCompletionProc!=NULL) {
		m_fnCompletionProc(dwRequestID, lResult);
		return true;
	}

	return false;
}

/**
*	Clean the tapi map
*/
void	ASTTapiProvider::CallMapCleanup()
{
	m_log.LogTrace(" CallMapCleanup() cleaning maps");
	// protect map access
	WaitForMapLock();

	// for each TapiLine
	for (Map_Xref2TapiLine::iterator i = m_mapXref2TapiLine.begin(); i!=m_mapXref2TapiLine.end(); i++) {
		TapiLine	*pTapiLine = i->second;
		delete pTapiLine ;
	}
	m_mapXref2TapiLine.clear();
	m_mapHtapiLine2xref.clear();

	// for each TapiCall
	for (Map_TapiCall::iterator i = m_mapTapiCall.begin(); i!=m_mapTapiCall.end(); i++) {
		TapiCall	*pTapiCall = i->second;
		delete pTapiCall ;
	}
	m_mapTapiCall.clear();
	m_mapHtapiCall2HdrvCall.clear();

	for (Map_CallId2HdrvCallSet::iterator i = m_mapCallId2HdrvCallSet.begin(); i!=m_mapCallId2HdrvCallSet.end(); i++) {
		Set_hdrvCall	*pSet_hdriver = i->second;
		pSet_hdriver->clear();
		delete pSet_hdriver;
	}
	m_mapCallId2HdrvCallSet.clear();

	ReleaseMapLock();
}


/**
 * ASTCstaListener implementation. 
 * 
 * \see http://windowssdk.msdn.microsoft.com/en-us/library/ms735543.aspx
 */

/**
 * create new TAPI call handler if none
 * sends LINECALLSTATE_DIALING and LINECALLSTATE_PROCEEDING 
 * Currently this listener is not used
 *	\param xref
 *	\param callId identifier of call
 *	\param originatingDevice who has originated the call
 *	\param callerId who is calling
 *	\param callerIdName
 *	\param calledDevice who has been called
 *	\param userData
 *	\param call_cause
 */
void	ASTTapiProvider::OnOriginated ( MonitorXref xref, DWORD callId, std::string originatingDevice, std::string callerId, std::string callerIdName, std::string calledDevice, std::string userData, call_cause cause)
{
	m_log.LogTrace("OnOriginated xref=%d callId=%d, originatingDevice=%s cause=%d", xref, callId, 
		originatingDevice.c_str(), cause);
	WaitForMapLock();

	TapiLine	*pTapiLine = GetTapiLine(xref);
	if (!pTapiLine) {
		ReleaseMapLock();
		m_log.LogTrace("  TapiLine not found");
		return ;	// :-(
	}

	LINEEVENT	fnCallback = pTapiLine->m_fnEventProc;
	HTAPILINE	htapiLine = pTapiLine->m_htapiLine;
	HTAPICALL	htapiCall = 0;

	m_log.LogTrace(" searching a TapiCall with callId %d in tapiLine 0x%x", callId, pTapiLine);
	TapiCall	*pTapiCall = GetTapiCall(callId, pTapiLine);
	if (pTapiCall){
		m_log.LogTrace("  pTapiCall found with callId %d and htCall 0x%x", pTapiCall->m_callId,pTapiCall->m_htapiCall);
		htapiCall = pTapiCall->m_htapiCall ;
	}else{
		ReleaseMapLock();
		m_log.LogTrace("  TapiCall not found");
		return ;	// :-(
	}

	if (pTapiCall->m_state == tapicallstate_idle){
		m_log.LogTrace("don't send events because call is idle");
		ReleaseMapLock();
		return;
	}

	pTapiCall->Update(callerId, callerIdName, calledDevice, "", tapicallstate_initialized);

	ReleaseMapLock();

}

/**
 *	create new TAPI call handler if none
 *	sends LINECALLSTATE_OFFERING for inbound or LINECALLSTATE_RINGBACK for outbound
  *	\param xref
 *	\param callId identifier of call
 *	\param alertingDevice what device is alerting
 *	\param callerId who is calling
 *	\param callerIdName
 *	\param calledDevice who has been called
 *	\param userData
 *	\param call_cause
 */
void	ASTTapiProvider::OnAlerting ( MonitorXref xref, DWORD callId, std::string alertingDevice, std::string callerId, std::string callerIdName, std::string calledDevice, std::string lastRedirectingDevice, std::string userData, call_cause cause)
{
	m_log.LogTrace("OnAlerting xref=%d callId=%d, alertingDevice=%s, callerId=%s, callerIdName=%s, calledID=%s "
		"userData=%s cause=%d", xref, callId, alertingDevice.c_str(), callerId.c_str(), callerIdName.c_str(), 
		calledDevice.c_str(), userData.c_str(), cause);

	// protect map access
	WaitForMapLock();
	// find xref
	TapiLine	*pTapiLine = GetTapiLine(xref);
	if (!pTapiLine) {
		ReleaseMapLock();
		m_log.LogTrace("  TapiLine not found");
		return ;	// :-(
	}

	LINEEVENT	fnCallback = pTapiLine->m_fnEventProc;
	HTAPILINE	htapiLine = pTapiLine->m_htapiLine;
	HTAPICALL	htapiCall = 0;

	m_log.LogTrace(" searching a TapiCall with callId %d in tapiLine 0x%x", callId, pTapiLine);
	TapiCall	*pTapiCall = GetTapiCall(callId, pTapiLine);
	if (!pTapiCall) {
		pTapiCall = new TapiCall();
		pTapiCall->m_callId = callId;
		// get a new htapicall
		m_log.LogTrace("  LINE_NEWCALL");
		fnCallback( htapiLine, 0, LINE_NEWCALL, (DWORD)(void*)pTapiCall, (DWORD)(void*)&(htapiCall), 0);
		if (!htapiCall) {	// handle this!
			m_log.LogWarn("  LINE_NEWCALL returned no handler");
			ReleaseMapLock();
			return ;
		}
		pTapiCall->m_htapiCall = htapiCall ;
		AddTapiCall(pTapiCall, pTapiLine, htapiCall, callId, alertingDevice, callerId, callerIdName, calledDevice, 
			lastRedirectingDevice,userData, tapicallstate_alerting);
	} else {
		m_log.LogTrace("  pTapiCall found with callId %d and htCall 0x%x", pTapiCall->m_callId,pTapiCall->m_htapiCall);
		htapiCall = pTapiCall->m_htapiCall ;
	}

	if (pTapiCall->m_state == tapicallstate_idle){
		m_log.LogTrace("don't send events because call is idle");
		ReleaseMapLock();
		return;
	} 

	pTapiCall->Update(callerId, callerIdName, calledDevice, lastRedirectingDevice, tapicallstate_alerting);

	std::string deviceID = AsteriskCall_GetDialableNumberFromChannel(pTapiLine->m_deviceID);
	m_log.LogTrace("  alertingDevice callerid = %s ", deviceID.c_str());
	bool	isFarEnd = ((alertingDevice != pTapiLine->m_deviceID ) && (deviceID != alertingDevice));
	if (isFarEnd) {
		pTapiCall->m_eLineCallOrigin = linecallorigin_outbound;
		if (alertingDevice.find("misdn") != -1){
			m_log.LogTrace("found channel with misdn");
			pTapiCall->m_redirectionId = calledDevice;
		} else {
			pTapiCall->m_redirectionId = alertingDevice;
		}
		if (pTapiCall->m_callerId == deviceID){
			pTapiCall->m_callerId = pTapiLine->m_deviceID;
		}
	} else {
		pTapiCall->m_eLineCallOrigin = linecallorigin_inbound;
		if (pTapiCall->m_calledDevice == deviceID){
			pTapiCall->m_calledDevice = pTapiLine->m_deviceID;
		}
		pTapiCall->m_redirectionId = pTapiLine->m_deviceID;
	}
	

	m_log.LogTrace("    htapiline=0x%x, deviceID=%s, htapicall=0x%x, callerid=%s, callerIdName=%s, calleddevice=%s, "
		"lastRedirection=%s userData=%s", htapiLine, pTapiLine->m_deviceID.c_str(), htapiCall, pTapiCall->m_callerId.c_str(), pTapiCall->m_callerIdName.c_str(), 
		pTapiCall->m_calledDevice.c_str(), pTapiCall->m_redirectingDevice.c_str(), pTapiCall->m_userData.c_str()); 

	ReleaseMapLock();	// release map lock 

	if (isFarEnd) {	// if outbound -> far-end  -> ringback
		m_log.LogTrace("  LINECALLSTATE_RINGBACK htapicall=0x%x", htapiCall);
		fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_RINGBACK, 0, LINEMEDIAMODE_INTERACTIVEVOICE);
	}else {	// if inbound ->  near-end -> accepted
		m_log.LogTrace("  LINECALLSTATE_OFFERING htapicall=0x%x", htapiCall);
		fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_OFFERING, 0, LINEMEDIAMODE_INTERACTIVEVOICE);
	}
}

/**
 * create new TAPI call handler if none
 * sends LINECALLSTATE_CONNECTED
 *	\param xref
 *	\param callId identifier of call
 *	\param establishedDevice who has picked up the phone
 *	\param callerId who is calling
 *	\param callerIdName
 *	\param calledDevice who has been called
 *	\param userData
 *	\param call_cause
 */
void	ASTTapiProvider::OnEstablished ( MonitorXref xref, DWORD callId, std::string establishedDevice, std::string callerId, std::string callerIdName, std::string calledDevice, std::string lastRedirectingDevice, std::string userData, call_cause cause)
{
	m_log.LogTrace("OnEstablished xref=%d callId=%d, establishedDevice=%s, callerId=%s, callerIdName=%s, "
		"calledDevice =%s, lastRedirectingDevice=%s, userData=%s cause=%d", xref, callId, establishedDevice.c_str(), 
		callerId.c_str(), callerIdName.c_str(), calledDevice.c_str(), lastRedirectingDevice.c_str(), 
		userData.c_str(), cause);

	// protect map access
	WaitForMapLock();
	// find xref
	TapiLine	*pTapiLine = GetTapiLine(xref);
	if (!pTapiLine) {
		ReleaseMapLock();
		m_log.LogTrace("  TapiLine not found");
		return ;	// :-(
	}

	LINEEVENT	fnCallback = pTapiLine->m_fnEventProc;
	HTAPILINE	htapiLine = pTapiLine->m_htapiLine;
	HTAPICALL	htapiCall = 0;

	m_log.LogTrace(" searching a TapiCall with callId %d in tapiLine 0x%x", callId, pTapiLine);
	TapiCall	*pTapiCall = GetTapiCall(callId, pTapiLine);
	if (!pTapiCall) {
		pTapiCall = new TapiCall();
		pTapiCall->m_callId = callId;
		// get a new htapicall
		m_log.LogTrace("  LINE_NEWCALL");
		fnCallback( htapiLine, 0, LINE_NEWCALL, (DWORD)(void*)pTapiCall, (DWORD)(void*)&(htapiCall), 0);
		if (!htapiCall) {	// handle this!
			m_log.LogWarn("  LINE_NEWCALL returned no handler");
			ReleaseMapLock();
			return ;
		}
		pTapiCall->m_htapiCall = htapiCall ;
		AddTapiCall(pTapiCall, pTapiLine, htapiCall, callId, establishedDevice, callerId, callerIdName, 
			calledDevice, lastRedirectingDevice, userData, tapicallstate_connected);
	}
	else {
		htapiCall = pTapiCall->m_htapiCall ;
	}

	if (pTapiCall->m_state == tapicallstate_idle){
		m_log.LogTrace("don't send events because call is idle");
		ReleaseMapLock();
		return;
	}

	// update TapiCall
	pTapiCall->Update(callerId, callerIdName, calledDevice, lastRedirectingDevice, tapicallstate_connected);

	std::string deviceID = AsteriskCall_GetDialableNumberFromChannel(pTapiLine->m_deviceID);
	bool	isFarEnd = (establishedDevice != pTapiLine->m_deviceID && deviceID != establishedDevice) ;
	if (isFarEnd) {
		pTapiCall->m_eLineCallOrigin = linecallorigin_outbound;
		if (pTapiCall->m_callerId == deviceID){
			pTapiCall->m_callerId = pTapiLine->m_deviceID;
		}
		pTapiCall->m_redirectionId = establishedDevice;
	}else {
		pTapiCall->m_eLineCallOrigin = linecallorigin_inbound;
		pTapiCall->m_eLineCallOrigin = linecallorigin_inbound;
		if (pTapiCall->m_calledDevice == deviceID){
			pTapiCall->m_calledDevice = pTapiLine->m_deviceID;
		}
		pTapiCall->m_redirectionId = pTapiLine->m_deviceID;
	}

	m_log.LogTrace("    htapiline=0x%x, deviceID=%s, htapicall=0x%x, callerid=%s, callerIdName=%s, calleddevice=%s, "
		"lastRedirection=%s, userData=%s", htapiLine, pTapiLine->m_deviceID.c_str(), htapiCall, pTapiCall->m_callerId.c_str(), pTapiCall->m_callerIdName.c_str(), 
		pTapiCall->m_calledDevice.c_str(), pTapiCall->m_redirectingDevice.c_str(), pTapiCall->m_userData.c_str()); 

	ReleaseMapLock();

	m_log.LogTrace("  LINECALLSTATE_CONNECTED  htapicall=0x%x", htapiCall);
	fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_CONNECTED, 0, LINEMEDIAMODE_INTERACTIVEVOICE);

}

/**
*	create new TAPI call handler if none
*	Sends LINECALLSTATE_DISCONNECTED_* depends on the cause. Currently the supported causes are:
*		- NORMAL This is a normal disconnect request by the remote party. The call was terminated normally.
*		- OUTOFORDER The call could not be connected or was disconnected because the destination device is out 
*		  of order (hardware failure).
*		- BUSY The remote user's station is busy.
*		- BADADDRESS The destination address is invalid.
*		- NOANSWER The remote user's station does not answer.
*		- UNKNOWN The reason for the disconnect request is unknown but may become known later.
*	Currently this listener is not used
*	\see http://msdn2.microsoft.com/en-us/library/ms735622.aspx 
*	\param xref
*	\param callId identifier of call
*	\param failingDevice 
*	\param callerId who is calling
 *	\param callerIdName
 *	\param calledDevice who has been called
*	\param userData
*	\param call_cause
*/
void	ASTTapiProvider::OnFailed ( MonitorXref xref, DWORD callId, std::string failingDevice, std::string callerId, std::string callerIdName, std::string calledDevice, std::string lastRedirectingDevice, std::string userData, call_cause cause)
{
	m_log.LogTrace("OnFailed xref=%d callId=%d, failingDevice=%s, callerId=%s, callerIdName=%s, calledDevice=%s, "
		"lastRedirectingDevice=%s, userData=%s, cause=%d", xref, callId, failingDevice.c_str(), callerId.c_str(), 
		callerIdName.c_str(), calledDevice.c_str(), lastRedirectingDevice.c_str(), userData.c_str(), cause);

	WaitForMapLock();

	TapiLine	*pTapiLine = GetTapiLine(xref);
	if (!pTapiLine) {
		ReleaseMapLock();
		m_log.LogTrace("  TapiLine not found");
		return ;	// :-(
	}

	LINEEVENT	fnCallback = pTapiLine->m_fnEventProc;
	HTAPILINE	htapiLine = pTapiLine->m_htapiLine;
	HTAPICALL	htapiCall = 0;

	m_log.LogTrace(" searching a TapiCall with callId %d in tapiLine 0x%x", callId, pTapiLine);
	TapiCall	*pTapiCall = GetTapiCall(callId, pTapiLine);
	if (pTapiCall){
		m_log.LogTrace("  pTapiCall found with callId %d and htCall 0x%x", pTapiCall->m_callId,pTapiCall->m_htapiCall);
		htapiCall = pTapiCall->m_htapiCall ;
	}else{
		ReleaseMapLock();
		m_log.LogTrace("  TapiCall not found");
		return ;	// :-(
	}

	if (pTapiCall->m_state == tapicallstate_idle){
		m_log.LogTrace("don't send events because call is idle");
		ReleaseMapLock();
		return;
	}

	//pTapiCall->Update(callerId, callerIdName, calledDevice, "", "failed");
	pTapiCall->Update("", "", "", "", tapicallstate_failed);

	ReleaseMapLock();

}


/**
 * create new TAPI call handler if none
 * sends LINECALLSTATE_ACCEPTED when a call has been entered into a queue or LINECALLSTATE_RINGBACK 
 * when the call is offering to an agent.
 *	\param xref
 *	\param callId identifier of call
 *	\param queuedDevice what device has been gone into queue
 *	\param queue
 *	\param position
 *	\param length
 *	\param callerId who is calling
 *	\param callerIdName
 *	\param calledDevice who has been called
 *	\param userData
 *	\param call_cause
 */
void	ASTTapiProvider::OnQueued ( MonitorXref xref, DWORD callId, std::string queuedDevice, std::string queue, int position_in_queue, int queue_len, std::string callerId, std::string callerIdName, std::string calledDevice, std::string userData, call_cause cause)
{
	m_log.LogTrace("OnQueued xref=%d callId=%d, queuedDevice=%s, queue=%s, callerId=%s, callerIdName=%s, "
		"userData=%s cause=%d", xref, callId, queuedDevice.c_str(), queue.c_str(), callerId.c_str(), 
		callerIdName.c_str(), userData.c_str(), cause);

	// protect map access
	WaitForMapLock();
	// find xref
	TapiLine	*pTapiLine = GetTapiLine(xref);
	if (!pTapiLine) {
		ReleaseMapLock();
		m_log.LogTrace("  TapiLine not found");
		return ;	// :-(
	}

	LINEEVENT	fnCallback = pTapiLine->m_fnEventProc;
	HTAPILINE	htapiLine = pTapiLine->m_htapiLine;
	HTAPICALL	htapiCall = 0;

	TapiCall	*pTapiCall = GetTapiCall(callId, pTapiLine);
	if (!pTapiCall) {
		pTapiCall = new TapiCall();
		pTapiCall->m_callId = callId;
		// get a new htapicall
		m_log.LogTrace("  LINE_NEWCALL");
		fnCallback( htapiLine, 0, LINE_NEWCALL, (DWORD)(void*)pTapiCall, (DWORD)(void*)&(htapiCall), 0);
		if (!htapiCall) {	// handle this!
			m_log.LogWarn("  LINE_NEWCALL returned no handler");
			ReleaseMapLock();
			return ;
		}
		pTapiCall->m_htapiCall = htapiCall ;
		AddTapiCall(pTapiCall, pTapiLine, htapiCall, callId, queuedDevice, callerId, callerIdName, calledDevice, 
			queue, userData, tapicallstate_queued);
	}else {
		htapiCall = pTapiCall->m_htapiCall ;
	}

	if (pTapiCall->m_state == tapicallstate_idle){
		m_log.LogTrace("don't send events because call is idle");
		ReleaseMapLock();
		return;
	}

	// update TapiCall
	std::string deviceID = AsteriskCall_GetDialableNumberFromChannel(pTapiLine->m_deviceID);
	bool	isQueue = (queue == pTapiLine->m_deviceID || deviceID == queue) ;	// i'm a queue and call has gone
	if (isQueue) {
		pTapiCall->m_eLineCallOrigin = linecallorigin_inbound;
	}

	if (isQueue) {	// when the call has been entered into a queue
		pTapiCall->Update(callerId, callerIdName, calledDevice, queue, tapicallstate_queued);
		pTapiCall->m_redirectionId = callerId;
		ReleaseMapLock();

		m_log.LogTrace("  LINECALLSTATE_ACCEPTED htapicall=0x%x htapiLine=0x%x", htapiCall, htapiLine);
		fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_ACCEPTED, 
			0, LINEMEDIAMODE_INTERACTIVEVOICE);
	}else{	// when the call is offering to an agent
		pTapiCall->Update(callerId, callerIdName, calledDevice,"", tapicallstate_alerting);
		ReleaseMapLock();

		m_log.LogTrace("  LINECALLSTATE_RINGBACK htapicall=0x%x", htapiCall);
		fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_RINGBACK, 
			0, LINEMEDIAMODE_INTERACTIVEVOICE);
	}

}

/**
 * create new TAPI call handler if none.
 * sends LINEDISCONNECTMODE_FORWARDED according to if the call is transferred to an agent.
 *	\param xref
 *	\param callId identifier of call
 *	\param divertedDevice
 *	\param divertingDevice
 *	\param callerId who is calling
 *	\param callerIdName
 *	\param calledDevice who has been called
 *	\param userData
 *	\param call_cause
 */
void	ASTTapiProvider::OnDiverted ( MonitorXref xref, DWORD callId, std::string divertedDevice, std::string divertingDevice, std::string callerId, std::string callerIdName, std::string calledDevice, std::string userData, call_cause cause)
{
	m_log.LogTrace("OnDiverted xref=%d callId=%d, divertedDevice=%s, divertingDevice=%s, callerId=%s, "
		"callerIdName=%s, userData=%s, cause=%d", xref, callId, divertedDevice.c_str(), divertingDevice.c_str(), 
		callerId.c_str(), callerIdName.c_str(), userData.c_str(), cause);

	// protect map access
	WaitForMapLock();
	// find xref
	TapiLine	*pTapiLine = GetTapiLine(xref);
	if (!pTapiLine) {
		ReleaseMapLock();
		m_log.LogTrace("  TapiLine not found");
		return ;	// :-(
	}
	m_log.LogTrace("  TapiLine found");

	LINEEVENT	fnCallback = pTapiLine->m_fnEventProc;
	HTAPILINE	htapiLine = pTapiLine->m_htapiLine;
	HTAPICALL	htapiCall = 0;

	m_log.LogTrace(" searching a TapiCall with callId %d in tapiLine 0x%x", callId, pTapiLine);
	TapiCall	*pTapiCall = GetTapiCall(callId, pTapiLine);
	if (!pTapiCall) {
		pTapiCall = new TapiCall();
		pTapiCall->m_callId = callId;
		// get a new htapicall
		m_log.LogTrace("  LINE_NEWCALL");
		fnCallback( htapiLine, 0, LINE_NEWCALL, (DWORD)(void*)pTapiCall, (DWORD)(void*)&(htapiCall), 0);
		if (!htapiCall) {	// handle this!
			m_log.LogWarn("  LINE_NEWCALL returned no handler");
			ReleaseMapLock();
			return ;
		}
		pTapiCall->m_htapiCall = htapiCall ;
		AddTapiCall(pTapiCall, pTapiLine, htapiCall, callId, divertedDevice, callerId, callerIdName, calledDevice, 
			divertingDevice, userData, tapicallstate_unknown);
	}
	else {
		m_log.LogTrace(" TapiCall found");
		htapiCall = pTapiCall->m_htapiCall ;
	}
	
	if (pTapiCall->m_state == tapicallstate_idle){
		m_log.LogTrace("don't send events because call is idle");
		ReleaseMapLock();
		return;
	}

	std::string deviceID = AsteriskCall_GetDialableNumberFromChannel(pTapiLine->m_deviceID);
	bool	isQueue = (divertingDevice == pTapiLine->m_deviceID || deviceID == divertingDevice) ;	// i'm a queue and call has gone

	m_log.LogTrace("    htapiline=0x%x, deviceID=%s, htapicall=0x%x, callerid=%s, callerIdName=%s, calleddevice=%s, "
		"lastRedirection=%s, userData=%s", htapiLine, pTapiLine->m_deviceID.c_str(), htapiCall, callerId.c_str(), callerIdName.c_str(), 
		calledDevice.c_str(), divertedDevice.c_str(), userData.c_str()); 

	if (isQueue) {		// if the call is accepted by the agent
		
		/* If a call is redirect to an agent, queue report a FORWARDED cause
		* LINEDISCONNECTMODE_FORWARDED: The call was forwarded by the switch.
		*/
		pTapiCall->Update(callerId, callerIdName, calledDevice, divertedDevice, tapicallstate_unknown /* previously was set to "forwarded" */);
		m_log.LogTrace("  LINEDISCONNECTMODE_FORWARDED htapicall=0x%x", htapiCall);
		fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_DISCONNECTED, LINEDISCONNECTMODE_FORWARDED, LINEMEDIAMODE_INTERACTIVEVOICE);

		pTapiCall->Update("", "", "", "", tapicallstate_idle);

		m_log.LogTrace("  LINECALLSTATE_IDLE FORWARDED htapicall=0x%x", htapiCall);
		fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_FORWARDED, LINEMEDIAMODE_INTERACTIVEVOICE);

	}
	ReleaseMapLock();

}


/**
 * Sends LINECALLSTATE_DISCONNECTED and LINECALLSTATE_IDLE. Create new TAPI call handler if none.
 * Currently this function only reports six differents causes: NORMAL, OUTOFORDER, BADADDRESS, REJECT, NOANSWER,
 * TEMPFAILURE, UNREACHABLE, UNALLOCATED, UNKNOWN
 * For more information about disconnect causes \see http://msdn2.microsoft.com/en-us/library/ms735622.aspx
 *
 *	\param xref 
 *	\param callId identifier of call
 *	\param releasingDevice what device has been hang up
 *	\param call_cause cause of disconnection
 */
void	ASTTapiProvider::OnConnectionCleared ( MonitorXref xref, DWORD callId, std::string releasingDevice, call_cause cause)
{
	m_log.LogTrace("OnConnectionCleared xref=%d callId=%d, releasingDevice=%s cause=%d", xref, callId, 
		releasingDevice.c_str(), cause);

	// protect map access
	WaitForMapLock();

	// find xref
	TapiLine	*pTapiLine = GetTapiLine(xref);
	if (!pTapiLine) {
		ReleaseMapLock();
		m_log.LogTrace("  TapiLine not found");
		return ;	// :-(
	}

	LINEEVENT	fnCallback = pTapiLine->m_fnEventProc;
	HTAPILINE	htapiLine = pTapiLine->m_htapiLine;
	HTAPICALL	htapiCall = 0;

	// handle only first disconnection and interpret it as both near and far-end disconnect.
	// remove tapicall from map and set tapiline to idle.
	m_log.LogTrace(" searching a TapiCall with callId %d in tapiLine 0x%x", callId, pTapiLine);
	TapiCall	*pTapiCall = GetTapiCall(callId, pTapiLine);
	if (!pTapiCall) {	// not in our map...
		m_log.LogWarn("  TapiCall {callId=%d, hdriverline=0x%x} not in TSP-call-map so ignore event.", 
			callId, pTapiLine);
		ReleaseMapLock();
		return ;
	}
	else {
		htapiCall = pTapiCall->m_htapiCall ;
	}

	if (pTapiCall->m_state == tapicallstate_idle){
		m_log.LogTrace("don't send events because call is idle");
		ReleaseMapLock();
		return;
	}

	//Force the new cause if it's known
	if (cause!=ast_cause_normal && cause!=ast_cause_unknown){
		pTapiCall->m_cause = cause;
	}

	//std::string deviceID = AsteriskCall_GetAddressFromChannel(pTapiLine->m_deviceID);
	m_log.LogTrace("    htapiline=0x%x, deviceID=%s, htapicall=0x%x, releasingDevice=%s, callerid=N/A, callerIdName=N/A, "
		"calleddevice=N/A, lastRedirection=N/A", htapiLine, pTapiLine->m_deviceID.c_str(), htapiCall, releasingDevice.c_str()); 

	/* FIX: [ 1563743 ] a call to ring-group is IDLED after a member answers */
	std::string deviceID = AsteriskCall_GetDialableNumberFromChannel(pTapiLine->m_deviceID);

	bool	isLocalRelease = (releasingDevice == pTapiLine->m_deviceID || releasingDevice == deviceID) ;	

	if (isLocalRelease) {
		pTapiCall->Update("", "", "", pTapiLine->m_deviceID, tapicallstate_idle);
		switch(pTapiCall->m_cause){
			case ast_cause_normal:
				m_log.LogTrace("  LINECALLSTATE_IDLE NORMAL htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_NORMAL, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_out_of_order:	//out of order
				m_log.LogTrace("  LINECALLSTATE_IDLE  OUTOFORDER htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_OUTOFORDER, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_user_busy:	//busy
				m_log.LogTrace("  LINECALLSTATE_IDLE BUSY htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_BUSY, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_unallocated:	//wrong number
				m_log.LogTrace("  LINECALLSTATE_IDLE BADADDRESS htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_BADADDRESS, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_alerting_noanswer:	//no answer
				m_log.LogTrace("  LINECALLSTATE_IDLE NOANSWER htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_NOANSWER, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_reject:	// the call is rejected by the user or the agent
				m_log.LogTrace("  LINECALLSTATE_IDLE REJECT htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_REJECT, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_unreachable:	//no route tu destination
				m_log.LogTrace("  LINECALLSTATE_IDLE UNREACHABLE htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_UNREACHABLE, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_temp_failure:	
				//The call could not be connected or was disconnected because of a temporary failure in the network
				m_log.LogTrace("  LINECALLSTATE_IDLE TEMPFAILURE htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_TEMPFAILURE, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_answering_machine:
				m_log.LogTrace("  LINECALLSTATE_IDLE ACTIVATSP_ANSWERING_MACHINE htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_ACTIVATSP_ANSWERING_MACHINE, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_modem_fax:
				m_log.LogTrace("  LINECALLSTATE_IDLE ACTIVATSP_MODEM_FAX htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_ACTIVATSP_MODEM_FAX, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_congestion:
				m_log.LogTrace("  LINECALLSTATE_IDLE CONGESTION htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_CONGESTION, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			default:	//unknown
				m_log.LogTrace("  LINECALLSTATE_IDLE UNKNOWN htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_UNKNOWN, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
		}

	} else {	// in a multiparty others can release but call does not drop.
		pTapiCall->Update("", "","", releasingDevice, tapicallstate_disconnected);
		pTapiCall->m_redirectionId = releasingDevice;
		pTapiCall->m_cause = cause;
		switch(cause){
			case ast_cause_normal:
				m_log.LogTrace("  LINEDISCONNECTMODE_NORMAL htapicall=0x%x",htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_DISCONNECTED, LINEDISCONNECTMODE_NORMAL, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_out_of_order:	//out of order
				m_log.LogTrace("  LINEDISCONNECTMODE_OUTOFORDER htapicall=0x%x",htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_DISCONNECTED, LINEDISCONNECTMODE_OUTOFORDER, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_user_busy:	//busy
				m_log.LogTrace("  LINEDISCONNECTMODE_BUSY  htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_DISCONNECTED, LINEDISCONNECTMODE_BUSY, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_unallocated:	//wrong number
				m_log.LogTrace("  LINEDISCONNECTMODE_BADADDRESS htapicall=0x%x",htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_DISCONNECTED, LINEDISCONNECTMODE_BADADDRESS, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_alerting_noanswer:	//no answer
				m_log.LogTrace("  LINEDISCONNECTMODE_NOANSWER htapicall=0x%x",htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_DISCONNECTED, LINEDISCONNECTMODE_NOANSWER, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_reject:	// the call is rejected by the user or the agent
				m_log.LogTrace("  LINEDISCONNECTMODE_REJECT htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_DISCONNECTED, LINEDISCONNECTMODE_REJECT, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_unreachable:	//no route tu destination
				m_log.LogTrace("  LINEDISCONNECTMODE_UNREACHABLE htapicall=0x%x",htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_DISCONNECTED, LINEDISCONNECTMODE_UNREACHABLE, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_temp_failure:	
				//The call could not be connected or was disconnected because of a temporary failure in the network
				m_log.LogTrace("  LINEDISCONNECTMODE_TEMPFAILURE htapicall=0x%x",htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_DISCONNECTED, LINEDISCONNECTMODE_TEMPFAILURE, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_answering_machine:
				m_log.LogTrace("  LINECALLSTATE_IDLE ACTIVATSP_ANSWERING_MACHINE htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_ACTIVATSP_ANSWERING_MACHINE, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_modem_fax:
				m_log.LogTrace("  LINECALLSTATE_IDLE ACTIVATSP_MODEM_FAX htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_ACTIVATSP_MODEM_FAX, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			case ast_cause_congestion:
				m_log.LogTrace("  LINECALLSTATE_IDLE CONGESTION htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_CONGESTION, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
			default:	//unknown
				m_log.LogTrace("  LINEDISCONNECTMODE_UNKNOWN htapicall=0x%x",htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_DISCONNECTED, LINEDISCONNECTMODE_UNKNOWN, LINEMEDIAMODE_INTERACTIVEVOICE);
				break;
		}
	}
	ReleaseMapLock();
}

/**
 * Sends LINECALLSTATE_IDLE. Does not create new TAPI call handler if none.
 *
 *	\param xref
 *	\param callId identifier of call
 *	\param userData
 *	\param call_cause
 */
void	ASTTapiProvider::OnCallCleared ( MonitorXref xref, DWORD callId, std::string userData, call_cause cause)
{
	m_log.LogTrace("OnCallCleared xref=%d callId=%d userData=%s cause=%d", xref, callId, userData.c_str(), cause);

	// protect map access
	WaitForMapLock();
	// find xref
	TapiLine	*pTapiLine = GetTapiLine(xref);
	if (!pTapiLine) {
		ReleaseMapLock();
		m_log.LogTrace("  TapiLine not found");
		return ;	// :-(
	}

	LINEEVENT	fnCallback = pTapiLine->m_fnEventProc;
	HTAPILINE	htapiLine = pTapiLine->m_htapiLine;
	HTAPICALL	htapiCall = 0;

	// handle disconnection and interpret it as both near and far-end disconnect.
	// remove tapicall from map and set tapiline to idle.
	m_log.LogTrace(" searching a TapiCall with callId %d in tapiLine 0x%x", callId, pTapiLine);
	TapiCall	*pTapiCall = GetTapiCall(callId, pTapiLine);
	if (!pTapiCall) {	// not in our map...
		m_log.LogWarn("  TapiCall {callId=%d, hdriverline=0x%x} not in TSP-call-map so ignore event.", callId, pTapiLine);
		ReleaseMapLock();
		return ;
	}
	else {
		htapiCall = pTapiCall->m_htapiCall ;
	}

	if (pTapiCall->m_state == tapicallstate_idle){
		m_log.LogTrace("don't send events because call is idle");
		ReleaseMapLock();
		return;
	}

	//Force the new cause if it's known
	if (cause!=ast_cause_normal && cause!=ast_cause_unknown){
		pTapiCall->m_cause = cause;
	}

	m_log.LogTrace("    htapiline=0x%x, deviceID=%s, htapicall=0x%x, callerid=N/A, callerIdName=N/A, calleddevice=N/A, "
		"lastRedirection=N/A, userData=%s", htapiLine, pTapiLine->m_deviceID.c_str(), htapiCall, pTapiCall->m_userData.c_str()); 

	pTapiCall->Update("", "", "", "", tapicallstate_idle);

	switch(pTapiCall->m_cause){
		case ast_cause_normal:
			m_log.LogTrace("  LINECALLSTATE_IDLE NORMAL htapicall=0x%x", htapiCall);
			fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_NORMAL, LINEMEDIAMODE_INTERACTIVEVOICE);
			break;
		case ast_cause_out_of_order:	//out of order
			m_log.LogTrace("  LINECALLSTATE_IDLE  OUTOFORDER htapicall=0x%x", htapiCall);
			fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_OUTOFORDER, LINEMEDIAMODE_INTERACTIVEVOICE);
			break;
		case ast_cause_user_busy:	//busy
			m_log.LogTrace("  LINECALLSTATE_IDLE BUSY htapicall=0x%x", htapiCall);
			fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_BUSY, LINEMEDIAMODE_INTERACTIVEVOICE);
			break;
		case ast_cause_unallocated:	//wrong number
			m_log.LogTrace("  LINECALLSTATE_IDLE BADADDRESS htapicall=0x%x", htapiCall);
			fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_BADADDRESS, LINEMEDIAMODE_INTERACTIVEVOICE);
			break;
		case ast_cause_alerting_noanswer:	//no answer
			m_log.LogTrace("  LINECALLSTATE_IDLE NOANSWER htapicall=0x%x", htapiCall);
			fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_NOANSWER, LINEMEDIAMODE_INTERACTIVEVOICE);
			break;
		case ast_cause_reject:	// the call is rejected by the user or the agent
			m_log.LogTrace("  LINECALLSTATE_IDLE REJECT htapicall=0x%x", htapiCall);
			fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_REJECT, LINEMEDIAMODE_INTERACTIVEVOICE);
			break;
		case ast_cause_unreachable:	//no route tu destination
			m_log.LogTrace("  LINECALLSTATE_IDLE UNREACHABLE htapicall=0x%x", htapiCall);
			fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_UNREACHABLE, LINEMEDIAMODE_INTERACTIVEVOICE);
			break;
		case ast_cause_temp_failure:	
			//The call could not be connected or was disconnected because of a temporary failure in the network
			m_log.LogTrace("  LINECALLSTATE_IDLE TEMPFAILURE htapicall=0x%x", htapiCall);
			fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_TEMPFAILURE, LINEMEDIAMODE_INTERACTIVEVOICE);
			break;
		case ast_cause_answering_machine:
			m_log.LogTrace("  LINECALLSTATE_IDLE ACTIVATSP_ANSWERING_MACHINE htapicall=0x%x", htapiCall);
			fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_ACTIVATSP_ANSWERING_MACHINE, LINEMEDIAMODE_INTERACTIVEVOICE);
			break;
		case ast_cause_modem_fax:
			m_log.LogTrace("  LINECALLSTATE_IDLE ACTIVATSP_MODEM_FAX htapicall=0x%x", htapiCall);
			fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_ACTIVATSP_MODEM_FAX, LINEMEDIAMODE_INTERACTIVEVOICE);
			break;
		case ast_cause_congestion:
			m_log.LogTrace("  LINECALLSTATE_IDLE CONGESTION htapicall=0x%x", htapiCall);
			fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_CONGESTION, LINEMEDIAMODE_INTERACTIVEVOICE);
			break;
		default:	//unknown
			m_log.LogTrace("  LINECALLSTATE_IDLE UNKNOWN htapicall=0x%x", htapiCall);
			fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_UNKNOWN, LINEMEDIAMODE_INTERACTIVEVOICE);
			break;
	}
	ReleaseMapLock();

}

/**
*	Held a call and send LINECALLSTATE_ONHOLD
*	\param xref
*	\param callId identifier of call
*	\param holdingDevice
*	\param call_cause
*/
void	ASTTapiProvider::OnHold(MonitorXref xref, DWORD callId, std::string holdingDevice, call_cause cause)
{
	m_log.LogTrace("Onhold xref=%d callId=%d holdingDevice=%s", xref, callId, holdingDevice.c_str());
	WaitForMapLock();
	TapiLine	*pTapiLine = GetTapiLine(xref);
	if (!pTapiLine) {
		ReleaseMapLock();
		m_log.LogTrace("  TapiLine not found");
		return ;	// :-(
	}

	LINEEVENT	fnCallback = pTapiLine->m_fnEventProc;
	HTAPILINE	htapiLine = pTapiLine->m_htapiLine;
	HTAPICALL	htapiCall = 0;

	m_log.LogTrace(" searching a TapiCall with callId %d in tapiLine 0x%x", callId, pTapiLine);
	TapiCall	*pTapiCall = GetTapiCall(callId, pTapiLine);
	if (!pTapiCall) {	// not in our map...
		m_log.LogWarn("  TapiCall {callId=%d, hdriverline=0x%x} not in TSP-call-map so ignore event.", 
			callId, pTapiLine);
		ReleaseMapLock();
		return ;
	}
	else {
		htapiCall = pTapiCall->m_htapiCall ;
	}

	if (pTapiCall->m_state == tapicallstate_idle){
		m_log.LogTrace("don't send events because call is idle");
		ReleaseMapLock();
		return;
	}

	std::string deviceID = AsteriskCall_GetDialableNumberFromChannel(pTapiLine->m_deviceID);
	bool	isLocalHold = (holdingDevice == pTapiLine->m_deviceID || holdingDevice == deviceID) ;

	if (isLocalHold){
		pTapiCall->m_state = tapicallstate_held;
		pTapiCall->m_redirectionId = pTapiLine->m_deviceID;
		m_log.LogTrace("  LINECALLSTATE_ONHOLD htapicall=0x%x", htapiCall);
		fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_ONHOLD, 0, LINEMEDIAMODE_INTERACTIVEVOICE);
	}

	ReleaseMapLock();
}

/**
*	Retreive a call and send LINECALLSTATE_CONNECTED
*	\param xref
*	\param callId identifier of call
*	\param unholdingDevice
*	\param call_cause
*/

void	ASTTapiProvider::OnUnhold ( MonitorXref xref, DWORD callId, std::string unholdingDevice, call_cause cause)
{
	m_log.LogTrace("OnUnhold xref=%d callId=%d unholdingDevice=%s", xref, callId, unholdingDevice.c_str());

	WaitForMapLock();
	TapiLine	*pTapiLine = GetTapiLine(xref);
	if (!pTapiLine) {
		ReleaseMapLock();
		m_log.LogTrace("  TapiLine not found");
		return ;	// :-(
	}

	LINEEVENT	fnCallback = pTapiLine->m_fnEventProc;
	HTAPILINE	htapiLine = pTapiLine->m_htapiLine;
	HTAPICALL	htapiCall = 0;

	m_log.LogTrace(" searching a TapiCall with callId %d in tapiLine 0x%x", callId, pTapiLine);
	TapiCall	*pTapiCall = GetTapiCall(callId, pTapiLine);
	if (!pTapiCall) {	// not in our map...
		m_log.LogWarn("  TapiCall {callId=%d, hdriverline=0x%x} not in TSP-call-map so ignore event.", 
			callId, pTapiLine);
		ReleaseMapLock();
		return ;
	}
	else {
		htapiCall = pTapiCall->m_htapiCall ;
	}


	if (pTapiCall->m_state == tapicallstate_idle){
		m_log.LogTrace("don't send events because call is idle");
		ReleaseMapLock();
		return;
	}

	std::string deviceID = AsteriskCall_GetDialableNumberFromChannel(pTapiLine->m_callerId);
	bool	isLocalUnhold = (unholdingDevice == pTapiLine->m_deviceID || unholdingDevice == deviceID) ;

	if (isLocalUnhold){
		pTapiCall->m_state = tapicallstate_connected;
		pTapiCall->m_redirectionId = pTapiLine->m_deviceID;
		m_log.LogTrace("  LINECALLSTATE_CONNECTED  htapicall=0x%x", htapiCall);
		fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_CONNECTED, 0, LINEMEDIAMODE_INTERACTIVEVOICE);
	}

	ReleaseMapLock();
}

/**
 * closes each TapiLine.
 * \see http://msdn.microsoft.com/library/default.asp?url=/library/en-us/tapi/tspi/line_close.asp
 */
void	ASTTapiProvider::OnLinkDown()
{
	m_log.LogTrace("OnLinkDown");

	// protect map access
	WaitForMapLock();

	for (Map_Xref2TapiLine::iterator i = m_mapXref2TapiLine.begin(); i!=m_mapXref2TapiLine.end(); i++) {
		TapiLine	*pTapiLine = i->second;
		LINEEVENT	fnCallback = pTapiLine->m_fnEventProc;
		m_log.LogTrace("  sent LINE_CLOSE to htapiline=0x%x", pTapiLine->m_htapiLine);
		fnCallback(pTapiLine->m_htapiLine, 0, LINE_CLOSE, 0, 0, 0);
	}

	ReleaseMapLock();

	//The cleanup is done by the last lineclose!! CallMapCleanup();
}

/**
*	Function TSPI_lineSetupTransfer
*	Initiates a transfer of the call specified by hdCall. 
*	It establishes a consultation call, lphdConsultCall, 
*	on which the party can be dialed that can become the destination of the transfer.
*	\see http://msdn2.microsoft.com/en-us/library/ms725609.aspx
*	\param dwRequestID
*	\param hdCall
*	\param htConsultCall
*	\param lphdConsultCall
*	\param lpCallParams
*	\return requestId
*/
LONG ASTTapiProvider::TSPI_lineSetupTransfer(
	DRV_REQUESTID	dwRequestID,
	HDRVCALL hdCall, 
	HTAPICALL htConsultCall,
	LPHDRVCALL lphdConsultCall, 
	LPLINECALLPARAMS const lpCallParams
	)
{
	m_log.LogTrace("TSPI_lineSetupTransfer() hdcall =0x%x htConsultCall = 0x%x", hdCall,htConsultCall);
	if (hdCall == 0){
		m_log.LogTrace(" not exist any Call with hdCall = 0 (FEATURE EVENT)");
		return 0;
	}
	WaitForMapLock();
	TapiCall	*pTapiCall = GetTapiCall(hdCall);
	if (!pTapiCall) {
		m_log.LogError("  LINEERR_INVALCALLHANDLE");
		ReleaseMapLock();
		return LINEERR_INVALCALLHANDLE;
	}
	TapiLine	*pTapiLine = GetTapiLine((MonitorXref)pTapiCall->m_pTapiLine->m_xref);
	if (pTapiLine) {
		TapiCall *pNewTapiCall = new TapiCall();
		pNewTapiCall->m_htapiCall = htConsultCall ;
		pNewTapiCall->m_pTapiLine = pTapiLine;
		pNewTapiCall->m_realCallId = pTapiCall->m_callId;
		m_log.LogTrace("  creating a new TapiCall=0x%x", pNewTapiCall);

		AddTapiCall(pNewTapiCall, pTapiLine, htConsultCall, 0, pTapiLine->m_deviceID,pTapiCall->m_callerId, 
			pTapiCall->m_callerIdName,pTapiCall->m_calledDevice, "",pTapiCall->m_userData, tapicallstate_unknown );


		*lphdConsultCall = (HDRVCALL) pNewTapiCall ;

		LINEEVENT	fnCallback = pTapiLine->m_fnEventProc;
		HTAPILINE	htapiLine = pNewTapiCall->m_pTapiLine->m_htapiLine;
		ReleaseMapLock();
		
		CompletionProc(dwRequestID, 0);
		fnCallback(htapiLine, htConsultCall, LINE_CALLSTATE, LINECALLSTATE_DIALTONE, 
			0, LINEMEDIAMODE_INTERACTIVEVOICE);
		return dwRequestID;
	}else{
		ReleaseMapLock();
		m_log.LogError(" LINEERR_INVALLINESTATE");
		return LINEERR_INVALLINESTATE;
	}
	
	return LINEERR_OPERATIONFAILED;
}

/**
*	Function TSPI_lineDial
*	Dials the specified dialable number on the specified call.
*	\see http://msdn2.microsoft.com/en-us/library/ms725542.aspx
*	\param dwRequestID
*	\param hdCall
*	\param lpszDestAddress
*	\param dwCountryCode
*	\return dwRequestId or LINEERR_OPERATIONFAILED
*/
LONG	ASTTapiProvider::TSPI_lineDial(
	DRV_REQUESTID dwRequestID,
	HDRVCALL hdCall, 
	LPCWSTR lpszDestAddress, 
	DWORD dwCountryCode)
{

	std::string	calleddevice = wstr2string(lpszDestAddress);
	m_log.LogTrace("TSPI_lineDial() dwRequestID=0x%x, hdCall=0x%x, calledDevice=%s", dwRequestID, hdCall, calleddevice.c_str() );
	
	std::string sFeature = ContainsVar(calleddevice,"FEATURE");
	if(sFeature.compare("")!=0){
		//It's a feature!
		WaitForMapLock();
		TapiCall	*pTapiCall = GetTapiCall(hdCall);
		if (pTapiCall) {
			DWORD		dwCallId = pTapiCall->m_callId;
			std::string deviceID = AsteriskCall_GetAddressFromChannel(pTapiCall->m_address);
			ReleaseMapLock();
			if (_stricmp(sFeature.c_str(),FEATURE_RECORDING_START)==0)
			{
				//StartRecording
				AC_RESULT rc = m_cstaProvider->StartRecording(dwCallId, deviceID, ContainsVar(calleddevice,"FILE","activatspRecording-"+deviceID), ContainsVar(calleddevice,"FORMAT","wav"), ContainsVar(calleddevice,"MIX","true"));
				if (rc==ok){
					CompletionProc(dwRequestID, 0);
					return (dwRequestID);
				}
				else{
					return (LINEERR_OPERATIONFAILED);
				}
			}
			if (_stricmp(sFeature.c_str(),FEATURE_RECORDING_STOP)==0)
			{
				//StopRecording
				AC_RESULT rc = m_cstaProvider->StopRecording(dwCallId, deviceID);
				if (rc==ok){
					CompletionProc(dwRequestID, 0);
					return (dwRequestID);
				}
				else{
					return (LINEERR_OPERATIONFAILED);
				}
			}
			if (_stricmp(sFeature.c_str(),FEATURE_RECORDING_PAUSE)==0)
			{
				//PauseRecording
				AC_RESULT rc = m_cstaProvider->PauseRecording(dwCallId, deviceID);
				if (rc==ok){
					CompletionProc(dwRequestID, 0);
					return (dwRequestID);
				}
				else{
					return (LINEERR_OPERATIONFAILED);
				}
			}
			if (_stricmp(sFeature.c_str(),FEATURE_RECORDING_UNPAUSE)==0)
			{
				//UnpauseRecording
				AC_RESULT rc = m_cstaProvider->UnpauseRecording(dwCallId, deviceID);
				if (rc==ok){
					CompletionProc(dwRequestID, 0);
					return (dwRequestID);
				}
				else{
					return (LINEERR_OPERATIONFAILED);
				}
			}
			if (_stricmp(sFeature.c_str(),FEATURE_RECORDING_CHANGE_FILENAME)==0)
			{
				//ChangeRecordingFileName
				AC_RESULT rc = m_cstaProvider->ChangeRecordingFileName(dwCallId, deviceID, ContainsVar(calleddevice,"FILE","activatspRecording-"+deviceID+"renamed"));
				if (rc==ok){
					CompletionProc(dwRequestID, 0);
					return (dwRequestID);
				}
				else{
					return (LINEERR_OPERATIONFAILED);
				}
			}
		}
		else
		{
			ReleaseMapLock();
			m_log.LogError("  no TapiCall is active for tapihandler");
			return LINEERR_INVALCALLSTATE;
		}
	}
	else{
		//it's a transfer
		std::string newCalleddevice;
		for (size_t i=0; i<calleddevice.size(); i++) {
			switch(calleddevice[i]) {
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
				case '*':
				case '#':
				case 'A':
				case 'B':
				case 'C':
				case 'D':
					newCalleddevice+=calleddevice[i];
					break;
				case '\'':	// escape character, add the remaining of the calleddevice string
					newCalleddevice.append(calleddevice, i+1, calleddevice.size()-i);
					i=calleddevice.size();	// force continue
					break;
				default:
					break;
			}
		}
		m_log.LogTrace("  filtered calledDevice %s => %s", calleddevice.c_str(), newCalleddevice.c_str());
		if (newCalleddevice[0] == '#'){
			m_log.LogTrace("  blindTransfer to %s",newCalleddevice.c_str());
			WaitForMapLock();
			TapiCall	*pTapiCall = GetTapiCall(hdCall);
			if (pTapiCall) {
				TapiLine *pTapiLine = GetTapiLine((MonitorXref)pTapiCall->m_pTapiLine->m_xref);
				if (!pTapiLine){
					ReleaseMapLock();
					return LINEERR_INVALCALLSTATE;
				}
				std::string toDevice = newCalleddevice.substr(1,newCalleddevice.size());
				std::string deviceID = AsteriskCall_GetAddressFromChannel(pTapiLine->m_deviceID);
				DWORD realCallId = pTapiCall->m_realCallId;
				LINEEVENT	fnCallback = pTapiLine->m_fnEventProc;
				HTAPILINE htapiLine = pTapiLine->m_htapiLine;
				HTAPICALL htapiCall = pTapiCall->m_htapiCall;
				ReleaseMapLock();
				
				AC_RESULT rc = m_cstaProvider->FastTransfer(realCallId,deviceID,toDevice,registryGetOutgoingContext("from-internal"));
				if (rc==ok){
					CompletionProc(dwRequestID, 0);
					
					WaitForMapLock();
					pTapiCall = GetTapiCall(hdCall);
					if (!pTapiCall) {
						ReleaseMapLock();
						return (LINEERR_OPERATIONFAILED);
					}

					if (pTapiCall->m_state != tapicallstate_idle) {
						pTapiCall->m_state = tapicallstate_idle;
						m_log.LogTrace("  LINECALLSTATE_IDLE" );
						fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_NORMAL, LINEMEDIAMODE_INTERACTIVEVOICE);
					}
					ReleaseMapLock();
					return dwRequestID;
				}
				else
				{
					ReleaseMapLock();
					return (LINEERR_OPERATIONFAILED);
				}
			}	
		}else{
			//Consultation transfer
			m_log.LogTrace("  consultation transfer to %s not supported",newCalleddevice.c_str());
		}
	}

	return LINEERR_OPERATIONUNAVAIL;
}

/**
*	In Features find if address contains the spcified key - value pair
*	\param sEncodedString
*	\param sKey
*	\param sDefaultValue
*	\return Value, if the key is not found in the sEncodedString returns sDefaultValue
*/
std::string ASTTapiProvider::ContainsVar(std::string sEncodedString, std::string sKey, std::string sDefaultValue)
{
	DWORD callid = -1;

	size_t pos = -1;
	sKey="#"+sKey+"=";
	std::string sValue="";
	if ((pos = sEncodedString.find(sKey))!= -1){
		std::string aux = sEncodedString.substr(pos+sKey.size(), sEncodedString.size());
		pos = -1;
		if ((pos = aux.find("#")) != -1){
			sValue = aux.substr(0,pos).c_str();
		}else{
			sValue = aux.c_str();
		}
		m_log.LogTrace0("EncodedString '%s' contain the key:%s%s",sEncodedString.c_str(),sKey.c_str(),sValue.c_str());
	}else{
		m_log.LogTrace0("EncodedString '%s' does not contain the key:%s",sEncodedString.c_str(),sKey.c_str());
	}

	if(sValue.compare("")==0){
		m_log.LogTrace0("Value for key %s isn't found then assign default value = %s",sKey.c_str(),sDefaultValue.c_str());
		sValue=sDefaultValue;
	}

	return sValue;
}

/**
*	In Features find if address contains the spcified key - value pair
*	\param sEncodedString
*	\param sKey
*	\return Value, if the key is not found in the sEncodedString returns ""
*/
std::string ASTTapiProvider::ContainsVar(std::string sEncodedString, std::string sKey)
{
	return ContainsVar(sEncodedString,sKey,"");
}


/**
*	Function TSPI_lineHold
*	places the specified call on hold
*	\see http://msdn2.microsoft.com/en-us/library/ms725575.aspx
*	\param dwRequestID
*	\param hdCall
*	\return dwRequestId or LINEERR_OPERATIONFAILED
*/
LONG	ASTTapiProvider::TSPI_lineHold(DRV_REQUESTID dwRequestID, HDRVCALL hdCall)
{
	m_log.LogTrace("TSPI_lineHold dwRequestID=0x%x, hdCall=0x%x", dwRequestID, hdCall);
	if (hdCall == 0){
		m_log.LogTrace(" not exist any Call with hdCall = 0 (FEATURE EVENT)");
		return 0;
	}
	WaitForMapLock();
	TapiCall	*pTapiCall = GetTapiCall(hdCall);	
	if (!pTapiCall){
		ReleaseMapLock();
		m_log.LogError("LINEERR_INVALCALLHANDLE");
		return LINEERR_INVALCALLHANDLE;
	}

	TapiLine	*pTapiLine = GetTapiLine((MonitorXref)pTapiCall->m_pTapiLine->m_xref);
	if (!pTapiLine){
		ReleaseMapLock();
		m_log.LogError("LINEERR_RESOURCEUNAVAIL");
		return LINEERR_RESOURCEUNAVAIL;
	}
	std::string deviceID = AsteriskCall_GetAddressFromChannel(pTapiCall->m_pTapiLine->m_deviceID);
	AC_RESULT rc = m_cstaProvider->HoldCall(pTapiCall->m_callId,deviceID,
		registryGetOutgoingContext("from-internal"));
	if (rc == ok){
		LINEEVENT	fnCallback = pTapiLine->m_fnEventProc;
		CompletionProc(dwRequestID, 0);
		fnCallback(pTapiLine->m_htapiLine, pTapiCall->m_htapiCall, LINE_CALLSTATE, LINECALLSTATE_ONHOLD, 0, 
			LINEMEDIAMODE_INTERACTIVEVOICE);
		ReleaseMapLock();
		return dwRequestID;

	}
	ReleaseMapLock();
	m_log.LogError("LINEERR_OPERATIONFAILED");
	return (LINEERR_OPERATIONFAILED);
}

/**
*	TSPI_lineUnHold
*	retrieves the specified held call.
*	\see http://msdn2.microsoft.com/en-us/library/ms725612.aspx
*	\param dwRequestID request identifier
*	\param hdCall	hdCall identifier
*	\return dwRequestId or LINEERR_OPERATIONFAILED
*/
LONG	ASTTapiProvider::TSPI_lineUnHold(DRV_REQUESTID dwRequestID,HDRVCALL hdCall)
{
	m_log.LogTrace("TSPI_lineUnHold dwRequestID=0x%x, hdCall=0x%x", dwRequestID, hdCall);
	if (hdCall == 0){
		m_log.LogTrace(" not exist any Call with hdCall = 0 (FEATURE EVENT)");
		return 0;
	}
	WaitForMapLock();
	TapiCall	*pTapiCall = GetTapiCall(hdCall);	
	if (!pTapiCall){
		ReleaseMapLock();
		m_log.LogError("LINEERR_INVALCALLHANDLE");
		return LINEERR_INVALCALLHANDLE;
	}

	TapiLine	*pTapiLine = GetTapiLine((MonitorXref)pTapiCall->m_pTapiLine->m_xref);
	if (!pTapiLine){
		ReleaseMapLock();
		m_log.LogError("LINEERR_RESOURCEUNAVAIL");
		return LINEERR_RESOURCEUNAVAIL;
	}
	std::string deviceID = AsteriskCall_GetAddressFromChannel(pTapiLine->m_deviceID);
	AC_RESULT rc = m_cstaProvider->RetrieveCall(pTapiCall->m_callId,deviceID,
		registryGetOutgoingContext("from-internal"));
	if (rc == ok){
		LINEEVENT	fnCallback = pTapiLine->m_fnEventProc;
		CompletionProc(dwRequestID, 0);
		fnCallback(pTapiLine->m_htapiLine, pTapiCall->m_htapiCall, LINE_CALLSTATE, LINECALLSTATE_CONNECTED, 
			0, LINEMEDIAMODE_INTERACTIVEVOICE);
		ReleaseMapLock();
		return dwRequestID;

	}
	ReleaseMapLock();
	m_log.LogError("LINEERR_OPERATIONFAILED");
	return (LINEERR_OPERATIONFAILED);
}

/**
 * Function TSPI_lineMakeCall.
 *
 * This function is called by TAPI to initialize a new outgoing call. 
 *
 * Originates a call via AMI originate action.
 *
 * TAPI standard compatible applications should send calledNumber in 'standard dialable form'
 * \see http://windowssdk.msdn.microsoft.com/en-us/library/ms726017.aspx 
 *
 * ... but begining in 1.2 and for flexibility reasons Activa tsp will filter out characters as following:
 *	- accepted characters: 0-9, A-D, *, #
 *	- escape character [ ' ] (apostrophe): stops filtering and accepts all remaining characters.
 *
 * sample:
 *	- calledNumber = +1 (555) 43-21=> 15554321
 *	- calledNumber = 'extension0001 => extension0001
 *
 * Some dialed sequences are interpreted as the following feature-codes:
 *	- FEATURE_LOGIN		*7361\<queue>	QueueAdd queue=\<queue>, paused=yes. Example: *7361555
 *	- FEATURE_LOGOUT	*7360\<queue>	QueueRemove queue=\<queue>. Example: *7360555
 *	- FEATURE_READY		*7351\<queue>	QueuePause queue=\<queue> paused=yes. Example: *7351555
 *	- FEATURE_NOTREADY	*7350\<queue>	QueuePause queue=\<queue> paused=no. Example: *7350555
 *
 * Adds new call to callmap. 
 *
 *	If LINECALLPARAMFLAGS_PREDICTIVEDIAL flag on pCallParams is activated, a makePredictiveCall will be executed.
 *	\see http://msdn2.microsoft.com/en-us/library/ms735534.aspx
 *
 * If ASTProvider returns ok_feature_dialed meaning that make call succeded but no real 
 * call was initiated, call is inmediately erased from map, and LINECALLSTATE_DIALTONE + LINECALLSTATE_IDLE events
 * are sent to tapi.
 * \see http://msdn.microsoft.com/library/default.asp?url=/library/en-us/tapi/tspi/tspi_linemakecall.asp
 *	\param dwRequestID
 *	\param hdLine
 *	\param htCall
 *	\param phdCall
 *	\param pszDestAddress
 *	\param dwCountryCode
 *	\param pCallParams
 */
LONG ASTTapiProvider::TSPI_lineMakeCall(
    DRV_REQUESTID       dwRequestID,
    HDRVLINE            hdLine,
    HTAPICALL           htCall,
    LPHDRVCALL          phdCall,
    LPCWSTR             pszDestAddress,
    DWORD               dwCountryCode,
    LPLINECALLPARAMS    const pCallParams
    )
{
	DWORD	callId = 0;
	LONG tr = 0;
	*phdCall = (HDRVCALL) 0 ;	// init 
	HTAPILINE	hTapiLine = 0;
	LINEEVENT	fnCallback = 0 ;
    
	std::string	calleddevice = wstr2string(pszDestAddress);
	m_log.LogTrace("TSPI_lineMakeCall() hdLine=0x%x, htCall=0x%x, phdCall=0x%x, calledDevice=%s", hdLine, htCall, phdCall, calleddevice.c_str() );

	WaitForMapLock();		// lock tsp map 
	TapiLine	*pTapiLine = GetTapiLine((MonitorXref)hdLine);
	if (pTapiLine) {
		std::string sDeviceID = pTapiLine->m_deviceID;
		std::string sCallerIdName = pTapiLine->m_callerIdName; 
											

		// some features...
		std::string	subjectDeviceId = pTapiLine->m_deviceID;
		std::string sFeature = ContainsVar(calleddevice,"FEATURE");
		hTapiLine = pTapiLine->m_htapiLine ;	// we copy to some method scope vars just to release map asap ...
		fnCallback = pTapiLine->m_fnEventProc;

		ReleaseMapLock();	
		if (sFeature.size()>0)
		{
			if (_stricmp(sFeature.c_str(),FEATURE_LOGIN)==0) 
			{
				std::string	queue = ContainsVar(calleddevice,"QUEUE");
				m_log.LogTrace("  FEATURE_LOGIN");
				AC_RESULT rc  = m_cstaProvider->SetAgentStatus(subjectDeviceId, queue, LOGIN, registryGetAgentContext("from-internal"));
				if (rc==ok) {
					// insert a dummy tapiCall in the map 
					callId = m_cstaProvider->GetNewCallId();
					WaitForMapLock();
					pTapiLine = GetTapiLine((MonitorXref)hdLine);	// line could have been deleted from map while it was unloked so get it again
					if (pTapiLine != NULL) {
						TapiCall *pTapiCall = new TapiCall();
						*phdCall = (HDRVCALL) pTapiCall ;	// init driver's call handler 
						pTapiCall->m_htapiCall = htCall;
						pTapiCall->m_callId = callId ;
						AddTapiCall ( pTapiCall, pTapiLine, htCall, callId, pTapiLine->m_deviceID /*address*/, pTapiLine->m_deviceID /*callerId*/, "" /*calleridname*/, calleddevice /*calleddevice*/, "" /*lastredirectiondevice*/, "", tapicallstate_unknown );

						CompletionProc(dwRequestID, 0);	
						
						m_log.LogTrace(" simulating TapiCall 0x%x with htCall 0x%x, callId=%d to LOGIN device %s", pTapiCall, htCall, callId, (LPCSTR) subjectDeviceId.c_str());

						fnCallback(hTapiLine, htCall, LINE_CALLSTATE, LINECALLSTATE_DIALTONE, 0, LINEMEDIAMODE_INTERACTIVEVOICE);
						fnCallback(hTapiLine, htCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_NORMAL, LINEMEDIAMODE_INTERACTIVEVOICE);
						ReleaseMapLock();	// release lock asap
						m_cstaProvider->DropCallId(callId);
						return dwRequestID;
					}
					else {	// line has been closed !!!
						ReleaseMapLock();	// release lock
						m_log.LogTrace(" line has been closed! (LINEERR_INVALLINEHANDLE)");
						return LINEERR_INVALLINEHANDLE ;
					}
				}
				else {	
					return LINEERR_OPERATIONUNAVAIL;
				}
			}
			else if (_stricmp(sFeature.c_str(),FEATURE_LOGOUT)==0) {
				std::string	queue = ContainsVar(calleddevice,"QUEUE");
				m_log.LogTrace("  FEATURE_LOGOUT");
				AC_RESULT rc  = m_cstaProvider->SetAgentStatus(subjectDeviceId, queue, LOGOUT, registryGetAgentContext("from-internal"));
				if (rc==ok) {
					// insert a dummy tapiCall in the map 
					callId = m_cstaProvider->GetNewCallId();
					WaitForMapLock();
					pTapiLine = GetTapiLine((MonitorXref)hdLine);	// line could have been deleted from map while it was unloked so get it again
					if (pTapiLine != NULL) {
						TapiCall *pTapiCall = new TapiCall();
						*phdCall = (HDRVCALL) pTapiCall ;	// init driver's call handler 
						pTapiCall->m_htapiCall = htCall;
						pTapiCall->m_callId = callId ;
						AddTapiCall ( pTapiCall, pTapiLine, htCall, callId, pTapiLine->m_deviceID /*address*/, pTapiLine->m_deviceID /*callerId*/, "" /*calleridname*/, calleddevice /*calleddevice*/, "" /*lastredirectiondevice*/, "", tapicallstate_unknown );

						CompletionProc(dwRequestID, 0);	
						
						m_log.LogTrace(" simulating TapiCall 0x%x with htCall 0x%x, callId=%d to LOGOUT device %s", pTapiCall, htCall, callId, (LPCSTR) subjectDeviceId.c_str());

						fnCallback(hTapiLine, htCall, LINE_CALLSTATE, LINECALLSTATE_DIALTONE, 0, LINEMEDIAMODE_INTERACTIVEVOICE);
						fnCallback(hTapiLine, htCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_NORMAL, LINEMEDIAMODE_INTERACTIVEVOICE);
						ReleaseMapLock();	// release lock asap
						m_cstaProvider->DropCallId(callId);
						return dwRequestID;
					}
					else {	// line has been closed !!!
						ReleaseMapLock();	// release lock
						m_log.LogTrace(" line has been closed! (LINEERR_INVALLINEHANDLE)");
						return LINEERR_INVALLINEHANDLE ;
					}
				}
				else {	
					/// \todo: more detailed error handling on FEATURE_LOGOUT
					return LINEERR_OPERATIONUNAVAIL;
				}
			}
			else if (_stricmp(sFeature.c_str(),FEATURE_READY)==0) {
				std::string	queue = ContainsVar(calleddevice,"QUEUE");
				m_log.LogTrace("  FEATURE_READY");
				AC_RESULT rc  = m_cstaProvider->SetAgentStatus(subjectDeviceId, queue, READY, registryGetAgentContext("from-internal"));
				if (rc==ok) {
					// insert a dummy tapiCall in the map 
					callId = m_cstaProvider->GetNewCallId();
					WaitForMapLock();
					pTapiLine = GetTapiLine((MonitorXref)hdLine);	// line could have been deleted from map while it was unloked so get it again
					if (pTapiLine != NULL) {
						TapiCall *pTapiCall = new TapiCall();
						*phdCall = (HDRVCALL) pTapiCall ;	// init driver's call handler 
						pTapiCall->m_htapiCall = htCall;
						pTapiCall->m_callId = callId ;
						AddTapiCall ( pTapiCall, pTapiLine, htCall, callId, pTapiLine->m_deviceID /*address*/, pTapiLine->m_deviceID /*callerId*/, "" /*calleridname*/, calleddevice /*calleddevice*/, "" /*lastredirectiondevice*/, "", tapicallstate_unknown );

						CompletionProc(dwRequestID, 0);	
						
						m_log.LogTrace(" simulating TapiCall 0x%x with htCall 0x%x, callId=%d to READY device %s", pTapiCall, htCall, callId, (LPCSTR) subjectDeviceId.c_str());

						fnCallback(hTapiLine, htCall, LINE_CALLSTATE, LINECALLSTATE_DIALTONE, 0, LINEMEDIAMODE_INTERACTIVEVOICE);
						fnCallback(hTapiLine, htCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_NORMAL, LINEMEDIAMODE_INTERACTIVEVOICE);
						ReleaseMapLock();	// release lock asap
						m_cstaProvider->DropCallId(callId);
						return dwRequestID;
					}
					else {	// line has been closed !!!
						ReleaseMapLock();	// release lock
						m_log.LogTrace(" line has been closed! (LINEERR_INVALLINEHANDLE)");
						return LINEERR_INVALLINEHANDLE ;
					}
				}
				else {	
					/// \todo: more detailed error handling on FEATURE_READY
					return LINEERR_OPERATIONUNAVAIL;
				}
			}
			else if (_stricmp(sFeature.c_str(),FEATURE_NOTREADY)==0) {
				std::string	queue = ContainsVar(calleddevice,"QUEUE");
				m_log.LogTrace("  FEATURE_NOTREADY");
				AC_RESULT rc  = m_cstaProvider->SetAgentStatus(subjectDeviceId, queue, NOTREADY, registryGetAgentContext("from-internal"));
				if (rc==ok) {
					// insert a dummy tapiCall in the map 
					callId = m_cstaProvider->GetNewCallId();
					WaitForMapLock();
					pTapiLine = GetTapiLine((MonitorXref)hdLine);	// line could have been deleted from map while it was unloked so get it again
					if (pTapiLine != NULL) {
						TapiCall *pTapiCall = new TapiCall();
						*phdCall = (HDRVCALL) pTapiCall ;	// init driver's call handler 
						pTapiCall->m_htapiCall = htCall;
						pTapiCall->m_callId = callId ;
						AddTapiCall ( pTapiCall, pTapiLine, htCall, callId, pTapiLine->m_deviceID /*address*/, pTapiLine->m_deviceID /*callerId*/, "" /*calleridname*/, calleddevice /*calleddevice*/, "" /*lastredirectiondevice*/, "", tapicallstate_unknown );

						CompletionProc(dwRequestID, 0);	
						
						m_log.LogTrace(" simulating TapiCall 0x%x with htCall 0x%x, callId=%d to NOTREADY device %s", pTapiCall, htCall, callId, (LPCSTR) subjectDeviceId.c_str());

						fnCallback(hTapiLine, htCall, LINE_CALLSTATE, LINECALLSTATE_DIALTONE, 0, LINEMEDIAMODE_INTERACTIVEVOICE);
						fnCallback(hTapiLine, htCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_NORMAL, LINEMEDIAMODE_INTERACTIVEVOICE);
						ReleaseMapLock();	// release lock asap
						m_cstaProvider->DropCallId(callId);
						return dwRequestID;
					}
					else {	// line has been closed !!!
						ReleaseMapLock();	// release lock
						m_log.LogTrace(" line has been closed! (LINEERR_INVALLINEHANDLE)");
						return LINEERR_INVALLINEHANDLE ;
					}
				}
				else {	
					/// \todo: more detailed error handling on FEATURE_NOTREADY
					return LINEERR_OPERATIONUNAVAIL;
				}
			}
			else if (_stricmp(sFeature.c_str(),FEATURE_DUMP)==0) {
				m_log.LogTrace("  FEATURE_DUMP");

				DumpMaps();

				// insert a dummy tapiCall in the map 
				callId = m_cstaProvider->GetNewCallId();
				WaitForMapLock();
				pTapiLine = GetTapiLine((MonitorXref)hdLine);	// line could have been deleted from map while it was unloked so get it again
				if (pTapiLine != NULL) {
					TapiCall *pTapiCall = new TapiCall();
					*phdCall = (HDRVCALL) pTapiCall ;	// init driver's call handler 
					pTapiCall->m_htapiCall = htCall;
					pTapiCall->m_callId = callId ;
					AddTapiCall ( pTapiCall, pTapiLine, htCall, callId, pTapiLine->m_deviceID /*address*/, pTapiLine->m_deviceID /*callerId*/, "" /*calleridname*/, calleddevice /*calleddevice*/, "" /*lastredirectiondevice*/, "", tapicallstate_unknown );

					CompletionProc(dwRequestID, 0);	
					
					m_log.LogTrace(" simulating TapiCall 0x%x with htCall 0x%x, callId=%d to DUMP device %s", pTapiCall, htCall, callId, (LPCSTR) subjectDeviceId.c_str());

					fnCallback(hTapiLine, htCall, LINE_CALLSTATE, LINECALLSTATE_DIALTONE, 0, LINEMEDIAMODE_INTERACTIVEVOICE);
					fnCallback(hTapiLine, htCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_NORMAL, LINEMEDIAMODE_INTERACTIVEVOICE);
					ReleaseMapLock();	// release lock asap
					m_cstaProvider->DropCallId(callId);
					return dwRequestID;
				}
				else {	// line has been closed !!!
					ReleaseMapLock();	// release lock
					m_log.LogTrace(" line has been closed! (LINEERR_INVALLINEHANDLE)");
					return LINEERR_INVALLINEHANDLE ;
				}
			}
		}

			// filter characters in pszDestAddress/calledDevice
		// http://windowssdk.msdn.microsoft.com/en-us/library/ms726017.aspx
		// we only understand 0-9, A-D, *, #
		std::string newCalleddevice;
		for (size_t i=0; i<calleddevice.size(); i++) {
			switch(calleddevice[i]) {
				case '0':
				case '1':
				case '2':
				case '3':
				case '4':
				case '5':
				case '6':
				case '7':
				case '8':
				case '9':
				case '*':
				case '#':
				case 'A':
				case 'B':
				case 'C':
				case 'D':
					newCalleddevice+=calleddevice[i];
					break;
				case '\'':	// escape character, add the remaining of the calleddevice string
					newCalleddevice.append(calleddevice, i+1, calleddevice.size()-i);
					i=calleddevice.size();	// force continue
					break;
				default:
					break;
			}
		}

		m_log.LogTrace("  filtered calledDevice %s => %s", calleddevice.c_str(), newCalleddevice.c_str());
		calleddevice = newCalleddevice;
		
		/*	To know if a call is predictive or not, we have to see some flags in LPLINECALLPARAMS const pCallParams
		*	\see http://msdn2.microsoft.com/en-us/library/ms735534.aspx
		*	IMPORTANT FLAGS:
		*	- dwCallParamFlags: Collection of Boolean call-setup parameters. 
		*	\see http://msdn2.microsoft.com/en-us/library/ms735533.aspx
		*	- dwNoAnswerTimeout: Number of seconds, after the completion of dialing, that the call should be 
		*	allowed to wait in the PROCEEDING or RINGBACK states, before it is automatically abandoned by the 
		*	service provider with a LINECALLSTATE_DISCONNECTED and LINEDISCONNECTMODE_NOANSWER. A value of 0 
		*	indicates that the application does not desire automatic call abandonment.
		*
		*	FIX: outlook2000 don't initialited pCallParams
		*/
		AC_RESULT rc;
		std::string userData;
		if (pCallParams){
			m_log.LogTrace(" Reading NOANSWERTIMEOUT %d",pCallParams->dwNoAnswerTimeout);
		
			// pCallParams contains UserData from dwUserUserInfoOffset to dwUserUserInfoSize positions
			if (pCallParams->dwUserUserInfoSize != 0){
				//
				char	userDataChr[4*1024] ;
				if (pCallParams->dwUserUserInfoSize +1 < (4*1024)) {
					CopyBuffer (userDataChr,(LPSTR)pCallParams+pCallParams->dwUserUserInfoOffset, pCallParams->dwUserUserInfoSize);
					userDataChr[pCallParams->dwUserUserInfoSize] = 0; // string end
					userData = ToString(userDataChr);
				}

				m_log.LogTrace(" UserData found = %s with length = %d",userData.c_str(), pCallParams->dwUserUserInfoSize);
			}else{
				m_log.LogTrace(" UserData not found");
			}
		} // if (pCallParams)

		/**
		*	FIX: Dead Lock when events and make predictive call runs concurrently 	
		*/

		callId = m_cstaProvider->GetNewCallId();

		WaitForMapLock();
		/**
		*	FIX: Race condition: one TapiLine haven't do more than one calls at the same time.
		*	To solve it, we create a new TapiCall and get the new callId previously to make call or makepredictive
		*	call. 
		*/
		TapiCall	*pTapiCall = GetTapiCall(htCall);

		if (!pTapiCall){
			pTapiCall = new TapiCall();
			m_log.LogTrace(" creating TapiCall 0x%x with htCall 0x%x", pTapiCall, htCall);
			pTapiCall->m_htapiCall = htCall;
			pTapiCall->m_callId = callId;
			AddTapiCall(pTapiCall, pTapiLine, htCall, callId, pTapiLine->m_deviceID, pTapiLine->m_deviceID, "", calleddevice, "",userData, tapicallstate_unknown );
		} else {
			pTapiCall->m_htapiCall = htCall;
			pTapiCall->m_callId = callId;
		}
		

		if (pCallParams && (pCallParams->dwCallParamFlags & LINECALLPARAMFLAGS_PREDICTIVEDIAL)){
			pTapiCall->m_isPredictive = true;
			ReleaseMapLock();

			m_log.LogTrace(" Executing MakePredictiveCall");
			rc = m_cstaProvider->MakePredictiveCall( sDeviceID,
													registryGetOutgoingContext("from-internal"), 
													calleddevice,
													registryGetOutgoingContext("from-internal"),
													callId,
													calleddevice,
													userData,
													pCallParams->dwNoAnswerTimeout,
													registryGetOriginateVariable(""),
													true);
		} 
		else 
		{
			ReleaseMapLock();
			m_log.LogTrace(" Executing MakeCall");
			rc  = m_cstaProvider->MakeCall(	sDeviceID, 
											registryGetOutgoingContext("from-internal"),
											calleddevice, 
											callId, 
											registryGetOutgoingContext("from-internal"), 
											sCallerIdName, 
											userData, 
											registryGetOriginateVariable(""),
											true);
		}
		if (rc == ok ) {
			// lock tsp call map again
			m_log.LogTrace(" makecall or makepredictivecall success");
			WaitForMapLock();

			/*
			* MakeCall() or MakePredictiveCall() originates a new call and reports new events.
			* - it is possible that an event arrives before MakeCall returns
			*	in that case pTapiCall will already be updated in the map
			* - otherwise we insert it because we need to return pTapiCall as hdrvcall
			* - pTapiLine->m_originatingHTapiCall should be cleared when event arrives and
			*	is updated in the callmap
			*/

			pTapiCall = GetTapiCall(htCall);
			*phdCall = (HDRVCALL) pTapiCall ;
			DWORD	dwCallId = pTapiCall->m_callId ;
			ReleaseMapLock();		// release TSP map
			
			CompletionProc(dwRequestID, 0);	// tbd: more error handling
			m_cstaProvider->UnLockReader();
			m_log.LogTrace0("CompletionProc Done and the reader was unlocked");
			
			m_log.LogTrace("  NEW callId=%d for htapicall=0x%x and hdcall=0x%x", dwCallId, htCall, phdCall);
			m_log.LogTrace("  returning dwRequestId %d", dwRequestID);
			
		}
		else if (rc == ok_feature_dialed ) {
			// make call succeded but no real call was initiated.
			WaitForMapLock();
			pTapiCall = GetTapiCall(htCall);
			RemoveTapiCall(pTapiCall);
			delete pTapiCall;
			ReleaseMapLock();

			*phdCall = (HDRVCALL) NULL ;

			CompletionProc(dwRequestID, 0);
			m_cstaProvider->UnLockReader();
			m_log.LogTrace0("CompletionProc Done and the reader was unlocked");
			
			m_log.LogTrace("  make call succeded but no real call was initiated");
			
			fnCallback(hTapiLine, htCall, LINE_CALLSTATE, LINECALLSTATE_DIALTONE, 0, LINEMEDIAMODE_INTERACTIVEVOICE);
			fnCallback(hTapiLine, htCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_NORMAL, LINEMEDIAMODE_INTERACTIVEVOICE);
		}
		else { //makecall failed

			//FIX: report the failed to TAPI now LINE_CLOSE is allowed 
			CompletionProc(dwRequestID, LINEERR_OPERATIONUNAVAIL);
			if(rc != time_out)
			{
				m_cstaProvider->UnLockReader();
				m_log.LogTrace0("CompletionProc Done and the reader was unlocked");
			}
			m_log.LogWarn("  MakeCall failed, returning LINEERR_OPERATIONUNAVAIL" );

			WaitForMapLock();
			pTapiCall = GetTapiCall(htCall);
			RemoveTapiCall(pTapiCall);
			delete pTapiCall;
			ReleaseMapLock();		// release TSP map before MakeCall
			
			/***
			***WARN: if we have done a CompletionProc reporting the error(negative number), 
			***we should return the dwRequestID, it can cause a TapiSrv AV
			***/
			return dwRequestID ;
		}
	} else { // pTapiLine does not exist
		ReleaseMapLock();		// release TSP map

		m_log.LogWarn("TSP", "  did not found a tapiline for xref=hdLine=" + ToString(hdLine) );
		return LINEERR_INVALLINEHANDLE;
	}

	return dwRequestID;
}


/**
 * Function TSPI_lineOpen
 * Connects to asterisk (it may alredy be connected so a fails_already_connected is treated as ok)
 * and a new monitor is opened on the deviceID.
 * Use pointer to TapiLine as opaque HDRVLINE.
 *	\see http://msdn2.microsoft.com/en-us/library/ms725583.aspx
 *	\param dwDeviceID
 *	\param htLine
 *	\param phdLine
 *	\param dwTSPIVersion
 *	\param pfnEventProc
 */
LONG ASTTapiProvider::TSPI_lineOpen(
    DWORD               dwDeviceID,
    HTAPILINE           htLine,
    LPHDRVLINE          phdLine,
    DWORD               dwTSPIVersion,
    LINEEVENT           pfnEventProc
    )
{
	m_log.LogTrace("TSPI_lineOpen() dwDeviceID=0x%x htLine=0x%x", dwDeviceID, htLine );

	std::string	sHost, sUser, sPassword;
	DWORD		dwDeviceIdBase = GetLineDeviceIDBase();
	//std::string	sLineDN = registryGetDn(dwDeviceID-dwDeviceIdBase);
	std::string		sLineDN = m_vOutDnVector.at(dwDeviceID - dwDeviceIdBase);
	std::string	sLineCallerIdName = registryGetCallerIdName(dwDeviceID-dwDeviceIdBase);
	DWORD	nPort,	nAuthType;
	AUTH_TYPE eAuthType;

	sHost = registryGetHostIp("192.168.0.67");
	nPort = registryGetHostPort(5038);
	sUser = registryGetLoginUser("admin");
	sPassword = registryGetLoginPassword("amp111");
	nAuthType = registryGetLoginAuthType(at_md5);

	if(nAuthType==0)
		eAuthType=at_plaintext;
	else
		eAuthType=at_md5;

	//open lock
	WaitForLineOpenCloseLock();

	AC_RESULT	rc = Connect(sHost, nPort, sUser, sPassword, eAuthType);
	if ( (rc!=ok) && (rc!=fails_already_connected)) {
		//open unlock
		ReleaseLineOpenCloseLock();
		m_log.LogTrace("  LINEERR_NOTREGISTERED");
		return LINEERR_NOTREGISTERED;
	}

	MonitorXref	xref = OpenMonitor(sLineDN);		

	TapiLine	*pxTapiLine = new TapiLine() ;
	pxTapiLine->m_xref = xref ;
	pxTapiLine->m_dwDeviceID = dwDeviceID ;
	size_t extPos = -1;
	if ((extPos = sLineDN.find("/")) != -1){
		pxTapiLine->m_callerId = sLineDN.substr(extPos+1, sLineDN.size());
	} else {
		pxTapiLine->m_callerId = dwDeviceID;
	}
	pxTapiLine->m_deviceID = sLineDN;	// ex: "201" or "SIP/201"
	pxTapiLine->m_callerIdName = sLineCallerIdName ;
	pxTapiLine->m_htapiLine = htLine ;
	pxTapiLine->m_fnEventProc = pfnEventProc ;

	// add to map
	AddTapiLine(xref, htLine, pxTapiLine);

	//open unlock
	ReleaseLineOpenCloseLock();

	*phdLine = (HDRVLINE) xref ;	// we use xref as driver handler

	return 0;
}


/**
*	Function TSPI_lineDrop drops a call
*	It returns success if related AsteriskCall can be droped or if it had already been dropped.
*	Must IDLE the TapiCall.
*	FIX: outlook 2000 call this function when the call is finished. Asterisk report an error (no such channel).
*	\see http://msdn2.microsoft.com/en-us/library/ms725543.aspx
*	\param dwRequestID
*	\param hdCall
*	\param lpsUserInfo
*	\param dwSize
*/
LONG ASTTapiProvider::TSPI_lineDrop(
    DRV_REQUESTID       dwRequestID,
    HDRVCALL            hdCall,
    LPCSTR              lpsUserInfo,
    DWORD               dwSize
    )
{
	m_log.LogTrace("TSPI_lineDrop() hdCall=0x%x", hdCall);

	if (hdCall == 0){
		m_log.LogTrace(" not exist any Call with hdCall = 0 (FEATURE EVENT)");
		return 0;
	}
	// locks call map access 
	WaitForMapLock();

	TapiCall	*pTapiCall = GetTapiCall(hdCall);

	m_log.LogTrace("  got TapiCall=0x%x", pTapiCall);
	if (!pTapiCall) {
		ReleaseMapLock();
		m_log.LogError("  no TapiCall is active for tapihandler");
		return LINEERR_INVALCALLHANDLE;
	}

	//mostrar trazas del mapa de handlers
	DumpMaps();

	TapiLine	*pTapiLine = pTapiCall->m_pTapiLine;
	HTAPILINE	htapiLine = pTapiCall->m_pTapiLine->m_htapiLine;
	HTAPICALL	htapiCall = pTapiCall->m_htapiCall;
	DWORD		dwCallId = pTapiCall->m_callId;
	std::string address = pTapiCall->m_address;
	LINEEVENT	fnCallback = 0;

	if (pTapiLine) {
		htapiLine = pTapiLine->m_htapiLine;
		fnCallback = pTapiLine->m_fnEventProc;
	}

	m_log.LogTrace("  got callID=%d", dwCallId);
	if (!dwCallId) {
		ReleaseMapLock();
		m_log.LogTrace("  no AsteriskCall callId is known for tapihandler but handlercall removed anyway!");
		CompletionProc(dwRequestID, 0);	// tbd more error handling
		return dwRequestID;	// this signals OK rc
	}
	ReleaseMapLock();

	std::string deviceID = AsteriskCall_GetAddressFromChannel(address);
	AC_RESULT rc = m_cstaProvider->ClearConnection(dwCallId, deviceID);
	if ( rc == ok) {	// connectionClearedEvent will idle the tapi call
		WaitForMapLock();
		pTapiCall = GetTapiCall(hdCall);
		if (!pTapiCall){
			ReleaseMapLock();
			m_log.LogError("  no TapiCall is active for tapihandler");
			return LINEERR_INVALCALLHANDLE;
		}

		CompletionProc(dwRequestID, 0);	

		// do not send IDLE if already idled 
		if (pTapiCall->m_state != tapicallstate_idle) {
			pTapiCall->m_state = tapicallstate_idle;
			m_log.LogTrace("  ClearConnection() returned ok. Sending LINECALLSTATE_IDLE  htapicall=0x%x", htapiCall);
			fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_NORMAL, LINEMEDIAMODE_INTERACTIVEVOICE);
		}
		else {
			m_log.LogTrace("  ClearConnection() return ok. Line is already IDLE, do not send IDLE again!");
		}

		ReleaseMapLock();
		return dwRequestID;	// this signals OK rc

	} else if (rc == invalid_callid) {
		m_log.LogWarn("  ClearConnection() returned invalid_callid to address %s. Try to clearcall.", address.c_str());
		rc = m_cstaProvider->ClearCall(dwCallId);
		if (rc == ok){
			WaitForMapLock();
			pTapiCall = GetTapiCall(hdCall);
			if (!pTapiCall){
				ReleaseMapLock();
				m_log.LogError("  no TapiCall is active for tapihandler");
				return LINEERR_INVALCALLHANDLE;
			}

			CompletionProc(dwRequestID, 0);

			// do not send IDLE if already idled already
			if (pTapiCall->m_state != tapicallstate_idle) {
				pTapiCall->m_state = tapicallstate_idle;
				m_log.LogTrace("  ClearConnection() returned ok. Sending LINECALLSTATE_IDLE  htapicall=0x%x", htapiCall);
				fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_NORMAL, LINEMEDIAMODE_INTERACTIVEVOICE);
			}
			else {
				m_log.LogTrace("  ClearConnection() return ok. Line is already IDLE, do not send IDLE again!");
			}

			ReleaseMapLock();
			return dwRequestID;	// this signals OK rc
		}

		//if ClearCall returns != ok
		WaitForMapLock();
		pTapiCall = GetTapiCall(hdCall);
		if (!pTapiCall){
			ReleaseMapLock();
			m_log.LogError("  no TapiCall is active for tapihandler");
			return LINEERR_INVALCALLHANDLE;
		}
		if (pTapiCall->m_state != tapicallstate_idle){
			pTapiCall->m_state = tapicallstate_idle;
			m_log.LogError("  ClearCall() returned invalid_callid. LINECALLSTATE_IDLE  htapicall=0x%x", htapiCall);
			CompletionProc(dwRequestID, LINEERR_INVALCALLHANDLE);
			fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_NORMAL, LINEMEDIAMODE_INTERACTIVEVOICE);
			ReleaseMapLock();	// release lock asap
			
			/***
			***WARN: if we have done a CompletionProc reporting the error(negative number), 
			***we should return the dwRequestID, it can cause a TapiSrv AV
			***/
			return dwRequestID;
		}
		ReleaseMapLock();	// release lock asap
		return LINEERR_INVALCALLHANDLE;
	} else if (rc == no_such_channel){

		CompletionProc(dwRequestID, 0);	

		WaitForMapLock();
		pTapiCall = GetTapiCall(hdCall);
		if (!pTapiCall){
			ReleaseMapLock();
			m_log.LogError("  no TapiCall is active for tapihandler");
			return LINEERR_INVALCALLHANDLE;
		}

		if (pTapiCall->m_state != tapicallstate_idle){
			pTapiCall->m_state = tapicallstate_idle;
			m_log.LogWarn("  ClearConnection() returned no such channel for this tapihandler but send idle anyway!");
			m_log.LogTrace("  LINECALLSTATE_IDLE  htapicall=0x%x", htapiCall);
			fnCallback(htapiLine, htapiCall, LINE_CALLSTATE, LINECALLSTATE_IDLE, LINEDISCONNECTMODE_NORMAL, LINEMEDIAMODE_INTERACTIVEVOICE);
		} else {
			m_log.LogTrace(" ClearConnection() returned no such channel for this tapihandler and it is already idled!");	
		}
		ReleaseMapLock();
		return dwRequestID;	// this signals OK rc	
	}

	m_log.LogTrace("  LINEERR_OPERATIONFAILED");
	return (LINEERR_OPERATIONFAILED);
}


/**
 * Function TSPI_lineClose
 * Called by TAPI when a line is no longer required.
 * We close the monitor opened on the device.
 * \see http://windowssdk.msdn.microsoft.com/en-us/library/ms725531.aspx
 *	\param hdLine
 */
LONG ASTTapiProvider::TSPI_lineClose(HDRVLINE hdLine)
{
	//HDRVLINE is a pointer to our asterisk manager object.
	//this should free any resources, such as the socket, 
	m_log.LogTrace("TSPI_lineClose() hdLine=0x%x", hdLine);

	//lineClose lock
	WaitForLineOpenCloseLock();

	// protect map access with mutex
	WaitForMapLock();
	TapiLine	*pTapiLine = RemoveTapiLine(hdLine);
	if (pTapiLine) {
		CloseMonitor(pTapiLine->m_xref);
		delete pTapiLine;
	}
	//Fix disconect don't need MapLock
	ReleaseMapLock();

	if(m_mapXref2TapiLine.size()==0) {
		m_log.LogInfo("Disconnect socked and clear callmaps because no line is open.");
		Disconnect();
	}

	//lineCloseunlock
	ReleaseLineOpenCloseLock();

	return 0;
}


/**
*	lineForward mainly used to simulate do-not-disturb=yes/no
*	if lpForwardList==NULL we treat as DND=OFF, DND=ON otherwise
*	\see http://msdn.microsoft.com/library/default.asp?url=/library/en-us/tapi/tspi/tspi_lineforward.asp?frame=true
*	\see http://msdn2.microsoft.com/en-us/library/ms725546.aspx
*	\param dwRequestID
*	\param hdLine
*	\param bAllAddresses
*	\param dwAddressID
*	\param lpForwardList
*	\param dwNumRingsNoAnswer
*	\param htConsultCall
*	\param lphdConsultCall
*	\param lpCallParams
*/
LONG ASTTapiProvider::TSPI_lineForward(
    DRV_REQUESTID       dwRequestID,
    HDRVLINE            hdLine,
    DWORD               bAllAddresses,
    DWORD               dwAddressID,
    LPLINEFORWARDLIST   const lpForwardList,
    DWORD               dwNumRingsNoAnswer,
    HTAPICALL           htConsultCall,
    LPHDRVCALL          lphdConsultCall,
    LPLINECALLPARAMS    const lpCallParams
    )
{
	m_log.LogTrace("TSPI_lineForward() hdLine=0x%x, htConsultCall=0x%x", hdLine, htConsultCall );

	bool	bDND_on = true;
	if (lpForwardList==NULL) {
		bDND_on = false;
	}

	// no consultation call will be initiated so...
	*lphdConsultCall = (HDRVCALL) NULL ;

	WaitForMapLock();		// lock tsp map 
	TapiLine	*pTapiLine = GetTapiLine((MonitorXref)hdLine);
	if (pTapiLine) {
		std::string	deviceId = pTapiLine->m_deviceID;
		ReleaseMapLock();		// release TSP map

		AC_RESULT rc  = SetDoNotDisturb(deviceId, bDND_on, registryGetAgentContext("from-internal"));
		if (rc == ok ) {
			m_log.LogTrace("  ok!");
			CompletionProc(dwRequestID, 0);	// tbd: more error handling
			return dwRequestID;
		}
		else {
			m_log.LogError("  failed!");
		}
	}	// pTapiLine
	else {
		ReleaseMapLock();		// release TSP map
		m_log.LogError("  did not found a tapiline for xref=hdLine=0x%x", hdLine);
		return LINEERR_INVALLINEHANDLE;
	}

	return (LINEERR_OPERATIONFAILED);
}


/**
 * Function TSPI_lineGetDevCaps
 * Allows TAPI to check our line capabilities before placing a call
 * \see http://msdn.microsoft.com/library/default.asp?url=/library/en-us/tapi/tspi/tspi_service_provider_functions.asp
 *	\see http://msdn2.microsoft.com/en-us/library/ms725568.aspx
 *	\param dwDeviceID
 *	\param dwTSPIVersion
 *	\param dwExtVersion
 *	\param pldc
 */
LONG ASTTapiProvider::TSPI_lineGetDevCaps(
    DWORD           dwDeviceID,
    DWORD           dwTSPIVersion,
    DWORD           dwExtVersion,
    LPLINEDEVCAPS   pldc
    )
{ 
	LONG            tr = 0;
    const wchar_t   szProviderInfo[] = L"CallBox TSP";
	wchar_t			szLineName[1024]; 
	DWORD			dwDeviceIdBase = GetLineDeviceIDBase();
	//std::string		sLineDN = registryGetDn(dwDeviceID - dwDeviceIdBase);
	std::string		sLineDN = m_vOutDnVector.at(dwDeviceID - dwDeviceIdBase);
	std::string		sLineName = registryGetDnNamePrefix("AST");
	sLineName += sLineDN ;
	int iSize = MultiByteToWideChar(CP_ACP, 0, (LPCSTR) sLineName.c_str(), -1, NULL, 0);
	MultiByteToWideChar(CP_ACP, 0, (LPCSTR) sLineName.c_str(), -1, szLineName, iSize);

	m_log.LogTrace("TSPI_lineGetDevCaps() dwDeviceIdBase=0x%x dwDeviceID=0x%x", dwDeviceIdBase, dwDeviceID );
	
    pldc->dwNeededSize = sizeof(LINEDEVCAPS) +
                         sizeof(szProviderInfo) +
                         iSize*sizeof(wchar_t);
    
    if( pldc->dwNeededSize <= pldc->dwTotalSize ) {
        pldc->dwUsedSize = pldc->dwNeededSize;

        pldc->dwProviderInfoSize    = sizeof(szProviderInfo);
        pldc->dwProviderInfoOffset  = sizeof(LINEDEVCAPS) + 0;
        wchar_t* pszProviderInfo = (wchar_t*)((BYTE*)pldc + pldc->dwProviderInfoOffset);
        wcscpy_s(pszProviderInfo,sizeof(szProviderInfo), szProviderInfo);
        
        pldc->dwLineNameSize        = sizeof(szLineName);
        pldc->dwLineNameOffset      = sizeof(LINEDEVCAPS) + sizeof(szProviderInfo);
        wchar_t* pszLineName = (wchar_t*)((BYTE*)pldc + pldc->dwLineNameOffset);
		/// fix: wcscpy_s second parameter makes an access violation. Now that is only the content lenght of szLineName
		wcscpy_s(pszLineName,iSize, szLineName);
    }
    else {
        pldc->dwUsedSize = sizeof(LINEDEVCAPS);
    }

    pldc->dwStringFormat      = STRINGFORMAT_ASCII;
    pldc->dwPermanentLineID   = MAKEPERMLINEID(GetPermanentProviderID(), dwDeviceID - GetLineDeviceIDBase());
    pldc->dwAddressModes      = LINEADDRESSMODE_ADDRESSID;
    pldc->dwNumAddresses      = 1;
    pldc->dwBearerModes       = LINEBEARERMODE_VOICE;
    pldc->dwMediaModes        = LINEMEDIAMODE_INTERACTIVEVOICE;
    pldc->dwGenerateDigitModes= LINEDIGITMODE_DTMF;
    pldc->dwDevCapFlags       = 0;	//LINEDEVCAPFLAGS_CLOSEDROP;
    pldc->dwMaxNumActiveCalls = 1;
    pldc->dwLineFeatures      = LINEFEATURE_MAKECALL;	// tbd add new caps

    // DialParams
    pldc->MinDialParams = g_dpMin;
    pldc->MaxDialParams = g_dpMax;
    pldc->DefaultDialParams = g_dpDef;
    
    return tr ;
}

/**
 * Removes TapiCall and deletes from map.
 * Does not hangup nor change Asterisk call status.
 *
 * Implementation note: TAPI specs say that 'The status information for the call is retained until the function 
 * TSPI_lineCloseCall is called' so in future revisions we won't delete TapiCall here.
 *
 * Anyway we have found that some buggy tapi apps do not call release handlers appropiately so
 * tapi never calls TSPI_lineCloseCall(), and a driver resource-leak arises.
 *
 * For now we delete our TapiCall objects in OnConnectionCleared() handler, and in a future revision we will 
 * implement some kind of 'dirty' cache to retain information while not suffering from buggy tapi apps.
 *	\see http://msdn2.microsoft.com/en-us/library/ms725532.aspx
 *	\param hdCall
 */
LONG ASTTapiProvider::TSPI_lineCloseCall(HDRVCALL hdCall)
{
	m_log.LogTrace("TSPI_lineCloseCall() hdCall=0x%x", hdCall);

	if (hdCall == 0){
		m_log.LogTrace(" call with hdCall = 0 (FEATURE EVENT)");	//// TBD canviar tarce!!!
		return 0;
	}

	// this call locks call map access 
	WaitForMapLock();
	TapiCall	*pTapiCall = GetTapiCall(hdCall);
	if (!pTapiCall) {
		m_log.LogError("  no pTapiCall in map");
		ReleaseMapLock();
		return LINEERR_RESOURCEUNAVAIL;
		//return 0;
	}

	RemoveTapiCall(pTapiCall);

	// apuntar a atributs y moure trace fora  CS 
	m_log.LogTrace("  deleting pTapiCall addrs=%s htapiCall=0x%x", (LPCSTR) pTapiCall->m_address.c_str(), (DWORD) pTapiCall->m_htapiCall);
	delete pTapiCall;

	ReleaseMapLock();
	return (0);
}


/**
 * Function TSPI_lineGetCallInfo
 * \see http://windowssdk.msdn.microsoft.com/en-us/library/ms725566.aspx
 *
 * CallerID, CalledID, ConnectedID, RedirectingID, RedirectionID
 * \see http://groups.google.es/group/microsoft.public.win32.programmer.tapi/browse_frm/thread/4f67342dbdf53d07/1d258a4ac426e38c?lnk=st&q=connectedID&rnum=11&hl=es#1d258a4ac426e38c
 *	\param hdCall
 *	\param lpCallInfo
 */
LONG ASTTapiProvider::TSPI_lineGetCallInfo(
    HDRVCALL            hdCall,
    LPLINECALLINFO      lpCallInfo
    )
{
	//m_log.LogTrace("TSPI_lineGetCallInfo hdCall=%x", hdCall);

	if (hdCall == 0){
		m_log.LogTrace("TSPI_lineGetCallInfo: not exist any Call with hdCall = 0 (FEATURE EVENT)");
		return 0;
	}
 
	//These are the items that a TSP is required to fill out - other items we must preserve
	// - that is they are used by TAPI.

	lpCallInfo->dwNeededSize = sizeof (LINECALLINFO);
	lpCallInfo->dwUsedSize = sizeof (LINECALLINFO);
	lpCallInfo->dwLineDeviceID = 0;  
	lpCallInfo->dwAddressID = 0;
	lpCallInfo->dwBearerMode = LINEBEARERMODE_SPEECH;
	//lpCallInfo->dwRate;
	// set right mediamode
	// http://groups.google.es/group/microsoft.public.win32.programmer.tapi/browse_frm/thread/f4a636a41cd5f8e4/9991c3220e6dab7b?lnk=st&q=TSPI_lineGetCallInfo+TSPI_lineCloseCall&rnum=3&hl=es#9991c3220e6dab7b
	lpCallInfo->dwMediaMode = LINEMEDIAMODE_INTERACTIVEVOICE;	

	// dwCallStates: Value that specifies the call states of type LINECALLSTATE_ for which the application can be 
	// notified on this call. The dwCallStates member is constant in LINECALLINFO and does not change depending on 
	// the call state
	// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/wcecomm5/html/wce50lrfLINECALLINFO.asp
    lpCallInfo->dwCallStates = LINECALLSTATE_IDLE | 
								LINECALLSTATE_OFFERING | 
								LINECALLSTATE_ACCEPTED | 
								LINECALLSTATE_RINGBACK | 
								LINECALLSTATE_DIALING | 
								LINECALLSTATE_DIALTONE | 
								LINECALLSTATE_PROCEEDING |
								LINECALLSTATE_DISCONNECTED |
								LINECALLSTATE_CONNECTED |
								LINECALLSTATE_UNKNOWN;

	lpCallInfo->dwOrigin = LINECALLORIGIN_UNKNOWN;	
	lpCallInfo->dwReason = LINECALLREASON_UNKNOWN;
	lpCallInfo->dwCompletionID = 0;
	lpCallInfo->dwCountryCode = 0;   // 0 = unknown
	lpCallInfo->dwTrunk = 0xFFFFFFFF; // 0xFFFFFFFF = unknown

	// For the next section we need to find our call relating to this.
	lpCallInfo->dwCallerIDOffset = 0;
	lpCallInfo->dwCallerIDSize = 0;
	lpCallInfo->dwCallerIDFlags = 0;
	lpCallInfo->dwCallerIDNameOffset = 0;
	lpCallInfo->dwCallerIDNameSize = 0;

	lpCallInfo->dwCalledIDOffset = 0;
	lpCallInfo->dwCalledIDSize = 0;
	lpCallInfo->dwCalledIDFlags = 0;
	lpCallInfo->dwCalledIDNameOffset = 0;
	lpCallInfo->dwCalledIDNameSize = 0;

	lpCallInfo->dwConnectedIDSize = 0;
	lpCallInfo->dwConnectedIDOffset = 0;
	lpCallInfo->dwConnectedIDFlags = 0;
	lpCallInfo->dwConnectedIDNameSize = 0;
	lpCallInfo->dwConnectedIDNameOffset = 0;

	lpCallInfo->dwRedirectingIDSize = 0;
	lpCallInfo->dwRedirectingIDOffset =0;
	lpCallInfo->dwRedirectingIDFlags=0;
	lpCallInfo->dwRedirectingIDNameSize =0;
	lpCallInfo->dwRedirectingIDNameOffset =0;
	
	lpCallInfo->dwRedirectionIDSize =0;
	lpCallInfo->dwRedirectionIDOffset =0;
	lpCallInfo->dwRedirectionIDFlags =0;
	lpCallInfo->dwRedirectionIDNameSize =0;
	lpCallInfo->dwRedirectionIDNameOffset =0;

	lpCallInfo->dwUserUserInfoSize =0;
	lpCallInfo->dwUserUserInfoOffset=0;

	lpCallInfo->dwDisplaySize =0;
	lpCallInfo->dwHighLevelCompSize=0;
	lpCallInfo->dwLowLevelCompSize=0;
	lpCallInfo->dwChargingInfoSize=0;
	lpCallInfo->dwTerminalModesSize=0;
	lpCallInfo->dwDevSpecificSize=0;

	// this method locks call map access 
	WaitForMapLock();
	TapiCall	*pTapiCall = GetTapiCall(hdCall);
	if (pTapiCall) {	// found call
		// fill in call info
		lpCallInfo->dwOrigin = LINECALLORIGIN_UNKNOWN ; // LINECALLORIGIN_INBOUND / LINECALLORIGIN_OUTBOUND
		lpCallInfo->dwCallID =			pTapiCall->m_callId;
		std::string callerID =			pTapiCall->m_callerId;
		std::string callerIDName =		pTapiCall->m_callerIdName ;
		std::string calledID =			pTapiCall->m_calledDevice;
		std::string calledIDName =		pTapiCall->m_calledDevice;
		std::string connectedID =		"";
		std::string connectedIDName =	"";
		std::string redirectingID =		pTapiCall->m_redirectingDevice;	
		std::string redirectingIDName =	pTapiCall->m_redirectingDevice;	
		std::string redirectionID =		"";		// used as alerting/established device by some apps 
		std::string userData =			pTapiCall->m_userData;
		bool		isPredictive =		pTapiCall->m_isPredictive;
		DWORD address =					atoi(pTapiCall->m_address.c_str());

		//[ 1548889 ] CID on incomming calls from zap trunk gives unknown callerID
		if (AsteriskCall_IsUnknown(callerIDName)) {	
			m_log.LogTrace("TSPI_lineGetCallInfo: changing callerIDName=%s by callerId=%s because is unknown", callerIDName.c_str(), callerID.c_str());
			callerIDName = callerID;
		}

		switch (pTapiCall->m_eLineCallOrigin) {
			case linecallorigin_inbound:
//				m_log.LogTrace("    LINECALLORIGIN_INBOUND");
				lpCallInfo->dwOrigin = LINECALLORIGIN_INBOUND ;
				connectedID =		pTapiCall->m_callerId;
				connectedIDName =	pTapiCall->m_callerIdName;	
				redirectionID =		pTapiCall->m_address;		// used as alerting/established device by some apps 
				if (pTapiCall->m_redirectionId != ""){
					redirectionID = pTapiCall->m_redirectionId;
				}
				break;
			case linecallorigin_outbound:
//				m_log.LogTrace("    LINECALLORIGIN_OUTBOUND");
				lpCallInfo->dwOrigin = LINECALLORIGIN_OUTBOUND ;
				connectedID =		pTapiCall->m_calledDevice;
				connectedIDName =	pTapiCall->m_calledDevice;	
				redirectionID =		pTapiCall->m_calledDevice;	// used as alerting/established device by some apps 
				if (pTapiCall->m_redirectionId != ""){
					redirectionID = pTapiCall->m_redirectionId;
				}
				break;
			default:
//				m_log.LogTrace("    LINECALLORIGIN_UNKNOWN");
				break;

		}
		
		m_log.LogTrace("TSPI_lineGetCallInfo hdCall=0x%x callerId=%s, callerIdName=%s, calledId=%s, calledIdName=%s, "
			"connectedId=%s, connectedIdName=%s, redirectingId=%s, redirectingIdName=%s, redirectionId=%s, "
			"userData=%s, isPredictive=%s, dwOrigin=0x%x", hdCall, callerID.c_str(), callerIDName.c_str(),calledID.c_str(), calledIDName.c_str(), 
			connectedID.c_str(), connectedIDName.c_str(), redirectingID.c_str(), redirectingIDName.c_str(), 
			redirectionID.c_str(), userData.c_str(),ToString(isPredictive).c_str(), lpCallInfo->dwOrigin );

//		m_log.LogTrace("  address=%d", address);
		if (address > 0){
			lpCallInfo->dwLineDeviceID = address;
		}

//		m_log.LogTrace("  callerid=%s",callerID.c_str());
		if ( callerID != "" ) {
			// http://windowssdk.msdn.microsoft.com/en-us/library/ms735536.aspx
			lpCallInfo->dwCallerIDFlags |= LINECALLPARTYID_ADDRESS;	
			AddDataBlock (lpCallInfo, lpCallInfo->dwCallerIDOffset, lpCallInfo->dwCallerIDSize,
				(LPCTSTR)callerID.c_str());
		}

//		m_log.LogTrace("  callerIDName=%s",callerIDName.c_str());
		if ( callerIDName != "" ) {
			lpCallInfo->dwCallerIDFlags |= LINECALLPARTYID_NAME;
			AddDataBlock (lpCallInfo, lpCallInfo->dwCallerIDNameOffset, lpCallInfo->dwCallerIDNameSize,
				(LPCTSTR)callerIDName.c_str());
		}

//		m_log.LogTrace("  calledID=%s", calledID.c_str());
		if ( calledID != "" ) {
			lpCallInfo->dwCalledIDFlags |= LINECALLPARTYID_ADDRESS;
			AddDataBlock (	lpCallInfo, lpCallInfo->dwCalledIDOffset, lpCallInfo->dwCalledIDSize, 
				(LPCTSTR)calledID.c_str());
		}

//		m_log.LogTrace("  calledIDName=%s", calledIDName.c_str());
		if ( calledIDName != "" ) {
			lpCallInfo->dwCalledIDFlags |= LINECALLPARTYID_NAME;
			AddDataBlock (	lpCallInfo, lpCallInfo->dwCalledIDNameOffset, lpCallInfo->dwCalledIDNameSize, 
				(LPCTSTR)calledIDName.c_str());
		}
	
//		m_log.LogTrace("  connectedID=%s", connectedID.c_str());
		if ( connectedID != "" ) {
			lpCallInfo->dwConnectedIDFlags |= LINECALLPARTYID_ADDRESS;
			AddDataBlock (	lpCallInfo, lpCallInfo->dwConnectedIDOffset, lpCallInfo->dwConnectedIDSize, 
				(LPCTSTR)connectedID.c_str());
		}

//		m_log.LogTrace("  connectedIDName=%s", connectedIDName.c_str());
		if ( connectedIDName != "" ) {
			lpCallInfo->dwConnectedIDFlags |= LINECALLPARTYID_NAME;
			AddDataBlock (	lpCallInfo, lpCallInfo->dwConnectedIDNameOffset, lpCallInfo->dwConnectedIDNameSize, 
				(LPCTSTR)connectedIDName.c_str());
		}

//		m_log.LogTrace("  redirectionID=%s", redirectionID.c_str());	//used as alerting/established device by some apps
		if ( redirectionID != "" ) {
			lpCallInfo->dwRedirectionIDFlags |= LINECALLPARTYID_ADDRESS;
			AddDataBlock (	lpCallInfo, lpCallInfo->dwRedirectionIDOffset, lpCallInfo->dwRedirectionIDSize, 
				(LPCTSTR)redirectionID.c_str());
		}

//		m_log.LogTrace("  redirectingID=%s", redirectingID.c_str());
		if ( redirectingID != "" ) {
			lpCallInfo->dwRedirectingIDFlags |= LINECALLPARTYID_ADDRESS;
			AddDataBlock (	lpCallInfo, lpCallInfo->dwRedirectingIDOffset, lpCallInfo->dwRedirectingIDSize, 
				(LPCTSTR)redirectingID.c_str());
		}

		if (pTapiCall->m_isPredictive){
//			m_log.LogTrace("  isPredictiveCall");
			lpCallInfo->dwCallParamFlags |= LINECALLPARAMFLAGS_PREDICTIVEDIAL;
		}

//		m_log.LogTrace("  userData=%s", userData.c_str());
		if ( userData != ""){
			// UserData should be sent as a ASCII array
			/* AddDataBlock */
			AddDataBlock ( lpCallInfo, lpCallInfo->dwUserUserInfoOffset, lpCallInfo->dwUserUserInfoSize,  userData.c_str(), userData.size() );
		}
	
		ReleaseMapLock();
	}else{
		m_log.LogError("  pTapiCall NOT FOUND");
		ReleaseMapLock();
		return LINEERR_INVALCALLHANDLE;
	}

	return 0;
	
}

/**
 * Function TSPI_lineGetAddressCaps
 * Allows TAPI to check our line capabilities before placing a call
 *	\see http://msdn2.microsoft.com/en-us/library/ms725560.aspx
 *	\param dwDeviceID
 *	\param dwAddressID
 *	\param dwTSPIVersion
 *	\param dwExtVersion
 *	\param pac
 */
LONG ASTTapiProvider::TSPI_lineGetAddressCaps(
    DWORD   dwDeviceID,
    DWORD   dwAddressID,
    DWORD   dwTSPIVersion,
    DWORD   dwExtVersion,
    LPLINEADDRESSCAPS  pac)
{
	m_log.LogTrace("TSPI_lineGetAddressCaps() dwDeviceID=0x%x dwAddressID=0x%x", dwDeviceID, dwAddressID);

	pac->dwNeededSize = sizeof(LINEADDRESSCAPS);
    pac->dwUsedSize = sizeof(LINEADDRESSCAPS);
    
    pac->dwLineDeviceID = dwDeviceID;
    pac->dwAddressSharing = LINEADDRESSSHARING_PRIVATE;
    pac->dwCallInfoStates = LINECALLINFOSTATE_MEDIAMODE | LINECALLINFOSTATE_APPSPECIFIC;
    pac->dwCallerIDFlags = LINECALLPARTYID_ADDRESS | LINECALLPARTYID_UNKNOWN;
    pac->dwCalledIDFlags = LINECALLPARTYID_ADDRESS | LINECALLPARTYID_UNKNOWN;
    pac->dwRedirectionIDFlags = LINECALLPARTYID_ADDRESS | LINECALLPARTYID_UNKNOWN ;
    pac->dwCallStates = LINECALLSTATE_IDLE | 
						LINECALLSTATE_OFFERING | 
						LINECALLSTATE_ACCEPTED | 
						LINECALLSTATE_RINGBACK | 
						LINECALLSTATE_DIALING | 
						LINECALLSTATE_DIALTONE |
						LINECALLSTATE_PROCEEDING |
						LINECALLSTATE_DISCONNECTED |
						LINECALLSTATE_CONNECTED;
    
    pac->dwDialToneModes = LINEDIALTONEMODE_UNAVAIL;
    pac->dwBusyModes = LINEDIALTONEMODE_UNAVAIL;
    pac->dwSpecialInfo = LINESPECIALINFO_UNAVAIL;
    
    pac->dwDisconnectModes = LINEDISCONNECTMODE_UNAVAIL;
    
    /* TBD */
    pac->dwMaxNumActiveCalls = 1;
    
    pac->dwAddrCapFlags = LINEADDRCAPFLAGS_DIALED | LINEADDRCAPFLAGS_ACCEPTTOALERT;
    pac->dwCallFeatures = LINECALLFEATURE_DIAL | LINECALLFEATURE_DROP | LINECALLFEATURE_GENERATEDIGITS;
    pac->dwAddressFeatures = LINEADDRFEATURE_MAKECALL | LINEADDRFEATURE_PICKUP;

    return 0;
}

/**
 * Function TSPI_providerShutdown
 * Disconnects from asterisk. Shutdown and clean up.
 *	\param dwTSPIVersion
 *	\param dwPermanentProviderID
 */
LONG ASTTapiProvider::TSPI_providerShutdown(
    DWORD               dwTSPIVersion,
    DWORD               dwPermanentProviderID                   // TSPI v2.0
    )
{
	m_log.LogTrace("TSPI_providerShutdown()");
	Disconnect();
    
	return 0;
}


/**
 * dump call maps
 */
void	ASTTapiProvider::DumpMaps()
{
	// protect map access with mutex
	WaitForMapLock();

	// clear m_id2call map
	m_log.LogTrace0("================================================================= start callmap dump...");
	m_log.LogTrace0("DUMP OF TSP CALLS MAP: [HDRVCALL] => call info:");
	for (Map_TapiCall::iterator i = m_mapTapiCall.begin(); i!=m_mapTapiCall.end(); i++) {
		TapiCall	*pTapiCall = i->second;
		if (pTapiCall) 
			m_log.LogTrace0("  [0x%x] %s", pTapiCall, pTapiCall->DumpToString().c_str());
	}

	m_log.LogTrace0("DUMP OF CALL-ID MAP: {CALL_ID} => hdrivercall");
	// for each device close all its monitors
	for (Map_CallId2HdrvCallSet::iterator i = m_mapCallId2HdrvCallSet.begin(); i!=m_mapCallId2HdrvCallSet.end(); i++) {
		DWORD callId = i->first;
		Set_hdrvCall *pset_hdrivercall  = i->second;
		for (Set_hdrvCall::iterator j = pset_hdrivercall->begin(); j!=pset_hdrivercall->end(); j++) {
			HDRVCALL hdrivercall = *j;
			TapiCall	*pTapiCall = GetTapiCall(hdrivercall);
			if (pTapiCall) {
				m_log.LogTrace0("      [%d] => %s", pTapiCall->m_callId, pTapiCall->DumpToString().c_str());
			}
			else {
				m_log.LogTrace0("      [%d] => NULL", hdrivercall);
			}
		}
	}

	m_log.LogTrace0("================================================================= end callmap dump...");

	ReleaseMapLock();
}

/**
*	Get expresion dn from registy, process it and store dns in vector
*	\param pout_dn_vector Vector with the extension numbers.
*	\param logger BasicLogger to trace
*	\return nlines Number of lines = pout_dn_vector.size()
*/
DWORD ASTTapiProvider::ProcessDnExp(std::vector<std::string> &out_dn_vector, BasicLogger &logger)
{

	if(out_dn_vector.size()>0) return ((DWORD)out_dn_vector.size()); //Proces is done only return number of dns

    #define NVALIDCHARS 13
	#define VALIDCHARS {'0','1','2','3','4','5','6','7','8','9',' ',',','-'}
	
	char validChar[NVALIDCHARS] = VALIDCHARS;
	int i,j,v,w,lastInvalidExp;
	std::string subexp, exp, temp, start, end;
	size_t found, found2;
	char temp2[255];
	bool valid_char, valid_subexp;

	//make dn expresion from registry
	j=(int)registryGetNumLines(1);
	for(i=0;i<j;i++){
		subexp=registryGetDn(i);
	    //process subexpresion to validate if it have valid characters else put it in vector directly
		valid_subexp=true;
		for (v=0; v<(int)subexp.size() && valid_subexp ;v++) {
			valid_char=false;
			for(w=0;w<NVALIDCHARS;w++)	if(subexp[v] == validChar[w]) valid_char=true;
			if(!valid_char) valid_subexp=false;
		}
		//subexp have valid chars, it will be added to exp and processed
		if(valid_subexp) {
			exp+=subexp;
			exp+=",";	
		}
		else { //else it will be not processed and added directly to vector for compativility for non number digits
			logger.LogTrace("TSP ProcessDnExp"," Expresion:"+subexp+" HAVE NO VALID CHARACTER FOR BE A VALID EXPRESSION");
			logger.LogTrace("TSP ProcessDnExp"," Added DN = "+subexp);
			out_dn_vector.push_back(subexp);
		}
	}
	lastInvalidExp=(int)out_dn_vector.size();

	logger.LogTrace("TSP ProcessDnExp","Dn expression to filter: "+exp);
	
	//delete spaces (trim)
	found=exp.find(' ');
	while (found!=exp.npos){
		exp.erase(found,1);
		found=exp.find(' ');
	}

	//delete initial ','
	if (exp[0]==',') exp.erase(0,1);  

	//delete final ',' . Size should be bigger than 0 
	if (exp.size()>0 && exp[exp.size()-1]==',') exp.erase(exp.size()-1,1);  

	//avoid invalid sequence ",,", two or more
	found=exp.find(",,");
	while (found!=exp.npos){
		exp.erase(found,1);
		found=exp.find(",,");
	}

	//avoid invalid sequence "--", two or more
    found=exp.find("--");
	while (found!=exp.npos){
		exp.erase(found,1);
		found=exp.find("--");
	}

    //erase '-' in "-,"
    found=exp.find("-,");
	while (found!=exp.npos){
		exp.erase(found,1);
		found=exp.find("-,");
	}

	//erase '-' in ",-" 
	found=exp.find(",-");
	while (found!=exp.npos){
		exp.erase(found+1,1);
		found=exp.find(",-");
	}

	logger.LogTrace("TSP ProcessDnExp","Valid expression: "+exp);

	//proces comma expresions 
	do{
		found=exp.find(',');
		//interval string
		temp=exp.substr(0,found);
		//search - for interval substr
		found2=temp.find('-');
		if (found2!=temp.npos){
			//process interval substr
			start=temp.substr(0,found2);
			end=temp.substr(found2+1);
			j=atoi(end.c_str());
			for(i=atoi(start.c_str()); i<=j; i++){
				sprintf_s(temp2,"%d",i);
				out_dn_vector.push_back(std::string(temp2));
			}
		}
		else{
			if(temp.size() != 0) {
				out_dn_vector.push_back(temp);
			}
		}
		
		//erase procesed substr
		exp.erase(0,found+1);

	} while (found!=exp.npos);

	//delete repeated and blank dn after proces all dn string
	for(i=lastInvalidExp; i<(int)out_dn_vector.size();i++){
		if (out_dn_vector[i]==""){
			out_dn_vector.erase(out_dn_vector.begin()+i);
			logger.LogTrace("TSP ProcessDnExp", "blank entry erased from list");
		}
		else{
			for(j=i; j<(int)out_dn_vector.size();j++) {
				if (i!=j && (out_dn_vector[i]==out_dn_vector[j]) ){
					out_dn_vector.erase(out_dn_vector.begin()+j);
					logger.LogTrace("TSP ProcessDnExp", "duplicate entry erased from list: "+out_dn_vector[i]);
					j--; //becuse actual element was erased
				}
			}
		}
	}

	//print results
	char num[100];
	sprintf_s(num , "Total DN number = %d", (int)out_dn_vector.size());
	logger.LogTrace("TSP ProcessDnExp", num );
	for (i=0; i<(int)out_dn_vector.size();i++) 
		logger.LogTrace("TSP ProcessDnExp"," DN = "+out_dn_vector.at(i));

	return ((DWORD)out_dn_vector.size());

}
