/*
	Name: asttapiprovider.h
	Copyright: Under the Mozilla Public License Version 1.1 or later
	Author: ACTIVA
	Date: MAY/06
*/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Activa asttapiprovider.
 *
 * The Initial Developer of the Original Code is ACTIVA.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): none
 *
 * ***** END LICENSE BLOCK ***** */

#pragma once

#ifndef TAPIMODEL_H
#define TAPIMODEL_H

#include <string>
#include <map>
#include <vector>
#include <set>
#include <tspi.h>
#include "../activaAPI/activaAPI.h"
#include "../astprovider/src/astlogger.h"

/// TSP level model of a tapi line
/**
 * TapiLine
 * uniqueID is monitor_xref == hdriverLine
 */
class TapiLine
{
public:
	MonitorXref	m_xref;				///< we use xref as HDRVLINE
	HTAPILINE	m_htapiLine;		///< tapi line handler 
	DWORD		m_dwDeviceID;		///< tapi dwDeviceID
	std::string	m_deviceID;			///< line dn
	std::string	m_callerIdName;		
	std::string m_callerId;			
	LINEEVENT	m_fnEventProc;

	TapiLine();
};

typedef	enum {	
			linecallorigin_unknown = 0,
			linecallorigin_inbound = 1,
			linecallorigin_outbound = 2
			} TapiLine_LineCallOrigin ;

typedef	enum {	
			tapicallstate_unknown = 0,
			tapicallstate_idle = 1,
			tapicallstate_initialized = 2,
			tapicallstate_alerting = 3,
			tapicallstate_connected = 4,
			tapicallstate_failed = 5,
			tapicallstate_queued = 6,
			tapicallstate_disconnected = 7,
			tapicallstate_held = 8
			} eTapiCall_state ;


//need to send disconnect events with cause AMD or FAX 
#define LINEDISCONNECTMODE_ACTIVATSP_ANSWERING_MACHINE			0x80000000
#define LINEDISCONNECTMODE_ACTIVATSP_MODEM_FAX					0x40000000

/**
 * TapiCall
 *	Links tapi call attributes to AsteriskCall.
 *  TSP will use opaque pointer to TapiCall as HDRCALL
 */
/// TSP level model of a call
class TapiCall 
{
public:
	TapiLine		*m_pTapiLine;			/// pointer to tapiLine
	HTAPICALL		m_htapiCall;			/// driver call
	DWORD			m_callId;				/// call id
	std::string		m_address;				/// subject device 
	std::string		m_callerId;				/// callerID
	std::string		m_callerIdName;			/// callerIDName
	std::string		m_calledDevice;			/// called device
	std::string		m_redirectingDevice;	/// last redirecting device
	std::string		m_userData;				/// userData
	bool			m_isPredictive;			/// identify if the call is predictive or not
	eTapiCall_state	m_state;				/// tapi call state
	DWORD			m_realCallId;			/// used to identify the callId in fastTransfer
	std::string		m_redirectionId;		/// alerting/connected/releasing device
	TapiLine_LineCallOrigin	m_eLineCallOrigin ;
	call_cause		m_cause;				/// Cause of disconnection

					TapiCall();
	void			Update(std::string callerId, std::string callerIdName, std::string calledDevice, std::string lastRedirectingDevice, eTapiCall_state eState);
	std::string		DumpToString();
};


// MAP TYPES
/**
*	map from monitors to TapiLines
*	In TAPI is only valid one monitor of each tapiLines
*/
typedef	std::map < MonitorXref, TapiLine* > Map_Xref2TapiLine;
typedef	std::map < HTAPILINE, MonitorXref > Map_HTapiLine2Xref;

// TSP uses TapiCall as driver's call representation
// we map asteriskcall => tapiCall
typedef	std::map < HDRVCALL, TapiCall* > Map_TapiCall;	// opaque TapiCall* => real TapiCall* 
typedef	std::map < HTAPICALL, HDRVCALL > Map_HtapiCall2HdrvCall;

typedef	std::set <HDRVCALL> Set_hdrvCall ;				//< set of HDRVCALL
typedef	std::map < DWORD, Set_hdrvCall* > Map_CallId2HdrvCallSet;


/**
 * ASTTapiProvider 
 * uses ASTCstaProvider to integrate TAPI to Asterisk pbx
 * \see http://msdn2.microsoft.com/en-us/library/ms736535.aspx
 */
/// tsp-level abstraction on top of ASTCstaProvider
class	ASTTapiProvider : public ASTCstaListener 
{
protected:
		ASTTapiProvider();
public:
	/// factory method
	static ASTTapiProvider* Create(	ASTLogListener *logCallBack,
									ASYNC_COMPLETION fnAsincCompletionProc,     
									DWORD dwPermanentProviderID,
                                    DWORD dwLineDeviceIDBase,
									std::vector<std::string> v_vOutDnVector);
		~ASTTapiProvider();

	/// shutsdown and disconnects from asterisk
	LONG	TSPI_providerShutdown(DWORD dwTSPIVersion, DWORD  dwPermanentProviderID);

	/// returns device caps
	LONG	TSPI_lineGetDevCaps(DWORD dwDeviceID, DWORD dwTSPIVersion, DWORD dwExtVersion, LPLINEDEVCAPS pldc);

	/// returns address caps
	LONG	TSPI_lineGetAddressCaps(DWORD dwDeviceID, DWORD dwAddressID, DWORD dwTSPIVersion, DWORD dwExtVersion, LPLINEADDRESSCAPS pac);

	/// opens a tapi line
	LONG	TSPI_lineOpen(DWORD dwDeviceID, HTAPILINE htLine, LPHDRVLINE phdLine, DWORD dwTSPIVersion, LINEEVENT pfnEventProc);

	///closes a tapi line
	LONG	TSPI_lineClose(HDRVLINE hdLine);

	///makes a call
	LONG	TSPI_lineMakeCall(DRV_REQUESTID dwRequestID, HDRVLINE hdLine, HTAPICALL htCall, LPHDRVCALL phdCall, LPCWSTR pszDestAddress, DWORD dwCountryCode, LPLINECALLPARAMS const pCallParams);

	///hangups a call
	LONG	TSPI_lineDrop(DRV_REQUESTID dwRequestID, HDRVCALL hdCall, LPCSTR lpsUserInfo, DWORD dwSize);

	///deallocates a call
	LONG	TSPI_lineCloseCall(HDRVCALL hdCall);

	///sets line worward options
	LONG	TSPI_lineForward(DRV_REQUESTID dwRequestID, HDRVLINE hdLine, DWORD bAllAddresses, DWORD dwAddressID, LPLINEFORWARDLIST const lpForwardList, DWORD dwNumRingsNoAnswer, HTAPICALL htConsultCall, LPHDRVCALL lphdConsultCall, LPLINECALLPARAMS const lpCallParams );

	/// returns call info to tapi
	LONG	TSPI_lineGetCallInfo(HDRVCALL hdCall, LPLINECALLINFO lpCallInfo);

	/// setup transfer
	LONG	TSPI_lineSetupTransfer(DRV_REQUESTID dwRequestID,HDRVCALL hdCall, HTAPICALL htConsultCall,LPHDRVCALL lphdConsultCall, LPLINECALLPARAMS const lpCallParams);

	/// Dial
	LONG	TSPI_lineDial(DRV_REQUESTID dwRequestID,HDRVCALL hdCall, LPCWSTR lpszDestAddress, DWORD dwCountryCode);

	/// hold call
	LONG	TSPI_lineHold(DRV_REQUESTID dwRequestID, HDRVCALL hdCall);

	/// retrieve call
	LONG	TSPI_lineUnHold(DRV_REQUESTID dwRequestID,HDRVCALL hdCall);

	/// get dns from registry expresions. STATIC!!!
	static DWORD ProcessDnExp(std::vector<std::string> &out_dn_vector, BasicLogger &logger);

private:
	/// connect to pbx
	AC_RESULT	Connect(std::string	sHost, int nPort, std::string user, std::string pwd, AUTH_TYPE eAuthType);

	/// disconnect from pbx
	bool	Disconnect();

	/// open a monitor on deviceId
	MonitorXref	OpenMonitor(std::string	deviceId);
	
	///close opened monitor
	bool	CloseMonitor(MonitorXref xref);

	/// set do-not-disturb flag
	AC_RESULT	SetDoNotDisturb(std::string deviceId, bool bDND_on, std::string dndContext);
	
	DWORD		GetPermanentProviderID() { return m_dwPermanentProviderID; };
	DWORD		GetLineDeviceIDBase() { return m_dwLineDeviceIDBase; };

	// access to TapiLine maps
	void		AddTapiLine(MonitorXref xref, HTAPILINE htapiLine, TapiLine *pLineInfo);
	TapiLine *	GetTapiLine(MonitorXref xref);	// hdriverLine=xref
	TapiLine *	GetTapiLine(HDRVLINE hdriverLine) { return GetTapiLine((MonitorXref)hdriverLine);};	
	TapiLine *	GetTapiLine(HTAPILINE htapiLine);	 

	// caller responsible of freeing TapiLine memory
	TapiLine *	RemoveTapiLine(HDRVLINE hdriverLine);

	// access to TapiCall maps
	void		AddTapiCall(TapiCall *pNewTapiCall, TapiLine *pTapiLine, HTAPICALL htapiCall, DWORD callId, std::string address, std::string callerId, std::string callerIdName, std::string calledDevice, std::string lastRedirectionDevice, std::string userData, eTapiCall_state state);
	void		SetTapiCall(TapiCall *pTapiCallToUpdate, DWORD callId, std::string address, std::string callerId, std::string callerIdName, std::string calledDevice, std::string lastRedirectionDevice);
	TapiCall *	GetTapiCall(TapiCall* pTapiCall);	//may return null if not in map
	TapiCall *	GetTapiCall(DWORD callId, TapiLine *pTapiLine);
	TapiCall *	GetTapiCall(HDRVCALL hdriverCall) {return GetTapiCall((TapiCall*)hdriverCall);};
	TapiCall *	GetTapiCall(HTAPICALL htapiCall);
	// caller responsible of freeing TapiCall memory
	TapiCall *	RemoveTapiCall(TapiCall* pTapiCall);

	bool		GetMonitorSet(DWORD callId, Set_xref & setXref) {if (m_cstaProvider) {return m_cstaProvider->GetMonitorSet(callId, setXref);} return false;}

	bool		CompletionProc(DRV_REQUESTID dwRequestID, LONG lResult);

	void		WaitForMapLock() { 
					EnterCriticalSection(&m_csMapLock); 
					return; 
				};
	void		ReleaseMapLock() { 
					LeaveCriticalSection(&m_csMapLock);
					return; 
				};
	void		WaitForLineOpenCloseLock() { 
					EnterCriticalSection(&m_csLineOpenCloseLock); 
					return; 
				};
	void		ReleaseLineOpenCloseLock() { 
					LeaveCriticalSection(&m_csLineOpenCloseLock);
					return; 
				};

	/// returns the value of a key in a encodedString, # separated
	std::string ContainsVar(std::string sEncodedString, std::string sKey, std::string sDefaultValue);
	std::string ContainsVar(std::string sEncodedString, std::string sKey);
protected:
	int		m_nLockDebugCounter;

private:
	ASTCstaProviderInterface	*m_cstaProvider;
	ASYNC_COMPLETION	m_fnCompletionProc ;		//Function call for asyncrounous functions
	DWORD				m_dwPermanentProviderID ;
	DWORD				m_dwLineDeviceIDBase ;
	CRITICAL_SECTION	m_csMapLock;
	CRITICAL_SECTION	m_csLineOpenCloseLock;
	ASTLogger			m_log;
	std::vector<std::string> m_vOutDnVector;

	// tsp line map and helper maps
	Map_Xref2TapiLine			m_mapXref2TapiLine ;
	Map_HTapiLine2Xref			m_mapHtapiLine2xref ;	//helper map
	
	// tsp call map and helper maps
	Map_TapiCall				m_mapTapiCall;				// hdrvrcall -> *pTapiCall 
	Map_HtapiCall2HdrvCall		m_mapHtapiCall2HdrvCall;	// htapicall -> hdrvcall (== *pTapiCall)
	Map_CallId2HdrvCallSet		m_mapCallId2HdrvCallSet ;	// callid -> set of hdrvrcall (== *pTapiCall)

	void	CallMapCleanup();
	void	DumpMaps();

	// csta listener implementation
	void	OnOriginated ( MonitorXref xref, DWORD callId, std::string originatingDevice, std::string callerid, std::string calleridname, std::string calledDevice, std::string userData, call_cause cause);
	void	OnAlerting ( MonitorXref xref, DWORD callId, std::string alertingDevice, std::string callerid, std::string calleridname, std::string calledDevice, std::string lastRedirectingDevice, std::string userData, call_cause cause);
	void	OnEstablished ( MonitorXref xref, DWORD callId, std::string establishedDevice, std::string callerid, std::string calleridname, std::string calledDevice, std::string lastRedirectingDevice, std::string userData, call_cause cause);
	void	OnFailed ( MonitorXref xref, DWORD callId, std::string failingDevice, std::string callerid, std::string calleridname, std::string calledDevice, std::string lastRedirectingDevice, std::string userData, call_cause cause);
	void	OnQueued ( MonitorXref xref, DWORD callId, std::string queuedDevice, std::string queue, int position_in_queue, int queue_len, std::string callerId, std::string callerIdName, std::string calledDevice, std::string userData, call_cause cause) ;
	void	OnDiverted ( MonitorXref xref, DWORD callId, std::string divertedDevice, std::string divertingDevice, std::string callerId, std::string callerIdName, std::string calledDevice, std::string userData, call_cause cause) ;
	void	OnConnectionCleared ( MonitorXref xref, DWORD callId, std::string releasingDevice, call_cause cause);
	void	OnCallCleared ( MonitorXref xref, DWORD callId, std::string userData, call_cause cause);
	void	OnHold ( MonitorXref xref, DWORD callId, std::string holdingDevice, call_cause cause);
	void	OnUnhold ( MonitorXref xref, DWORD callId, std::string unholdingDevice, call_cause cause);
	void	OnLinkDown ();
};

#endif	// TAPIMODEL_H
