/*
	Name: astcall.cpp
	Copyright: Under the Mozilla Public License Version 1.1 or later
	Author: ACTIVA
	Date: JULY/06
*/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Activa.
 *
 * The Initial Developer of the Original Code is ACTIVA TEAM.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): none
 *
 * ***** END LICENSE BLOCK ***** */

#include "astinterface.h"
#include "call_map.h"	// CSTAState2Str()

#include <algorithm>	// transform


/**
*	create new AsteriskCall with uniqueId and callerId for this connection
*	\param _uniqueId identifier of AsteriskCall
*	\param _callerId 
*	\param _callId identifier of call
*	\param _userData
*	\return a new AsteriskCall
*/
AsteriskCall*	AsteriskCall::Create(std::string _uniqueId, std::string _callerId, DWORD _callId, std::string _userData)
{
	AsteriskCall *call = new AsteriskCall();
	if (!call)
		return NULL;
	call->m_uniqueId = _uniqueId;
	call->m_callerid = _callerId;
	call->m_callId = _callId;
	call->m_userData = _userData;
	call->m_predictive = false;
	call->m_cause = ast_cause_normal;
	return call;
}

/**
*	Constructs an AsteriskCall
*/
AsteriskCall::AsteriskCall() 
{
	m_uniqueId = "" ;
	m_callerid = "";
	m_channel = "";
	m_state = call_state_not_known;
	m_sourceUniqueId = "";
}

/**
*	Set a Channel and address in AsteriskCall
*	\param newChannel
*/
void		AsteriskCall::SetChannel(std::string newChannel) 
{ 
	m_channel = newChannel; 
	m_address = GetAddressFromChannel(newChannel);
	m_numDevice = GetDialableNumberFromChannel(newChannel);
}	


/**
*	This link is only used when exist two differents CSTA Calls that are the same audio link
*	\return true if sourceUniqueId has the prefix CSTA
*/
bool	AsteriskCall::HasCSTALink()
{
	size_t npos = -1;
	if ((npos = this->GetSourceUniqueId().find("CSTA")) != -1){
		return true;
	}
	return false;
}

/**
*	links to destCall
*	\param pSourceCall
*	FIX: AsteriskCall::link makes an inverted link
*/
void		AsteriskCall::Link(AsteriskCall *pSourceCall)
{
	if (!pSourceCall)
		return;

	//We don't assign the SourceUniqueId if the same of the SourceCall 
	if(pSourceCall->GetUniqueId()==this->GetUniqueId())
		return;

	m_sourceUniqueId = pSourceCall->GetUniqueId();
}

/**
*	Gets extension from channel specification
*	- HEADER: channel DATA: Local/311\@default-3d4a,2		=> 311
*	- HEADER: channel DATA: SIP/311-f260					=> SIP/311
*	- HEADER: channel DATA: mISDN/1-1						=> mISDN/1-1
*	- HEADER: channel DATA: AsyncGoto/SIP/311-fw60		=> SIP/311
*	- HEADER: channel DATA: CAPI/ISDN1/319-2				=> CAPI/ISDN1/319
*	- HEADER: channel DATA: Transfered/SIP/419-08269c40   => SIP/419
*	- other generic formats decoded as XXX/address-XXX	=> XXX/address
*
*	\param channel A pure address can also be passed. 
*	\return string with the address
*/
std::string	AsteriskCall::GetAddressFromChannel(std::string channel)
{
	std::string	address = channel;
	size_t extPos = -1;

	std::transform(channel.begin(), channel.end(), channel.begin(), tolower);	//tolower
	const std::string local="local/";
	const std::string sip="sip/";
	const std::string capi="capi/isdn1/";

	if ((extPos=channel.find(local)) != -1) {	// local channel type
		address = channel.substr(extPos+local.size(), channel.size());
		if ((extPos=address.find("@")) != -1) {	// @context found
			address = address.substr(0, extPos);
		}
		else if ((extPos=address.find("/n")) != -1){	//find "/n" at the end
			address = address.substr(0,extPos);
		}
	}
	else if ((extPos = channel.find(sip)) != -1) {	// sip channel type
		address = channel.substr(extPos+sip.size(), channel.size());
		if ((extPos=address.find_last_of("-")) != -1) {	// '-' found
			address = address.substr(0, extPos);
		}
	}
	else if ((extPos = channel.find("misdn/")) != -1) {	// mISDN/1-1
		address = channel;
	}
	else if ((extPos = channel.find("asyncgoto/")) != -1){	//AsyncGoto
		address = channel.substr(extPos+1, channel.size());
		if ((extPos=address.find_last_of("-")) != -1) {	// '-' found
			address = address.substr(0, extPos);
		}
	}
	else if ((extPos = channel.find("transfered/")) != -1){	//Transfered
		address = channel.substr(extPos+1, channel.size());
		if ((extPos=address.find_last_of("-")) != -1) {	// '-' found
			address = address.substr(0, extPos);
		}
	}	
	else if ((extPos = channel.find(capi)) != -1){	//CAPI/ISDN1
		address = channel.substr(extPos+capi.size(), channel.size());
		if ((extPos=address.find_last_of("-")) != -1) {	// '-' found
			address = address.substr(0,extPos);
		}
	}
	else if ((extPos = channel.find("/")) != -1) {	// generic channel handling
		address = channel.substr(extPos+1, channel.size());
		if ((extPos=address.find_last_of("-")) != -1) {	// '-' found
			address = address.substr(0, extPos);
		}
	}else{
		//if not contains a channel
		address = channel;
	}

	return address;
}

/**
*	Gets extension from channel specification
*	- HEADER: channel DATA: Local/311\@default-3d4a,2		=> 311
*	- HEADER: channel DATA: SIP/311-f260					=> 311
*	- HEADER: channel DATA: mISDN/1-1						=> mISDN/1-1
*	- HEADER: channel DATA: AsyncGoto/SIP/311-fw60		=> SIP/311
*	- HEADER: channel DATA: CAPI/ISDN1/319-2				=> 319
*	- HEADER: channel DATA: Transfered/SIP/419-08269c40   => SIP/419
*	- other generic formats decoded as XXX/address-XXX	=> address
*
*	\param channel A pure address can also be passed. 
*	\return string with the address
*/
std::string	AsteriskCall::GetDialableNumberFromChannel(std::string channel)
{
	std::string	address = channel;
	size_t extPos = -1;

	std::transform(channel.begin(), channel.end(), channel.begin(), tolower);	//tolower
	const std::string local="local/";
	const std::string sip="sip/";
	const std::string capi="capi/isdn1/";

	if ((extPos=channel.find(local)) != -1) {	// local channel type
		address = channel.substr(extPos+local.size(), channel.size());
		if ((extPos=address.find("@")) != -1) {	// @context found
			address = address.substr(0, extPos);
		}
		else if ((extPos=address.find("/n")) != -1){	//find "/n" at the end
			address = address.substr(0,extPos);
		}
	}
	else if ((extPos = channel.find(sip)) != -1) {	// sip channel type
		address = channel.substr(extPos+sip.size(), channel.size());
		if ((extPos=address.find_last_of("-")) != -1) {	// '-' found
			address = address.substr(0, extPos);
		}
	}
	else if ((extPos = channel.find("misdn/")) != -1) {	// mISDN/1-1
		address = channel;
	}
	else if ((extPos = channel.find("asyncgoto/")) != -1){	//AsyncGoto
		address = channel.substr(extPos+1, channel.size());
		if ((extPos=address.find_last_of("-")) != -1) {	// '-' found
			address = address.substr(0, extPos);
		}
	}
	else if ((extPos = channel.find("transfered/")) != -1){	//Transfered
		address = channel.substr(extPos+1, channel.size());
		if ((extPos=address.find_last_of("-")) != -1) {	// '-' found
			address = address.substr(0, extPos);
		}
	}	
	else if ((extPos = channel.find(capi)) != -1){	//CAPI/ISDN1
		address = channel.substr(extPos+capi.size(), channel.size());
		if ((extPos=address.find_last_of("-")) != -1) {	// '-' found
			address = address.substr(0,extPos);
		}
	}
	else if ((extPos = channel.find("/")) != -1) {	// generic channel handling
		address = channel.substr(extPos+1, channel.size());
		if ((extPos=address.find_last_of("-")) != -1) {	// '-' found
			address = address.substr(0, extPos);
		}
	}else{
		//if not contains a channel
		address = channel;
	}

	return address;
}

/**
*	Return true if channel is Local
*	i.e.: Local/320\@from-internal-6f9
*	\return true if channel is local, otherwise false
*/
bool	AsteriskCall::IsLocalChannel()
{
	const std::string local="local/";
	std::string	channel = m_channel;

	std::transform(channel.begin(), channel.end(), channel.begin(), tolower);	//tolower

	if (channel.find(local) != -1) {	// local channel type
		return true;
	}

	return false;
}

/**
*	Return true if channel is localMaster
*	i.e.: Local/320\@from-internal-6f9b,1
*	\return true if is a master channel, otherwise false
*/
bool	AsteriskCall::IsLocalMasterChannel()
{
	std::string	channel = m_channel;
	size_t extPos = -1;

	std::transform(channel.begin(), channel.end(), channel.begin(), tolower);	//tolower
	const std::string local="local/";

	if ((extPos=channel.find(local)) != -1) {	// local channel type
		if (channel.find(";1")!=-1 || channel.find(",1")!=-1) {	// master
			return true;
		}
	}

	return false;
}

/**
*	Return true if channel is localOutbound
*	i.e.: Local/320\@from-internal-6f9b,2
*	\return true if the channel is an Outbound channel, otherwise false
*/
bool	AsteriskCall::IsLocalOutboundChannel()
{
	std::string	channel = m_channel;
	size_t extPos = -1;

	std::transform(channel.begin(), channel.end(), channel.begin(), tolower);	//tolower
	const std::string local="local/";

	if ((extPos=channel.find(local)) != -1) {	// local channel type
		if (channel.find(";2")!=-1 || channel.find(",2")!=-1) {	// outbound
			return true;
		}
	}

	return false;
}

/**
*	Return getOutboundChannelFromMasterChannel ( ast14 ->",1 =>,2" ; ast16->";1 =>;2" )
*	i.e.: Local/320\@from-internal-6f9b,1 -> Local/320\@from-internal-6f9b,2
*	\return getOutboundChannelFromMasterChannel, otherwise ""
*/
std::string	AsteriskCall::getOutboundChannelFromMasterChannel(std::string channel)
{
	size_t extPos = -1;
	if ((extPos=channel.find(";1")) != -1)	// master ast16 leg separator
		return (channel.substr(0,extPos)+";2");
	else if ((extPos=channel.find(",1")) != -1)	// master ast14 leg separator
		return (channel.substr(0,extPos)+",2");

	return ("");
}
	
/**
*	Return getMasterChannelFromOutboundChannel ( ast14 ->",2 =>,1" ; ast16->";2 =>;1" )
*	i.e.: Local/320\@from-internal-6f9b,2 -> Local/320\@from-internal-6f9b,1
*	\return getMasterChannelFromOutboundChannel, otherwise ""
*/
std::string	AsteriskCall::getMasterChannelFromOutboundChannel(std::string channel)
{
	size_t extPos = -1;
	if ((extPos=channel.find(";2")) != -1)	// master ast16 leg separator
		return (channel.substr(0,extPos)+";1");
	else if ((extPos=channel.find(",2")) != -1)	// master ast14 leg separator
		return (channel.substr(0,extPos)+",1");

	return ("");
}



/**
*	Return CallerIdName prefix
*	CIDName prefix has to be separated by ':'
*	i.e.: 111: => callerIdName prefix is 111
*	\param calleridname
*	\return string with CIDNamePrefix
*/
std::string	AsteriskCall::GetCIDNamePrefix(std::string calleridname)
{
	std::string	prefix;
	size_t extPos;
	if ((extPos=calleridname.find(":")) != -1) {	
		prefix = calleridname.substr(0, extPos);
	}
	
	return prefix;
}

/**
*	Return true if device is \<unknown> or ""
*	\param device
*	\return true if device is unknown, otherwise false
*/
bool	AsteriskCall::IsUnknown(std::string device)
{
	std::transform(device.begin(), device.end(), device.begin(), tolower);	//tolower
	const std::string unknown="<unknown>";

	if (device=="" || device==unknown) {	
		return true;
	}
	return false;
}

/**
*	Dump AsteriskCall
*	\param log
*/
void	AsteriskCall::Dump(ASTLogger &log)
{
	log.LogTrace("_____________________________________________________ AsteriskCall DUMP BEGIN");
	log.LogTrace("  {%s}.%s callId=%d callerId=%s", m_uniqueId.c_str(), m_address.c_str(), m_callId, m_callerid.c_str());
	log.LogTrace("    state=%s, source=%s, lastRedir=%s, cause=%d", GetStateString().c_str(), m_sourceUniqueId.c_str(), m_lastRedirectingDevice.c_str(), m_cause);
	log.LogTrace("_____________________________________________________ AsteriskCall DUMP END");
}

/**
*	Converts an AsteriskCall to String
*	\return string with the AsterisCall atributes
*/
std::string	AsteriskCall::DumpToString()
{
	std::string	str;
	str = "{"+m_uniqueId+"}." +m_address;
	str += " numDevice=" + m_numDevice;
	str += " callID=" + ToString(m_callId);
	str += " state=" + GetStateString();
	str += " chann=" + m_channel;
	str += " CID=" + m_callerid;
	str += " CIDName=" + m_calleridname;
	str += " ConLineNum=" + m_connectedLineNum;
	str += " ConLineName=" + m_connectedLineName;
	str += " lastRedir=" + m_lastRedirectingDevice;
	str += " source=" + m_sourceUniqueId;
	str += " userData=" + m_userData;
	str += " cause=" +ToString(m_cause);
	return str;
}

/**
*	Returns type of channel.
*	Recognise types:
*	- Local
*	- Sip
*	- misdn
*	- asyncgoto
*	- capi
*	- transfered
*	- generic types
*	\return string type of channel
*/
std::string		AsteriskCall::GetTypeChannel()
{
	std::string type = "";
	std::string channel = m_channel;

	size_t extPos = -1;

	std::string local = "local/";
	std::string sip = "sip/";
	std::string misdn = "misdn/";
	std::string asyncgoto = "asyncgoto/";
	std::string capi = "capi/";
	std::string others = "/";
	std::string tranfered = "transfered/";

	std::transform(channel.begin(), channel.end(), channel.begin(), tolower);	//tolower

	if ((extPos=channel.find(local)) != -1) {	// local channel type
		type = channel.substr(0, local.size()-1);
	}
	else if ((extPos = channel.find(sip)) != -1) {	// sip channel type
		type = channel.substr(0, sip.size()-1);
	}
	else if ((extPos = channel.find(misdn)) != -1) {	// mISDN/1-1
		type = channel.substr(0, misdn.size()-1);
	}
	else if ((extPos = channel.find(asyncgoto)) != -1){	//AsyncGoto
		type = channel.substr(0, asyncgoto.size()-1);
	}
	else if ((extPos = channel.find(capi)) != -1){	//CAPI/ISDN1
		type = channel.substr(0, capi.size() -1);
	}
	else if ((extPos = channel.find(tranfered)) != -1){ //tranfered
		type = channel.substr(0, tranfered.size() -1);
	}
	else if ((extPos = channel.find(others)) != -1) {	// generic channel handling
		type = channel.substr(0, others.size()-1);
	}

	return type;
}
