/* 
	Name: astcstaprovider.cpp
	Copyright: Under the Mozilla Public License Version 1.1 or later
	Author: ACTIVA
*/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is astprovider.
 *
 * The Initial Developer of the Original Code is ACTIVA.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): none
 *
 * ***** END LICENSE BLOCK ***** */

#include "astcstaprovider.h"


#define DUMP_EVENT(m_log, ev)		m_log.LogTrace("  Event Info => \r\n\tuniqueId=%s \r\n\tsrcUniqueId=%s \r\n\tdestUniqueId=%s \r\n\taddress=%s \r\n\textension=%s \r\n\tpriority=%s \r\n\tchannel=%s \r\n\tcallerid=%s \r\n\tcause=%s \r\n\tcontext=%s \r\n\tstate=%s \r\n\t", \
										ev.m_uniqueId.c_str(),	\
										ev.m_srcUniqueId.c_str(),	\
										ev.m_destUniqueId.c_str(),	\
										ev.m_address.c_str(),	\
										ev.m_extension.c_str(),		\
										ev.m_priority.c_str(),		\
										ev.m_channel.c_str(),	\
										ev.m_callerId.c_str(),	\
										ev.m_cause.c_str(),		\
										ev.m_context.c_str(),	\
										ev.m_state.c_str());			



ASTCstaProvider::ASTCstaProvider() : m_callMap(NULL)
{
	m_log.SetOrigin("ASTCstaProvider");
	m_currentXref = (MonitorXref) 1;
}

ASTCstaProvider::~ASTCstaProvider()
{
	// relese memory and clean maps
	m_callMap.CleanupMaps();
}

	
/**
 * ASTCstaProvider factory method.
 * Creates a ASTCstaProvider and attach a listener.
 * \param cstaListener csta event listener
 * \param logger logger, can be null
 * \return Returns NULL if cannot create or listener is null.
 */
ASTCstaProvider	*ASTCstaProvider::Create(ASTCstaListener *cstaListener, ASTLogListener *logger)
{
	if (cstaListener==NULL) {
		return NULL;
	}

	ASTCstaProvider	*instance = new ASTCstaProvider();
	if (instance==NULL) {
		return NULL;
	}

	instance->m_cstaListener = cstaListener;

	ASTProvider	*astListener = ASTProvider::Create(instance, logger);
	if (astListener==NULL) { // ups... rollback
		delete instance;
		return NULL;
	}

	instance->m_astProvider = astListener;

	instance->m_log.SetLogger(logger);
	instance->m_callMap.SetLogger(logger);
	instance->m_log.LogInfo("__________________ ASTCstaProvider::Create __________________");
	instance->m_log.LogInfo(" %20s ",SVN_REVISION);
	instance->m_log.LogInfo("_____________________________________________________________");
	return instance;
}

/**
*	Connects to asterisk 
*	\param sHost ip host address of asterisk box
*	\param nPort ip port asterisk manaegr api is listening to. Usually default asterisk configuration port is 5038
*	\param user user, ie: admin
*	\param pwd password, ie: amp111
*	\param eAuthType authentification type 0-plaintext 1-md5
*	\return Returns AC_RESULT
*/
AC_RESULT	ASTCstaProvider::Connect(std::string sHost, int nPort, std::string user, std::string pwd, AUTH_TYPE eAuthType)
{
	return m_astProvider->Connect(sHost, nPort, user, pwd, eAuthType);
}

/**
*	Disconnect from Asterisk
*	\return bool
*/
bool	ASTCstaProvider::Disconnect()
{
	bool	rc = m_astProvider->Disconnect();
	if (rc) {
		m_callMap.CleanupMaps();
	}

	return rc;
}

/**
*	Opens a new monitor interested in events on specific deviceID.
*	DeviceId can be an address/extension (example: 311) or a full channel spec (example: sip/311)
*	and it is parsed to extract the address part.
*	\param channel	what channel, address or extension do you want to monitoring
*	\return number of monitor
*/
MonitorXref	ASTCstaProvider::OpenMonitor(std::string channel)
{
	const std::string	address = AsteriskCall::GetAddressFromChannel(channel);
	m_log.LogTrace("ASTCstaProvider::OpenMonitor device = %s => %s", channel.c_str(), address.c_str());

	// protect map access with mutex
	m_callMap.__WaitForMapLock();

	MonitorXref	xref = m_currentXref;

	m_currentXref++;
	
	// add to m_mapExtension2Xref multimap
	m_callMap.AddXref(address, xref);

	m_callMap.__ReleaseMapLock();

	return xref;
}
/**
*	Close a monitor with reference xref
*	\param xref the monitor you want to close
*	\return bool if monitor is closed correctly or not
*/
bool	ASTCstaProvider::CloseMonitor(MonitorXref xref)
{
	m_log.LogTrace("ASTCstaProvider::CloseMonitor xref = %d", xref);

	bool	rc = m_callMap.RemoveXref(xref);
	m_log.LogTrace("  return %x!", rc);
	return rc;
}

/**
*	CallerIdName may set caller_id_name info
*	\param callingDevice Device who makes the call
*	\param calledDevice Device who is called
*	\param callid if callId is 0, it generates a new one
*	\param context	
*	\param callerIdName	
*	\param userData
*	\return AC_RESULT: ok, generic failure, timeout
*/
AC_RESULT	ASTCstaProvider::MakeCall(std::string callingDevice, std::string contextCalling, std::string calledDevice, DWORD &callid, std::string contextCalled, std::string callerIdName, std::string userData,  std::string variable, bool bLockReaderAfterResponse)
{
	std::string	channel = callingDevice;
	if (channel.find("/") == -1) {	// callingDevice is not a full-specified channel
		channel = "Local/" + callingDevice + "@" + contextCalling;;
	}

//	callerIdName = callerIdName + "'                                                  ";
	callerIdName = callerIdName + "      ";
	if (userData != ""){
		callerIdName = callerIdName + "#U=" + userData;
	}

	// if callId != 0, do not generate a callID
	if (callid == 0){
		callid = m_callMap.GetNewCallId();
		m_log.LogTrace(" NEW callid=%d", callid);
	}

	// because to delete a tapicall needs an entry in the CSTA callMap
	m_callMap.DumpMaps();
	m_callMap.AddConnectionToCallModel(callid, callingDevice, INITIALIZED /*"INITIALITED"*/);

	return m_astProvider->Originate(	channel, 
										callingDevice,	// callerId
										callerIdName,	
										calledDevice,	// calledExtension
										callid, contextCalled,30,
										variable,bLockReaderAfterResponse);
}

/**
*	A predictive call is a call that originates in callingDevice an is transferred to callerDevice
*	\param callingDevice Device who makes the call
*	\param contextCalling
*	\param calledDevice Device who is called
*	\param contextCalled
*	\param callid if callId is 0, it generates a new one
*	\param callerId
*	\param userData
*	\param timeout
*	\return AC_RESULT: ok, generic failure, timeout
*/
AC_RESULT	ASTCstaProvider::MakePredictiveCall(std::string callingDevice, std::string contextCalling, std::string calledDevice, std::string contextCalled, DWORD &callid, std::string callerId, std::string userData, DWORD timeout, std::string variable, bool bLockReaderAfterResponse)
{

	std::string channel = calledDevice;
	if(channel.find("/") == -1){
		channel = "Local/" + channel + "@" + contextCalled;
	}

	size_t pos = -1;
//	std::string callerIdName = "'                                                  ";
	std::string callerIdName = "      ";

	if ((pos = callerId.find("<")) != -1){
		callerIdName = callerId.substr(0, pos-1);
		callerId = callerId.substr(pos+1,callerId.size());
		pos = callerId.find(">");
		callerId = callerId.substr(0,pos);
	}

	if (userData != ""){
		callerIdName = callerIdName + "#U=" + userData;
	}

	callerIdName = callerIdName + "#D";

	//	if callId != 0, do not generate a callID
	if (callid == 0){
		callid = m_callMap.GetNewCallId();
		m_log.LogTrace(" NEW callid=%d", callid);
	}

	m_callMap.DumpMaps();
	m_callMap.AddConnectionToCallModel(callid, callingDevice, INITIALIZED /*"INITIALITED"*/);

	AC_RESULT rc = m_astProvider->Originate(	channel,	
										callerId,
										callerIdName,
										callingDevice,
										callid,
										contextCalling, timeout, 
										variable, bLockReaderAfterResponse);

	return rc;
}

/**
*	Hang up an established call
*	\param callid identifier of call
*	\return AC_RESULT: ok, no such channel, generic failure, timeout
*/
AC_RESULT	ASTCstaProvider::ClearCall(DWORD callid)
{
	m_log.LogTrace("ASTCstaProvider::ClearCall callID=%d", callid);

	// protect map access with mutex
	m_callMap.__WaitForMapLock();

	// Find all devices with the same callId
	Set_deviceId setDevices;
	m_callMap.GetSetDevicesFromCallId(callid,setDevices);
	
	m_callMap.__ReleaseMapLock();

	if (setDevices.size() == 0){
		m_log.LogError(" not exist any call with callid %d", callid);
		return invalid_callid;
	}
	for (Set_deviceId::iterator i = setDevices.begin(); i != setDevices.end(); i++){
		const std::string device = i->data();
		m_callMap.__WaitForMapLock();
		AsteriskCall *pAC = m_callMap.FindCallByCallIdAndDevice(callid,device);
		std::string sChannel ;
		if(pAC)
			sChannel = pAC->GetChannel();
		
		m_callMap.__ReleaseMapLock();
		AC_RESULT rc = m_astProvider->Hangup(sChannel);
	}

	return ok;
}

/**
*	Hang up an established Connection
*	\param callid identifier of call
*	\param address device
*	\return AC_RESULT: ok, no such channel, generic failure, timeout
*/
AC_RESULT	ASTCstaProvider::ClearConnection(DWORD callid, std::string address)
{
	m_log.LogTrace("ASTCstaProvider::ClearConnection callID=%d address=%s", callid, address.c_str());

	// protect map access with mutex
	m_callMap.__WaitForMapLock();

	// lookup channel
	AsteriskCall *pAC = m_callMap.FindCallByCallIdAndDevice(callid,address);
	if (!pAC){
		m_callMap.__ReleaseMapLock();
		m_log.LogWarn("  not found any call with callid %d and device %s", callid, address.c_str());
		return invalid_callid;
	}

	const std::string	sChannel = pAC->GetChannel();

	// release map lock
	m_callMap.__ReleaseMapLock();

	m_log.LogTrace("  Hangup callid=%d,channel=%s", callid, sChannel.c_str());

	return m_astProvider->Hangup(sChannel);
}
/**
*	Do a FastTransfer from one device to another.
*	\param callId identifier of call
*	\param transferringDevice
*	\param destination
*	\param destinationContext
*	\return AC_RESULT: ok, generic failure, timeout
*/
AC_RESULT	ASTCstaProvider::FastTransfer(DWORD callId, std::string transferringDevice, std::string destination, std::string destinationContext)
{
	m_log.LogTrace("ASTCstaProvider::FastTransfer callID=%d from=%s destination=%s", callId,transferringDevice.c_str(), destination.c_str());

	// protect map access with mutex
	m_callMap.__WaitForMapLock();
	
	m_callMap.DumpMaps();
	AsteriskCall *pAC = m_callMap.FindCallByCallIdAndDevice(callId,transferringDevice);
	m_log.LogTrace(" found call {%s}.%s", pAC->GetUniqueId().c_str(), transferringDevice.c_str());
	// searching source call
	
	std::string	sChannel ;
	if (pAC->HasCSTALink()){
		//we found the other branch
		std::string uniqueId = GetCSTALink(pAC->GetSourceUniqueId());
		AsteriskCall *pSourceCall = m_callMap.FindCallByUniqueId(uniqueId);
		if (!pSourceCall){
			m_callMap.__ReleaseMapLock();
			return error_not_implemented;
		}
		sChannel = pSourceCall->GetChannel();
	} // pAC->HasCSTALink()
	else {
		Set_deviceId setDevices;
		m_callMap.GetSetDevicesFromCallId(callId,setDevices);

		//******************** 
		//DBR TRANSFER...

		// Dont count initialized devices
		for (Set_deviceId::iterator i = setDevices.begin(); i != setDevices.end();i++){
			std::string device = i->data();
			if (device.compare(transferringDevice) != 0){
				if(m_callMap.GetStateOfDevice(callId,device)==csta_states::INITIALIZED)
				{
					m_log.LogTrace0("Erase, device %s from transfer list.",device.c_str());
					setDevices.erase(i);
				}
			}
		}

		//********************

		if (setDevices.size() != 2){
			m_callMap.__ReleaseMapLock();
			m_log.LogTrace("  too much devices => error_not_implemented");
			return error_not_implemented;
		}
		// Find what device is not transferring device
		for (Set_deviceId::iterator i = setDevices.begin(); i != setDevices.end();i++){
			std::string device = i->data();
			if (device.compare(transferringDevice) != 0){
				pAC = m_callMap.FindCallByCallIdAndDevice(callId,device);
				break;
			}
		}
		if (!pAC) {
			m_callMap.__ReleaseMapLock();
			m_log.LogTrace("  callid not in map => invalid_callid");
			return invalid_callid;
		}

		sChannel = pAC->GetChannel();
	} // else pAC->HasCSTALink()

	m_log.LogTrace("  redirecting channel %s to destination %s",sChannel.c_str(), destination.c_str());

	// release map lock
	m_callMap.__ReleaseMapLock();
	AC_RESULT rc = m_astProvider->Redirect(sChannel, "", destination, destinationContext);
	ClearConnection(callId, transferringDevice);
	return rc;
}

/**
*	Put a call in HOLD state
*	\param callId identifier of call
*	\param holdingDevice
*	\param holdContext
*	\return AC_RESULT: ok, generic failure, timeout
*/
AC_RESULT ASTCstaProvider::HoldCall(DWORD callId, std::string holdingDevice, std::string holdContext)
{
	m_log.LogTrace("ASTCstaProvider::HoldCall callID=%d hold=%s", callId, holdingDevice.c_str());

	// protect map access with mutex
	m_callMap.__WaitForMapLock();
	
	// Find all devices with the same callId
	Set_deviceId setDevices;
	m_callMap.GetSetDevicesFromCallId(callId,setDevices);

	// At the moment, the conferences are not supported
	if (setDevices.size() != 2){
		m_callMap.__ReleaseMapLock();
		m_log.LogTrace("  too much devices => error_not_implemented");
		return error_not_implemented;
	}

	AsteriskCall *pAC = NULL;
	for (Set_deviceId::iterator i = setDevices.begin(); i != setDevices.end();i++){
		std::string device = i->data();
		if (device.compare(holdingDevice) != 0){
			pAC = m_callMap.FindCallByAddress(device);
		}
	}
	
	if (!pAC) {
		m_callMap.__ReleaseMapLock();
		m_log.LogTrace("  callid not in map => invalid_callid");
		return invalid_callid;
	}

	std::string	sChannelMon;
	const std::string	channel = pAC->GetChannel();
	
	AsteriskCall *pACmon = NULL;
	m_log.LogTrace("  call with callid=%d has channel=%s", callId, channel.c_str());
	if (pAC->HasSourceUniqueId()) {
		pACmon = m_callMap.GetSourceCall(pAC);
		m_log.LogTrace("  its sourceCall has uniqueid=%s has channel=%s", pACmon->GetUniqueId().c_str(), pACmon->GetChannel().c_str());
		sChannelMon = pACmon->GetChannel();
	}else{
		pACmon = m_callMap.FindCallByAddress(holdingDevice);
		m_log.LogTrace("  its sourceCall has uniqueid=%s has channel=%s", pACmon->GetUniqueId().c_str(), pACmon->GetChannel().c_str());
		sChannelMon = pACmon->GetChannel();
	}

	// release map lock
	m_callMap.__ReleaseMapLock();

	m_log.LogTrace(" redirecting channel %s to hold extension", channel.c_str());

	m_astProvider->SetVar(channel,"ChannelFrom", sChannelMon);
	m_callMap.AddConnectionToCallModel(callId, holdingDevice, HELD /*"HELD"*/);
	return m_astProvider->Redirect(channel, "", "69", holdContext);
}

/**
*	Retrieve a hold call
*	\param callId identifier of call
*	\param retreivingDevice
*	\param retrieveContext
*	\return AC_RESULT: ok, generic failure, timeout
*/
AC_RESULT ASTCstaProvider::RetrieveCall(DWORD callId, std::string retreivingDevice, std::string retrieveContext)
{
	m_log.LogTrace("ASTCstaProvider::RetrieveCall callID=%d hold=%s context=%s", callId,retreivingDevice.c_str(), retrieveContext.c_str());

	// protect map access with mutex
	m_callMap.__WaitForMapLock();
	
	m_callMap.DumpMaps();
	Set_deviceId setDevices;
	m_callMap.GetSetDevicesFromCallId(callId,setDevices);

	if (setDevices.size() < 1){
		m_callMap.__ReleaseMapLock();
		m_log.LogTrace("  too much devices => error_not_implemented");
		return error_not_implemented;
	}

	AsteriskCall *pAC = NULL;
	for (Set_deviceId::iterator i = setDevices.begin(); i != setDevices.end();i++){
		std::string device = i->data();
		if (m_callMap.GetStateOfDevice(callId,device) != HELD /*"HELD"*/){
			pAC = m_callMap.FindCallByAddress(device);
		}
	}

	const std::string sChannel = pAC?pAC->GetChannel() : "";

	// release map lock
	m_callMap.__ReleaseMapLock();
	return m_astProvider->Redirect(sChannel, "",retreivingDevice , retrieveContext);
}


/**
*	SetDoNotDisturb
*	actually implemented as SetAgentStatus (READY/NOTREADY)
*	\param deviceId
*	\param bDND_on
*	\param dndContext
*	\return AC_RESULT: ok, already logged in, queue not specified, no such queue, generic failure
*/
AC_RESULT	ASTCstaProvider::SetDoNotDisturb(std::string deviceId, bool bDND_on, std::string dndContext)
{
	m_log.LogTrace("ASTCstaProvider::SetDoNotDisturb [%s] => %d", deviceId.c_str(), bDND_on);
	AGENT_STATUS	newAgentStatus = bDND_on?NOTREADY:READY;

	AC_RESULT rc = SetAgentStatus(deviceId, "", newAgentStatus, dndContext);
	
	if (rc == interface_not_found) {	// hack! somewhere setdonotdisturb is used _before- loggin, but we dont want to send a failure
		rc = ok;
	}

	return rc;
}


/**
*	Set agentState.
*	If deviceId is not a full channel identifier including technology will default to local/deviceId
*	\param deviceId
*	\param queue
*	\param newStatus
*	\param agentContext
*	\return AC_RESULT: ok, already logged in, queue not specified, no such queue, generic failure
*/
AC_RESULT	ASTCstaProvider::SetAgentStatus(std::string deviceId, std::string queue, AGENT_STATUS newStatus, std::string agentContext)
{
	std::string	channel = AsteriskCall::GetAddressFromChannel(deviceId);
	channel = "Local/" + channel + "@"+agentContext+"/n";


	switch(newStatus) {
		case LOGIN:
			return m_astProvider->QueueAdd(queue, channel, 0, true);
			break;
		case LOGOUT:
			return m_astProvider->QueueRemove(queue, channel);
			break;
		case NOTREADY:
			return m_astProvider->QueuePause(channel, true);
			break;
		case READY:
			return m_astProvider->QueuePause(channel, false);
			break;
		default:
			return error_not_implemented;
	}
}

/**
*	Start recording
*	\param callid identifier of call
*	\param address device
*	\param RecordingFileName Recording file name
*	\return AC_RESULT: ok, no such channel, generic failure, timeout
*/
AC_RESULT	ASTCstaProvider::StartRecording(DWORD callid, std::string address, std::string RecordingFileName, std::string format, std::string mix)
{
	m_log.LogTrace("ASTCstaProvider::StartRecording callID=%d address=%s", callid, address.c_str());

	// protect map access with mutex
	m_callMap.__WaitForMapLock();

	// lookup channel
	AsteriskCall *pAC = m_callMap.FindCallByCallIdAndDevice(callid,address);
	if (!pAC){
		m_callMap.__ReleaseMapLock();
		m_log.LogWarn("  not found any call with callid %d and device %s", callid, address.c_str());
		return invalid_callid;
	}
	std::string	sChannel ;
	sChannel = pAC->GetChannel();
	// release map lock
	m_callMap.__ReleaseMapLock();

	m_log.LogTrace("  call with callid=%d has channel=%s", callid, sChannel.c_str());

	return m_astProvider->Monitor(sChannel,RecordingFileName,format,mix);
}

/**
*	Stop recording
*	\param callid identifier of call
*	\param address device
*	\return AC_RESULT: ok, no such channel, generic failure, timeout
*/
AC_RESULT	ASTCstaProvider::StopRecording(DWORD callid, std::string address)
{
	m_log.LogTrace("ASTCstaProvider::StopRecording callID=%d address=%s", callid, address.c_str());

	// protect map access with mutex
	m_callMap.__WaitForMapLock();

	// lookup channel
	AsteriskCall *pAC = m_callMap.FindCallByCallIdAndDevice(callid,address);
	if (!pAC){
		m_callMap.__ReleaseMapLock();
		m_log.LogWarn("  not found any call with callid %d and device %s", callid, address.c_str());	
		return invalid_callid;
	}
	std::string	sChannel ;
	sChannel = pAC->GetChannel();
	// release map lock
	m_callMap.__ReleaseMapLock();

	m_log.LogTrace("  call with callid=%d has channel=%s", callid, sChannel.c_str());

	return m_astProvider->StopMonitor(sChannel);
}

/**
*	Pause recording
*	\param callid identifier of call
*	\param address device
*	\return AC_RESULT: ok, no such channel, generic failure, timeout
*/
AC_RESULT	ASTCstaProvider::PauseRecording(DWORD callid, std::string address)
{
	m_log.LogTrace("ASTCstaProvider::PauseRecording callID=%d address=%s", callid, address.c_str());

	// protect map access with mutex
	m_callMap.__WaitForMapLock();

	// lookup channel
	AsteriskCall *pAC = m_callMap.FindCallByCallIdAndDevice(callid,address);
	if (!pAC){
		m_callMap.__ReleaseMapLock();
		m_log.LogWarn("  not found any call with callid %d and device %s", callid, address.c_str());
		return invalid_callid;
	}
	std::string	sChannel ;
	sChannel = pAC->GetChannel();
	// release map lock
	m_callMap.__ReleaseMapLock();

	m_log.LogTrace("  call with callid=%d has channel=%s", callid, sChannel.c_str());
	
	return m_astProvider->PauseMonitor(sChannel);
}

/**
*	Continue recording after pause
*	\param callid identifier of call
*	\param address device
*	\return AC_RESULT: ok, no such channel, generic failure, timeout
*/
AC_RESULT	ASTCstaProvider::UnpauseRecording(DWORD callid, std::string address)
{
	m_log.LogTrace("ASTCstaProvider::UnpauseRecording callID=%d address=%s", callid, address.c_str());

	// protect map access with mutex
	m_callMap.__WaitForMapLock();

	// lookup channel
	AsteriskCall *pAC = m_callMap.FindCallByCallIdAndDevice(callid,address);
	if (!pAC){
		m_callMap.__ReleaseMapLock();
		m_log.LogWarn("  not found any call with callid %d and device %s", callid, address.c_str());
		return invalid_callid;
	}
	std::string	sChannel ;
	sChannel = pAC->GetChannel();
	// release map lock
	m_callMap.__ReleaseMapLock();

	m_log.LogTrace("  call with callid=%d has channel=%s", callid, sChannel.c_str());

	return m_astProvider->UnpauseMonitor(sChannel);
}

/**
*	Change recording file name
*	\param callid identifier of call
*	\param address device
*	\param NewRecordingFileName Recording file name
*	\return AC_RESULT: ok, no such channel, generic failure, timeout
*/
AC_RESULT	ASTCstaProvider::ChangeRecordingFileName(DWORD callid, std::string address, std::string NewRecordingFileName)
{
	m_log.LogTrace("ASTCstaProvider::ChangeRecordingFileName callID=%d address=%s", callid, address.c_str());

	// protect map access with mutex
	m_callMap.__WaitForMapLock();

	// lookup channel
	AsteriskCall *pAC = m_callMap.FindCallByCallIdAndDevice(callid,address);
	if (!pAC){
		m_callMap.__ReleaseMapLock();
		m_log.LogWarn("  not found any call with callid %d and device %s", callid, address.c_str());
		return invalid_callid;
	}
	std::string	sChannel ;
	sChannel = pAC->GetChannel();
	// release map lock
	m_callMap.__ReleaseMapLock();
	
	m_log.LogTrace("  call with callid=%d has channel=%s", callid, sChannel.c_str());

	return m_astProvider->ChangeMonitor(sChannel,NewRecordingFileName);
}

/**
 * handles provider link-down
 */
void	ASTCstaProvider::OnLinkDown()
{
	// cleanup maps!
	m_cstaListener->OnLinkDown();
	m_callMap.CleanupMaps();
}

/**
*	Channel Event. Currently not used
*	\param ev
*/
void	ASTCstaProvider::OnChannelEvent(ManagerEvent ev)
{
	// do nothing
}

/**
*	Dial generates new call leg
*	DialEvent info
*	srcUniqueId = original call uniqueId
*	destUniqueId = new call leg uniqueId
*	\param ev
*/
void	ASTCstaProvider::OnDialEvent(ManagerEvent ev)
{
	m_log.LogTrace("OnDialEvent:CFT: ev.m_srcUniqueId=\"%s\"; ev.m_destUniqueId=\"%s\"; ev.m_callerIdName=\"%s\"; ev.m_callerId=\"%s\"; m_pxCstaProvider->OnDialEvent(ev);", ev.m_srcUniqueId.c_str(), ev.m_destUniqueId.c_str(), ev.m_callerIdName.c_str(), ev.m_callerId.c_str());

	// protect map access with mutex
	m_callMap.__WaitForMapLock();
	AsteriskCall *pCall = m_callMap.FindCallByUniqueId(ev.m_srcUniqueId);	
	if (!pCall) {
      	// we don't care if something happens on channels we don't know
		m_callMap.__ReleaseMapLock();
		m_log.LogTrace(" no exist any call with srcUniqueId %s", ev.m_srcUniqueId.c_str());
      	return;
	}

	AsteriskCall *pDestCall = m_callMap.GetOrCreateCall(ev.m_destUniqueId); // register new destUniqueId call created by Dial

	pDestCall->Link(pCall);			// links original call to new one
	m_log.LogTrace("  linked calls {%s} <- {%s}", pCall->GetUniqueId().c_str(), pDestCall->GetUniqueId().c_str() );

	// update calleridname
	DWORD callId = -1;
	if (!AsteriskCall::IsUnknown(ev.m_callerIdName)){
		const std::string userData = ContainsUserData(ev.m_callerIdName);
		if (userData != ""){
			//if exist userData, set it into Asterisk Call
			pDestCall->SetUserData(userData);
			m_log.LogTrace("  {%s}.%s has userData=%s", pDestCall->GetUniqueId().c_str(), pDestCall->GetAddress().c_str(), pDestCall->GetUserData().c_str());
			pCall->SetUserData(userData);
			m_log.LogTrace("  {%s}.%s has userData=%s", pCall->GetUniqueId().c_str(), pCall->GetAddress().c_str(), pCall->GetUserData().c_str());
		}
		callId = ContainsCallId(ev.m_callerIdName);

		if (callId != -1){
			DWORD lastCallId = pCall->GetCallId();
			m_log.LogTrace("  overriding {%s}.%d to new callId %d", pCall->GetUniqueId().c_str(), lastCallId, callId);
			if (lastCallId != callId){
				m_callMap.DropCallId(lastCallId);
                pCall->SetCallId(callId);
			}
			lastCallId = pDestCall->GetCallId();
			m_log.LogTrace("  overriding {%s}.%d to new callId %d", pDestCall->GetUniqueId().c_str(), lastCallId, callId);
			if (lastCallId != callId){
				m_callMap.DropCallId(lastCallId);
                pDestCall->SetCallId(callId);
			}
		}

		if (ContainsPredictiveCall(ev.m_callerIdName)){
			pCall->SetPredictiveCall();
			pDestCall->SetPredictiveCall();
		}

		pDestCall->SetCallerIdName(ev.m_callerIdName);
		pCall->SetCallerIdName(ev.m_callerIdName);
	}

	//security net in case of ev.m_callerIdName incorrect format
	if (callId == -1){
		//try to find the callid by searching the logicCallId
		m_log.LogTrace("Dial event don't contain callid, we try to get it using GetLogicCallId");
		DWORD	logicCallId = m_callMap.GetLogicCallId(pCall);
		
		if (logicCallId != -1){
			BOOL bApplyLogicCall=TRUE;
			m_log.LogTrace("Call logicCallId(): %d",logicCallId);
			
			/*
			Don't link if the destination device has fired an alerting event and the iax2 security net was not applied (alerting -> minimization rule 4)
			*/
			Set_xref set_Monitors;
			/*pCall->GetCallerId() != pDestCall->GetCallerId() complements alerting -> minimization rule 4 for iax2*/
			if(pCall->GetCallerId() != pDestCall->GetCallerId())
			{
				if(logicCallId!=pDestCall->GetCallId() && GetMonitorSet(pDestCall->GetCallId(), set_Monitors) && set_Monitors.size() > 0 )
				{
					for (Set_xref::iterator i = set_Monitors.begin(); i!=set_Monitors.end(); i++) {
						if (m_callMap.GetStateOfDevice(pDestCall->GetCallId(),m_callMap.GetDeviceFromMonitor(*i))==csta_states::ALERTING)
						{
							/*The call was monitorized and the alerting was fired*/
							bApplyLogicCall=FALSE;
							pDestCall->Unlink();
						}
					}
				}
			}

			AsteriskCall* logicCall = m_callMap.FindCallByCallId(logicCallId);
			if(bApplyLogicCall)
			{
				DWORD lastCallId = pCall->GetCallId();
				if (lastCallId != logicCallId){
					m_log.LogTrace("  overriding {%s}.%d to new callId %d", pCall->GetUniqueId().c_str(), lastCallId, logicCallId);
					m_callMap.DropCallId(lastCallId);
					//if call is used in the CSTA call state map delete it
					m_callMap.RemoveCallFromCallModel(lastCallId);
					//update callid
					pCall->SetCallId(logicCallId);
				}
				lastCallId = pDestCall->GetCallId();
				if (lastCallId != logicCallId){
					m_log.LogTrace("  overriding {%s}.%d to new callId %d", pDestCall->GetUniqueId().c_str(), lastCallId, logicCallId);
					m_callMap.DropCallId(lastCallId);
					//if call is used in the CSTA call state map delete it
					m_callMap.RemoveCallFromCallModel(lastCallId);
					//update callid
					pDestCall->SetCallId(logicCallId);
				}
				if (logicCall && logicCall->IsPredictiveCall()){
					m_log.LogTrace("Call is predictive");
					pCall->SetPredictiveCall();
					pDestCall->SetPredictiveCall();
				}
			}
			//WARN this callerid name haven't correct format
			m_log.LogTrace("overriding {%s}. to new calleridname %s",pDestCall->GetUniqueId().c_str(),ev.m_callerIdName.c_str()); 
			pDestCall->SetCallerIdName(ev.m_callerIdName);
			m_log.LogTrace("overriding {%s}. to new calleridname %s",pCall->GetUniqueId().c_str(),ev.m_callerIdName.c_str()); 
			pCall->SetCallerIdName(ev.m_callerIdName);
		}
	}

	/* update lastRedirectingDevice */
	pDestCall->SetLastRedirectionDevice(pCall->GetLastRedirectionDevice());

	/**
	 *	/Fix: Asterisk 1.4 compatibility 
	 *	update source callerid 
	 */

	if (ev.m_callerId!="") {
		pCall->SetCallerId(ev.m_callerId);
		m_log.LogTrace("overriding {%s}. to new callerid %s",pCall->GetUniqueId().c_str(),pCall->GetCallerId().c_str()); 	
	}

	/* if source address and destination address are diferent than callerId, we must search a call with the same 
	 * callId and callerId to link them*/
	if ((pCall->GetAddress() == pDestCall->GetAddress()) && pCall->GetAddress() != ev.m_callerId){
		/**
		*	\bug: if a new call is incoming into an ACD, fails because don't exist another call with the same callid 
		*	and callerid
		*/
		AsteriskCall *pAux = m_callMap.FindCallByCallIdAndDevice(pCall->GetCallId(),ev.m_callerId);
		pCall = m_callMap.GetSourceCall(pCall);

		pCall->Link(pAux);
	}

	ConnectionAlerting	xConnectionAlerting;
	if (pDestCall->GetState() == call_state_ringing){
		m_log.LogTrace("  state = RINGING => Send ALERTING to {%s}.address=%s", pDestCall->GetUniqueId().c_str(), pDestCall->GetAddress().c_str());
		bool	bSend = handleConnectionAlerting(pDestCall, ast_cause_normal, xConnectionAlerting);
		m_callMap.__ReleaseMapLock();	// release lock before callback
		
		if (bSend) 
			fireOnAlertingCallback(xConnectionAlerting);
		
		return ;
	}

	m_callMap.__ReleaseMapLock();
}

/**
*	Two channels got linked. 
*	This event is received when the audio channels are connected (too late). 
*	For this reason it can't be used to link calls
*	\param ev
*/
void	ASTCstaProvider::OnLinkEvent(ManagerEvent ev)
{
	m_log.LogTrace("OnLinkEvent:CFT: ev.m_srcUniqueId=\"%s\"; ev.m_destUniqueId=\"%s\"; m_pxCstaProvider->OnLinkEvent(ev);", ev.m_srcUniqueId.c_str(), ev.m_destUniqueId.c_str());

	// protect map access with mutex
	m_callMap.__WaitForMapLock();
	AsteriskCall *pCall = m_callMap.FindCallByUniqueId(ev.m_srcUniqueId);	
	if (!pCall) {
      	// we don't care if something happens on channels we don't know
		m_callMap.__ReleaseMapLock();
		m_log.LogTrace(" no exist any call with srcUniqueId %s", ev.m_srcUniqueId.c_str());
      	return;
	}

	AsteriskCall *pDestCall = m_callMap.FindCallByUniqueId(ev.m_destUniqueId); // register new destUniqueId call created by Dial

	if (!pDestCall){
		// we don't care if something happens on channels we don't know
		m_callMap.__ReleaseMapLock();
		m_log.LogTrace("  destUniqueId {%s} NOT FOUND in map", ev.m_destUniqueId.c_str());
		return;
	}

	/* link/unlink are received too late/early in the event-flow and they are not used to link calls. 
	Only dial event and local channel master/outgoing legs relationship are used to relate calls to each other 
	If destCall hasn't got sourceUniqueId, we can set a CSTALink
	*/
	if (!pDestCall->HasSourceUniqueId()){
		pDestCall->LinkCSTACall(pCall->GetUniqueId());
	}

	// do not update callerid because it may not be correct from asterisk

	// update lastRedirectingDevice
	pDestCall->SetLastRedirectionDevice(pCall->GetLastRedirectionDevice());
	m_log.LogTrace("  set {%s}.lastRedirectionDevice=%s", pDestCall->GetUniqueId().c_str(), pDestCall->GetLastRedirectionDevice().c_str());

	m_callMap.__ReleaseMapLock();
}


/**
*	Channels link was  broken
*	This event is received when the audio channels are broken (too early).
*	For this reason this event cannot be used to unlink calls.
*	\param ev
*/
void	ASTCstaProvider::OnUnlinkEvent(ManagerEvent ev)
{
}


/**
*	Hangup event
*	\param ev
*/
void	ASTCstaProvider::OnHangup(ManagerEvent ev)
{
	m_log.LogTrace("OnHangup:CFT: ev.m_uniqueId=\"%s\"; ev.m_cause=\"%s\"; m_pxCstaProvider->OnHangupEvent(ev);", ev.m_uniqueId.c_str(), ev.m_cause.c_str());

	// protect map access with mutex
	m_callMap.__WaitForMapLock();

	AsteriskCall *pCall = m_callMap.FindCallByUniqueId(ev.m_uniqueId);
	if (!pCall) {
		m_callMap.__ReleaseMapLock();
		m_log.LogTrace(" no exist any call with uniqueId %s", ev.m_uniqueId.c_str());
      	return;
	}

	DWORD	dwCallId = pCall->GetCallId();
	AsteriskCall	*pSourceAC = m_callMap.FindCallByUniqueId( pCall->GetSourceUniqueId() );

	DWORD	logicCallId = m_callMap.GetLogicCallId(pCall);

	csta_states	eStatus;
	if (pSourceAC) {
		eStatus = m_callMap.GetStateOfDevice(logicCallId, pSourceAC->GetAddress());
	}
	else {
		eStatus = m_callMap.GetStateOfDevice(logicCallId, pCall->GetAddress());
	}

	call_cause eCause;
	if (pCall->GetCause() == ast_cause_normal || pCall->GetCause() == ast_cause_unknown) {
		eCause = Str2CallCause(ev.m_cause);
		pCall->SetCause(eCause); 
	}
	else {
		eCause = pCall->GetCause();
	}

	bool	bSend = false;
	ConnectionDisconnected	xConnectionDisconnected;
	if(eStatus == FAILED ){
		m_log.LogTrace("  handleConnectionDisconnected {%s} to %s with status %d and cause %d", pCall->GetUniqueId().c_str(), pCall->GetAddress().c_str(), eStatus, eCause);
		if (pSourceAC) {
			bSend = handleConnectionDisconnected(pCall, pSourceAC->GetCause(), xConnectionDisconnected);
		}
		else {
			bSend = handleConnectionDisconnected(pCall, pCall->GetCause(), xConnectionDisconnected);
		}
	} else if (!pCall->HasSourceUniqueId()){
		m_log.LogTrace("  handleConnectionDisconnected {%s} to %s", pCall->GetUniqueId().c_str(), pCall->GetAddress().c_str());
		bSend =  handleConnectionDisconnected(pCall, eCause, xConnectionDisconnected);
		// continue		
	} else if(m_callMap.Size() > 1){
		m_log.LogTrace(" callMap.size == %d", m_callMap.Size());
		m_log.LogTrace("  handleConnectionDisconnected {%s} to %s", pCall->GetUniqueId().c_str(), pCall->GetAddress().c_str());
		bSend = handleConnectionDisconnected(pCall, Str2CallCause(ev.m_cause), xConnectionDisconnected);
	}

	const std::string	sAddress = pCall->GetAddress();
	const std::string	sHangupCallUniqueId = pCall->GetUniqueId();
	std::string	sRootSourceUniqueId = "";
	if (pSourceAC)
		sRootSourceUniqueId = pSourceAC->GetUniqueId();

	m_callMap.__ReleaseMapLock();	// release callmap lock before call back

	if (bSend) {
		fireOnConnectionClearedCallback(xConnectionDisconnected);

		// can't RemoveConnectionFromCallModel() before fireOnConnectionClearedCallback() because nothing would get fired...
		// also: remove it conditionally
		m_callMap.RemoveConnectionFromCallModel(dwCallId, sAddress);
	}
	
	// continue		

	// transmit callid to the following in call-linked-list
	bSend = false;
	CallDisconnected	xCallDisconnected;

	m_callMap.__WaitForMapLock();

	Set_uniqueId	setUniqueid;
	m_callMap.GetDestinationCallSet(sHangupCallUniqueId, setUniqueid);
	for (Set_uniqueId::iterator i = setUniqueid.begin(); i!=setUniqueid.end(); i++) {
		std::string	uniqueId = *i;
		AsteriskCall	*pDestAC = m_callMap.FindCallByUniqueId(uniqueId);
		AsteriskCall	*pRootSourceAC = m_callMap.FindCallByUniqueId(sRootSourceUniqueId);
		if(!pDestAC) {
			m_log.LogTrace(" pCall for destination %s not found", uniqueId.c_str());
			continue;
		}
		if (pDestAC->HasCSTALink()){
			m_log.LogTrace(" call %s has a CSTALink. Don't transfer callid or any other parameters", pDestAC->GetUniqueId().c_str());
			continue;
		}

		DWORD	dwPreviousCallId = pDestAC->GetCallId();
		m_log.LogTrace("  transferring callid and other parameters to its destination asteriskCall in the 'call walk path' currently destination call {%s}.callid = %d", i->c_str(), dwPreviousCallId);

		pDestAC->SetCallId(dwCallId);
		if (eCause!=ast_cause_normal && eCause!=ast_cause_unknown)
			pDestAC->SetCause(eCause);

		pDestAC->Unlink();
		// link to previous!
		if (pDestAC) {
			pDestAC->Link(pRootSourceAC);	// pRootSourceAC can be NULL	
		}

		m_log.LogTrace("    now destination call {%s}.callid = %d and sourceUniqueId=''", i->c_str(), pDestAC->GetCallId());
		if (!m_callMap.FindCallByCallId(dwPreviousCallId)) {	// was last one
			m_log.LogTrace("  handleCallDisconnected {%d} due to call-map adjustment", dwPreviousCallId);
			
			// pRootSourceAC can be null
			bSend = handleCallDisconnected(pRootSourceAC, dwPreviousCallId, Str2CallCause(ev.m_cause), xCallDisconnected);
			m_callMap.RemoveCallFromCallModel(dwPreviousCallId);
			m_log.LogTrace("  removed entry with callId=%d from MAP due to call-map adjustment", dwCallId);
			m_callMap.DropCallId(dwPreviousCallId);
		}
	}

	m_log.LogTrace("  removed uniqueId=%s call from MAP", ev.m_uniqueId.c_str());
	m_callMap.RemoveCallByUniqueId(ev.m_uniqueId);

	// free memory
	delete pCall;
	pCall = NULL;

	m_callMap.__ReleaseMapLock();		// unlock

	if (bSend) 
		fireOnCallClearedCallback(xCallDisconnected);

	if (!m_callMap.FindCallByCallId(dwCallId)) {	// was last one
		m_log.LogTrace("  handleCallDisconnected {%d} was the last one", dwCallId);
		
		//handleCallDisconnected(pCall,dwCallId, normal);
		CallDisconnected	xLastCallDisconnected;
		xLastCallDisconnected.m_dwCallId = dwCallId;
		xLastCallDisconnected.m_eCause = eCause;//ast_cause_normal;
		fireOnCallClearedCallback(xLastCallDisconnected);
		m_callMap.RemoveCallFromCallModel(dwCallId);		// this already calls m_callMap.DropCallId(dwCallId);
		m_log.LogTrace("  removed entry with callId=%d from MAP", dwCallId);
	}
}


/** 
*	CallerId event is generated if the CallerId is set by the extensions configuration
*	\param ev
*/
void	ASTCstaProvider::OnNewCallerIdEvent(ManagerEvent ev)
{
	m_log.LogTrace("OnNewCallerIdEvent:CFT: ev.m_uniqueId=\"%s\"; ev.m_callerId=\"%s\"; ev.m_callerIdName=\"%s\"; m_pxCstaProvider->OnNewCallerIdEvent(ev);", ev.m_uniqueId.c_str(), ev.m_callerId.c_str(), ev.m_callerIdName.c_str());

	m_callMap.__WaitForMapLock();	// lock callmap

	AsteriskCall *pCall = m_callMap.GetOrCreateCall(ev.m_uniqueId);
	if (pCall->GetCallerId() == ""){
		m_log.LogTrace("  overriding {%s}.%s to new callerId %s", pCall->GetUniqueId().c_str(), pCall->GetCallerId().c_str(), ev.m_callerId.c_str());
		pCall->SetCallerId(ev.m_callerId);
	}

	if (!AsteriskCall::IsUnknown(ev.m_callerIdName)) {	// after a dial has updated callerid/calleridname on destUniqueid call,
											// a new newcalleridevent to dest overrides again and 
											// sets it to <Unknown>, so to try to avoid that we only update if 
											// newcalleridname is nor unknown
		std::string userData;
		if ((userData = ContainsUserData(ev.m_callerIdName)) != ""){
			//if userData exists, update it into the AsteriskCall
			pCall->SetUserData(userData);
			m_log.LogTrace("  {%s}.%s has userData=%s", pCall->GetUniqueId().c_str(), pCall->GetAddress().c_str(), pCall->GetUserData().c_str());
		}

		DWORD callId = ContainsCallId(ev.m_callerIdName);
		if (callId != -1){
			DWORD lastCallId = pCall->GetCallId();
			m_log.LogTrace("  overriding {%s}.%d to new callId %d", pCall->GetUniqueId().c_str(), lastCallId, callId);
			if (lastCallId != callId){
				m_callMap.DropCallId(lastCallId);
                pCall->SetCallId(callId);
			}
		}

		if (ContainsPredictiveCall(ev.m_callerIdName)) {
			pCall->SetPredictiveCall();
		}

		if (ev.m_callerIdName.size() != 0){
			m_log.LogTrace("  overriding {%s}.%s to new calleridname %s", pCall->GetUniqueId().c_str(), pCall->GetCallerIdName().c_str(), ev.m_callerIdName.c_str());
			pCall->SetCallerIdName(ev.m_callerIdName);
		}
	}

	m_callMap.__ReleaseMapLock();	// release lock
	return;
}

/**
*	NewExten event used for dialplan. Currently not used
*	\param ev
*/
void	ASTCstaProvider::OnNewExten(ManagerEvent ev) 
{
	// do nothing
	return;
}

/**
*	NewChannel event
*	\param ev
*/
void	ASTCstaProvider::OnNewChannelEvent(ManagerEvent ev)
{
	m_log.LogTrace("OnNewChannelEvent:CFT: ev.m_uniqueId=\"%s\"; ev.m_callerId=\"%s\"; ev.m_callerIdName=\"%s\"; ev.m_state=\"%s\"; ev.m_channel=\"%s\"; m_pxCstaProvider->OnNewChannelEvent(ev);", ev.m_uniqueId.c_str(), ev.m_callerId.c_str(), ev.m_callerIdName.c_str(), ev.m_state.c_str(), ev.m_channel.c_str());

	// Sometimes asterisk sends newchannelevent with empty "channel" data
	// in this case it will never get hang so a callmap leak occurs
	// so we just ignore this crazy events
	if (ev.m_channel.size()==0) {
		m_log.LogTrace("  empty channel!");
		return;
	}

	m_callMap.__WaitForMapLock();		// lock call map	

	AsteriskCall *pCall = m_callMap.GetOrCreateCall(ev.m_uniqueId);

	// populate some call info
	if (ev.m_callerId.size()>0 && !AsteriskCall::IsUnknown(ev.m_callerId)) {
		pCall->SetCallerId(ev.m_callerId);
	}

	if (!AsteriskCall::IsUnknown(ev.m_callerIdName)) {
		const std::string userData = ContainsUserData(ev.m_callerIdName);
		if (userData != ""){
			//if exist userData, set it into Asterisk Call
			pCall->SetUserData(userData);
			m_log.LogTrace("  {%s}.%s has userData=%s", pCall->GetUniqueId().c_str(), pCall->GetAddress().c_str(), pCall->GetUserData().c_str());
		}
		DWORD callId = ContainsCallId(ev.m_callerIdName);
		if (callId != -1 && ev.m_state != "Up"){
			DWORD lastCallId = pCall->GetCallId();
			m_log.LogTrace("  overriding {%s}.%d to new callId %d", pCall->GetUniqueId().c_str(), lastCallId, callId);
			if (lastCallId != callId){
				m_callMap.DropCallId(lastCallId);
                pCall->SetCallId(callId);
			}
		}
		if (ContainsPredictiveCall(ev.m_callerIdName))
			pCall->SetPredictiveCall();

		pCall->SetCallerIdName(ev.m_callerIdName);
	}
	
	m_log.LogTrace("  setting {%s}.callerID=%s, callerIdName=%s", pCall->GetUniqueId().c_str(), pCall->GetCallerId().c_str(), pCall->GetCallerIdName().c_str());

	if (ev.m_state.size()>0) {
		pCall->SetState(ev.m_state);
	}

	if (ev.m_channel.size()>0) {
		std::string	channel = ev.m_channel;

		pCall->SetChannel(channel);
		m_log.LogTrace("  setting {%s}.channel=%s, address=%s", pCall->GetUniqueId().c_str(), channel.c_str(), pCall->GetAddress().c_str());

		// if local channel type we may need to link master->outbound_channel
		/* ,1 => master_channel / ,2 => outbound_channel */ 
		if (pCall->IsLocalOutboundChannel()) {		// ,2
			std::string master_channel = AsteriskCall::getMasterChannelFromOutboundChannel(channel);
			if (master_channel!="")
			{
				AsteriskCall	*pMaster = m_callMap.FindCallByChannel(master_channel);
				if (pMaster) {
					pCall->Link(pMaster);	// master -> outbound
					m_log.LogTrace(" link call {%s} with call {%s}", pCall->GetUniqueId().c_str(), pCall->GetSourceUniqueId().c_str());
				}
			}
		} // isLocalOutbound
		else if (pCall->IsLocalMasterChannel()) {	// ,1
			std::string outbound_channel = AsteriskCall::getOutboundChannelFromMasterChannel(channel);
			if(outbound_channel!="")
			{
				AsteriskCall	*pOutbound = m_callMap.FindCallByChannel(outbound_channel);
				if (pOutbound) {
					pOutbound->Link(pCall);	// master -> outbound
					m_log.LogTrace(" link call {%s} with call {%s}", pOutbound->GetUniqueId().c_str(), pOutbound->GetSourceUniqueId().c_str());
				}
			}
		} // isLocalOutbound
	}
	
	// we get address from channel instead of using event.
	// callerid, which is sometimes wrong in newchannel/newstate astreisk events
	const std::string	address = AsteriskCall::GetAddressFromChannel(ev.m_channel);	

	if (ev.m_state.compare("Ring")==0) {		// Ring 
		m_log.LogTrace("  state = RING => Send ORIGINATED to {%s}.address=%s", pCall->GetUniqueId().c_str(), address.c_str());
		ConnectionOriginated	xConnectionOriginated;
		bool bSend = handleConnectionOriginated(pCall, ast_cause_normal, xConnectionOriginated);

		m_callMap.__ReleaseMapLock();	// release call map lock

		if (bSend) 
			fireOnOriginatedCallback(xConnectionOriginated);
		
		return;
	} 
	else if (ev.m_state.compare("Ringing")==0) {
		m_log.LogTrace("  state = RINGING => Send ALERTING to {%s}.address=%s", pCall->GetUniqueId().c_str(), address.c_str());
		ConnectionAlerting xConnectionAlerting;
		bool bSend = handleConnectionAlerting(pCall, ast_cause_normal, xConnectionAlerting);
		
		m_callMap.__ReleaseMapLock();	// release call map lock

		if (bSend) 
			fireOnAlertingCallback(xConnectionAlerting);
		
		return;
    }
	else if (ev.m_state.compare("Up")==0) {	
		m_log.LogTrace("  state = UP => Send ESTABLISHED to {%s}.address=%s", pCall->GetUniqueId().c_str(), address.c_str());
		EncapsulateChannel(pCall);
		ConnectionConnected	xConnectionConnected;
		bool bSend = handleConnectionConnected(pCall, ast_cause_normal, xConnectionConnected);
		
		m_callMap.__ReleaseMapLock();	// release call map lock

		if (bSend) 
			fireOnEstablishedCallback(xConnectionConnected);
		
		return;
    } 
	else if (ev.m_state.compare("Down")==0) {	
		m_log.LogTrace("  state = DOWN => do nothing");
    } 
	else if (ev.m_state.compare("Busy")==0) {
		m_log.LogTrace("  state = BUSY => do nothing");
    } 

	m_callMap.__ReleaseMapLock();	// release call map lock
}

/**
*	Rename event
*	\param ev
*/
void	ASTCstaProvider::OnRenameEvent(ManagerEvent ev)
{
	m_log.LogTrace("OnRenameEvent:CFT: ev.m_uniqueId=\"%s\"; ev.m_oldname=\"%s\"; ev.m_newname=\"%s\"; m_pxCstaProvider->OnRenameEvent(ev);", ev.m_uniqueId.c_str(), ev.m_oldname.c_str(), ev.m_newname.c_str());

	m_callMap.__WaitForMapLock();	// protect map access with mutex
	AsteriskCall	*pAC = m_callMap.FindCallByUniqueId(ev.m_uniqueId);
	if (!pAC) {
		// in transferred calls, rename events has a uniqueId that not exist
		m_log.LogTrace(" creating a new AsteriskCall with uniqueId %s", ev.m_uniqueId.c_str());
		pAC = m_callMap.GetOrCreateCall(ev.m_uniqueId);
		pAC->SetChannel(ev.m_oldname);
		const std::string address = AsteriskCall::GetAddressFromChannel(ev.m_oldname);
		AsteriskCall *pAux = m_callMap.FindCallByChannel(ev.m_newname);
		if (!pAux){
			m_log.LogTrace(" not found any call with channel %s",ev.m_newname.c_str());
			m_callMap.__ReleaseMapLock();
			return;
		}
		m_log.LogTrace(" found call %s",pAux->GetUniqueId().c_str());
		pAC->SetAddress(address);
		pAC->SetState(pAux->GetState());
		pAC->SetCallerId(pAux->GetCallerId());
		pAC->SetUserData(pAux->GetUserData());
		if (pAux->IsPredictiveCall()){
			pAC->SetPredictiveCall();
		}
		pAC->Link(pAux);
	}

	ConnectionDisconnected	xConnectionDisconnected;
	bool	bSend = false;
	if (!pAC->IsLocalChannel() && pAC->IsPredictiveCall() ){
		/// \todo: when the support of transferences will be included, should be send a transfer event.
		;//bSend = handleConnectionDisconnected(pAC, ast_cause_normal, xConnectionDisconnected);
	}

	m_log.LogTrace("  rename channel [%s] = > [%s]", pAC->GetChannel().c_str(), ev.m_newname.c_str());
	pAC->SetChannel(ev.m_newname) ;	// do not update pAC->address to maintain it inmutable

	m_callMap.__ReleaseMapLock();

	if (bSend) 
		fireOnConnectionClearedCallback(xConnectionDisconnected);
}

/**
*	NewState Event
*	\param ev
*/
void	ASTCstaProvider::OnNewStateEvent(ManagerEvent ev)
{
	m_log.LogTrace("OnNewStateEvent:CFT: ev.m_uniqueId=\"%s\"; ev.m_callerIdName=\"%s\"; ev.m_state=\"%s\"; ev.m_channel=\"%s\"; m_pxCstaProvider->OnNewStateEvent(ev);", ev.m_uniqueId.c_str(), ev.m_callerIdName.c_str(), ev.m_state.c_str(), ev.m_channel.c_str());

	m_callMap.__WaitForMapLock();	//lock call map
	AsteriskCall *pCall = m_callMap.GetOrCreateCall(ev.m_uniqueId);

	// populate some call info
	if (!AsteriskCall::IsUnknown(ev.m_callerIdName)){
		std::string userData;
		if ((userData = ContainsUserData(ev.m_callerIdName)) != ""){
			//if exist userData, set it into Asterisk Call
			pCall->SetUserData(userData);
			m_log.LogTrace("  {%s}.%s has userData=%s", pCall->GetUniqueId().c_str(), pCall->GetAddress().c_str(), pCall->GetUserData().c_str());
		}

		DWORD callId = ContainsCallId(ev.m_callerIdName);
		if (callId != -1 && ev.m_state != "Up"){
			DWORD lastCallId = pCall->GetCallId();
			m_log.LogTrace("  overriding {%s}.%d to new callId %d", pCall->GetUniqueId().c_str(), lastCallId, callId);
			if (lastCallId != callId){
				m_callMap.DropCallId(lastCallId);
                pCall->SetCallId(callId);
			}
		}

		/*
			With iax-trunk experimentaly we deduce that:
			1.- Newstate and dial event arrives with inverted order, and we need the minimization rule 4 on newstate event, and dial will fire the alerting
			2.- OnNewState the call is not linked pCall->GetSourceUniqueId()!="" and we need to don't touch predictive false for minimizaton rule4
		*/
		if (ContainsPredictiveCall(ev.m_callerIdName) && pCall->GetSourceUniqueId()!="")
			pCall->SetPredictiveCall();

		pCall->SetCallerIdName(ev.m_callerIdName);
		m_log.LogTrace("  setting {%s}.callerID=%s, callerIdName=%s", pCall->GetUniqueId().c_str(), pCall->GetCallerId().c_str(), pCall->GetCallerIdName().c_str());
	}

	// connectedLineNum/Name The AMI Newstate event contains different information between v1.4 and v1.8
	// since 1.8.5.0 The ConnectedLineNum and ConnectedLineName headers were added to many AMI events/responses 
	// if the CallerIDNum/CallerIDName headers were also present.
	if(!AsteriskCall::IsUnknown(ev.m_connectedLineNum))
	{
		pCall->SetConnectedLineNum(ev.m_connectedLineNum);
		m_log.LogTrace("  setting {%s} ConnectedLineNum=%s", pCall->GetUniqueId().c_str(), pCall->GetConnectedLineNum().c_str());
	}
	if(!AsteriskCall::IsUnknown(ev.m_connectedLineName))
	{
		pCall->SetConnectedLineName(ev.m_connectedLineName);
		m_log.LogTrace("  setting {%s} ConnectedLineName=%s", pCall->GetUniqueId().c_str(), pCall->GetConnectedLineName().c_str());
		std::string userData;
		if ((userData = ContainsUserData(ev.m_connectedLineName)) != ""){
			//if exist userData, set it into Asterisk Call
			pCall->SetUserData(userData);
			m_log.LogTrace("  {%s}.%s has userData=%s", pCall->GetUniqueId().c_str(), pCall->GetAddress().c_str(), pCall->GetUserData().c_str());
		}
	}

	if (ev.m_state.size()>0) {
		pCall->SetState(ev.m_state);
	}

	// pAC->address should be inmutable so do not update *unless void* .
	// [ 1560547 ] A network loss disables ActivaTSP communication to asterisk.
	if (ev.m_channel.size()>0 && pCall->GetChannel()=="") {
		std::string	channel = ev.m_channel;

		pCall->SetChannel(channel);
		m_log.LogTrace("  setting {%s}.channel=%s, address=%s", pCall->GetUniqueId().c_str(), channel.c_str(), pCall->GetAddress().c_str());
	}

	// we get callerid from channel instead of using event. 
	// Callerid, wich is sometimes wrong in newchannel/newstate astreisk events
	std::string	address = AsteriskCall::GetAddressFromChannel(ev.m_channel);	

	if (ev.m_state.compare("Ring")==0) {		// Ring 
		m_log.LogTrace("  state = RING => Send ORIGINATED to {%s}.address=%s", pCall->GetUniqueId().c_str(), address.c_str());
		ConnectionOriginated	xConnectionOriginated;
		bool bSend = handleConnectionOriginated(pCall, ast_cause_normal, xConnectionOriginated);
		
		m_callMap.__ReleaseMapLock();	// release call map lock
		if (bSend) 
			fireOnOriginatedCallback(xConnectionOriginated);
		
		return;
	} 
	else if (ev.m_state.compare("Ringing")==0) {
		m_log.LogTrace("  state = RINGING => Send ALERTING to {%s}.address=%s", pCall->GetUniqueId().c_str(), address.c_str());
		
		ConnectionAlerting xConnectionAlerting;
		bool bSend = handleConnectionAlerting(pCall, ast_cause_normal, xConnectionAlerting);
		
		m_callMap.__ReleaseMapLock();	// release call map lock
		if (bSend) 
			fireOnAlertingCallback(xConnectionAlerting);
		
		return;
    }
	else if (ev.m_state.compare("Up")==0) {	
		m_log.LogTrace("  state = UP => Send ESTABLISHED to {%s}.address=%s", pCall->GetUniqueId().c_str(), address.c_str());
		EncapsulateChannel(pCall);
		ConnectionConnected xConnectionConnected;
		bool bSend = handleConnectionConnected(pCall, ast_cause_normal, xConnectionConnected);
		
		m_callMap.__ReleaseMapLock();	// release call map lock
		if (bSend) 
			fireOnEstablishedCallback(xConnectionConnected);
		
		return;
    }
	else if (ev.m_state.compare("Down")==0) {	
		m_log.LogTrace("  state = DOWN => do nothing");
    } 
	else if (ev.m_state.compare("Busy")==0) {
		m_log.LogTrace("  state = BUSY => do nothing");
    } 
	else if (ev.m_state.compare("Dialing") == 0){
		m_log.LogTrace("  state = DIALING => do nothing");
	}

	m_callMap.__ReleaseMapLock();	// release map lock
	return;
}

/**
*	Join Event: used when a call enter on a queue
*	HEADER: event DATA: Join
*	HEADER: privilege DATA: call,all
*	HEADER: channel DATA: mISDN/1-1
*	HEADER: callerid DATA: 317
*	HEADER: calleridname DATA: unknown
*	HEADER: queue DATA: 555
*	HEADER: position DATA: 1
*	HEADER: count DATA: 1
*	\param ev
*/
void	ASTCstaProvider::OnJoinEvent(ManagerEvent ev)
{
	m_log.LogTrace("OnJoinEvent:CFT: ev.m_channel=\"%s\"; ev.m_queue=\"%s\"; ev.m_position=\"%s\"; ev.m_count=\"%s\"; m_pxCstaProvider->OnJoinEvent(ev);", ev.m_channel.c_str(), ev.m_queue.c_str(), ev.m_position.c_str(), ev.m_count.c_str());

	m_callMap.__WaitForMapLock();

	AsteriskCall	*pAC = m_callMap.FindCallByChannel(ev.m_channel);
	if(!pAC) {
		m_callMap.__ReleaseMapLock();
		m_log.LogTrace(" no exist any call with channel %s", ev.m_channel.c_str());
		return;
	}

	pAC->SetLastRedirectionDevice(ev.m_queue);
	m_log.LogTrace("  set {%s}.lastRedirectionDevice=%s", pAC->GetUniqueId().c_str(), pAC->GetLastRedirectionDevice().c_str());
	ConnectionQueued	xConnectionQueued;
	bool	bSend = handleConnectionQueued(pAC, ev.m_queue, atoi(ev.m_position.c_str()), atoi(ev.m_count.c_str()), ast_cause_normal, xConnectionQueued);
	
	m_callMap.__ReleaseMapLock();

	if (bSend)
		fireOnQueuedCallback(xConnectionQueued);
}

/**
*	OnLeave event: Used when a call leave a queue
*	\param ev
*/
void	ASTCstaProvider::OnLeaveEvent(ManagerEvent ev)
{
	m_log.LogTrace("OnLeaveEvent:CFT: ev.m_channel=\"%s\"; ev.m_queue=\"%s\"; m_pxCstaProvider->OnLeaveEvent(ev);", ev.m_channel.c_str(), ev.m_queue.c_str() );

	m_callMap.__WaitForMapLock();

	AsteriskCall	*pAC = m_callMap.FindCallByChannel(ev.m_channel);
	if (!pAC){
		m_callMap.__ReleaseMapLock();
		m_log.LogTrace(" no exist any call with channel %s", ev.m_channel.c_str());
		return;
	}

	const std::string	sUniqueId = pAC->GetUniqueId();

	if (pAC->GetCause() == ast_cause_abandon_call){
		m_callMap.__ReleaseMapLock();
		m_log.LogTrace(" the call is leaved by the user");
		return;
	}

	// if it has an AsteriskCall with different address, this call is leaving to an agent,
	// else the call is leaved by the user.
	Set_uniqueId	setUniqueid;
	m_callMap.GetDestinationCallSet(pAC->GetUniqueId(),setUniqueid);
	m_log.LogTrace(" %d calls has sourceUniqueId %s",setUniqueid.size(), pAC->GetUniqueId().c_str());

	// if call is not a predictive call, there are no destination calls and the call is forwarded to an agent anyway
	if (setUniqueid.size() == 0){		// set.size == 0
		m_log.LogTrace(" the call is forwarded to an agent");
		const std::string	address = pAC->GetLastRedirectionDevice() ;
		///m_log.LogTrace("ASTCstaProvider::handleConnectionDequeued uniqueId=%s, address=%s, queue=%s", pCall->GetUniqueId().c_str(), address.c_str(), queue.c_str());

		AsteriskCall	*pACSource = m_callMap.GetSourceCall(pAC);
		DWORD dwCallId = m_callMap.GetLogicCallId(pAC);

		// RULE-9: Do not send EVENTS to local channels except for localCalls with monitors
		if (pAC->IsLocalChannel() && (!pACSource->IsPredictiveCall() || m_callMap.IsLocalCall(pAC->GetCallId()))) {
			/*
				for (Set_xref::iterator i = set_Monitors.begin(); i!=set_Monitors.end(); i++) {
					m_log.LogTrace("    sending CALLCLEARED to xref=%d, {%s}.%s", *i, pCall->GetUniqueId().c_str(), address.c_str());
					m_cstaListener->OnCallCleared(*i,dwCallId,"",cause);
				}
				m_callMap.RemoveConnectionFromCallModel(dwCallId, queue);	
			*/
			CallDisconnected xCallDisconnected;
			xCallDisconnected.m_dwCallId = dwCallId;
			xCallDisconnected.m_eCause = ast_cause_normal;

			m_callMap.__ReleaseMapLock();	// unlock 
			
			//fire always
			fireOnCallClearedCallback(xCallDisconnected);

			// remove *after* firing
			m_callMap.RemoveConnectionFromCallModel(dwCallId, ev.m_queue);	
			return;
		}
		else {
			ConnectionDiverted	xConnectionDiverted;
			bool bSend = handleConnectionDiverted(pAC, ev.m_queue, ast_cause_normal, xConnectionDiverted);
			
			m_callMap.__ReleaseMapLock();	// unlock 

			if (bSend) 
				fireOnDivertedCallback(xConnectionDiverted);
			// remove *after* firing
			m_callMap.RemoveConnectionFromCallModel(dwCallId, ev.m_queue);	
			return;
		}
		
		return;
	} // if set.size == 0
	else if (setUniqueid.size() == 1){ // set.size == 1
		m_callMap.__ReleaseMapLock();	// unlock 
		m_log.LogTrace(" the call is leaved by the user");
		return;
	}
	
	const std::string	address = pAC->GetAddress();
	m_callMap.__ReleaseMapLock();	// unlock 

	// set.size>1
	for (Set_uniqueId::iterator i = setUniqueid.begin(); i != setUniqueid.end(); i++){
		std::string	uniqueId = *i;

		m_callMap.__WaitForMapLock();

		AsteriskCall	*pDestCall = m_callMap.FindCallByUniqueId(uniqueId);
		AsteriskCall	*pAC2 = m_callMap.FindCallByUniqueId(sUniqueId);
		if (pAC2 && pDestCall && pDestCall->GetAddress() != address){
			// this call is leaving to an agent
			m_log.LogTrace(" the call is forwarded to an agent");

			AsteriskCall	*pACSource = m_callMap.GetSourceCall(pAC2);
			DWORD dwCallId = m_callMap.GetLogicCallId(pAC2);

			// RULE-9: Do not send EVENTS to local channels except for localCalls with monitors
			if (pAC2->IsLocalChannel() && (!pACSource->IsPredictiveCall() || m_callMap.IsLocalCall(pAC2->GetCallId()))) {
				CallDisconnected xCallDisconnected;
				xCallDisconnected.m_dwCallId = dwCallId;
				xCallDisconnected.m_eCause = ast_cause_normal;
				
				m_callMap.__ReleaseMapLock();	// unlock 

				fireOnCallClearedCallback(xCallDisconnected);	// fire always
				
				// if we removed this call before firing, it would fire NOTHING!!!!
				// also: remove always
				m_callMap.RemoveConnectionFromCallModel(dwCallId, ev.m_queue);		
			}
			else {
				ConnectionDiverted	xConnectionDiverted;
				bool bSend = handleConnectionDiverted(pAC2, ev.m_queue, ast_cause_normal, xConnectionDiverted);
				m_callMap.__ReleaseMapLock();	// unlock 

				if (bSend) 
					fireOnDivertedCallback(xConnectionDiverted);
				
				// if we removed this call before firing, it would fire NOTHING!!!!
				// also: remove always
				m_callMap.RemoveConnectionFromCallModel(dwCallId, ev.m_queue);	
			}
			// map is unlocked here
		}
		else {
			m_callMap.__ReleaseMapLock();	// unlock 
		}
	}
}

/**
*	UnparkedCall event: used when a call leave a park extension
*	FIX: if the call before to be parked had a CSTALink, the call after unparked will have it too
*/
void	ASTCstaProvider::OnUnparkedCallEvent(ManagerEvent ev)
{
	m_log.LogTrace("OnUnparkedCallEvent:CFT: ev.m_channel=\"%s\"; ev.m_from=\"%s\"; m_pxCstaProvider->OnUnparkedCallEvent(ev);", ev.m_channel.c_str(), ev.m_from.c_str() );

	m_callMap.__WaitForMapLock();

	AsteriskCall *pACPark = m_callMap.FindCallByChannel(ev.m_channel);
	if (!pACPark){
		m_callMap.__ReleaseMapLock();
		m_log.LogTrace(" no exist any call with channel %s", ev.m_channel.c_str());
		return;
	}
	m_log.LogTrace(" CallPark has uniqueId %s", pACPark->GetUniqueId().c_str());
	AsteriskCall *pACFrom = m_callMap.FindCallByChannel(ev.m_from);
	if (!pACFrom){
		m_callMap.__ReleaseMapLock();
		m_log.LogTrace(" don't find fromcall");
		return;
	}

	m_log.LogTrace(" CallFrom has uniqueId %s", pACFrom->GetUniqueId().c_str());

	/// it is possible that they have been sent events to a new call. To correct it, we send a hangup of pACFrom and then link
	///	the two calls.

	ConnectionDisconnected	xConnectionDisconnected;
	bool	bSend = false;
	if (!pACFrom->HasSourceUniqueId()){
		m_log.LogTrace(" Send a hangup to %s to adjust callMaps",pACFrom->GetUniqueId().c_str());
		bSend = handleConnectionDisconnected(pACFrom, ast_cause_normal, xConnectionDisconnected);
	}

	pACFrom->Link(pACPark);
	m_log.LogTrace(" Link fromCall{%s} has sourceUniqueId parkedCall{%s}",pACFrom->GetUniqueId().c_str(), pACFrom->GetSourceUniqueId().c_str());

	m_callMap.__ReleaseMapLock();

	if (bSend) 
		fireOnConnectionClearedCallback(xConnectionDisconnected);
}

/**
*	OriginateFailure event. It's produced when a callback fails
*/
void	ASTCstaProvider::OnOriginateFailureEvent(ManagerEvent ev)
{
	m_log.LogTrace("OnOriginateFailureEvent:CFT: ev.m_channel=\"%s\"; ev.m_callerIdName=\"%s\"; ev.m_reason=\"%s\"; m_pxCstaProvider->OnOriginateFailureEvent(ev);", ev.m_channel.c_str(), ev.m_callerIdName.c_str(), ev.m_reason.c_str() );

	m_callMap.__WaitForMapLock();
	
	AsteriskCall *pCall = NULL;

	//First search by channel. Works in Ast1.4
	int CallId = this->ContainsCallId(ev.m_callerIdName);
	if (CallId!=-1)
		pCall = m_callMap.FindCallByCallId(CallId);

	//Work on both 1.4 1.2 but less precise, one channel can have more than one active call.
	//-2 to delete /n
	if(!pCall)
		pCall = m_callMap.FindCallByChannel( ev.m_channel.substr(0,ev.m_channel.size()-2) );

	ConnectionFailed	xConnectionFailed;
	bool	bSend = false;
	if (pCall){
		pCall->SetCause(Str2OriginateFailure(ev.m_reason));
		m_log.LogTrace(" Call {%s} has cause %d",pCall->GetUniqueId().c_str(), pCall->GetCause());
		bSend = handleConnectionFailed(pCall,Str2OriginateFailure(ev.m_reason), xConnectionFailed);
	} else {
		m_log.LogTrace(" Don't exist any call with CallId %d or channel %s", CallId, ev.m_channel.c_str());
	}

	m_callMap.__ReleaseMapLock();		// release lock

	if (bSend)
		fireOnFailedCallback(xConnectionFailed);
}

/**
*	QueueCallerAbandon event. It's produce in Asterisk 1.4 when a caller abandon a queue
*/
void ASTCstaProvider::OnQueueCallerAbandonEvent(ManagerEvent ev)
{
	m_log.LogTrace("OnQueueCallerAbandonEvent:CFT: ev.m_uniqueId=\"%s\"; m_pxCstaProvider->OnQueueCallerAbandonEvent(ev);", ev.m_uniqueId.c_str() );

	m_callMap.__WaitForMapLock();

	AsteriskCall *pCall = m_callMap.FindCallByUniqueId(ev.m_uniqueId);
	if (pCall){
		pCall->SetCause(ast_cause_abandon_call);
	} else {
		m_log.LogTrace(" no exist any call with uniqueId %s", ev.m_uniqueId.c_str());
	}
	m_callMap.__ReleaseMapLock();
}

/**
*	OnHold event. It's produce when a user press Hold button in phone
*/
void ASTCstaProvider::OnHoldEvent(ManagerEvent ev)
{
	m_log.LogTrace("OnHoldEvent:CFT: ev.m_uniqueId=\"%s\"; m_pxCstaProvider->OnHoldEvent(ev);", ev.m_uniqueId.c_str() );

	m_callMap.__WaitForMapLock();

	AsteriskCall *pCall = m_callMap.FindCallByUniqueId(ev.m_uniqueId);
	if (!pCall){
		m_callMap.__ReleaseMapLock();
		m_log.LogTrace(" no exist any call with uniqueId %s", ev.m_uniqueId.c_str());
		return;
	}
	ConnectionHeld	xConnectionHeld;
	bool bSend = handleConnectionHeld(pCall, ast_cause_held, xConnectionHeld);
	
	m_callMap.__ReleaseMapLock();

	if(bSend)
		fireOnHeldCallback(xConnectionHeld);
}

/**
*	OnUnhold event. It's produced when a user press unhold button in phone
*/
void	ASTCstaProvider::OnUnholdEvent(ManagerEvent ev)
{
	m_log.LogTrace("OnUnholdEvent:CFT: ev.m_uniqueId=\"%s\"; m_pxCstaProvider->OnUnholdEvent(ev);", ev.m_uniqueId.c_str() );

	m_callMap.__WaitForMapLock();

	AsteriskCall *pCall = m_callMap.FindCallByUniqueId(ev.m_uniqueId);
	if (!pCall){
		m_callMap.__ReleaseMapLock();
		m_log.LogTrace(" no exist any call with uniqueId %s", ev.m_uniqueId.c_str());
		return;
	}
	ConnectionHeld	xConnectionHeld;
	bool bSend = handleConnectionUnhold(pCall, ast_cause_normal, xConnectionHeld);
	
	m_callMap.__ReleaseMapLock();
	
	if (bSend)
		fireOnRetrievedCallback(xConnectionHeld);
}

/**
*	Find if pCall encapsulate another channel and link AsteriskCalls
*	\param pCall
*/
void	ASTCstaProvider::EncapsulateChannel(AsteriskCall *pCall)
{
	// when called this function is called the map CS is always already locked so we don't need to lock again
	std::string	channel = pCall->GetChannel();
	size_t extPos = -1;

	const std::string asyncgoto="AsyncGoto/";
	const std::string transfered="Transfered/";

	if ((extPos=channel.find(asyncgoto)) != -1) {	// if contains AsyncGoto
		std::string chanOrg = channel.substr(extPos+asyncgoto.size(), channel.size());
		m_log.LogTrace("Channel %s has an encapsulate channel %s with ASYNCGOTO prefix", pCall->GetChannel().c_str(), chanOrg.c_str());
		AsteriskCall *pACorg = m_callMap.FindCallByChannel(chanOrg);
		if (!pACorg){
			m_log.LogError("Asterisk Call not found");
			return;
		}
		AsteriskCall *pSourceCall;
		if (pACorg->HasSourceUniqueId()){
			pSourceCall = m_callMap.GetSourceCall(pACorg);
		}else{
			pSourceCall = pACorg;
		}

		/// Links AsyncGoto channel with it's encapsulate channel
		m_log.LogTrace("Link pCall %s with pSourceCall %s", pCall->GetSourceUniqueId().c_str(),pACorg->GetSourceUniqueId().c_str());
		pCall->Link(pACorg);
		pCall->SetCallerId(pACorg->GetCallerId());
		pCall->SetLastRedirectionDevice(pACorg->GetLastRedirectionDevice());
		if(pACorg->IsPredictiveCall()){
			pCall->SetPredictiveCall();
		}
		pCall->SetUserData(pACorg->GetUserData());
		pCall->SetCallerIdName(pACorg->GetCallerIdName());
		return;
	}
	if ((extPos=channel.find(transfered))!= -1){
		std::string chanOrg = channel.substr(extPos+transfered.size(), channel.size());
		m_log.LogInfo("Channel %s has an encapsulate channel %s with TRANSFERED prefix", pCall->GetChannel().c_str(), chanOrg.c_str());
		return;
	}
}

/**
*	Find if callerIdName contains UserData
*	\param callerIdName
*	\return userData
*/
std::string ASTCstaProvider::ContainsUserData(std::string &callerIdName)
{
	std::string userData = "";

	size_t pos = -1;
	std::string capUserData = "#U=";
	if ((pos = callerIdName.find(capUserData))!= -1){
		std::string aux = callerIdName.substr(pos+capUserData.size(),callerIdName.size());
		callerIdName = callerIdName.substr(0,pos);
		pos = -1;
		if ((pos = aux.find("#")) != -1){
			userData = aux.substr(0,pos);
			callerIdName = callerIdName + aux.substr(pos,aux.size());
		}else{
			userData = aux;
		}
		m_log.LogTrace("Call Contains UserData(%s)",userData.c_str());
	}else{
		m_log.LogTrace("Call does not contains UserData");
	}
	return userData;
}

/**
*	Find if callerIdName contains callId
*	\param callerIdName
*	\return callId
*/
DWORD ASTCstaProvider::ContainsCallId(std::string &callerIdName)
{
	DWORD callid = -1;

	size_t pos = -1;
	std::string capCallId = "#Id=";
	if ((pos = callerIdName.find(capCallId))!= -1){
		std::string aux = callerIdName.substr(pos+capCallId.size(),callerIdName.size());
		callerIdName = callerIdName.substr(0,pos);
		pos = -1;
		if ((pos = aux.find("#")) != -1){
			callid = atoi(aux.substr(0,pos).c_str());
			callerIdName = callerIdName + aux.substr(pos,aux.size());
		}else{
			callid = atoi(aux.c_str());
		}
		m_log.LogTrace("Call containsCallId(%d)",callid);
	}else{
		m_log.LogTrace("Call does not containsCallId");
	}

	return callid;
}

/**
*	Find if callerIdName contains the string #D
*/
bool ASTCstaProvider::ContainsPredictiveCall(std::string &callerIdName)
{
	bool dialer = false;

	size_t pos = -1;
	std::string capDialer = "#D";
	if ((pos = callerIdName.find(ToString(capDialer)))!= -1){
		std::string aux = callerIdName.substr(pos+capDialer.size(),callerIdName.size());
		callerIdName = callerIdName.substr(0,pos);
		pos = -1;
		if ((pos = aux.find("#")) != -1)
			callerIdName = callerIdName + aux.substr(pos,aux.size());
		dialer = true;
		m_log.LogTrace("Call is predictive");
	}else{
		m_log.LogTrace("Call is NOT predictive");
	}
	return dialer;
}

std::string ASTCstaProvider::GetCSTALink(std::string sourceUniqueId)
{
	size_t pos = -1;
	std::string csta = "CSTA";
	if ((pos = sourceUniqueId.find(csta)) != -1){
		return sourceUniqueId.substr(csta.size(), sourceUniqueId.size());
	}
	return sourceUniqueId;
}
/**
*	Converts hangup cause to call_cause enum
*	\param causeStr
*	\return call_cause
*/
call_cause ASTCstaProvider::Str2CallCause(std::string causeStr)
{
	int causeInt = atoi(causeStr.c_str());
	switch(causeInt){
		case 16:	return ast_cause_normal;
		case 1:		return ast_cause_unallocated;
		case 2:		return ast_cause_new_call;
		case 27:	return ast_cause_out_of_order;
		case 0:		return ast_cause_unknown;		
		case 5:		
		case 17:	
			return ast_cause_user_busy;
		case 4:		return ast_cause_active;
		case 19:	return ast_cause_alerting_noanswer;
		case 3:		return ast_cause_unreachable;	//alerting_noanswer;
		case 21:	return ast_cause_temp_failure;
		case 38:	return ast_cause_reject;
		case 203:	return ast_cause_answering_machine;
		case 204:	return ast_cause_modem_fax;
		case 34: 
		case 42:	return ast_cause_congestion;
		default: return ast_cause_unknown;
	}
}

/**
*	Converts OriginateFailure cause to call_cause enum
*	\param causeStr
*	\return call_cause
*/
call_cause ASTCstaProvider::Str2OriginateFailure(std::string causeStr)
{
	int causeInt = atoi(causeStr.c_str());
	switch(causeInt){
		case 3: return ast_cause_alerting_noanswer;
		case 5: return ast_cause_user_busy;
		case 4: return ast_cause_active;
		default: return ast_cause_unknown;
	}
}
