/*  
	Name: astcstaprovider.h 
	Copyright: Under the Mozilla Public License Version 1.1 or later
	Author: ACTIVA
*/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is astprovider.
 *
 * The Initial Developer of the Original Code is ACTIVA.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): none
 *
 * ***** END LICENSE BLOCK ***** */


/*! \mainpage
 * Activa brings the Asterisk IP PBX to the call center. 
 * 
 * This is the Activa framework: a set of C++ classes that enables integration of third party CTI applications to Asterisk.
 * There is a basic C++ class abstraction level and a ASTProviderTest application that shows how the framework can be used.
 * A working TAPI Service Provider is also included.
 *
 * Main Activa framework classes are ASTProvider, ASTCstaProvider and ASTTapiProvider.
 * 
 * ASTCstaProvider class provides a csta-like API that is aimed to help you to develop CTI apps on top of Asterisk. 
 * It implements some methods and fires csta-like events through the ASTCstaListener interface.
 *
 * ASTProvider is a thin c++ wrapper on top os Asterisk Management Interface (AMI), intended to 
 * serve the needs of upper layer Activa framework classes such as ASTCstaProvider. It also fires AMI events through the ASTListener interface.
 *
 * The TAPI Service Provider implementation is based on ASTTapiProvider.
 *
 * About the Activa Project: Built on top of Asterisk, Activa components enable successful call center implementations adding value in areas such as computer telephony, screenpop&click2dial, agent control, automatic dialing...   
 * 
 * http://activa.sourceforge.net
 *
 *   The Activa team
 * 
 */
   
#pragma once

#ifndef CSTAMODEL_H
#define CSTAMODEL_H

#include <string>
#include <map>
#include <vector>
#include <set>

#include "astprovider.h"
#include "call_map.h"

/// helper structure
class	ConnectionAlerting
{
public:
	DWORD		m_dwCallId;
	std::string	m_sAddress;
	std::string	m_sCallerId;
	std::string	m_sCallerIdName;
	std::string	m_sCalled;
	std::string	m_sLastRedirection;
	std::string	m_sUserData;
	call_cause	m_eCause;

	ConnectionAlerting() { m_dwCallId=-1; m_eCause = ast_cause_normal;};
};

/// helper structure
class	ConnectionOriginated
{
public:
	DWORD		m_dwCallId;
	std::string	m_sAddress;
	std::string	m_sCallerId;
	std::string	m_sCallerIdName;
	std::string	m_sCalled;
	call_cause	m_eCause;

	ConnectionOriginated() { m_dwCallId=-1; m_eCause = ast_cause_normal;};
};

/// helper structure
class	ConnectionConnected
{
public:
	DWORD		m_dwCallId;
	std::string	m_sAddress;
	std::string	m_sCallerId;
	std::string	m_sCallerIdName;
	std::string	m_sCalled;
	std::string	m_sLastRedirection;
	std::string	m_sUserData;
	call_cause	m_eCause;

	ConnectionConnected() { m_dwCallId=-1; m_eCause = ast_cause_normal;};
};

/// helper structure
class	ConnectionDisconnected
{
public:
	DWORD		m_dwCallId;
	std::string	m_sAddress;
	call_cause	m_eCause;

	ConnectionDisconnected() { m_dwCallId=-1; m_eCause = ast_cause_normal;};
};


/// helper structure
class	CallDisconnected
{
public:
	DWORD		m_dwCallId;
	std::string	m_sAddress;
	call_cause	m_eCause;

	CallDisconnected() { m_dwCallId=-1; m_eCause = ast_cause_normal;};
};


/// helper structure
class	ConnectionFailed
{
public:
	DWORD		m_dwCallId;
	std::string	m_sAddress;
	std::string	m_sCallerId ;
	std::string	m_sCallerIdName ;
	std::string	m_sCalled ;
	std::string	m_sLastRedirection ;
	call_cause	m_eCause;

	ConnectionFailed() { m_dwCallId=-1; m_eCause = ast_cause_normal;};
};

/// helper structure
class	ConnectionQueued
{
public:
	DWORD		m_dwCallId;
	std::string	m_sAddress;
	std::string	m_sCallerId ;
	std::string	m_sCallerIdName ;
	std::string	m_sCalled ;
	std::string	m_sQueue;
	int			m_nPosition;
	int			m_nCount;
	std::string	m_sUserData;
	call_cause	m_eCause;

	ConnectionQueued() { m_dwCallId=-1; m_eCause = ast_cause_normal;};
};


/// helper structure
class	ConnectionDiverted
{
public:
	DWORD		m_dwCallId;
	std::string	m_sAddress;
	std::string	m_sQueue;
	std::string	m_sCallerId ;
	std::string	m_sCallerIdName ;
	std::string	m_sCalled ;
	std::string	m_sUserData;
	call_cause	m_eCause;

	ConnectionDiverted() { m_dwCallId=-1; m_eCause = ast_cause_normal;};
};

/// helper structure
class	ConnectionHeld
{
public:
	DWORD		m_dwCallId;
	std::string	m_sAddress;
	std::string	m_sCallerId ;
	std::string	m_sCallerIdName ;
	std::string	m_sCalled ;
	std::string	m_sLastRedirection ;
	call_cause	m_eCause;

	ConnectionHeld() { m_dwCallId=-1; m_eCause = ast_cause_normal;};
};


/**
 * This is the main class of Activa framework. 
 * Provides a csta-like programming interface.
 * To create an instance you can call the Create() factory method 
 * Events are reported through a listener class implementing ASTCstaListener interface.
 * A custom logger can be attached via ASTLogListener
 */

/// This is the main class of Activa framework. Provides a csta-like programming interface.
class ASTCstaProvider : public ASTCstaProviderInterface 
{
protected:
		ASTCstaProvider();
public:
	/// factory method
	static ASTCstaProvider* Create(ASTCstaListener	*listener, ASTLogListener *logger);
		~ASTCstaProvider();

	/// Connect to Asterisk 
	AC_RESULT	Connect(std::string	sHost, int nPort, std::string user, std::string pwd, AUTH_TYPE eAuthType = at_md5);

	/// Disconnect from Asterisk
	bool		Disconnect();

	/// Opens a monitor to a device. Monitors are refererd by an opaque int handler known as monitor xref 
	/// (monitor cross-reference)
	MonitorXref	OpenMonitor(std::string	deviceId);

	/// Close a previously opened monitor
	bool		CloseMonitor(MonitorXref xref);

	/// Originates a call from callingDevice to calleddevice using 'context'.
	AC_RESULT	MakeCall(std::string callingDevice, std::string contextCalling, std::string calledDevice, DWORD &callid, std::string contextCalled, std::string callerIdName, std::string userData, std::string variable, bool bLockReaderAfterResponse = false);

	///	Originate a predictive call from callingDevice to calledDevice using 'context'
	AC_RESULT	MakePredictiveCall(std::string callingDevice, std::string contextCalling, std::string calledDevice, std::string contextCalled, DWORD &callid, std::string callerId, std::string userData, DWORD timeout, std::string variable, bool bLockReaderAfterResponse = false);

	/// Disconnects a call identified by uniqueid/callid
	AC_RESULT	ClearCall(DWORD callId);

	/// Disconnects a Connection identified by callId and device)
	AC_RESULT	ClearConnection(DWORD callid, std::string address);

	/// Redirects a call to an extsnsion
	AC_RESULT	FastTransfer(DWORD callId, std::string transferringDevice, std::string destination, std::string destinationContext);

	/// Sets Do-not-disturb flag on/off
	AC_RESULT	SetDoNotDisturb(std::string deviceId, bool bDND_on, std::string dndContext);

	/// Set agentState
	AC_RESULT	SetAgentStatus(std::string deviceId, std::string queue, AGENT_STATUS newStatus, std::string agentContext);

	// Hold a call
	AC_RESULT	HoldCall(DWORD callId, std::string holdingDevice, std::string holdContext);

	// Retrieve call
	AC_RESULT	RetrieveCall(DWORD callId, std::string retreivingDevice, std::string retrieveContext);
	
	/// Start Recording
	AC_RESULT	StartRecording(DWORD callid, std::string address, std::string RecordingFileName, std::string format, std::string mix);

	/// Stop Recording
	AC_RESULT	StopRecording(DWORD callid, std::string address);

	/// Pause Recording
	AC_RESULT	PauseRecording(DWORD callid, std::string address);

	/// Unpause Recording
	AC_RESULT	UnpauseRecording(DWORD callid, std::string address);

	/// Change Recording File Name
	AC_RESULT	ChangeRecordingFileName(DWORD callid, std::string address, std::string NewRecordingFileName);

	/**
	*	\return New Callid
	*	This method is implemented ONLY to allow ASTTapiProvider::TSPI_lineMakeCall() to get a real unique callid before Originate(). 
	*/
	DWORD		GetNewCallId() { return m_callMap.GetNewCallId(); };

	/**
	*	\return result
	*	This method is implemented ONLY to allow ASTTapiProvider::TSPI_lineMakeCall() to drop a idCall previously acquired through GetCallId()
	*/
	bool		ASTCstaProvider::DropCallId(DWORD idCall) { return m_callMap.DropCallId( idCall ); };

	/// returns the monitor set linked to the given callid
	bool	GetMonitorSet(DWORD callId, Set_xref & setXref) {
		return m_callMap.GetMonitorSetFromCallModel(callId, setXref);
	};
	
	/**
	*	UnLocks Reader Thread
	*
	**/
	void		UnLockReader() { if (m_astProvider!=NULL) m_astProvider->UnLockReader();};

	/// test helpers methods
#ifdef	__ACTIVATSP_CPPUNIT_BUILD
	AsteriskCall	*GetAsteriskCall(std::string uniqueId) { return m_callMap.FindCallByUniqueId(uniqueId);};
	size_t			GetAsteriskNumberOfCallsInCallMap() { return m_callMap.Size();};
	csta_states		GetCSTAStateOfDeviceInCallMap(DWORD callid,std::string device) {return m_callMap.GetStateOfDevice(callid,device);};
	ASTProvider		*GetAstProvider(){return(m_astProvider);};
	size_t			GetAsteriskNumberOfCallsInCSTAMap() { return m_callMap.CSTAStateMapSize();};
#endif

private:
	ASTLogger		m_log;
	ASTProvider		*m_astProvider;
	ASTCstaListener	*m_cstaListener;
	MonitorXref		m_currentXref;

	/// callmap
	ASTCallMap	m_callMap ;

	/// If pCall has encapsulateChannel, it links both
	void	EncapsulateChannel(AsteriskCall *pCall);

	void	fireOnAlertingCallback(ConnectionAlerting &xConnectionAlerting);
	void	fireOnOriginatedCallback(ConnectionOriginated &xConnectionOriginated);
	void	fireOnEstablishedCallback(ConnectionConnected &xConnectionConnected);
	void	fireOnConnectionClearedCallback(ConnectionDisconnected &xConnectionDisconnected);
	void	fireOnCallClearedCallback(CallDisconnected &xCallDisconnected);
	void	fireOnFailedCallback(ConnectionFailed &xConnectionFailed);
	void	fireOnQueuedCallback(ConnectionQueued &xConnectionQueued);
	void	fireOnDivertedCallback(ConnectionDiverted &xConnectionDiverted);
	void	fireOnHeldCallback(ConnectionHeld &xConnectionHeld);
	void	fireOnRetrievedCallback(ConnectionHeld &xConnectionHeld);

	/// handlers
	bool	handleConnectionAlerting (AsteriskCall *pCall, call_cause cause, ConnectionAlerting &xConnectionAlerting);
	bool	handleConnectionConnected (AsteriskCall *pCall, call_cause cause, ConnectionConnected &xConnectionConnected);
	bool	handleConnectionOriginated (AsteriskCall *pCall, call_cause cause, ConnectionOriginated &xConnectionOriginated);
	bool	handleConnectionDisconnected (AsteriskCall *pCall, call_cause cause, ConnectionDisconnected &xConnectionDisconnected);
	bool	handleCallDisconnected (AsteriskCall *pCall, DWORD dwCallId, call_cause cause, CallDisconnected &xCallDisconnected);
	bool	handleConnectionFailed (AsteriskCall *pCall, call_cause cause, ConnectionFailed &xConnectionFailed);
	bool	handleConnectionQueued (AsteriskCall *pCall, std::string queue, int position, int count, call_cause cause, ConnectionQueued &xConnectionQueued);
	bool	handleConnectionDiverted (AsteriskCall *pCall, std::string queue, call_cause cause, ConnectionDiverted &xConnectionDiverted);
	bool	handleConnectionHeld(AsteriskCall* pCall,call_cause cause, ConnectionHeld &xConnectionHeld);
	bool	handleConnectionUnhold(AsteriskCall *pCall, call_cause cause, ConnectionHeld &xConnectionHeld);

	/// to find if exist UserData and CallId
	std::string ContainsUserData(std::string &callerIdName);
	DWORD ContainsCallId(std::string &callerIdName);
	bool ContainsPredictiveCall(std::string &callerIdName);
	std::string GetCSTALink(std::string sourceUniqueId);

	/// Converts string to call_cause
	call_cause Str2CallCause(std::string causeStr);
	call_cause Str2OriginateFailure(std::string causeStr);

// ASTListener manager api interface
public:
	/// OnNewStateEvent handler 
	void		OnNewStateEvent(ManagerEvent ev);
	/// OnNewCallerIdEvent handler 
	void		OnNewCallerIdEvent(ManagerEvent ev);
	/// OnNewExten handler 
	void		OnNewExten(ManagerEvent ev) ;
	/// OnChannelEvent handler 
	void		OnChannelEvent(ManagerEvent ev);
	/// OnHangup handler 
	void		OnHangup(ManagerEvent ev);
	/// OnNewChannelEvent handler 
	void		OnNewChannelEvent(ManagerEvent ev);
	/// OnDialEvent handler 
	void		OnDialEvent(ManagerEvent ev);
	/// OnRenameEvent handler 
	void		OnRenameEvent(ManagerEvent ev);
	/// OnLinkEvent handler 
	void		OnLinkEvent(ManagerEvent ev);
	/// OnUnLinkEvent handler 
	void		OnUnlinkEvent(ManagerEvent ev);
	/// OnJoinEvent handler 
	void		OnJoinEvent(ManagerEvent ev);
	/// OnLeaveEvent handler 
	void		OnLeaveEvent(ManagerEvent ev);
	/// OnLinkDown handler
	void		OnLinkDown();
	/// OnUnparked handler
	void		OnUnparkedCallEvent(ManagerEvent ev);
	/// OnOriginateFailure handler
	void		OnOriginateFailureEvent(ManagerEvent ev);
	/// OnQueueCallerAbandon handler Asterisk 1.4
	void		OnQueueCallerAbandonEvent(ManagerEvent ev);
	/// OnHold event
	void		OnHoldEvent(ManagerEvent ev);
	/// OnUnhold event
	void		OnUnholdEvent(ManagerEvent ev);

	friend class ASTTapiProvider;
	friend class ActivaTSP_Test_Adapter;
};

#endif	// CSTAMODEL_H
