/*
	Name: astinterface.h
	Copyright: Under the Mozilla Public License Version 1.1 or later
	Author: ACTIVA
	Date: MAY/06
*/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is astprovider.
 *
 * The Initial Developer of the Original Code is ACTIVA.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): none
 *
 * ***** END LICENSE BLOCK ***** */

#pragma once

#ifndef ASTINTERFACE_H
#define ASTINTERFACE_H

#include <string>
#include <map>
#include <sstream>

#include "astlogger.h"
#include <stdlib.h>
#include <algorithm>	// transform

typedef enum {
	call_state_not_known = 0,
	call_state_ringing = 1,
	call_state_ring = 2,
	call_state_down = 3
} activa_call_state;

/**
 * Manager version
 */
typedef enum {
	manager_v_unsupported = 0,
	manager_v1_0 = 1,
	manager_v1_1 = 2,
} MANAGER_VERSION;

typedef	void*	Opaque_call_reference;

/// Internal representation of asterisk calls
class	AsteriskCall {
private:
	std::string m_uniqueId;				/// connection uniqueId
	std::string m_callerid;				/// callerId <- asterisk channel.callerid
	std::string m_calleridname;			/// callerIdName <- asterisk channel.calleridname
	std::string m_connectedLineNum;		/// connectedLineId <- asterisk channel.connectedlineid
	std::string m_connectedLineName;	/// connectedLineIdName <- asterisk channel.connectedlineidname
	std::string m_channel;				/// channel
	std::string m_numDevice;			/// dialable device
	std::string m_address;				/// address
	activa_call_state	m_state;				/// as reported in newchannel and newstate asterisk events
	std::string	m_sourceUniqueId;		/// back-links to source call, if there is one
	std::string	m_lastRedirectingDevice;	/// last redirecting device
	DWORD		m_callId;				/// callId
	std::string m_userData;				/// userData
	bool		m_predictive;			/// true if call was making by a dialer (makepredictivecall)
	call_cause	m_cause;				/// cause of originatefailure

	AsteriskCall();

public:
	/// factory method
	static AsteriskCall*	Create(std::string uniqueId, std::string callerId, DWORD callId, std::string userData);

	/// Destruct an asterisk call
	~AsteriskCall(){};

	/// gets uniqueid attribute
	std::string		GetUniqueId() { return m_uniqueId; };	

	/**
	*	Sets associated address. We differentiate between address and callerid since callerid may change. 
	*	Usually read 'address' parsing the channel description.
	*/
	void			SetAddress(std::string address) {m_address = address;};	
	
	/**
	*	gets associated address. We differentiate between address and callerid since callerid may change. 
	*	Usually read 'address' parsing the channel description.
	*/
	std::string		GetAddress() {return m_address;};

	std::string		GetNumDevice() {return m_numDevice;};

	/// set AsteriskCall callerId
	void			SetCallerId(std::string callerid){m_callerid = callerid;};		
	
	/// sets callerId 
	std::string		GetCallerId() { return m_callerid; }	
	
	/**
	*	That function is currently not used
	*	return true if callerid attribute is different than ""
	*/
	bool			HasCallerId() { return (m_callerid.size()>0); }	

	/// set AsteriskCall callerIdName
	void			SetCallerIdName(std::string calleridname) {m_calleridname=calleridname;};		
	
	/// sets callerIdName 
	std::string		GetCallerIdName() { return m_calleridname; }

	/// sets connectedLineId 
	std::string		GetConnectedLineNum() { return m_connectedLineNum; }

	/// set AsteriskCall connectedLineId
	void			SetConnectedLineNum(std::string connectedLineNum){m_connectedLineNum = connectedLineNum;};		
	
	/// sets connectedLineId 
	std::string		GetConnectedLineName() { return m_connectedLineName; }

	/// set AsteriskCall connectedLineId
	void			SetConnectedLineName(std::string connectedLineName){m_connectedLineName = connectedLineName;};	
	
	/**
	*	That function is currently not used.
	*	return true if calleridname attribute is different than ""
	*/
	bool			HasCallerIdName() { return (m_calleridname.size()>0); }	

	/// gets associated asterisk channel name (ie: local/201)
	std::string		GetChannel() { return m_channel; }		
	
	/// sets associated asterisk channel name (ie: local/201)
	void			SetChannel(std::string newChannel);	

	/// gets state 
	activa_call_state	GetState() { return m_state; };			
		
	/// gets state string
	std::string			GetStateString() { switch(m_state) {
											case (call_state_ring): return "Ring"; break;
											case (call_state_ringing): return "Ringing"; break;
											case (call_state_down): return "Down"; break;
											case (call_state_not_known): return "not known"; break;
											default: return "?";} };			

	/// sets state 
	void			SetState(std::string newState) { 
		// tolower comparisons
		std::transform(newState.begin(), newState.end(), newState.begin(), tolower);
		if (newState=="ring") {
			m_state = call_state_ring;
		} else if (newState=="ringing") {
			m_state = call_state_ringing;
		} else if (newState=="down") {
			m_state = call_state_down;
		} else { m_state = call_state_not_known; }
	};	

	void			SetState(activa_call_state eNewState) { m_state=eNewState; }; 

	/// links this call to a destination call 
	void			Link(AsteriskCall * pDestCall);			
	
	/// unlinks this call from a destination call
	void			Unlink(){m_sourceUniqueId = "";};	

	/// links two different CSTA Calls
	void			LinkCSTACall(std::string cstaUniqueId){m_sourceUniqueId = "CSTA"+cstaUniqueId;};

	/// return true if exist a CSTA link
	bool			HasCSTALink();

	///	set userData 
	void			SetUserData(std::string userData)	{m_userData = userData;};

	/// Get userData
	std::string		GetUserData()	{return m_userData;};

	/// \return true if AsteriskCall has a source call
	bool			HasSourceUniqueId(){return (m_sourceUniqueId.length()>0);};
	
	/// gets reference to source call 
	std::string		GetSourceUniqueId(){return m_sourceUniqueId;};					

	/// get device of the last redirection
	std::string		GetLastRedirectionDevice() { return m_lastRedirectingDevice;};
	
	/// set device of the last redirection
	void			SetLastRedirectionDevice(std::string redirecting) { m_lastRedirectingDevice=redirecting;};

	/// Gets address from channel specification
	static std::string	GetAddressFromChannel(std::string channel);

	/// Gets dialable number from Channel
	static std::string GetDialableNumberFromChannel(std::string channel);

	/// Is local channel?
	bool			IsLocalChannel();
	
	/// Is local master leg? of the form local/xxx\@yyy,1
	bool			IsLocalMasterChannel();

	/// Is local outbound leg? of the form local/xxx\@yyy,2
	bool			IsLocalOutboundChannel();

	/// getMasterChannelFromOutboundChannel
	static std::string	getMasterChannelFromOutboundChannel(std::string channel);
	
	/// getOutboundChannelFromMasterChannel
	static std::string	getOutboundChannelFromMasterChannel(std::string channel);

	/// set a new callid to call
	void			SetCallId(DWORD newCallId) {m_callId=newCallId;};
	/// return callId
	DWORD			GetCallId() {return m_callId;};

	/// parses cidname and if it has the form xxx:yyyyyy it returns  xxx
	static std::string	GetCIDNamePrefix(std::string calleridname);

	/// true if device is void or equal to \<unknown>
	static bool		AsteriskCall::IsUnknown(std::string device);

	void			Dump(ASTLogger &log);
	std::string		DumpToString();

	std::string		GetTypeChannel();

	/// return true if call is predictive
	bool			IsPredictiveCall(){return m_predictive;};
	/// set the call as predictive
	void			SetPredictiveCall(){m_predictive = true;};

	/// return the cause of call
	call_cause		GetCause(){return m_cause;};

	/// set any cause to call
	void			SetCause(call_cause cause){m_cause = cause;};

};


class	ManagerResponse
{
public:
	std::string m_actionId;
    std::string m_response;
	std::string m_challenge;
    std::string m_message;
    std::string m_uniqueId;
    std::map<std::string, std::string> m_attributes;
};



		
/*
 * ManActionKeyNames name action (not used why not needed)
 */
/*class	ManActionKeys
{
public:
	std::string m_actionId;
    std::string m_actionNameOriginate;
	std::string m_actionParam;
    
	///Establishes the Name of the parameters expected for the request
	void setAstManResponseNames(MANAGER_VERSION dManagerVersion)
	{
		switch (dManagerVersion)
		{
			case (manager_v1_0)//asterisk 1.2 and 1.4
			{
				break;
			}
			default //asterisk 1.6 or unknown
			{
				break;	
			}
		}
	};
};*/


/*
 * ManResponseKeyNames name response parameters (not used why not needed)
 */
/*
class	ManResponseKeys
{
public:
	std::string m_actionId;
    std::string m_response;
    std::string m_message;
    std::string m_uniqueId;
    
	///Establishes the Name of the parameters expected for the request
	void setAstManResponseKeys(MANAGER_VERSION dManagerVersion)
	{
		switch (dManagerVersion)
		{
			case (manager_v1_0)//asterisk 1.2 and 1.4
			{
				m_actionId	= "actionid";
				m_response	= "response";
				m_message	= "message";
				m_uniqueId	= "uniqueid";
				break;
			}
			default //asterisk 1.6 or unknown
			{
				m_actionId	= "actionid";
				m_response	= "response";
				m_message	= "message";
				m_uniqueId	= "uniqueid";
				break;
			}
		}
	};
};
*/

/*
 * ManagerEventKeyNames name event parameters
 */
class ManEventKeys
{
public:
	// asterisk EventKeys
    std::string m_uniqueId;			/// uniqueId of event EventKey
    std::string m_destUniqueId;		/// destination uniqueId of event EventKey
    std::string m_srcUniqueId;		/// source uniqueId of event EventKey
	std::string m_context;			/// context of event EventKey
    std::string m_extension;		/// extension of event EventKey
    std::string m_application;		/// application of event EventKey
    std::string m_appData;			/// application data of event EventKey
    std::string m_priority;			/// priority of event EventKey
    std::string m_channel;			/// channel of event EventKey
	std::string m_from;				/// UnparkedCall EventKey
    std::string m_oldname;			/// rename event EventKey
    std::string m_newname;			/// rename event EventKey
	std::string m_state;			/// state of event EventKey
	std::string m_callerId;			/// callerId of event EventKey
	std::string m_callerIdNum;		/// callerIdnum of event EventKey
	std::string m_callerIdName;		/// callerId name of event EventKey
	std::string	m_connectedLineNum;	/// m_connectedLineNum of event, new in asterisk 1.8.5.0
	std::string	m_connectedLineName;/// m_connectedLineName of event, new in asterisk 1.8.5.0
	std::string m_cause;			/// cause of event EventKey
	std::string m_causeTxt;			/// cause TXT of event EventKey
	std::string m_queue;			/// queue of event EventKey
	std::string m_count;			/// number of clients in queue EventKey
	std::string m_position;			/// position of client in queue EventKey
	std::string m_origposition;		/// Original position of client in queue Asterisk 1.4 EventKey
	std::string m_actionId;			/// identifier of action EventKey
	std::string m_reason;			/// reason of event EventKey
	std::string m_response;			/// response of originate Asterisk 1.4 EventKey
	std::string m_holdtime;			/// time in hold Asterisk 1.4 EventKey
	std::string m_uniqueid1;		
	std::string m_uniqueid2;
	std::string m_callerid1;
	std::string m_challenge;		/// challenge from Asterisk EventKey

	//new 1.6
	std::string m_SubEvent;			/// Dial response Begin or End
	std::string m_BridgeState;		/// values link or unlink

	void setAstManEventKeys(MANAGER_VERSION dManagerVersion)
	{
		///set default values comatible with Ast 1.4 / Ast 1.2
		m_uniqueId			= "uniqueid";	
		m_destUniqueId		= "destuniqueid";		
		m_srcUniqueId		= "srcuniqueid";	
		m_context			= "context";		
		m_extension			= "exten";		
		m_application		= "application";		
		m_appData			= "appdata";			
		m_priority			= "priority";			
		m_channel			= "channel";			
		m_from				= "from";				
		m_oldname			= "oldname";			
		m_newname			= "newname";			
		m_state				= "state";			
		m_callerId			= "callerid";	
		m_callerIdNum		= "calleridnum";
		m_callerIdName		= "calleridname";
		m_cause				= "cause";			
		m_causeTxt			= "cause-txt";			
		m_queue				= "queue";			
		m_count				= "count";			
		m_position			= "position";			
		m_origposition		= "origposition";		
		m_actionId			= "actionid";			
		m_reason			= "reason";			
		m_response			= "response";		
		m_holdtime			= "holdtime";
		m_uniqueid1			= "uniqueid1";
		m_uniqueid2			= "uniqueid2";
		m_callerid1			= "callerid1";

		m_SubEvent			= "";
		m_BridgeState		= "";
		m_connectedLineNum = "";
		m_connectedLineName =	"";

		m_challenge			= "challenge";

		switch (dManagerVersion)
		{
		case (manager_v1_0)://asterisk 1.2 and 1.4
			{
				//default values are correct
				break;
			}
		default: //asterisk 1.6 or unknown
			{
				m_srcUniqueId		= "uniqueid";			
				m_state				= "channelstatedesc";			
				m_callerId			= "calleridnum";
				m_callerIdNum		= "calleridnum";
				m_SubEvent			= "subevent";
				m_oldname			= "channel";
				m_BridgeState		= "bridgestate";

				//asterisk 1.8
				m_connectedLineNum	= "connectedlinenum";
				m_connectedLineName = "connectedlinename";
				break;
			}
		}
	};
};	

/*
 * AstManTrans Translate name of the action, events and parameters to use by the provider algorithm.
 * Helper class to add backward compatibility on the action, events and parameters to handle.
 */
class AstManTrans
{
private:
	MANAGER_VERSION m_dManagerVersion;

public:
	ManEventKeys m_Event;
	//ManActionKeys m_Action;(not used why not needed, no diferences)
	//ManResponseKeys m_Response;(not used why not needed, no diferences)

	AstManTrans(){setManagerVersion(manager_v1_0);};

	MANAGER_VERSION getManagerVersion(){return(m_dManagerVersion);};

	void setManagerVersion(MANAGER_VERSION dManagerVersion)
	{
		m_dManagerVersion=dManagerVersion;
		m_Event.setAstManEventKeys(dManagerVersion);
		//m_Action.setAstManActionKeys(m_dManagerVersion);
		//m_Response.setAstManResponseKeys(m_dManagerVersion);
	};
};
			
#endif	// ASTINTERFACE_H