/*
	Name: astprovider.cpp
	Copyright: Under the Mozilla Public License Version 1.1 or later
	Author: ACTIVA
	Date: MAY/06
*/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Activa astprovider.
 *
 * The Initial Developer of the Original Code is ACTIVA.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): none
 *
 * ***** END LICENSE BLOCK ***** */

#include <list>
#include <map>
#include <algorithm>
#include <cctype>

#include "astprovider.h"
#include "md5.h"

#define	AST_TIMEOUT	30	// seconds

/**
*	Constructor
*/
ASTProvider::ASTProvider(void)
{
	m_socket = INVALID_SOCKET;
	m_bDie = false;
	m_hReaderThread = NULL;
	m_bProcessingCommandResult=false;
	m_log.SetOrigin("ASTProvider");
}


/**
*	Factory method allows for initialization. 
*	Returns an initialized ASTProvider* instance
*	or returns NULL in case of init failure.
*	\param listener
*	\param logger
*	\return ASTProvider
*/
ASTProvider*	ASTProvider::Create( ASTListener* listener, ASTLogListener* logger ) 
{

	WORD wVersionRequested = MAKEWORD( 2, 2 );
	WSADATA wsaData;
	WSAStartup( wVersionRequested, &wsaData );

	if (listener==NULL) {
		return NULL;
	}

	ASTProvider*	pInstance = new ASTProvider();
	if (pInstance!=NULL) {
		pInstance->m_listener = listener;
		pInstance->m_hMutex = CreateMutex(NULL, FALSE, NULL);
		pInstance->m_log.SetLogger(logger);

		pInstance->m_log.LogInfo("__________________ ASTProvider::Create __________________");
		pInstance->m_log.LogInfo("*** ASTProvider::Create() ***");
	}

	return pInstance;
}

/**
*	Destructor
*/
ASTProvider::~ASTProvider(void)
{
	m_bDie = true;
	m_log.LogInfo("ASTProvider::destroy will disconnect()...");
	Disconnect();

	CloseHandle(m_hMutex);
}

/**
*	Connects to asterisk AMI port
*	\param sHost ip host address of asterisk box
*	\param nPort ip port asterisk manager api is listening to. 
*	\param user ie: admin
*	\param pwd  ie: amp111
*	\eAuthType authentification type 0-plaintext 1-md5
*	\return fails_already_connected, ok, rc, socket_error
*/
AC_RESULT	ASTProvider::Connect(std::string sHost, int nPort, std::string user, std::string pwd, AUTH_TYPE eAuthType)
{
	m_log.LogInfo("Connect(%s,%d,%s,%s)", sHost.c_str(), nPort, user.c_str(), pwd.c_str());

	if ( m_socket != INVALID_SOCKET ) {
        return fails_already_connected;
	}

	if (ConnectSocket(sHost, nPort)) {
		AC_RESULT rc = Login(user, pwd, eAuthType);
		if (rc==ok) {
			m_log.LogTrace("Connect() returns ok");
			return ok;
		}
		else {
			Disconnect();
			m_log.LogTrace("Connect() returns ko=%d",rc);
			return rc ;
		}
	}

	m_log.LogTrace("Connect() returns socket_error");
	return socket_error ;
}

/**
 *	connects socket and if success starts reader thread
 *	\param host
 *	\param port
 *	\return true if the socket is correctly connected, otherwise false
 */
bool ASTProvider::ConnectSocket(std::string host, int port)
{
	// create socket
    struct sockaddr_in sa;
    struct hostent     *hp;
	int	wsaLasterror;

	if ( m_socket != INVALID_SOCKET )
	{
		m_log.LogTrace("m_socket != INVALID_SOCKET");
        closesocket(m_socket);
        m_socket = INVALID_SOCKET;
        return false;
	}
    
    hp = gethostbyname(host.c_str());
    if (hp == NULL) // we don't know who this host is
    {
		wsaLasterror = WSAGetLastError();
		m_log.LogTrace("We don't know who this host is, gethostbyname=%d error code",wsaLasterror);
		struct in_addr  addr;
		int	len = sizeof(addr);
		hp = gethostbyaddr(host.c_str(), len, AF_INET);
		if (hp==NULL) {
			wsaLasterror = WSAGetLastError();
			m_log.LogTrace("We don't know who this host is, gethostbyaddr error code=%d",wsaLasterror);
			m_socket = INVALID_SOCKET;
			return false;
		}
    }

    memset(&sa,0,sizeof(sa));
    memcpy((char *)&sa.sin_addr, hp->h_addr, hp->h_length);   // set address 
    sa.sin_family = hp->h_addrtype;
    sa.sin_port = htons((u_short)port);

    m_socket = socket(hp->h_addrtype, SOCK_STREAM, 0);
    if (m_socket == INVALID_SOCKET)
    {
        m_socket = INVALID_SOCKET;
        return false;
    }

    // try to connect to the specified socket 
    if (connect(m_socket, (struct sockaddr *)&sa, sizeof sa) == SOCKET_ERROR) 
    {
		wsaLasterror = WSAGetLastError();
		m_log.LogTrace("Connect fails, error code=%d",wsaLasterror);
        closesocket(m_socket);
        m_socket = INVALID_SOCKET;
        return false;
    }

	//create the background thread we use to monitor the asterisk manager for events
	HANDLE ourThread = CreateThread(NULL, 0, ProviderThreadProc, this, CREATE_SUSPENDED, NULL);
	SetThreadPriority(ourThread, THREAD_PRIORITY_NORMAL);
	SetThreadHandle(ourThread);
	m_bDie = false;
	ResumeThread(ourThread);

	return true;
}

/**
*	Wait and read the information of the thread
*/
void	ASTProvider::WaitReaderThread() 
{
	if ( m_hReaderThread != NULL ) {
		switch (WaitForSingleObject( m_hReaderThread, 50000))
		{
		case WAIT_OBJECT_0:
			//  The thread has terminated - do something
			m_log.LogTrace("WaitReaderThread => Thread terminated nicely\r\n");
			break;

		case WAIT_TIMEOUT:
			TerminateThread(m_hReaderThread,0);
			m_log.LogTrace("WaitReaderThread => killed the thread\r\n");
			//  The timeout has elapsed but the thread is still running
			//  do something appropriate for a timeout
			break;
		}

		CloseHandle(m_hReaderThread);
	}
}

/**
 *	disconnects tcp/ip from asterisk ami port
 *	\return true if the socket was closed correctly, otherwise false
 */
bool	ASTProvider::Disconnect()
{
	int err, rsoc;
	if ( m_socket != INVALID_SOCKET )
	{
		m_bDie = true;
		rsoc = closesocket(m_socket);
		err=WSAGetLastError();
		if (rsoc==0) err=0;
        switch (err)
		{
			case (0): m_log.LogInfo("Socket closed without errors.");
				break;
			case (WSANOTINITIALISED): m_log.LogInfo("Error code %d, A successful WSAStartup call must occur before using this function.",err);
				break;
			case (WSAENETDOWN): m_log.LogInfo("Error code %d, The network subsystem has failed.",err);
				break;
			case (WSAEFAULT): m_log.LogInfo("Error code %d, The buf parameter is not completely contained in a valid part of the user address space.",err);
				break;
			case (WSAENOTCONN): m_log.LogInfo("Error code %d, The socket is not connected.",err);
				break;
			case (WSAEINTR): m_log.LogInfo("Error code %d, The (blocking) call was canceled through WSACancelBlockingCall.",err);
				break;
			case (WSAEINPROGRESS): m_log.LogInfo("Error code %d, A blocking Windows Sockets 1.1 call is in progress, or the service provider is still processing a callback function.",err);
				break;
			case (WSAENETRESET): m_log.LogInfo("Error code %d, The connection has been broken due to the keep-alive activity detecting a failure while the operation was in progress.",err);
				break;
			case (WSAENOTSOCK): m_log.LogInfo("Error code %d, The descriptor is not a socket.",err);
				break;
			case (WSAEOPNOTSUPP): m_log.LogInfo("Error code %d, MSG_OOB was specified, but the socket is not stream-style such as type SOCK_STREAM, OOB data is not supported in the communication domain associated with this socket, or the socket is unidirectional and supports only send operations.",err);
				break;
			case (WSAESHUTDOWN): m_log.LogInfo("Error code %d, The socket has been shut down; it is not possible to receive on a socket after shutdown has been invoked with how set to SD_RECEIVE or SD_BOTH.",err);
				break;
			case (WSAEWOULDBLOCK): m_log.LogInfo("Error code %d, The socket is marked as nonblocking and the receive operation would block.",err);
				break;
			case (WSAEMSGSIZE): m_log.LogInfo("Error code %d, The message was too large to fit into the specified buffer and was truncated.",err);
				break;
			case (WSAEINVAL): m_log.LogInfo("Error code %d, The socket has not been bound with bind, or an unknown flag was specified, or MSG_OOB was specified for a socket with SO_OOBINLINE enabled or (for byte stream sockets only) len was zero or negative.",err);
				break;
			case (WSAECONNABORTED): m_log.LogInfo("Error code %d, The virtual circuit was terminated due to a time-out or other failure. The application should close the socket as it is no longer usable.",err);
				break;
			case (WSAETIMEDOUT): m_log.LogInfo("Error code %d, The connection has been dropped because of a network failure or because the peer system failed to respond.",err);
				break;
			case (WSAECONNRESET): m_log.LogInfo("Error code %d, The virtual circuit was reset by the remote side executing a hard or abortive close. The application should close the socket as it is no longer usable. On a UPD-datagram socket this error would indicate that a previous send operation resulted in an ICMP 'Port Unreachable' message.",err);
				break;
			default: m_log.LogInfo("Error code %d.",err);

		}
		WaitReaderThread();
        m_socket = INVALID_SOCKET;
	}

	return true;
}

/**
 * sends login action
 *	\param user username
 *	\param pwd password
 *	\eAuthType authentification type 0-plaintext 1-md5
 *	\return time_out, ok, generic_failure
 * Example: 
 *	\code
 *	Action: Login
 *	UserName: admin
 *	Secret: amp111
 *	ActionID: 1
 *	\endcode
 */
AC_RESULT ASTProvider::Login(std::string user, std::string pwd, AUTH_TYPE eAuthType)
{
    std::string command;

	std::string	invokeId = m_invokeIdManager.NewInvokeId();
	ManagerResponse	result ;
	
	if(eAuthType==at_md5)
	{
		MD5 md5sum;

		command = "Action: Challenge\r\n";
		command += "AuthType: MD5\r\n";
		command += "ActionID: ";
		command += invokeId;
		command += "\r\n\r\n";

		SendCommand(command);

		if (!m_invokeIdManager.WaitResponse(invokeId, result, AST_TIMEOUT)) {
			return time_out;
		}

		if (result.m_response.compare("success")==0) {
			command = "Action: Login\r\n";
			command +="AuthType: MD5\r\n";
			command += "UserName: ";
			command += user;
			command += "\r\n";
			command += "Key: ";
			//Take the md5sum of challenge+pwd
			command += md5sum.digestString((char*)(result.m_challenge.append(pwd).c_str()));
			command += "\r\n";
			command += "ActionID: ";
			command += invokeId;
			command += "\r\n\r\n";
			
			SendCommand(command);

			if (result.m_response.compare("success")==0){
				return ok;
			}
		}
	}
	else
	{
		command = "Action: Login\r\n";
		command += "UserName: ";
		command += user;
		command += "\r\n";
		command += "Secret: ";
		command += pwd;
		command += "\r\n";
		command += "ActionID: ";
		command += invokeId;
		command += "\r\n\r\n";

		SendCommand(command);

		if (!m_invokeIdManager.WaitResponse(invokeId, result, AST_TIMEOUT)) {
			return time_out;
		}

		if (result.m_response.compare("success")==0) {
			return ok;
		}
	}
	/// \todo: more error handling
	return generic_failure;
}
/**
* Set a global or local channel variable.
*	\see http://www.voip-info.org/wiki/index.php?page=Asterisk+Manager+API+Action+SetVar
*	\param channel
*	\param variable
*	\param value
*	\return time_out, ok, generic_failure
*	Example:
*	\code
*	Action: Setvar
*	Channel: Zap/1-1
*	Variable: AtestVariable
*	Value: This is now set
*	\endcode
*/
AC_RESULT ASTProvider::SetVar(std::string channel, std::string variable, std::string value)
{
	m_log.LogTrace("ASTProvider::SetVar channel:%s variable:%s value:%s", 
		channel.c_str(), variable.c_str(), value.c_str());

	std::string	invokeId = m_invokeIdManager.NewInvokeId();
	ManagerResponse	result ;
    std::string command;

	command = "Action: SetVar\r\n";
	if (channel != ""){
		command += "Channel: "+channel;
		command += "\r\n";
	}
    command += "Variable: " + variable;
    command += "\r\n";
    command += "Value: " + value;
    command += "\r\n";
	command += "ActionID: ";
	command += invokeId;
    command += "\r\n\r\n";

	SendCommand(command);

	if (m_invokeIdManager.WaitResponse(invokeId, result, AST_TIMEOUT)) {
		if (result.m_response.compare("success")==0) {
			m_log.LogTrace("  success");
			return ok;
		}
		else {
			/// \todo: more error handling
			m_log.LogTrace("  generic_failure");
			return generic_failure;
		}
	}

	m_log.LogTrace("  time_out");
	return time_out;
}

/**
*	Get a global or local channel variable.
*	\see http://www.voip-info.org/wiki/index.php?page=Asterisk+Manager+API+Action+GetVar
*	\param channel
*	\param variable
*	\param &value
*	\return time_out, ok, generic_failure
*	Example:
*	\code
*	ACTION: GetVar
*	Channel: SIP/5060-44d225d0
*	Variable: extension 
*	\endcode
*/
AC_RESULT	ASTProvider::GetVar(std::string channel, std::string variable, std::string &value)
{
	m_log.LogTrace("ASTProvider::GetVar channel:%s variable:%s", channel.c_str(), variable.c_str());

	std::string	invokeId = m_invokeIdManager.NewInvokeId();
	ManagerResponse	result ;
    std::string command;

	command = "Action: GetVar\r\n";
	if (channel != ""){
		command += "Channel: "+channel;
		command += "\r\n";
	}
    command += "Variable: " + variable;
    command += "\r\n";
    command += "ActionID: ";
	command += invokeId;
    command += "\r\n\r\n";

	SendCommand(command);

	if (m_invokeIdManager.WaitResponse(invokeId, result, AST_TIMEOUT)) {
		if (result.m_response.compare("success")==0) {
			m_log.LogTrace("  success");
			return ok;
		}
		else {
			m_log.LogTrace("  generic_failure");
			return generic_failure;
		}
	}

	m_log.LogTrace("  time_out");
	return time_out;
}

/**
*	connects two active channels
*	\param channel1
*	\param channel2
*	\param tone
*	\return time_out, ok, generic_failure
*/
AC_RESULT	ASTProvider::Bridge(std::string channel1, std::string channel2, bool tone)
{
	std::string toneStr;
	if (tone)
		toneStr = "yes";
	else
		toneStr = "no";
	m_log.LogTrace("ASTProvider::Bridge channel1:%s channel2:%s Tone:%s", 
		channel1.c_str(), channel2.c_str(), toneStr.c_str());

	std::string	invokeId = m_invokeIdManager.NewInvokeId();
	ManagerResponse	result ;
    std::string command;

	command = "Action: Bridge\r\n";
	command += "Channel1: "+channel1;
	command += "\r\n";
    command += "Channel2: "+channel2;
	command += "\r\n";
	if (tone){
		command += "Tone: " + toneStr;
		command += "\r\n";
	}
	command += "ActionID: ";
	command += invokeId;
    command += "\r\n\r\n";

	SendCommand(command);

	if (m_invokeIdManager.WaitResponse(invokeId, result, AST_TIMEOUT)) {
		if (result.m_response.compare("success")==0) {
			m_log.LogTrace("  success");
			return ok;
		}
		else {
			/// \todo: more error handling
			m_log.LogTrace("  generic_failure");
			return generic_failure;
		}
	}

	m_log.LogTrace("  time_out");
	return time_out;
}

/**
*	Adds an extension interface to a queue
*	\see http://www.voip-info.org/wiki/index.php?page=Asterisk+Manager+API+Action+QueueAdd
*	\param queue
*	\param channelInterface
*	\param penalty
*	\param bPaused
*	\return ok, already_logged_in, queue_not_specified, interface_not_specified, no_such_queue, time_out, generic_failure;
*
*	Example: 
*	\code
*	Action: QueueAdd 
*	Queue: 555 
*	Interface: local/201 
*	Penalty: 1 
*	Paused: true 
*	\endcode
*/

AC_RESULT	ASTProvider::QueueAdd(std::string queue, std::string channelInterface, int penalty, bool bPaused )
{
    std::string command;

	std::string	invokeId = m_invokeIdManager.NewInvokeId();
	ManagerResponse	result ;
	std::string	sPaused = (bPaused?"true":"false");

	command = "Action: QueueAdd\r\n";
	command += "Queue: " + queue;
    command += "\r\n";
    command += "Interface: " + channelInterface;
    command += "\r\n";
    command += "Penalty: " + ToString(penalty);
    command += "\r\n";
	command += "Paused: " + sPaused;
	command += "\r\n";
	command += "ActionID: " + invokeId;
	command += "\r\n";
	command += "Async: 1";
    command += "\r\n\r\n";

	SendCommand(command);

	if (!m_invokeIdManager.WaitResponse(invokeId, result, AST_TIMEOUT)) {
		return time_out;
	}

	if (result.m_response.compare("success")==0) {
		return ok;
	}
	else {
		if (result.m_message.compare("Unable to add interface: Already there")==0) {
			return already_logged_in;
		}
		else if (result.m_message.compare("'Queue' not specified.")==0) {
			return queue_not_specified;
		}
		else if (result.m_message.compare("'Interface' not specified.")==0) {
			return interface_not_specified;
		}
		else if (result.m_message.compare("Unable to add interface to queue: No such queue")==0) {
			return no_such_queue;
		}
	}
	return generic_failure;
}

/**
*	removes an extension interface to a queue
*	\see http://www.voip-info.org/wiki/index.php?page=Asterisk+Manager+API+Action+QueueRemove
*	\param queue
*	\param channelInterface
*	\return ok, not_logged_in, no_such_queue, time_out, generic_failure;
*
*	Example:
*	\code
*	Action: QueueRemove 
*	Queue: queue1 
*	Interface: sip/2600 
*	\endcode
*/
AC_RESULT	ASTProvider::QueueRemove(std::string queue, std::string channelInterface)
{
    std::string command;

	std::string	invokeId = m_invokeIdManager.NewInvokeId();
	ManagerResponse	result ;

	command = "Action: QueueRemove\r\n";
	command += "Queue: " + queue;
	command += "\r\n";
    command += "Interface: " + channelInterface;
	command += "\r\n";
	command += "ActionID: " + invokeId;
	command += "\r\n";
	command += "Async: 1";
    command += "\r\n\r\n";

	SendCommand(command);

	if (!m_invokeIdManager.WaitResponse(invokeId, result, AST_TIMEOUT)) {
		return time_out;
	}

	if (result.m_response.compare("success")==0) {
		return ok;
	}
	else {
		if (result.m_message.compare("Unable to remove interface: Not there")==0) {
			return not_logged_in;
		}
		else if (result.m_message.compare("Unable to remove interface from queue: No such queue")==0) {
			return no_such_queue;
		}
	}
	return generic_failure;
}

/**
*	pauses/unpauses an extension logged into any queue
*	\param channelInterface 
*	\param bPaused true/false
*	\return ok, not_logged_in, no_such_queue, time_out, generic_failure;
*
*	Example:
*	\code
*	Action: QueuePause 
*	Interface: local/2600
*	Paused: true
*	\endcode
*/
AC_RESULT	ASTProvider::QueuePause(std::string channelInterface, bool bPaused)
{
    std::string command;

	std::string	invokeId = m_invokeIdManager.NewInvokeId();
	ManagerResponse	result ;
	std::string	sPaused = (bPaused?"true":"false");

	command = "Action: QueuePause\r\n";
    command += "Interface: " + channelInterface;
	command += "\r\n";
    command += "Paused: " + sPaused;
	command += "\r\n";
	command += "ActionID: " + invokeId;
	command += "\r\n";
	command += "Async: 1";
    command += "\r\n\r\n";

	SendCommand(command);

	if (!m_invokeIdManager.WaitResponse(invokeId, result, AST_TIMEOUT)) {
		return time_out;
	}

	if (result.m_response.compare("success")==0) {
		return ok;
	}
	else {
		if (result.m_message.compare("Interface not found")==0) {
			return interface_not_found;
		}
	}
	return generic_failure;
}

/*
 * Function ProviderThreadProc
 *
 * this is the thread that receives asterisk messages: cofirmations and events
 *	\param lpParameter
 *	\return DWORD
 */
DWORD WINAPI ProviderThreadProc(LPVOID lpParameter)
{
	((ASTProvider*)lpParameter)->ReaderProc();

	return 0;
}


/**
 * Implements ProviderThreadProc 
 *	\return DWORD
 */
DWORD ASTProvider::ReaderProc(void)
{
	m_log.LogTrace("ASTProvider::ReaderProc");

	std::list<std::string>	commandResult ;
	std::string line;
    bool processingCommandResult = false;

    if (m_socket == INVALID_SOCKET) {
		m_log.LogTrace("Unable to run: socket is null.");
		m_log.LogTrace("ASTProvider::ReaderProc EXITS");
        return -1 ;
    }

	line = ReadLine();		// read from socket
	while (!m_bDie && line.size()!=0) {
		ParseLine(line);
		line = ReadLine();		// read from socket
	}

	if (!m_bDie) {
		m_log.LogWarn("Unexpected Link Down...");
        closesocket(m_socket);
        m_socket = INVALID_SOCKET;
		LinkDownProcedure();
	}

	m_log.LogTrace("ASTProvider::ReaderProc EXITS");
	return 0;
}


#define	AST_PROTOCOL_HEADER	"Asterisk Call Manager/"

/**
*	Parse a line received by the manager
*	\param line
*/
void	ASTProvider::ParseLine(std::string line) {
	// dirty hack for handling the CommandAction. Needs fix when manager protocol is enhanced.
	if (m_bProcessingCommandResult) {
		// in case of an error Asterisk sends a Usage: and an END COMMAND
		// that is prepended by a space :(
		if (line.compare("--END COMMAND--")==0 || line.compare(" --END COMMAND--")==0) {
			//do nothing for now!!!
		} 
		else {
			m_commandResult += line;
		}
		return ;
	}

	// Reponse: Follows indicates that the output starting on the next line until
	// --END COMMAND-- must be treated as raw output of a command executed by a
	// CommandAction.
	if (line.compare("Response: Follows")==0) {
		m_bProcessingCommandResult = true;
		m_commandResult.clear();
		return ;
	}

	// maybe we will find a better way to identify the protocol identifier but for now
	// this works quite well.
	if (line.find(AST_PROTOCOL_HEADER)!=-1) {
		m_log.LogTrace("Asterisk protocol header: %s", line.substr(0,line.size()-2).c_str());
		//Set the manager version to the version translation helper
		if(line.find("1.0")!=-1)
			m_AstManTrans.setManagerVersion(manager_v1_0);
		else if(line.find("1.1")!=-1)
			m_AstManTrans.setManagerVersion(manager_v1_1);
		else
			m_AstManTrans.setManagerVersion(manager_v1_1);
		return ;
	}

	// an empty line indicates completion of a normal response's or event's end so we build
	// the corresponding value object and dispatch it through the ManagerConnection.
	if (line=="\r\n") {
		if (m_bag.find("event") != m_bag.end()) {
			HandleEvent(m_bag);
		}
		else if (m_bag.find("response") != m_bag.end()) {
			m_log.LogTrace(" <== asterisk protocol response__________________________");
			HandleResponse(m_bag);
		}
		else {
			if (m_bag.size() > 0) {
				m_log.LogTrace("buffer contains neither response nor event");
			}
		}

		m_bag.clear();
	}
	else {	// continue handling parts of response/event
		//When we are down to the line level it is in the format
		//header: data
		size_t dataPos;
		if ( ( dataPos = line.find(":")) != -1 ) {
			std::string header = line.substr(0,dataPos);
			size_t	tail = line.length()-dataPos-4; // -4 elimina /n/r finales
			std::string data = line.substr(dataPos+2, tail);	
			
			std::transform(header.begin(), header.end(), header.begin(), tolower);

			// add it to own map!
			m_bag[header] = data;
			m_log.LogTrace0("HEADER: %s DATA: %s", header.c_str(), data.c_str());
		}
	}

	return ;
}

/**
*	Read a line of the manager
*	\return string with the line
*/
std::string	ASTProvider::ReadLine() {
    std::string	line;
	char		buffer[1024];
	int			numBytes = 0;
	int			err;

	numBytes = recv ( m_socket, buffer, 1, 0);
	err=WSAGetLastError();
	if ( numBytes == 0 || numBytes == SOCKET_ERROR || m_socket == INVALID_SOCKET )
	{
		m_log.LogInfo("Socket recv code = %d",numBytes);
		if (numBytes == SOCKET_ERROR){
			switch (err)
			{
				case (WSANOTINITIALISED): m_log.LogInfo("Error code %d, A successful WSAStartup call must occur before using this function.",err);
					break;
				case (WSAENETDOWN): m_log.LogInfo("Error code %d, The network subsystem has failed.",err);
					break;
				case (WSAEFAULT): m_log.LogInfo("Error code %d, The buf parameter is not completely contained in a valid part of the user address space.",err);
					break;
				case (WSAENOTCONN): m_log.LogInfo("Error code %d, The socket is not connected.",err);
					break;
				case (WSAEINTR): m_log.LogInfo("Error code %d, The (blocking) call was canceled through WSACancelBlockingCall.",err);
					break;
				case (WSAEINPROGRESS): m_log.LogInfo("Error code %d, A blocking Windows Sockets 1.1 call is in progress, or the service provider is still processing a callback function.",err);
					break;
				case (WSAENETRESET): m_log.LogInfo("Error code %d, The connection has been broken due to the keep-alive activity detecting a failure while the operation was in progress.",err);
					break;
				case (WSAENOTSOCK): m_log.LogInfo("Error code %d, The descriptor is not a socket.",err);
					break;
				case (WSAEOPNOTSUPP): m_log.LogInfo("Error code %d, MSG_OOB was specified, but the socket is not stream-style such as type SOCK_STREAM, OOB data is not supported in the communication domain associated with this socket, or the socket is unidirectional and supports only send operations.",err);
					break;
				case (WSAESHUTDOWN): m_log.LogInfo("Error code %d, The socket has been shut down; it is not possible to receive on a socket after shutdown has been invoked with how set to SD_RECEIVE or SD_BOTH.",err);
					break;
				case (WSAEWOULDBLOCK): m_log.LogInfo("Error code %d, The socket is marked as nonblocking and the receive operation would block.",err);
					break;
				case (WSAEMSGSIZE): m_log.LogInfo("Error code %d, The message was too large to fit into the specified buffer and was truncated.",err);
					break;
				case (WSAEINVAL): m_log.LogInfo("Error code %d, The socket has not been bound with bind, or an unknown flag was specified, or MSG_OOB was specified for a socket with SO_OOBINLINE enabled or (for byte stream sockets only) len was zero or negative.",err);
					break;
				case (WSAECONNABORTED): m_log.LogInfo("Error code %d, The virtual circuit was terminated due to a time-out or other failure. The application should close the socket as it is no longer usable.",err);
					break;
				case (WSAETIMEDOUT): m_log.LogInfo("Error code %d, The connection has been dropped because of a network failure or because the peer system failed to respond.",err);
					break;
				case (WSAECONNRESET): m_log.LogInfo("Error code %d, The virtual circuit was reset by the remote side executing a hard or abortive close. The application should close the socket as it is no longer usable. On a UPD-datagram socket this error would indicate that a previous send operation resulted in an ICMP 'Port Unreachable' message.",err);
					break;
				default: m_log.LogInfo("Error code %d.",err);
			}
			
		}
		//if we get here the chances are that the socket has
		//been closed down.
		return line;
	}

	buffer[numBytes] = 0;
	line += buffer;

	while ( line.find("\r\n") == -1 ) {
		numBytes = recv ( m_socket, buffer, 1, 0);
		if ( numBytes == 0 || numBytes == SOCKET_ERROR || m_socket == INVALID_SOCKET )
		{
			//if we get here the chances are that the socket has been closed down.
			return line;
		}

		buffer[numBytes] = 0;
		line += buffer;
	}
	
	return line;
}

/**
*	Sends a command to Asterisk Manager Interface
*	\param command 
*	\return true if the command was send correctly, otherwise false
*/
bool ASTProvider::SendCommand(std::string command)
{
	m_log.LogTrace("ASTProvider::SendCommand => \r\n%s", command.c_str());
    if ( SOCKET_ERROR == send(m_socket, command.c_str(), (int)command.length() , 0) ) {
		m_log.LogTrace("  ...return false");
		return false;
	}

	return true;
}

/**
*	Disconnect
*/
void	ASTProvider::LinkDownProcedure() 
{

	//FIX: signal all invokeid
	m_log.LogTrace("Sending all signals after linkdown for all pending invokeid");
	ManagerResponse response;
	response.m_response ="linkdown";
	int num=m_invokeIdManager.SignalAll(response);
	m_log.LogTrace("Signals sended = %d", num);

	//FIX: switch mahine for other process
	m_log.LogTrace("Switch machine to other threads");
	Sleep(0);
		//YieldProcessor(); 

	// fire event
	m_listener->OnLinkDown();
}
/**
*	Handle an event Response
*	\param bag
*/
void	ASTProvider::HandleResponse(std::map<std::string, std::string>	bag)
{
	ManagerResponse	response ;
	std::string	responseType = bag["response"];

	responseType = bag["response"];
	std::transform(responseType.begin(), responseType.end(), responseType.begin(), tolower);

/*	if (responseType=="error") {
	}
	else if (bag.find("challenge")!=bag.end()) {
	}
	else if ((bag.find("mailbox")!=bag.end()) && (bag.find("waiting")!=bag.end())) {
	}
	else if ((bag.find("mailbox")!=bag.end()) && (bag.find("newmessages")!=bag.end()) && (bag.find("oldmessages")!=bag.end())) {
	}
	else if ((bag.find("exten")!=bag.end()) && (bag.find("context")!=bag.end()) && (bag.find("hint")!=bag.end()) && (bag.find("status")!=bag.end())) {
	}
	else {
	}
*/
	response.m_response = responseType;

	response.m_challenge = bag["challenge"];

	//if (bag.find("actionid")!=bag.end()) {
	response.m_actionId = bag["actionid"];
	//}
	//if (bag.find("message")!=bag.end()) {
	response.m_message = bag["message"];
	//}
	//if (bag.find("uniqueid")!=bag.end()) {
	response.m_uniqueId = bag["uniqueid"];
	//}
	//if (bag.find("actionid")!=bag.end()) {
	response.m_actionId = bag["actionid"];
	//}
	
	m_invokeIdManager.SignalResponse(response, response.m_actionId);
}
/**
*	Manager a Handle Event
*	\param bag
*/
void	ASTProvider::HandleEvent(std::map<std::string, std::string>	bag)
{
	ManagerEvent	event ;
	std::string		eventType ;

	if (bag.find("event")==bag.end()) {
		// no event type sent!!
		m_log.LogTrace(" no event type send");
		return;
	}

	eventType = bag["event"];
	std::transform(eventType.begin(), eventType.end(), eventType.begin(), tolower);

	/*
	**	we handle: NewState, NewCallerId, NewExten, Channel, Hangup, NewChannel, dial, rename, link, 
	**  unlink, join, leave, UnParkedCall, originatefailure, originatesuccess
	*/

	if (eventType.compare("newstate") == 0) {
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
		// NewState channel, state, callerId, callerIdName, uniqueId, privilege
		event.m_channel = bag[m_AstManTrans.m_Event.m_channel]; ;
		event.m_state = bag[m_AstManTrans.m_Event.m_state];
		event.m_callerId = bag[m_AstManTrans.m_Event.m_callerId];
		event.m_callerIdName = bag[m_AstManTrans.m_Event.m_callerIdName];
		event.m_uniqueId = bag[m_AstManTrans.m_Event.m_uniqueId];
		if(m_AstManTrans.getManagerVersion()!=manager_v1_0)
		{
			//new in asterisk 1.8.5.0
			event.m_connectedLineNum = bag[m_AstManTrans.m_Event.m_connectedLineNum];
			event.m_connectedLineName = bag[m_AstManTrans.m_Event.m_connectedLineName];
		}

		m_listener->OnNewStateEvent(event);
		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	}
	else if (eventType.compare("newcallerid") == 0) {
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
		// NewCallerId callerid, calleridname, channel, uniqueId
		event.m_channel = bag[m_AstManTrans.m_Event.m_channel];
		event.m_callerId = bag[m_AstManTrans.m_Event.m_callerId];
		event.m_callerIdName = bag[m_AstManTrans.m_Event.m_callerIdName];
		event.m_uniqueId = bag[m_AstManTrans.m_Event.m_uniqueId];
		m_listener->OnNewCallerIdEvent(event);
		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	}
	else if (eventType.compare("newchannel") == 0) {
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
		// NewChannel channel, state, callerid, calleridname, uniqueid
		event.m_channel = bag[m_AstManTrans.m_Event.m_channel];
		event.m_state = bag[m_AstManTrans.m_Event.m_state];
		event.m_callerId = bag[m_AstManTrans.m_Event.m_callerId];
		if (event.m_callerId == ""){
			event.m_callerId = bag[m_AstManTrans.m_Event.m_callerIdNum];
		}
		event.m_callerIdName = bag[m_AstManTrans.m_Event.m_callerIdName];
		event.m_uniqueId = bag[m_AstManTrans.m_Event.m_uniqueId];
		m_listener->OnNewChannelEvent(event);
		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	}
	else if (eventType.compare("hangup") == 0) {
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
	// Hangup cause, channel, state, callerid, calleridname, uniqueid
		event.m_cause = bag[m_AstManTrans.m_Event.m_cause];
		event.m_causeTxt = bag[m_AstManTrans.m_Event.m_causeTxt];
		event.m_channel = bag[m_AstManTrans.m_Event.m_channel];
		event.m_state = bag[m_AstManTrans.m_Event.m_state];
		event.m_callerId = bag[m_AstManTrans.m_Event.m_callerId];
		event.m_callerIdName = bag[m_AstManTrans.m_Event.m_callerIdName];
		event.m_uniqueId = bag[m_AstManTrans.m_Event.m_uniqueId];
		m_listener->OnHangup(event);
		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	}
	else if (eventType.compare("newexten") == 0) {
		m_log.LogTrace0("ASTProvider::HandleEvent => %s", eventType.c_str());
		// NewExten uniqueId, context, extension, application, appData, priority, channel
		event.m_uniqueId = bag[m_AstManTrans.m_Event.m_uniqueId];
		event.m_context = bag[m_AstManTrans.m_Event.m_context];
		event.m_extension = bag[m_AstManTrans.m_Event.m_extension];
		event.m_application = bag[m_AstManTrans.m_Event.m_application];
		event.m_appData = bag[m_AstManTrans.m_Event.m_appData];
		event.m_priority = bag[m_AstManTrans.m_Event.m_priority];
		event.m_channel = bag[m_AstManTrans.m_Event.m_channel];
		m_listener->OnNewExten(event);
		m_log.LogTrace0(" <== asterisk protocol event _______________________________");
	}
	else if (eventType.compare("dial") == 0) {
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
		// Dial srcuniqueid, destuniqueid, calleraddress
		event.m_srcUniqueId = bag[m_AstManTrans.m_Event.m_srcUniqueId];
		event.m_destUniqueId = bag[m_AstManTrans.m_Event.m_destUniqueId];
		event.m_callerId = bag[m_AstManTrans.m_Event.m_callerId];
		event.m_callerIdName = bag[m_AstManTrans.m_Event.m_callerIdName];
		if(m_AstManTrans.getManagerVersion()==manager_v1_0 || bag[m_AstManTrans.m_Event.m_SubEvent]=="Begin")
			m_listener->OnDialEvent(event);
		//elseif (bag["SubEvent"]=="End") notifies the cause!!

		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	}
	else if (eventType.compare("bridge") == 0) {
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
		// Begin srcuniqueid, destuniqueid, calleraddress
		event.m_srcUniqueId = bag[m_AstManTrans.m_Event.m_uniqueid1];
		event.m_destUniqueId = bag[m_AstManTrans.m_Event.m_uniqueid2];
		event.m_callerId = bag[m_AstManTrans.m_Event.m_callerid1];
		if(bag[m_AstManTrans.m_Event.m_BridgeState]=="Link")
			m_listener->OnLinkEvent(event);
		else if(bag[m_AstManTrans.m_Event.m_BridgeState]=="Unlink")
			m_listener->OnUnlinkEvent(event);
		else
			m_log.LogWarn("Unknown bridgeState: %s",bag[m_AstManTrans.m_Event.m_BridgeState].c_str());
		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	}
	else if (eventType.compare("unlink") == 0) {
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
		// Dial srcuniqueid, destuniqueid, calleraddress
		event.m_srcUniqueId = bag[m_AstManTrans.m_Event.m_uniqueid1];
		event.m_destUniqueId = bag[m_AstManTrans.m_Event.m_uniqueid2];
		event.m_callerId = bag[m_AstManTrans.m_Event.m_callerid1];
		m_listener->OnUnlinkEvent(event);
		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	}
	else if (eventType.compare("link") == 0) {  //ast<=1.4
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
		// Dial srcuniqueid, destuniqueid, calleraddress
		event.m_srcUniqueId = bag[m_AstManTrans.m_Event.m_uniqueid1];
		event.m_destUniqueId = bag[m_AstManTrans.m_Event.m_uniqueid2];
		event.m_callerId = bag[m_AstManTrans.m_Event.m_callerid1];
		m_listener->OnLinkEvent(event);
		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	}
	else if (eventType.compare("rename") == 0) {
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
		// rename srcuniqueid, destuniqueid, calleraddress
		event.m_uniqueId = bag[m_AstManTrans.m_Event.m_uniqueId];
		event.m_oldname = bag[m_AstManTrans.m_Event.m_oldname];
		event.m_newname = bag[m_AstManTrans.m_Event.m_newname];
		m_listener->OnRenameEvent(event);
		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	}
	else if (eventType.compare("join") == 0) {
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
		// join channel, callerid, calleridname, queue, position, count
		event.m_channel = bag[m_AstManTrans.m_Event.m_channel];
		event.m_callerId = bag[m_AstManTrans.m_Event.m_callerId];
		event.m_callerIdName = bag[m_AstManTrans.m_Event.m_callerIdName];
		event.m_queue = bag[m_AstManTrans.m_Event.m_queue];
		event.m_position = bag[m_AstManTrans.m_Event.m_position];
		event.m_count = bag[m_AstManTrans.m_Event.m_count];
		m_listener->OnJoinEvent(event);
		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	}
	else if (eventType.compare("leave") == 0) {
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
		// leave channel, queue, count
		event.m_channel = bag[m_AstManTrans.m_Event.m_channel];
		event.m_queue = bag[m_AstManTrans.m_Event.m_queue];
		event.m_count = bag[m_AstManTrans.m_Event.m_count];
		m_listener->OnLeaveEvent(event);
		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	}
	else if (eventType.compare("originateresponse") == 0){ //ast > 1.2
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
		event.m_actionId = bag[m_AstManTrans.m_Event.m_actionId];
		event.m_response = bag[m_AstManTrans.m_Event.m_response];
		event.m_callerId = bag[m_AstManTrans.m_Event.m_callerId];
		if (event.m_callerId == ""){
			event.m_callerId = bag[m_AstManTrans.m_Event.m_callerIdNum];
		}
		event.m_callerIdName = bag[m_AstManTrans.m_Event.m_callerIdName];
		event.m_channel = bag[m_AstManTrans.m_Event.m_channel];
		event.m_context = bag[m_AstManTrans.m_Event.m_context];
		event.m_extension = bag[m_AstManTrans.m_Event.m_extension];
		event.m_reason = bag[m_AstManTrans.m_Event.m_reason];
		event.m_uniqueId = bag[m_AstManTrans.m_Event.m_uniqueId];
		if (event.m_response == "Failure") {
			m_listener->OnOriginateFailureEvent(event);
		}
		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	}
	else if (eventType.compare("queuecallerabandon") == 0){ //ast >= 1.4
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
		event.m_queue = bag[m_AstManTrans.m_Event.m_queue];
		event.m_uniqueId = bag[m_AstManTrans.m_Event.m_uniqueId];
		event.m_position = bag[m_AstManTrans.m_Event.m_position];
		event.m_origposition = bag[m_AstManTrans.m_Event.m_origposition];
		event.m_holdtime = bag[m_AstManTrans.m_Event.m_holdtime];
		m_listener->OnQueueCallerAbandonEvent(event);
		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	} 
	else if (eventType.compare("hold") == 0){
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
		event.m_channel = bag[m_AstManTrans.m_Event.m_channel];
		event.m_uniqueId = bag[m_AstManTrans.m_Event.m_uniqueId];
		m_listener->OnHoldEvent(event);
		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	} 
	else if (eventType.compare("unhold") == 0){
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
		event.m_channel = bag[m_AstManTrans.m_Event.m_channel];
		event.m_uniqueId = bag[m_AstManTrans.m_Event.m_uniqueId];
		m_listener->OnUnholdEvent(event);
		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	}
	else if (eventType.compare("unparkedcall") == 0){
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
		// unparked channel, parked by from
		event.m_extension = bag[m_AstManTrans.m_Event.m_extension];
		event.m_channel = bag[m_AstManTrans.m_Event.m_channel];
		event.m_from = bag[m_AstManTrans.m_Event.m_from];
		event.m_callerId = bag[m_AstManTrans.m_Event.m_callerId];
		event.m_callerIdName = bag[m_AstManTrans.m_Event.m_callerIdName];
		m_listener->OnUnparkedCallEvent(event);
		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	}
	else if (eventType.compare("channel") == 0) {  // only ast1.2
		m_log.LogTrace0("ASTProvider::HandleEvent => %s", eventType.c_str());
		// Channel channel, state, callerid, calleridname, uniqueid
		event.m_channel = bag[m_AstManTrans.m_Event.m_channel];
		event.m_state = bag[m_AstManTrans.m_Event.m_state];
		event.m_callerId = bag[m_AstManTrans.m_Event.m_callerId];
		event.m_callerIdName = bag[m_AstManTrans.m_Event.m_callerIdName];
		event.m_uniqueId = bag[m_AstManTrans.m_Event.m_uniqueId];
		m_listener->OnChannelEvent(event);
		m_log.LogTrace0(" <== asterisk protocol event _______________________________");
	}
	else if (eventType.compare("originatefailure") == 0){ // only ast1.2
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
		event.m_actionId = bag[m_AstManTrans.m_Event.m_actionId];
		event.m_channel = bag[m_AstManTrans.m_Event.m_channel];
		event.m_context = bag[m_AstManTrans.m_Event.m_context];
		event.m_extension = bag[m_AstManTrans.m_Event.m_extension];
		event.m_reason = bag[m_AstManTrans.m_Event.m_reason];
		event.m_uniqueId = bag[m_AstManTrans.m_Event.m_uniqueId];
		event.m_response = "failure";
		m_listener->OnOriginateFailureEvent(event);
		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	}
	else if (eventType.compare("originatesuccess") == 0){ // only ast1.2
		m_log.LogTrace("ASTProvider::HandleEvent => %s", eventType.c_str());
		event.m_actionId = bag[m_AstManTrans.m_Event.m_actionId];
		event.m_channel = bag[m_AstManTrans.m_Event.m_channel];
		event.m_context = bag[m_AstManTrans.m_Event.m_context];
		event.m_extension = bag[m_AstManTrans.m_Event.m_extension];
		event.m_reason = bag[m_AstManTrans.m_Event.m_reason];
		event.m_uniqueId = bag[m_AstManTrans.m_Event.m_uniqueId];
		event.m_response = "success";
		m_log.LogTrace(" <== asterisk protocol event _______________________________");
	}
	else{
		//only trace unhandled events in trace0 level
		m_log.LogTrace0("ASTProvider::HandleEvent => %s", eventType.c_str());
		m_log.LogTrace0(" <== asterisk protocol event _______________________________");
	}
}

/**
*	Originates a call via Asterisk API
*	\see http://www.voip-info.org/wiki/index.php?page=Asterisk+Manager+API+Action+Originate
*	\param channel 
*	\param callerId 
*	\param callerIdName 
*	\param calledExtension
*	\param callid Remarks: this version returns zero
*	\param context used to originate call
*	\param timeout
*	\return ok, generic_failure, time_out
*
*	\code
*	Action: Originate 
*	Channel: channel 
*	Exten: calledextension 
*	Priority: 1 
*	Callerid: "callerIdName" <callerId> 
*	Context: context 
*	ActionID: 2 
*	Async: 1 
*	\endcode
*/
AC_RESULT	ASTProvider::Originate(std::string channel, std::string callerId, std::string callerIdName, std::string calledExtension, DWORD callid, std::string context, DWORD timeout, std::string variable, bool bLockReaderAfterResponse)
{
	m_log.LogTrace("ASTProvider::Originate call %d channel:%s callerId:%s to calledExtension:%s in context:%s, "
		"callerIdName=%s timeout=%d", 
		callid, channel.c_str(), callerId.c_str(), calledExtension.c_str(), context.c_str(), callerIdName.c_str(),
		timeout);

	if (callid < 0){
		m_log.LogError(" callid has an invalid number");
		return invalid_callid;
	}

	std::string invokeId = m_invokeIdManager.NewInvokeId(bLockReaderAfterResponse) ;
	ManagerResponse	result ;

	AC_RESULT acr = OriginateByContext(channel, callerId, callerIdName, calledExtension, 
										callid, context, timeout, invokeId, variable);

	if (acr == invalid_callid){
		m_log.LogError(" OriginateByContext return invalid_callid");
		return invalid_callid;
	}

	if (m_invokeIdManager.WaitResponse(invokeId, result, AST_TIMEOUT)) {
		if (result.m_response.compare("success")==0) {
			m_log.LogTrace("  Originate success");
			return ok;
		}
		else {
			/// \todo: more error handling
			m_log.LogTrace("  generic_failure");
			return generic_failure;
		}
	}

	m_log.LogTrace("  time_out");
	return time_out;
}

/**
*	Hangup call in given channel
*	\see http://www.voip-info.org/wiki/index.php?page=Asterisk+Manager+API+Action+Hangup
*	\param channel
*	\return ok, generic_failure, time_out
*	Example:
*	\code
*	ACTION: Hangup
*	Channel: SIP/x7065558529-99a0 
*	\endcode
*/
AC_RESULT	ASTProvider::Hangup(std::string channel)
{
	m_log.LogTrace("ASTProvider::Hangup channel=%s", channel.c_str());

	std::string	invokeId = m_invokeIdManager.NewInvokeId();
	ManagerResponse	result ;

	std::string command;

	command = "Action: Hangup\r\n";
	command += "Channel: " + channel;
	command += "\r\n";
	command += "ActionID: " + invokeId;
	command += "\r\n";
	command += "Async: 1";
	command += "\r\n\r\n";
	SendCommand(command);

	if (m_invokeIdManager.WaitResponse(invokeId, result, AST_TIMEOUT)) {
		if (result.m_response.compare("success")==0) {
			m_log.LogTrace("  return = ok");
			return ok;
		}else {
			if (result.m_message.compare("No such channel") == 0){
				m_log.LogError("  return = no such channel");
				return no_such_channel;
			} else{
				m_log.LogTrace("  return = generic_failure");
				return generic_failure;
			}
		}
	}

	m_log.LogTrace("  return = time_out");
	return time_out;
}



/**
*	Originates an application via AMI originate
*	\param channel 
*	\param application 
*	\param data 
*	\param variable
*	\param timeout in ms
*	\return ok, generic_failure, time_out
*
*	\code
*	Action: Originate 
*	Channel: channel 
*	Application: app-name 
*	Data: app-data 
*	Variable:	variable_0 
*	ActionID: 2 
*	Async: 1 
*	\endcode
*/
AC_RESULT	ASTProvider::OriginateApplication(std::string channel, 
											std::string application, 
											std::string data, 
											std::string variable, DWORD timeout)
{
	m_log.LogTrace("ASTProvider::OriginateApplication channel:%s application:%s (%s) variable:%s timeout=%d", 
		channel.c_str(), application.c_str(), data.c_str(), variable.c_str(), timeout);

	std::string	invokeId = m_invokeIdManager.NewInvokeId();
	ManagerResponse	result ;
    std::string command;
	command = "Action: Originate\r\n";
	command += "Channel: " + channel ;
	command += "\r\n";
	command += "Application: " + application ;
	command += "\r\n";
	command += "Data: " + data ;
	command += "\r\n";
	command += "Variable: " + variable ;
	command += "\r\n";
	command += "ActionID: " + invokeId;
	command += "\r\n";
	if (timeout > 0){
		//timeout must be ms
		command += "Timeout: " + ToString(timeout*1000);
		command += "\r\n";
	}
	command += "Async: 0";		// synch for now
	command += "\r\n\r\n";
	SendCommand(command);

	if (m_invokeIdManager.WaitResponse(invokeId, result, AST_TIMEOUT)) {
		if (result.m_response.compare("success")==0) {
			m_log.LogTrace("  success");
			return ok;
		}
		else {
			/// \todo: more error handling
			m_log.LogTrace("  generic_failure");
			return generic_failure;
		}
	}

	m_log.LogTrace("  time_out");
	return time_out;
}


/**
*	Redirect (or transfer) a call
*	\param channel
*	\param extraChannel
*	\param extension
*	\param context
*	\return ok, generic_failure, time_out
*
*	\code
*	Action: Redirect 
*	Channel: channel
*	Exten: 70 
*	Priority: 1 
*	Context: from-internal 
*	ActionID: 2 
*	Async: 1 
*	\endcode
*	\see http://www.voip-info.org/wiki/index.php?page=Asterisk+Manager+API+Action+Redirect
*/
AC_RESULT	ASTProvider::Redirect(	std::string channel, 
									std::string extraChannel, 
									std::string extension, 
									std::string context)
{
	m_log.LogTrace("ASTProvider::Redirect channel:%s extraChannel:%s extension:%s context:%s", 
		channel.c_str(), extraChannel.c_str(), extension.c_str(), context.c_str());

	std::string	invokeId = m_invokeIdManager.NewInvokeId();
	ManagerResponse	result ;
    std::string command;

	command = "Action: Redirect\r\n";
	command += "Channel: " + channel ;
	command += "\r\n";
	if (extraChannel.size()>0) {
		command += "ExtraChannel: " + extraChannel ;
		command += "\r\n";
	}
	command += "Exten: " + extension ;
	command += "\r\n";
	command += "Context: " + context ;
	command += "\r\n";
	command += "Priority: 1" ;
	command += "\r\n";
	/* Redirect action does not support Async mode  */
	command += "ActionID: " + invokeId;
	command += "\r\n\r\n";
	SendCommand(command);

	if (m_invokeIdManager.WaitResponse(invokeId, result, AST_TIMEOUT)) {
		if (result.m_response.compare("success")==0) {
			m_log.LogTrace("  success");
			return ok;
		}
		else {
			/// \todo: more error handling
			m_log.LogTrace("  generic_failure");
			return generic_failure;
		}
	}

	m_log.LogTrace("  time_out");
	return time_out;
}


/**
*	Monitor given channel
*	\see http://www.voip-info.org/wiki/view/Asterisk+Manager+API+Action+Monitor
*	\param channel
*	\param file
*	\param format
*	\param mix
*	\return ok, no_such_channel, generic_failure, time_out
*	Example:
*	\code
*	ACTION: Monitor
*	Channel: SIP/x7062618529-643d
*	File: channelsavefile
*	Format: gsm
*	Mix: true 
*	\endcode
*/
AC_RESULT	ASTProvider::Monitor(std::string channel, std::string file, std::string format, std::string mix)
{
	m_log.LogTrace("ASTProvider::Monitor channel=%s file=%s format=%s mix=%s ", channel.c_str(), file.c_str(), format.c_str(), mix.c_str());

	std::string	invokeId = m_invokeIdManager.NewInvokeId();
	ManagerResponse	result ;

	std::string command;

	command = "Action: Monitor\r\n";
	command += "Channel: " + channel;
	command += "\r\n";
	command += "file: " + file;
	command += "\r\n";
	command += "format: " + format;
	command += "\r\n";
	command += "mix: " + mix;
	command += "\r\n";
	command += "ActionID: " + invokeId;
	command += "\r\n";
	command += "Async: 1";
	command += "\r\n\r\n";
	SendCommand(command);

	if (m_invokeIdManager.WaitResponse(invokeId, result, AST_TIMEOUT)) {
		if (result.m_response.compare("success")==0) {
			m_log.LogTrace("  return = ok");
			return ok;
		}else {
			if (result.m_message.compare("No such channel") == 0){
				m_log.LogError("  return = no such channel");
				return no_such_channel;
			} else{
				m_log.LogTrace("  return = generic_failure");
				return generic_failure;
			}
		}
	}

	m_log.LogTrace("  return = time_out");
	return time_out;
}

/**
*	StopMonitor given channel
*	\see http://www.voip-info.org/wiki/view/Asterisk+Manager+API+Action+StopMonitor
*	\param channel
*	\return ok, no_such_channel, generic_failure, time_out
*	Example:
*	\code
*	ACTION: StopMonitor
*	Channel: SIP/x7062618529-643d
*	\endcode
*/
AC_RESULT	ASTProvider::StopMonitor(std::string channel)
{
	m_log.LogTrace("ASTProvider::StopMonitor channel=%s", channel.c_str());

	std::string	invokeId = m_invokeIdManager.NewInvokeId();
	ManagerResponse	result ;

	std::string command;

	command = "Action: StopMonitor\r\n";
	command += "Channel: " + channel;
	command += "\r\n";
	command += "ActionID: " + invokeId;
	command += "\r\n";
	command += "Async: 1";
	command += "\r\n\r\n";
	SendCommand(command);

	if (m_invokeIdManager.WaitResponse(invokeId, result, AST_TIMEOUT)) {
		if (result.m_response.compare("success")==0) {
			m_log.LogTrace("  return = ok");
			return ok;
		}else {
			if (result.m_message.compare("No such channel") == 0){
				m_log.LogError("  return = no such channel");
				return no_such_channel;
			} else{
				m_log.LogTrace("  return = generic_failure");
				return generic_failure;
			}
		}
	}

	m_log.LogTrace("  return = time_out");
	return time_out;
}

/**
*	PauseMonitor given channel
*	\param channel
*	\return ok, no_such_channel, generic_failure, time_out
*	Example:
*	\code
*	ACTION: PauseMonitor
*	Channel: SIP/x7062618529-643d
*	\endcode
*/
AC_RESULT	ASTProvider::PauseMonitor(std::string channel)
{
	m_log.LogTrace("ASTProvider::PauseMonitor channel=%s", channel.c_str());

	std::string	invokeId = m_invokeIdManager.NewInvokeId();
	ManagerResponse	result ;

	std::string command;

	command = "Action: PauseMonitor\r\n";
	command += "Channel: " + channel;
	command += "\r\n";
	command += "ActionID: " + invokeId;
	command += "\r\n";
	command += "Async: 1";
	command += "\r\n\r\n";
	SendCommand(command);

	if (m_invokeIdManager.WaitResponse(invokeId, result, AST_TIMEOUT)) {
		if (result.m_response.compare("success")==0) {
			m_log.LogTrace("  return = ok");
			return ok;
		}else {
			if (result.m_message.compare("No such channel") == 0){
				m_log.LogError("  return = no such channel");
				return no_such_channel;
			} else{
				m_log.LogTrace("  return = generic_failure");
				return generic_failure;
			}
		}
	}

	m_log.LogTrace("  return = time_out");
	return time_out;
}

/**
*	UnpauseMonitor given channel
*	\param channel
*	\return ok, no_such_channel, generic_failure, time_out
*	Example:
*	\code
*	ACTION: PauseMonitor
*	Channel: SIP/x7062618529-643d
*	\endcode
*/
AC_RESULT	ASTProvider::UnpauseMonitor(std::string channel)
{
	m_log.LogTrace("ASTProvider::UnpauseMonitor channel=%s", channel.c_str());

	std::string	invokeId = m_invokeIdManager.NewInvokeId();
	ManagerResponse	result ;

	std::string command;

	command = "Action: UnpauseMonitor\r\n";
	command += "Channel: " + channel;
	command += "\r\n";
	command += "ActionID: " + invokeId;
	command += "\r\n";
	command += "Async: 1";
	command += "\r\n\r\n";
	SendCommand(command);

	if (m_invokeIdManager.WaitResponse(invokeId, result, AST_TIMEOUT)) {
		if (result.m_response.compare("success")==0) {
			m_log.LogTrace("  return = ok");
			return ok;
		}else {
			if (result.m_message.compare("No such channel") == 0){
				m_log.LogError("  return = no such channel");
				return no_such_channel;
			} else{
				m_log.LogTrace("  return = generic_failure");
				return generic_failure;
			}
		}
	}

	m_log.LogTrace("  return = time_out");
	return time_out;
}

/**
*	ChangeMonitor given channel
*	\see http://www.voip-info.org/wiki/view/Asterisk+Manager+API+Action+Monitor
*	\param channel
*	\param file
*	\return ok, no_such_channel, generic_failure, time_out
*	Example:
*	\code
*	ACTION: ChangeMonitor
*	Channel: SIP/x7062618529-643d
*	File: channelsavefile
*	\endcode
*/
AC_RESULT	ASTProvider::ChangeMonitor(std::string channel, std::string file)
{
	m_log.LogTrace("ASTProvider::ChangeMonitor channel=%s file=%s", channel.c_str(), file.c_str());

	std::string	invokeId = m_invokeIdManager.NewInvokeId();
	ManagerResponse	result ;

	std::string command;

	command = "Action: ChangeMonitor\r\n";
	command += "Channel: " + channel;
	command += "\r\n";
	command += "file: " + file;
	command += "\r\n";
	command += "ActionID: " + invokeId;
	command += "\r\n";
	command += "Async: 1";
	command += "\r\n\r\n";
	SendCommand(command);

	if (m_invokeIdManager.WaitResponse(invokeId, result, AST_TIMEOUT)) {
		if (result.m_response.compare("success")==0) {
			m_log.LogTrace("  return = ok");
			return ok;
		}else {
			if (result.m_message.compare("No such channel") == 0){
				m_log.LogError("  return = no such channel");
				return no_such_channel;
			} else{
				m_log.LogTrace("  return = generic_failure");
				return generic_failure;
			}
		}
	}

	m_log.LogTrace("  return = time_out");
	return time_out;
}