/*
	Name: call_map.cpp
	Copyright: Under the Mozilla Public License Version 1.1 or later
	Author: ACTIVA
	Date: August/06
*/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is astprovider.
 *
 * The Initial Developer of the Original Code is ACTIVA Team.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): none
 *
 * ***** END LICENSE BLOCK ***** */


#include "astcstaprovider.h"
#include "astinterface.h"
#include "call_map.h"

#define MAX_LINKED_CALLS	10	//< max levels of nested calls in ASTCallMap::GetSourceCall()
#define MAX_CALLID			10000	//< max generated callid


/**
*	Construct a map with Asterisk calls and CSTA calls
*/
ASTCallMap::ASTCallMap(ASTLogListener *logger)
{
	InitializeCriticalSection(&m_csCallMapLock);
	m_currentCallId = 1 ;
	m_log.SetOrigin("ASTCallMap");
	m_log.SetLogger(logger);
}

/**
*	destruct the map
*/
ASTCallMap::~ASTCallMap()
{
	// relese memory and clean maps
	CleanupMaps();
	DeleteCriticalSection(&m_csCallMapLock);
}

/**
*	Enter into a critical section
*/
void	ASTCallMap::__WaitForMapLock() 
{ 
	EnterCriticalSection(&m_csCallMapLock); 
};

/**
*	Leaves a critical section
*/
void	ASTCallMap::__ReleaseMapLock() 
{ 
	LeaveCriticalSection(&m_csCallMapLock);
};

/**
*	Add a new reference on m_mapDeviceId2Xref map
*	\param address device or extension
*	\param newxref number of monitor
*/
void	ASTCallMap::AddXref(std::string	address, MonitorXref newxref)
{
	__WaitForMapLock();

	// add to m_mapExtension2Xref multimap
	if (m_mapDeviceId2Xref.find(address)==m_mapDeviceId2Xref.end()) {	// none found
		m_log.LogTrace("  first monitor for this device: {%s,%d}", address.c_str(), newxref);
		Set_xref *pv = new Set_xref();
		pv->insert(newxref);
		m_mapDeviceId2Xref[address] = pv;
	}
	else {
		m_log.LogTrace("  add new monitor for this device");
		m_mapDeviceId2Xref[address]->insert(newxref);
	}

	__ReleaseMapLock();
}


/**
*	Removes a reference from the monitor map
*	\param xref reference to search
*	\return true if the reference was removed correctly
*/
bool	ASTCallMap::RemoveXref(MonitorXref xref)
{
	// protect map access with mutex
	__WaitForMapLock();

	// search xref in map and delete it
	Set_xref *setXref;
	// for each extension
	for (Map_deviceId2setXref::iterator i = m_mapDeviceId2Xref.begin(); i!=m_mapDeviceId2Xref.end(); i++) {
		setXref = i->second ;
		if (setXref->find(xref)!=setXref->end()) {	
			m_log.LogTrace("  erase xref map entry");
			setXref->erase(xref);
			if (setXref->size()==0) { // if set gets empty we need to erase the extension entry
				m_log.LogTrace("    erase extension map entry");
				delete setXref;
				m_mapDeviceId2Xref.erase(i);
			}
			__ReleaseMapLock();
			return true;
		}
	}

	__ReleaseMapLock();
	return false;
}

/**
*	Return the device of the monitor
*	\param xref monitor reference
*	\return string with device
*/
std::string	ASTCallMap::GetDeviceFromMonitor(MonitorXref xref)
{
	__WaitForMapLock();

	// search xref in map and return it
	Set_xref *setXref;
	// for each extension
	for (Map_deviceId2setXref::iterator i = m_mapDeviceId2Xref.begin(); i!=m_mapDeviceId2Xref.end(); i++) {
		setXref = i->second ;
		if (setXref->find(xref) != setXref->end()) {
			__ReleaseMapLock();
			return i->first;
		}
	}

	__ReleaseMapLock();
	m_log.LogTrace("  xref not found!");
	return "";
}


/**
*	Update CSTA call map	[callId] -> device -> state
*	\param callId identifier of the call
*	\param deviceId device
*	\param deviceState The state of device: can be Initialited, Alerting, Connected, Failed, Held.
*/
void	ASTCallMap::AddConnectionToCallModel(DWORD callId, std::string deviceId, csta_states deviceState)
{
	m_log.LogTrace("AddConnectionToCallModel {%d, %s} => %s", callId, deviceId.c_str(), CSTAState2Str(deviceState).c_str());

	// protect map access with mutex
	__WaitForMapLock();

	//find call
	if (m_mapCallId2DeviceId.find(callId)==m_mapCallId2DeviceId.end()) {
		//not exist any call with this callId
		Set_CSTADevice *pDeviceSet = new Set_CSTADevice();
		if (!pDeviceSet) {
			__ReleaseMapLock();
			m_log.LogError(" NOT ENOUGH MEMORY: new Set_CSTADevice() FAILED" );
			return;
		}
		// warning C6211: Leaking memory 'pDeviceSet' due to an exception. Consider using a local catch block to clean up memory: Lines: 165, 167, 170, 173, 175, 176, 181
		CSTADevice *pDevice = NULL;
		try {
			pDevice = new CSTADevice();
		}
		catch(...) {
			if (!pDevice) {
				m_log.LogError(" NOT ENOUGH MEMORY: new CSTADevice() FAILED" );
				delete pDeviceSet;
				__ReleaseMapLock();
				return;
			}
		}
		pDevice->deviceId = deviceId;
		pDevice->deviceState = deviceState;
		pDeviceSet->insert(pDevice);
		m_mapCallId2DeviceId[callId] = pDeviceSet;
	} else {
		Set_CSTADevice *pDeviceSet = m_mapCallId2DeviceId[callId];
		bool exist = false;
		for (Set_CSTADevice::iterator i = pDeviceSet->begin(); i != pDeviceSet->end(); i++){
			CSTADevice *pDevice (*i);
			if (pDevice->deviceId == deviceId){
				pDevice->deviceState = deviceState;
				exist = true;
				break;
			}
		}
		if (!exist){
			CSTADevice *pDevice  = new CSTADevice;
			pDevice->deviceId = deviceId;
			pDevice->deviceState = deviceState;
			pDeviceSet->insert(pDevice);	// this works even if deviceId already in set
		}
	}

	m_log.LogTrace(" done, now CALL map has %d items", m_mapCallId2DeviceId.size() );
	__ReleaseMapLock();
}

/**
*	Returns the state of device specify on CSTA call map.
*	\param callId identifier of the call
*	\param deviceId identifier of device
*	\return string with state of device
*/
csta_states	ASTCallMap::GetStateOfDevice(DWORD callId, std::string deviceId)
{
	m_log.LogTrace("GetStateOfDevice {%d, %s}", callId, deviceId.c_str());

	// protect map access with mutex
	__WaitForMapLock();

	//find call
	if (m_mapCallId2DeviceId.find(callId)==m_mapCallId2DeviceId.end()) {	// none found
		m_log.LogTrace(" call %d not found", callId);
	}else {
		Set_CSTADevice *pDeviceSet = m_mapCallId2DeviceId[callId];
		for (Set_CSTADevice::iterator i = pDeviceSet->begin(); i != pDeviceSet->end(); i++){
			CSTADevice *pDevice (*i);
			if (pDevice->deviceId == deviceId){
				m_log.LogTrace("   deviceState = %s", CSTAState2Str(pDevice->deviceState).c_str() );
				__ReleaseMapLock();
				return pDevice->deviceState;
			}
		}
		m_log.LogTrace(" not found device %s with callId %d", deviceId.c_str(), callId);
	}

	__ReleaseMapLock();
	return UNKNOWN ;
}


/** 
*	Removes a connection from CSTA call map.
*	\param callId identifier of the call
*	\param deviceId identifier of device
*/
void	ASTCallMap::RemoveConnectionFromCallModel(DWORD callId, std::string deviceId)
{
	m_log.LogTrace("RemoveConnection {%d}=> %s", callId, deviceId.c_str());

	// protect map access with mutex
	__WaitForMapLock();

	//find call
	if (m_mapCallId2DeviceId.find(callId)==m_mapCallId2DeviceId.end()) {	// none found
		__ReleaseMapLock();
		m_log.LogTrace("    {%d} NOT IN CALL-MONITOR MAP", callId);
		return;
	}
	else {
		Set_CSTADevice *pDeviceSet = m_mapCallId2DeviceId[callId];
		for (Set_CSTADevice::iterator i = pDeviceSet->begin(); i != pDeviceSet->end(); i++){
			CSTADevice *pDevice(*i);
			if (pDevice->deviceId == deviceId){
				pDeviceSet->erase(pDevice);
				break;
			}
		}
		if (pDeviceSet->size()==0) {
			pDeviceSet->clear();
			delete pDeviceSet;
			m_mapCallId2DeviceId.erase(callId);
		}
	}
    
	__ReleaseMapLock();
}

/**
*	Removes a call from CSTA call map.
*	\param callId
*/
void	ASTCallMap::RemoveCallFromCallModel(DWORD callId)
{
	m_log.LogTrace("RemoveCall {%d}", callId);

	// protect map access with mutex
	__WaitForMapLock();

	m_callIdSet.erase(callId);		// remove it from active callId set

	//find call
	if (m_mapCallId2DeviceId.find(callId)==m_mapCallId2DeviceId.end()) {	// none found
		__ReleaseMapLock();
		m_log.LogTrace("    {%d} NOT IN CALL-MONITOR MAP", callId);
		return;
	}
	else {
		Set_CSTADevice *pDeviceSet = m_mapCallId2DeviceId[callId];
		for (Set_CSTADevice::iterator i = pDeviceSet->begin(); i != pDeviceSet->end(); i++){
			delete *i;
		}
		pDeviceSet->clear();
		delete pDeviceSet;
		m_mapCallId2DeviceId.erase(callId);
	}
    
	__ReleaseMapLock();
}

/**
 * Get a set of reference with the same callId.
 *	\param callId identifier of call.
 *	\param setXref set of reference.
 *	\return true if exist any call with this callId, otherwise false
 */
bool	ASTCallMap::GetMonitorSetFromCallModel(DWORD callId, Set_xref & setXref)
{
	m_log.LogTrace("GetMonitorSetFromCallModel = %d",callId);

	setXref.clear();

	__WaitForMapLock();	// protect map access with mutex

	//find call
	if (m_mapCallId2DeviceId.find(callId)==m_mapCallId2DeviceId.end()) {	// none found
		__ReleaseMapLock();
		m_log.LogTrace("  callerId not found");
		return false;
	} else {
		Set_CSTADevice *pDeviceSet = m_mapCallId2DeviceId[callId];
		// ok so we have a pdeviceSet here...
		for (Set_CSTADevice::iterator i = pDeviceSet->begin(); i!=pDeviceSet->end(); i++) {
			CSTADevice *pDevice(*i);
			// get Set_xref for each deviceId and add to setXref
			if (m_mapDeviceId2Xref.find(pDevice->deviceId)!=m_mapDeviceId2Xref.end()) {
				Set_xref *pSet = m_mapDeviceId2Xref[pDevice->deviceId];
				setXref.insert(pSet->begin(),pSet->end());
			} 
			
			std::string deviceId;
			size_t extPos = -1;
			if ((extPos = (pDevice->deviceId).find("/")) != -1) {	// generic channel handling
				deviceId = (pDevice->deviceId).substr(extPos+1, (pDevice->deviceId).size());
			}

			if (m_mapDeviceId2Xref.find(deviceId)!=m_mapDeviceId2Xref.end()) {
				Set_xref *pSet = m_mapDeviceId2Xref[deviceId];
				setXref.insert(pSet->begin(),pSet->end());
			} 
		}
	} 

	__ReleaseMapLock();
	m_log.LogTrace("  num_monitors = %d", setXref.size());
	return true;
}


/**
 * Get a call with uniqueId specify. Otherwise, create one with a new callID.
 *	\param _uniqueId 
 *	\return an AsteriskCall with call
 */
AsteriskCall*	ASTCallMap::GetOrCreateCall(std::string _uniqueId) {
	AsteriskCall* _call ;

	// protect map access with mutex
	__WaitForMapLock();

	if (m_uniqueId2Asteriskcall.find(_uniqueId)==m_uniqueId2Asteriskcall.end()) {
		_call = AsteriskCall::Create(_uniqueId, "", GetNewCallId(),"");
  		m_uniqueId2Asteriskcall[_uniqueId] = _call;
		m_log.LogTrace("Create AsteriskCall with uniqueId = %s and callid =%d", _uniqueId.c_str(), _call->GetCallId());
	}
	else {
		_call = m_uniqueId2Asteriskcall[_uniqueId];
	}

	__ReleaseMapLock();
	return _call;
}


/**
 * Find an AsteriskCall with a specify uniqueId
 *	\param _uniqueId
 *	\return an AsteriskCall or null if does not exist any call in m_uniqueId2Asteriskcall.
 */
AsteriskCall*	ASTCallMap::FindCallByUniqueId(std::string _uniqueId) 
{
	AsteriskCall* _call ;
	
	// protect map access with mutex
	__WaitForMapLock();
	
	if (m_uniqueId2Asteriskcall.find(_uniqueId)==m_uniqueId2Asteriskcall.end()) {
		__ReleaseMapLock();
		return NULL;
	}
	else {
		_call = m_uniqueId2Asteriskcall[_uniqueId];
	}

	__ReleaseMapLock();
	return _call;
}


/**
 * Find an AsteriskCall with a specify channel
 *	\param _channel
 *	\return an AsteriskCall or null if it doesn't exist
 */
AsteriskCall*	ASTCallMap::FindCallByChannel(std::string _channel) {
	// protect map access with mutex
	__WaitForMapLock();

	for (Map_UniqueId2AsteriskCall::iterator i = m_uniqueId2Asteriskcall.begin(); i!=m_uniqueId2Asteriskcall.end(); i++) {
		AsteriskCall	*pAC = i->second;
		std::string channel = pAC->GetChannel().substr(0,_channel.size());
		if (_channel==channel) {
			__ReleaseMapLock();
			return pAC;
		}
	}

	__ReleaseMapLock();
	return NULL;
}
/**
*	Find an AsteriskCall with a specify Address.
*	\param address
*	\return an AsteriskCall or null if it doesn't exist.
*/
AsteriskCall* ASTCallMap::FindCallByAddress(std::string address)
{
	__WaitForMapLock();

	for (Map_UniqueId2AsteriskCall::iterator i = m_uniqueId2Asteriskcall.begin(); i!=m_uniqueId2Asteriskcall.end(); i++) {
		AsteriskCall	*pAC = i->second;
		if (address==pAC->GetAddress()) {
			__ReleaseMapLock();
			return pAC;
		}
	}

	__ReleaseMapLock();
	return NULL;
}

/**
 * Find an AsteriskCall with a specify callId.
 * \param callId
 * \return the fist AsteriskCall or null if it doesn't exist
 */
AsteriskCall*	ASTCallMap::FindCallByCallId(DWORD callId) {
	// protect map access with mutex
	__WaitForMapLock();

	for (Map_UniqueId2AsteriskCall::iterator i = m_uniqueId2Asteriskcall.begin(); i!=m_uniqueId2Asteriskcall.end(); i++) {
		AsteriskCall	*pAC = i->second;
		if (callId==pAC->GetCallId()) {
			__ReleaseMapLock();
			return pAC;
		}
	}

	__ReleaseMapLock();
	return NULL;
}

/**
 * Find an AsteriskCall with a specify callID and device.
 *	\param callId
 *	\param device
 *	\return the fist AsteriskCall or null if it doesn't exist
 */
AsteriskCall* ASTCallMap::FindCallByCallIdAndDevice(DWORD callId, std::string device)
{
	// protect map access with mutex
	__WaitForMapLock();

	for (Map_UniqueId2AsteriskCall::iterator i = m_uniqueId2Asteriskcall.begin(); i!=m_uniqueId2Asteriskcall.end(); i++){
		AsteriskCall *pAC = i->second;
//		if (atoi(device.c_str()) == atoi(pAC->GetAddress().c_str()) && pAC->GetCallId() == callId){
		if (device.c_str() == pAC->GetAddress().c_str() && pAC->GetCallId() == callId) {
			__ReleaseMapLock();
			return i->second;
		}
	}
	//if not found anything with direct CallId and device, search as a logic CallId.
	for (Map_UniqueId2AsteriskCall::iterator i = m_uniqueId2Asteriskcall.begin(); i!=m_uniqueId2Asteriskcall.end(); i++){
		AsteriskCall *pAC = i->second;
		if (atoi(device.c_str()) == atoi(pAC->GetAddress().c_str())){
			DWORD logicCallId = GetLogicCallId(pAC);
			if (logicCallId == callId){
				__ReleaseMapLock();
				return i->second;
			}
		}
	}

	__ReleaseMapLock();
	return NULL;
}

/**
*	Find another call (different of callid) with the same CID 
*	\param cid
*	\param callid
*	\return the fist AsteriskCall or null if it doesn't exist
*/
AsteriskCall*	ASTCallMap::_deprecated__FindCallByCID(std::string cid, DWORD callid)
{
	m_log.LogTrace("ASTCallMap::FindCallByCID(cid = %s, callid = %d)", cid.c_str(), callid);
	__WaitForMapLock();

	for (Map_UniqueId2AsteriskCall::iterator i = m_uniqueId2Asteriskcall.begin(); i!=m_uniqueId2Asteriskcall.end(); i++){
		AsteriskCall *pAC = i->second;
		if (callid != GetLogicCallId(pAC)){
			std::string callerId, callerIdName, userData;
			GetCIDInfoFromSourceCall(pAC, callerId, callerIdName, userData);
			if (cid == callerId){
				__ReleaseMapLock();
				return pAC;
			}
		}
	}

	__ReleaseMapLock();
	return NULL;
}

/**
 * Walks through Asterisk call map and returns the source/originator call. That call hasn't got sourceUniqueId.
 * we cannot assert there are no oops in the walk so implementation is not recursive.
 * walk limited to MAX_LINKED_CALLS deep by implementation.
 *	\param pAC
 *	\return Source AsteriskCall or null if there is one or itself is none
 */
AsteriskCall	* ASTCallMap::GetSourceCall(AsteriskCall *pAC) {
	m_log.LogTrace("ASTCallMap::GetSourceCall(0x%0x)", pAC);

	// protect map access with mutex
	__WaitForMapLock();
	if (!pAC) {
		__ReleaseMapLock();
		m_log.LogTrace(" GetSourceCall() with NULL argument");
		return NULL;
	}

	AsteriskCall	*pCurrentCall = pAC;
	int	nCounter = MAX_LINKED_CALLS ;
	// loop inside
	do {
		if (pCurrentCall && pCurrentCall->HasSourceUniqueId()) {
			std::string	sourceUniqueId = pCurrentCall->GetSourceUniqueId();
			AsteriskCall	*pInnerCall = FindCallByUniqueId(sourceUniqueId);
			if (pInnerCall) {
				pCurrentCall = pInnerCall;
			}
			else {	// for some reason hasourcecall is true but call not yet in the call-map. 
					// this can happen if two dial events related to same source call so second 'win'
					// So we keep last one
				m_log.LogTrace("  SourceCall is %s", pCurrentCall->GetUniqueId().c_str());
				__ReleaseMapLock();
				return pCurrentCall;
			}
		}
		else { // no more inner calls
			m_log.LogTrace("  SourceCall is %s", pCurrentCall->GetUniqueId().c_str());
			__ReleaseMapLock();
			return pCurrentCall;
		}
	} while (--nCounter > 0);	//

	// no real call so it is itself
	m_log.LogWarn("GetSourceCall TOO DEEP. Returning call 0x%0x", pCurrentCall);
	__ReleaseMapLock();

	return pCurrentCall;
}

/**
*	Find the callId of the call. The first AsteriskCall contains the logic CallId.
*	\param pAC
*	\return Returns the callId of the source call.
*/
DWORD		ASTCallMap::GetLogicCallId(AsteriskCall *pAC)
{
	m_log.LogTrace0("ASTCallMap::GetLogicCallId(0x%0x)", pAC);

	AsteriskCall	*pRootAC = pAC;
	int	nCounter = MAX_LINKED_CALLS ;

	__WaitForMapLock();
	while (nCounter>0 && pRootAC && pRootAC->HasSourceUniqueId()){
		if (pRootAC->HasCSTALink()){
			break;
		}
		pRootAC = FindCallByUniqueId(pRootAC->GetSourceUniqueId());
		nCounter--;
	}

	if (nCounter==0) {
		__ReleaseMapLock();
		m_log.LogWarn("GetLogicCallId TOO DEEP. Returning callId=-1");
		return (DWORD) -1;
	}

	if (!pRootAC) {
		__ReleaseMapLock();
		m_log.LogWarn("GetLogicCallId pRootAC=NULL");
		return (DWORD) -1;
	}

	DWORD	dwCallId = pRootAC->GetCallId();

	__ReleaseMapLock();

	return dwCallId;
}

std::string		ASTCallMap::GetLogicAddress(AsteriskCall *pAC)
{
	m_log.LogTrace("ASTCallMap::GetLogicAddress(0x%0x)", pAC);

	AsteriskCall	*pRootAC = pAC;
	int	nCounter = MAX_LINKED_CALLS ;

	__WaitForMapLock();
	while(nCounter>0 && pRootAC && pRootAC->HasSourceUniqueId()){
		if (pRootAC->HasCSTALink())
			break;
		AsteriskCall *pAux = FindCallByUniqueId(pRootAC->GetSourceUniqueId());
		if (pAux && (pAux->GetNumDevice() == pRootAC->GetNumDevice()))
			pRootAC = pAux;
		else
			break;

		nCounter--;
	}

	if (nCounter==0) {
		__ReleaseMapLock();
		m_log.LogWarn("GetLogicAddress TOO DEEP. Returning void address");
		return "";
	}

	if (!pRootAC) {
		__ReleaseMapLock();
		m_log.LogWarn("GetLogicAddress pRootAC=NULL");
		return "";
	}

	std::string	sAddress = pRootAC->GetAddress();	

	__ReleaseMapLock();

	return sAddress;
}

/**
*	Search all callId that one device can have got
*	\param device
*	\param setCallId set of callId with device specify
*/
void	ASTCallMap::GetSetCallIdByDevice(std::string device, Set_callId &setCallId)
{
	setCallId.clear();

	__WaitForMapLock();

	for(Map_callId2setDeviceId::iterator i = m_mapCallId2DeviceId.begin(); i!=m_mapCallId2DeviceId.end(); i++){
		Set_CSTADevice *pSetDevice = i->second;
		bool found = false;
		for (Set_CSTADevice::iterator j = pSetDevice->begin(); j != pSetDevice->end(); j++){
			CSTADevice *pDevice(*j);
			if (pDevice->deviceId == device){
				found = true;
				break;
			}
		}
		if (found){
			setCallId.insert(i->first);
		}
	}

	__ReleaseMapLock();
}

/**
*	Search  set devices with one callId
*	\param callId
*	\param setDevices by reference. 
*/
void ASTCallMap::GetSetDevicesFromCallId(DWORD callId, Set_deviceId &setDevices)
{
		// protect map access with mutex
	setDevices.clear();

	__WaitForMapLock();

	for (Map_callId2setDeviceId::iterator i = m_mapCallId2DeviceId.begin(); i != m_mapCallId2DeviceId.end(); i++){
		if (callId == i->first){
			Set_CSTADevice *pSetDevice = i->second;
			for (Set_CSTADevice::iterator j = pSetDevice->begin(); j != pSetDevice->end(); j++){
				CSTADevice *pDevice = *j;
				setDevices.insert(pDevice->deviceId);
			}
		}
	}

	__ReleaseMapLock();
}

/**
 *  Extract callerID, callerIDname and userData from a call
 *	\param pAC
 *	\param cid by reference
 *	\param cidname by reference
 *	\param userData by reference
 *	\return false if the AsteriskCall specify is null or if the map of Asterisk calls is too deep.
 */
bool	ASTCallMap::GetCIDInfoFromSourceCall(AsteriskCall *pAC, std::string &cid, std::string &cidname, std::string &userData) 
{
	// protect map access with mutex
	__WaitForMapLock();

	if (!pAC) {
		__ReleaseMapLock();
		m_log.LogTrace("EXIT ASTCallMap::GetCIDInfoFromSourceCall() pAC is NULL");
		return false;
	}

	m_log.LogTrace("GetCIDInfoFromSourceCall(%s)", pAC->GetUniqueId().c_str());

	//if Asterisk was informed it (since 1.8.5.0) 
	//it seems a better way to do it
	if(pAC->GetConnectedLineNum()!="" || pAC->GetConnectedLineName()!="")
	{
		cid			= pAC->GetConnectedLineNum();
		cidname		= pAC->GetConnectedLineName();
		userData	= pAC->GetUserData();
		__ReleaseMapLock();
		m_log.LogTrace("  ConnectedLineNum/Name found in current AsteriskCall.");
		m_log.LogTrace("  Exiting with cid/cidname/userData=%s/%s/%s", cid.c_str(), cidname.c_str(), userData.c_str());
		return true;
	}

	AsteriskCall	*pCurrentCall = pAC;
	int	nCounter = MAX_LINKED_CALLS ;
	std::string	newCid;
	std::string	newCidName;
	std::string newUserData;

	// loop inside
	do {
		if (pCurrentCall) {
			newCid = pCurrentCall->GetCallerId();
			newCidName = pCurrentCall->GetCallerIdName();
			newUserData = pCurrentCall->GetUserData();
			if (newCid != "") {	// update cid/cidname pair if cid is not <unknown>
				cid = newCid;
				cidname = newCidName;
				userData = newUserData;
			}
		}

		if (pCurrentCall && pCurrentCall->HasSourceUniqueId()) {
			std::string	sourceUniqueId = pCurrentCall->GetSourceUniqueId();
			pCurrentCall = FindCallByUniqueId(sourceUniqueId);
		}
		else { // no more inner calls
			__ReleaseMapLock();
			m_log.LogTrace("  Exiting with cid/cidname/userData=%s/%s/%s", cid.c_str(), cidname.c_str(), userData.c_str());
			return true;
		}
	} while (--nCounter > 0);	//

	// no real call so it is itself
	__ReleaseMapLock();

	m_log.LogTrace("  GetCIDInfoFromSourceCall TOO DEEP. Exiting with cid/cidname/userData=%s/%s/%s", cid.c_str(), cidname.c_str(), userData.c_str());
	return false;
}

/**
 * Looks for the set of calls (identified by its uniqueid) that have a given sourceUniqueId.
 * Search is based on list-scan so its complexity is O(n)
 *	\param sourceUniqueId
 *	\param setUniqueid by reference
 */
void	ASTCallMap::GetDestinationCallSet(std::string sourceUniqueId, Set_uniqueId & setUniqueid)
{
	setUniqueid.clear();

	__WaitForMapLock();	// protect map access with mutex

	// scan uniqueid table
	std::string	uniqueId;
	for (Map_UniqueId2AsteriskCall::iterator i = m_uniqueId2Asteriskcall.begin(); i!=m_uniqueId2Asteriskcall.end(); i++) {
		uniqueId = i->first;
		AsteriskCall *pAC = i->second;
		if (pAC && (pAC->GetSourceUniqueId()==sourceUniqueId || pAC->GetSourceUniqueId() == "CSTA"+sourceUniqueId) ) {	// found one of them
			setUniqueid.insert(uniqueId);
		}
	} // for

	__ReleaseMapLock();
}

/**
*	Find if a call specify with callId has only local channels
*	\param callId
*	\return True if only has local channels. Otherwise return false.
*/
bool	ASTCallMap::IsLocalCall(DWORD callId)
{
	__WaitForMapLock();

	for (Map_UniqueId2AsteriskCall::iterator i = m_uniqueId2Asteriskcall.begin(); i!=m_uniqueId2Asteriskcall.end(); i++) {
		AsteriskCall *pAC = i->second;
		if (GetLogicCallId(pAC) == callId){
			if (!pAC->IsLocalChannel()){
				__ReleaseMapLock();
				m_log.LogTrace("  IsLocalCall (%d) false.", callId);
				return false;
			}
		}
	}

	__ReleaseMapLock();
	m_log.LogTrace("  IsLocalCall (%d) true.", callId);
	return true;
}
/**
*	Scan if an AsteriskCall has a destination
*	\param pAC
*	\return true if has destination, otherwise return false
*/
bool	ASTCallMap::HasDestinationCall(AsteriskCall *pAC)
{
	__WaitForMapLock();	// protect map access with mutex

	// scan uniqueid table
	std::string	myUniqueId = pAC->GetUniqueId();
	for (Map_UniqueId2AsteriskCall::iterator i = m_uniqueId2Asteriskcall.begin(); i!=m_uniqueId2Asteriskcall.end(); i++) {
		AsteriskCall *pAC2 = i->second;
		if (pAC2 && (pAC2->GetSourceUniqueId()==myUniqueId) ) {	// found one of them
			__ReleaseMapLock();
			return true;
		}
	}

	__ReleaseMapLock();
	return false;
}
/**
*	Search if exist an entry in CSTAMap with the specific callId and device
*	\param callId
*	\param device
*	\return Retuns true or false if exist an entry in CSTAMap with the specific callId and device
*/
bool		ASTCallMap::_deprecated__ExistCSTACall(DWORD callId, std::string device)
{
	__WaitForMapLock();

	for (Map_callId2setDeviceId::iterator i = m_mapCallId2DeviceId.begin(); i != m_mapCallId2DeviceId.end(); i++){
		if (i->first == callId){
			Set_CSTADevice *pSetDevice;
			pSetDevice = i->second;
			for (Set_CSTADevice::iterator j = pSetDevice->begin(); j != pSetDevice->end(); j++){
				CSTADevice *pDevice(*j);
				if (pDevice->deviceId == device){
					__ReleaseMapLock();
					return true;
				}
			}
		}
	}

	__ReleaseMapLock();
	return false;
}

/**
*	Remove m_uniqueId2Asteriskcall map 
*	\param uniqueId
*/
void		ASTCallMap::RemoveCallByUniqueId(std::string uniqueId) 
{
	__WaitForMapLock();
	m_uniqueId2Asteriskcall.erase(uniqueId);
	__ReleaseMapLock();
}


/**
 *	Cleanup all call maps
 */
void	ASTCallMap::CleanupMaps()
{
	// protect map access with mutex
	__WaitForMapLock();

	// clear m_id2call map
	for (Map_UniqueId2AsteriskCall::iterator i = m_uniqueId2Asteriskcall.begin(); i!=m_uniqueId2Asteriskcall.end(); i++) {
		AsteriskCall	*pAC = i->second;
		delete pAC;
	}
	m_uniqueId2Asteriskcall.clear();

	// for each device close all its monitors
	for (Map_deviceId2setXref::iterator i = m_mapDeviceId2Xref.begin(); i!=m_mapDeviceId2Xref.end(); i++) {
		std::string	deviceID = i->first;
		Set_xref	*pSetXref = i->second;
		pSetXref->clear();
		delete pSetXref;
	}
	m_mapDeviceId2Xref.clear();

	for (Map_callId2setDeviceId::iterator k = m_mapCallId2DeviceId.begin(); k!=m_mapCallId2DeviceId.end(); k++) {
		DWORD	callId = k->first;
		Set_CSTADevice *pSetDeviceId = k->second;
		pSetDeviceId->clear();
		delete pSetDeviceId;
	}
	m_mapCallId2DeviceId.clear();

	__ReleaseMapLock();
}

/**
 *	Dump all call maps
 */
void	ASTCallMap::DumpMaps()
{
	// protect map access with mutex
	__WaitForMapLock();

	// clear m_id2call map
	m_log.LogTrace0("================================================================= start callmap dump...");
	m_log.LogTrace0("DUMP OF ASTERISK CALLS MAP: {UNIQUEID} => call info:");
	for (Map_UniqueId2AsteriskCall::iterator i = m_uniqueId2Asteriskcall.begin(); i!=m_uniqueId2Asteriskcall.end(); i++) {
		std::string	uniqueid = i->first;
		AsteriskCall	*pAC = i->second;
		if (pAC) 
			m_log.LogTrace0("  %s",pAC->DumpToString().c_str());
		else 
			m_log.LogTrace0("  {%s} = > NULL !!!!", uniqueid.c_str());
	}

	m_log.LogTrace0("DUMP OF MONITOR CALLS MAP: {CALL_ID} => DEVICE: => STATE");
	// for each device close all its monitors
	for (Map_callId2setDeviceId::iterator i = m_mapCallId2DeviceId.begin(); i!=m_mapCallId2DeviceId.end(); i++) {
		DWORD callId = i->first;
		Set_CSTADevice *pset_deviceId  = i->second;
		for (Set_CSTADevice::iterator j = pset_deviceId->begin(); j!=pset_deviceId->end(); j++) {
			CSTADevice *device(*j);
			m_log.LogTrace0("    [%d] => %s => %s", callId, device->deviceId.c_str(), CSTAState2Str(device->deviceState).c_str() );
		}
	}

	m_log.LogTrace0("================================================================= end callmap dump...");

	__ReleaseMapLock();
}

/**
*	Convert from csta_state to std::string
*/
std::string	ASTCallMap::CSTAState2Str(csta_states state)
{
	switch(state){
		case INITIALIZED:
			return "INITIALIZED";
		case ALERTING:
			return "ALERTING";
		case CONNECTED:
			return "CONNECTED";
		case FAILED:
			return "FAILED";
		case HELD:
			return "HELD";
		case QUEUED:
			return "QUEUED";
		default:
			return "?";
	}
	return "";
}

/** 
*	Generate unique callID
*	\return new CallID
*/
DWORD	ASTCallMap::GetNewCallId()
{
	__WaitForMapLock();

	if (m_callIdSet.size()>=MAX_CALLID) { // map is full :-(
		__ReleaseMapLock();
		return 0;
	}

	do {
		m_currentCallId++ ;
		if (m_currentCallId > MAX_CALLID) {
			m_currentCallId=1;
		}
	} while (m_callIdSet.find(m_currentCallId) != m_callIdSet.end());

	m_callIdSet.insert(m_currentCallId);
	DWORD	callId = m_currentCallId;

	__ReleaseMapLock();
    
    return callId;
}

bool	ASTCallMap::DropCallId(DWORD callId)
{
	__WaitForMapLock();
	m_callIdSet.erase(callId);		// remove it from active callId set
	__ReleaseMapLock();
	return true;
}
