/*
	Name: call_map.h
	Copyright: Under the Mozilla Public License Version 1.1 or later
	Author: ACTIVA
	Date: AUG/06
*/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is astprovider.
 *
 * The Initial Developer of the Original Code is ACTIVA.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): none
 *
 * ***** END LICENSE BLOCK ***** */


#pragma once

#ifndef CALL_MAP_H
#define CALL_MAP_H

#include <string>
#include <map>
#include <vector>
#include <set>

#include "astprovider.h"



/**
*	TAPI connection states
*/
typedef enum{
	UNKNOWN = -1,		///<unknown
	INITIALIZED = 1,	///< the state where the device is requesting service. Usually this results in the creation of a call
	ALERTING = 2,		///< the state where the device is alerting
	CONNECTED = 3,		///< the state where the device can comunicate with other devices on a call
	FAILED = 4,			///< the state where normal state progression has been aborted
	HELD = 5,			///< the call is on hold at the device
	QUEUED = 6			///< the state where normal state progression has been stalled. For example, the call is beeing processed by an ACD
} csta_states;

/* some typedefs */
//typedef struct {std::string deviceId; std::string deviceState;} CSTADevice; //< Device CSTA with its state
typedef struct {std::string deviceId; csta_states deviceState;} CSTADevice; //< Device CSTA with its state
typedef std::set<CSTADevice*> Set_CSTADevice;	//< set of CSTADevices
typedef	std::set<std::string> Set_uniqueId ;	//< set of uniqueids
typedef std::map< std::string, Set_xref* >	Map_deviceId2setXref ;	//< map: device_id to set_of_monitor_xrefs
typedef	std::set<std::string> Set_deviceId ;	//< Set of extensions/device_id
typedef std::map< DWORD, Set_CSTADevice* >	Map_callId2setDeviceId ;	//< map: callId to set_of_CSTADevice
typedef std::set<DWORD> Set_callId;


///	 This class administer all call maps of ActivaTSP

/** 
*	It encapsulates three different callmaps:
*
*	- call-map: A {uniqueID => AsteriskCall} call map.
*	- address-monitor-map: An {address => monitorcrossreference} map that stores the address's list of monitor-cross-references.
*	- call-monitor: really a call-model that mantains calls. Uniqueid related to the addresses interested in their events.
*/
class	ASTCallMap
{
public:
		ASTCallMap(ASTLogListener *pLog);
		~ASTCallMap();

	void	SetLogger(ASTLogListener *logger) {m_log.SetLogger(logger);};

    void	__WaitForMapLock() ;		// locks call map
	void	__ReleaseMapLock() ;		// releases call map

	// adds a new {deviceid, xref} on to the address-monitor-map 
	void	AddXref(std::string	deviceId, MonitorXref newxref);

	// removes a xref from the address-monitor-map
	bool	RemoveXref(MonitorXref xref);

	// get device from monitor
	std::string	GetDeviceFromMonitor(MonitorXref xref);

	// adds a connection to the 'call-monitor' map
	void	AddConnectionToCallModel(DWORD callId, std::string deviceId, csta_states deviceState);

	// set new status to call
	void	SetStatusToCallModel(DWORD callId,std::string status);

	// Get the state of Device in a call
	csta_states	GetStateOfDevice(DWORD callId, std::string device);

	// removes a connection from the 'call-monitor' map
	void	RemoveConnectionFromCallModel(DWORD callId, std::string deviceId);

	// completely removes a call from 'call-monitor' map
	void	RemoveCallFromCallModel(DWORD callId);

	// returns the monitor set linked to the given callid, from the 'interested extensions' map
	bool	GetMonitorSetFromCallModel(DWORD callId, Set_xref & setXref);

	//---------------------------------------------------------------------------
	// looks in call map and returns AsteriskCall; if not in map creates a new one
	AsteriskCall *	GetOrCreateCall(std::string _uniqueId) ;

	// returns AsteriskCall or null if it does not exist
	AsteriskCall *	FindCallByUniqueId(std::string uniqueId) ;

	// returns AsteriskCall or null if it does not exist
	AsteriskCall *	FindCallByChannel(std::string channel) ;

	// returns AsteriskCall or null if it does not exist
	AsteriskCall *	FindCallByAddress(std::string address);

	// returns AsteriskCall or null if it does not exist
	AsteriskCall *	FindCallByCallId(DWORD callId) ;

	// Find Call by a callId and a CallerId
	AsteriskCall*	FindCallByCallIdAndDevice(DWORD callId, std::string device);

	// Find call (different of callId) by CID 
	AsteriskCall*	_deprecated__FindCallByCID(std::string cid, DWORD callId);

	// returns 'source/originator call' if there is one, itself otherwise
	AsteriskCall *	GetSourceCall(AsteriskCall *pAC) ;

	// returns logic callId
	DWORD			GetLogicCallId(AsteriskCall *pAC);

	std::string		GetLogicAddress(AsteriskCall *pAC);

	//returns set of callId with the same device
	void			GetSetCallIdByDevice(std::string device, Set_callId &setCallId);

	// walks the call linked list and gets cid/cidname
	bool			GetCIDInfoFromSourceCall(AsteriskCall *pAC, std::string &cid, std::string &cidname, std::string &userData) ;

	// returns the set of calls wich have sourceUniqueId as its sourceUniqueId
	void			GetDestinationCallSet(std::string sourceUniqueId, Set_uniqueId & setUniqueid);

	// returns true if all AsteriskCall with logic CallId has Local channels
	bool			IsLocalCall(DWORD callId);

	// returns true if there is some other call pointing to this call
	bool			HasDestinationCall(AsteriskCall *pAC);

	// returns true if call exist in CSTAMap, otherwise retuns false
	bool			_deprecated__ExistCSTACall(DWORD callId, std::string device);

	// removes uniqueid call from map
	void			RemoveCallByUniqueId(std::string uniqueId) ;

	// Return set devices with one callId
	void			GetSetDevicesFromCallId(DWORD callId, Set_deviceId &setDevices);

	// Cleans connection and asteriskCall maps
	void			CleanupMaps();

	// dump to traces
	void			DumpMaps();

	// set/get state
	void			SetState(std::string state){m_state = state;};
	std::string		GetState(){return m_state;};

	static std::string		CSTAState2Str(csta_states state);

	size_t			Size(){return m_uniqueId2Asteriskcall.size();};

	size_t			CSTAStateMapSize(){return m_mapCallId2DeviceId.size();};

	virtual DWORD	GetNewCallId() sealed;
	bool			DropCallId(DWORD callid);


protected:
	ASTLogger			m_log;
	CRITICAL_SECTION	m_csCallMapLock;
	std::string			m_state;

	/// set of active callId. Used by GetNewCallId() and DropCallId(callid)
	Set_callId			m_callIdSet;
	DWORD				m_currentCallId;

	// 'interested extensions' map

	/// map of deviceId (extensions) to list of references (xref).
	Map_deviceId2setXref	m_mapDeviceId2Xref ;		
	/// map of callId to list of devices (extensions) 
	Map_callId2setDeviceId	m_mapCallId2DeviceId ;		

	/// map of {uniqueId, AsteriskCall}
	Map_UniqueId2AsteriskCall	m_uniqueId2Asteriskcall ;

};

#endif	// CALL_MAP_H
