/*
	Name: event_triggers.cpp
	Copyright: Under the Mozilla Public License Version 1.1 or later
	Author: ACTIVA
*/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is astprovider.
 *
 * The Initial Developer of the Original Code is ACTIVA.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): none
 *
 * ***** END LICENSE BLOCK ***** */


#include "astcstaprovider.h"
#include <algorithm>


/**
 * Event minimization rules:
 *
 * Asterisk Manager Interface (AMI) sends many information in a number of different (and may be not enough 
 * coordinated) events. So we found that direct transformation of AMI events to csta-like events lent us to a 
 * too 'verbose' or 'noisy' event-flow.
 *
 * Running a number of standard scenarios we looked-up for event-minimization-rules to reduce 
 * the unnecesary csta-events. Most of the rules are empirical but effective.
 *
 * Current rules:
 * RULE-1: Do not send ORIGINATED if callerid is unknown 
 * RULE-2: Do not send ALERTING/ESTABLISHED/FAILED to SOURCE when SOURCE.address==this.address
 * RULE-3: Do not send ESTABLISHED if callerid="" and calleddevice==address
 * RULE-4: Do not send ALERTING/ESTABLISHED/FAILED/QUEUED/DEQUEUED if caller==called==(address||numDevice)
 * RULE-5: Do not send CONNCLEARED to SOURCE if source.address==this.address
 * RULE-6: Do not send CONNCLEARED if there is a destination and destination.address==this.address
 * RULE-7: Do not send ALERTING/ESTABLISHED/FAILED when callerid is unknown and calling='': 
 *			if returns on calling=void acd callflow is bad; if do not return outbound is bad => so only returns 
 *			when both calling=void AND caller is unknown
 * RULE-8: Do not send ESTABLISHED if there is a destination call: second established is wrong 
 *			(indicates audio is up but established was already sent)
 * RULE-9: Do not send EVENTS to local channels except for queues.
 * RULE-10: Do not send DISCONNECTED to local channels. But send it if call has no destination.
 * RULE-11: Do not send EVENTS to AsyncGoto channels
 */


void	ASTCstaProvider::fireOnOriginatedCallback(ConnectionOriginated &xConnectionOriginated)
{
	Set_xref set_Monitors;
	GetMonitorSet(xConnectionOriginated.m_dwCallId, set_Monitors);
	if (set_Monitors.size() == 0){
		m_log.LogTrace("  do not send EVENTS because not exist any monitors in call %d", xConnectionOriginated.m_dwCallId);
		return;
	}

	for (Set_xref::iterator i = set_Monitors.begin(); i!=set_Monitors.end(); i++) {
		m_log.LogTrace("    sending event:CFTE:ORIGINATED to xref=%d, address=%s, callid=%d", *i, xConnectionOriginated.m_sAddress.c_str(), xConnectionOriginated.m_dwCallId);
		m_cstaListener->OnOriginated(*i, xConnectionOriginated.m_dwCallId, xConnectionOriginated.m_sAddress, xConnectionOriginated.m_sCallerId, xConnectionOriginated.m_sCallerIdName, "","", xConnectionOriginated.m_eCause);
	}
}


/**
 * A call is in progress in the address.
 */
bool	ASTCstaProvider::handleConnectionOriginated (AsteriskCall *pCall, call_cause cause, ConnectionOriginated &xConnectionOriginated) 
{
	std::string	address = pCall->GetAddress() ;
	m_log.LogTrace("ASTCstaProvider::handleConnectionOriginated uniqueId=%s, address=%s", pCall->GetUniqueId().c_str(), address.c_str());

	// RULE-9: Do not send EVENTS to local channels
	if (pCall->IsLocalChannel()) {
		m_log.LogTrace("  event minimization RULE-9: Do not send EVENTS to local channels");
		return false;
	}

	//*	RULE-11: Do not send EVENTS to AsyncGoto channels */
	if ((pCall->GetChannel()).find("AsyncGoto") != -1){
		m_log.LogTrace("  event minimization RULE-11: Do not send EVENTS to AsyncGoto channels");
		return false;
	}

	m_callMap.DumpMaps();
	m_callMap.AddConnectionToCallModel(pCall->GetCallId(), address, INITIALIZED /*"INITIALITED"*/);	

	// RULE-1: Do not send ORIGINATED if callerid is unknown
	if (pCall->GetCallerId()!="" && AsteriskCall::IsUnknown(pCall->GetCallerId())) {
		m_log.LogTrace("  event minimization RULE-1: caller is unknown");
		return false;
	}

	// fill in ConnectionOriginated
	xConnectionOriginated.m_dwCallId = pCall->GetCallId();
	xConnectionOriginated.m_sAddress = address;
	xConnectionOriginated.m_sCallerId = pCall->GetCallerId();
	xConnectionOriginated.m_sCallerIdName = pCall->GetCallerIdName();
	xConnectionOriginated.m_sCalled = "";
	xConnectionOriginated.m_eCause = cause;

	return true;
};



void	ASTCstaProvider::fireOnAlertingCallback(ConnectionAlerting &xConnectionAlerting)
{
	Set_xref set_Monitors ;
	GetMonitorSet(xConnectionAlerting.m_dwCallId, set_Monitors);	// get alerting device near-end set
	if (set_Monitors.size() == 0){
		m_log.LogWarn("  do not send EVENTS because not exist any monitors in call %d", xConnectionAlerting.m_dwCallId);
		return;
	}
	
	for (Set_xref::iterator i = set_Monitors.begin(); i!=set_Monitors.end(); i++) {
		m_log.LogTrace("    sending event:CFTE:ALERTING to xref=%d, address=%s, callid=%d, lastRedirection=%s, userData=%s", *i, xConnectionAlerting.m_sAddress.c_str(), xConnectionAlerting.m_dwCallId, xConnectionAlerting.m_sLastRedirection.c_str(), xConnectionAlerting.m_sUserData.c_str());
		m_cstaListener->OnAlerting(*i, xConnectionAlerting.m_dwCallId, xConnectionAlerting.m_sAddress, xConnectionAlerting.m_sCallerId, xConnectionAlerting.m_sCallerIdName, xConnectionAlerting.m_sCalled, xConnectionAlerting.m_sLastRedirection, xConnectionAlerting.m_sUserData, xConnectionAlerting.m_eCause);
	}
}

/**
 * Address is alerting.
 */
bool	ASTCstaProvider::handleConnectionAlerting( AsteriskCall *pCall, call_cause cause, ConnectionAlerting &xConnectionAlerting) 
{
	const std::string	address = pCall->GetAddress();
	m_log.LogTrace("ASTCstaProvider::handleConnectionAlerting uniqueId=%s, address=%s", pCall->GetUniqueId().c_str(), address.c_str());

	AsteriskCall	*pACSource = m_callMap.GetSourceCall(pCall);	// walk the call map

	DWORD	dwCallId = m_callMap.GetLogicCallId(pCall);

	// RULE-9: Do not send EVENTS to local channels
	//if (pCall->IsLocalChannel() && !pACSource->IsPredictiveCall()) {
	if (pCall->IsLocalChannel()) {
		m_log.LogTrace("  event minimization RULE-9: Do not send EVENTS to local channels");
		return false;
	}

	// FIX: It is giving 'trunk used' instead of called number
	std::string		callerid, calleridname, userData;
	std::string		called = pCall->GetCallerId();

	m_callMap.GetCIDInfoFromSourceCall(pCall, callerid, calleridname, userData);

	if (callerid == called && pACSource->IsPredictiveCall()){
		m_log.LogTrace("  overriding calledNumber from %s => N/A", called.c_str());
		called="";
	}
	
	// RULE-7: Do not send ALERTING/ESTABLISHED when callerid is unknown and calling=''
	if (pCall->GetCallerId()!="" && AsteriskCall::IsUnknown(pCall->GetCallerId())) {
		if (callerid.size()==0) {
			m_log.LogTrace("  event minimization RULE-7: callerid is unknown and calling==VOID");
			return false;
		}
	}

	// RULE-3: Do not send ALERTING/ESTABLISHED/FAILED if callerid="" and calleddevice==address
	if (pCall->GetCallerId()!="" && AsteriskCall::IsUnknown(pCall->GetCallerId()) && called==address) {
		m_log.LogTrace("  event minimization RULE-3: callerid is unknown and calleddevice==address==%s", pCall->GetAddress().c_str());
		return false;
	}

	//*	RULE-11: Do not send EVENTS to AsyncGoto channels */
	if ((pCall->GetChannel()).find("AsyncGoto") != -1){
		m_log.LogTrace("  event minimization RULE-11: Do not send EVENTS to AsyncGoto channels");
		return false;
	}

	// TBD for DEBUG purposes!!!
	m_callMap.DumpMaps();
	m_callMap.AddConnectionToCallModel(dwCallId, address, ALERTING /*"ALERTING"*/);

	// RULE-4: Do not send ALERTING/ESTABLISHED/FAILED/QUEUED/DEQUEUED if caller==called==(address||numDevice)
	if (callerid==called && !pACSource->IsPredictiveCall()) {
		m_log.LogTrace("  event minimization RULE-4: (callerid = %s) == (called = %s) == (address = %s)",callerid.c_str(), called.c_str(), pCall->GetAddress().c_str());
		return false;
	}

	std::string	lastRedirection = pACSource->GetLastRedirectionDevice();
	std::string cidnamePrefix = AsteriskCall::GetCIDNamePrefix(calleridname);

	// as a last resort we will read the cidname prefix, if any
	if (lastRedirection=="") {
		lastRedirection = cidnamePrefix;
	}

	// if there is a cidnamePrefix use it as called
	if (cidnamePrefix!="") {
		m_log.LogTrace("  overriding calledNumber from %s => %s", called.c_str(), cidnamePrefix.c_str());
		called = cidnamePrefix;
	}else{
		m_log.LogTrace("  overriding calledNumber from %s => %s", called.c_str(), pCall->GetCallerId().c_str());
		called = pCall->GetCallerId();
	}

	m_log.LogTrace("  source call is {%s} and lastRedirection=[%s]", pACSource->GetUniqueId().c_str(), lastRedirection.c_str());

	// fill xConnectionAlerting
	xConnectionAlerting.m_dwCallId = dwCallId;
	xConnectionAlerting.m_sAddress = address;
	xConnectionAlerting.m_sCallerId = callerid;
	xConnectionAlerting.m_sCallerIdName = calleridname;
	xConnectionAlerting.m_sCalled = called;
	xConnectionAlerting.m_sLastRedirection = lastRedirection;
	xConnectionAlerting.m_sUserData = userData;
	xConnectionAlerting.m_eCause = cause;

	return true;
};

void	ASTCstaProvider::fireOnEstablishedCallback(ConnectionConnected &xConnectionConnected)
{
	Set_xref set_Monitors;
	GetMonitorSet(xConnectionConnected.m_dwCallId, set_Monitors);
	if (set_Monitors.size() == 0){
		m_log.LogWarn("  do not send EVENTS because not exist any monitors in call %d", xConnectionConnected.m_dwCallId);
	}

	for (Set_xref::iterator i = set_Monitors.begin(); i!=set_Monitors.end(); i++) {
		m_log.LogTrace("    sending event:CFTE:ESTABLISHED to xref=%d, address=%s, callid=%d, lastRedirection=%s, userData=%s", *i, xConnectionConnected.m_sAddress.c_str(), xConnectionConnected.m_dwCallId, xConnectionConnected.m_sLastRedirection.c_str(), xConnectionConnected.m_sUserData.c_str());
		m_cstaListener->OnEstablished(*i, xConnectionConnected.m_dwCallId, xConnectionConnected.m_sAddress, xConnectionConnected.m_sCallerId, xConnectionConnected.m_sCallerIdName, xConnectionConnected.m_sCalled, xConnectionConnected.m_sLastRedirection, xConnectionConnected.m_sUserData, xConnectionConnected.m_eCause);
	}
}

/**
 * Address is connected.
 */
bool	ASTCstaProvider::handleConnectionConnected(	AsteriskCall *pCall, call_cause cause, ConnectionConnected &xConnectionConnected) 
{
	const std::string	address = pCall->GetAddress() ;
	m_log.LogTrace("ASTCstaProvider::handleConnectionConnected uniqueId=%s, address=%s", pCall->GetUniqueId().c_str(), address.c_str());

	AsteriskCall	*pACSource = m_callMap.GetSourceCall(pCall);
	if (!pACSource) {
		m_log.LogWarn("  pACSource is NULL");
		return false;
	}

	DWORD	dwCallId = pACSource->GetCallId();

	// RULE-9: Do not send EVENTS to local channels
	if (pCall->IsLocalChannel() && (!pACSource->IsPredictiveCall() || m_callMap.IsLocalCall(pCall->GetCallId()))) {
		m_log.LogTrace("  event minimization RULE-9: Do not send EVENTS to local channels");
		return false;
	}

	// FIX: It was giving 'trunk used' instead of called number
	std::string	called = pCall->GetCallerId();

	std::string	callerid, calleridname, userData;
	m_callMap.GetCIDInfoFromSourceCall(pCall, callerid, calleridname, userData);

	if (callerid == called && pACSource->IsPredictiveCall())
		called="";

	// RULE-7: Do not send ALERTING/ESTABLISHED when callerid is unknown and calling=''
	if (pCall->GetCallerId()!="" && AsteriskCall::IsUnknown(pCall->GetCallerId())) {
		if (callerid.size()==0) {
			m_log.LogTrace("  event minimization RULE-7: callerid is unknown and calling==VOID");
			return false;
		}
	}

	// RULE-3: Do not send ALERTING/ESTABLISHED/FAILED if callerid is unknown and calleddevice==address
	if (pCall->GetCallerId()!="" && AsteriskCall::IsUnknown(pCall->GetCallerId()) && called==address) {
		m_log.LogTrace("  event minimization RULE-3: callerid is unknown and calleddevice==address==%s", address.c_str());
		return false;
	}

	//* RULE-8: Do not send ESTABLISHED if there is a destination call: second established is wrong 
	// (indicates audio is up but established was already sent)
	if (m_callMap.HasDestinationCall(pCall)) {
		m_log.LogTrace("  event minimization RULE-8: Do not send ESTABLISHED if there is a destination call: second established is wrong (indicates audio is up but established was already sent)");
		return false;
	}

	//*	RULE-11: Do not send EVENTS to AsyncGoto channels */
	if ((pCall->GetChannel()).find("AsyncGoto") != -1){
		m_log.LogTrace("  event minimization RULE-11: Do not send EVENTS to AsyncGoto channels");
		return false;
	}

	// TBD for DEBUG purposes!!!
	m_callMap.DumpMaps();
	m_callMap.AddConnectionToCallModel(dwCallId, address, CONNECTED /*"CONNECTED"*/);	

	// RULE-4: Do not send ALERTING/ESTABLISHED/FAILED/QUEUED/DEQUEUED if caller==called==(address||numDevice)
	if (callerid==called && !pACSource->IsPredictiveCall()) {
		m_log.LogTrace("  event minimization RULE-4: (callerid = %s) == (called = %s) == (address = %s)",callerid.c_str(), called.c_str(), pCall->GetAddress().c_str());
		return false;
	}

	std::string	lastRedirection = pACSource->GetLastRedirectionDevice();
	std::string cidnamePrefix = AsteriskCall::GetCIDNamePrefix(calleridname);

	// as a last resort we will read the cidname prefix, if any
	if (lastRedirection=="") {
		lastRedirection = cidnamePrefix;
	}

	// if there is a cidnamePrefix use it as called
	if (cidnamePrefix!="") {
		m_log.LogTrace("  overriding calledNumber from %s => %s", called.c_str(), cidnamePrefix.c_str());
		called = cidnamePrefix;
	}

	m_log.LogTrace("  source call is {%s} and lastRedirection=%s", pACSource->GetUniqueId().c_str(), lastRedirection.c_str());

	// fill ConnectionConnected &xConnectionConnected
	xConnectionConnected.m_dwCallId = dwCallId;
	xConnectionConnected.m_sAddress = address;
	xConnectionConnected.m_sCallerId = callerid;
	xConnectionConnected.m_sCalled = called;
	xConnectionConnected.m_sCallerIdName = calleridname;
	xConnectionConnected.m_sLastRedirection = lastRedirection;
	xConnectionConnected.m_sUserData = userData;
	xConnectionConnected.m_eCause = cause;

	return true;
};


void	ASTCstaProvider::fireOnFailedCallback(ConnectionFailed &xConnectionFailed)
{
	Set_xref set_Monitors;
	GetMonitorSet(xConnectionFailed.m_dwCallId, set_Monitors);
	if (set_Monitors.size() == 0){
		m_log.LogWarn("  do not send EVENTS because not exist any monitors in call %d", xConnectionFailed.m_dwCallId);
	}

	for (Set_xref::iterator i = set_Monitors.begin(); i!=set_Monitors.end(); i++) {
		m_log.LogTrace("    sending event:CFTE:FAILED to xref=%d, address=%s, callid=%d", *i, xConnectionFailed.m_sAddress.c_str(), xConnectionFailed.m_dwCallId);
		m_cstaListener->OnFailed(*i, xConnectionFailed.m_dwCallId, xConnectionFailed.m_sAddress, xConnectionFailed.m_sCallerId, xConnectionFailed.m_sCallerIdName, xConnectionFailed.m_sCalled, xConnectionFailed.m_sLastRedirection,"", xConnectionFailed.m_eCause);
	}
}

/**
 * Call to Address has failed.
 */
bool	ASTCstaProvider::handleConnectionFailed(AsteriskCall *pCall, call_cause cause, ConnectionFailed & xConnectionFailed) 
{
	const std::string	address = pCall->GetAddress() ;
	m_log.LogTrace("ASTCstaProvider::handleConnectionFailed uniqueId=%s, address=%s, cause=%d", pCall->GetUniqueId().c_str(), address.c_str(), cause);

	AsteriskCall	*pACSource = m_callMap.GetSourceCall(pCall);
	if (!pACSource) {
		m_log.LogWarn("  pACSource is NULL");
		return false;
	}
	DWORD	dwCallId = pACSource->GetCallId();

	// RULE-9: Do not send EVENTS to local channels
	if (pCall->IsLocalChannel() && !pACSource->IsPredictiveCall()) {
		m_log.LogTrace("  event minimization RULE-9: Do not send EVENTS to local channels");
		return false;
	}

	// FIX: It is giving 'trunk used' instead of called number
	std::string	called = pCall->GetCallerId();

	std::string	callerid, calleridname, userData;
	m_callMap.GetCIDInfoFromSourceCall(pCall, callerid, calleridname, userData);

	// RULE-7: Do not send ALERTING/ESTABLISHED/FAILED when callerid is unknown and calling=''
	if (pCall->GetCallerId()!="" && AsteriskCall::IsUnknown(pCall->GetCallerId())){
		if (callerid.size()==0) {
			m_log.LogTrace("  event minimization RULE-7: callerid is unknown and calling==VOID");
			return false;
		}
	}

	// RULE-3: Do not send ALERTING/ESTABLISHED/FAILED if callerid is unknown and calleddevice==address
	if (pCall->GetCallerId()!="" && AsteriskCall::IsUnknown(pCall->GetCallerId()) && called==pCall->GetAddress()) {
		m_log.LogTrace("  event minimization RULE-3: callerid is unknown and calleddevice==address==%s", pCall->GetAddress().c_str());
		return false;
	}

	//	RULE-11: Do not send EVENTS to AsyncGoto channels 
	if ((pCall->GetChannel()).find("AsyncGoto") != -1){
		m_log.LogTrace("  event minimization RULE-11: Do not send EVENTS to AsyncGoto channels");
		return false;
	}

	// TBD for DEBUG purposes!!!
	m_callMap.DumpMaps();
	m_callMap.AddConnectionToCallModel(dwCallId, address, FAILED /*"FAILED"*/);

	// RULE-4: Do not send ALERTING/ESTABLISHED/FAILED/QUEUED/DEQUEUED if caller==called==(address||numDevice)
	if (callerid==called && (called==pCall->GetAddress() ||called==pCall->GetNumDevice())) {
		m_log.LogTrace("  event minimization RULE-4: (callerid = %s) == (called = %s) == (address = %s)",callerid.c_str(), called.c_str(), pCall->GetAddress().c_str());
		return false;
	}

	std::string	lastRedirection = pACSource->GetLastRedirectionDevice();
	std::string cidnamePrefix = AsteriskCall::GetCIDNamePrefix(calleridname);

	// as a last resort we will read the cidname prefix, if any
	if (lastRedirection=="") {
		lastRedirection = cidnamePrefix;
	}

	// if there is a cidnamePrefix use it as called
	if (cidnamePrefix!="") {
		m_log.LogTrace("  overriding calledNumber from %s => %s", called.c_str(), cidnamePrefix.c_str());
		called = cidnamePrefix;
	}

	m_log.LogTrace("  source call is {%s} and lastRedirection=%s", pACSource->GetUniqueId().c_str(), lastRedirection.c_str());

	// fill ConnectionFailed
	xConnectionFailed.m_dwCallId = dwCallId;
	xConnectionFailed.m_sAddress = address;
	xConnectionFailed.m_sCallerId = callerid;
	xConnectionFailed.m_sCallerIdName = calleridname;
	xConnectionFailed.m_sCalled = called;
	xConnectionFailed.m_sLastRedirection = lastRedirection;
	xConnectionFailed.m_eCause = cause;

	return true;
};



void	ASTCstaProvider::fireOnConnectionClearedCallback(ConnectionDisconnected &xConnectionDisconnected)
{
	Set_xref set_Monitors;
	GetMonitorSet(xConnectionDisconnected.m_dwCallId, set_Monitors);
	if (set_Monitors.size() == 0){
		m_log.LogWarn("  do not send EVENTS because not exist any monitors in call %d", xConnectionDisconnected.m_dwCallId);
	}

	for (Set_xref::iterator i = set_Monitors.begin(); i!=set_Monitors.end(); i++) {
		std::string deviceMonitor = m_callMap.GetDeviceFromMonitor(*i);
		m_log.LogTrace("    sending event:CFTE:CONNCLEARED to xref=%d, address=%s, callid=%d", *i, deviceMonitor.c_str(), xConnectionDisconnected.m_dwCallId);
		m_cstaListener->OnConnectionCleared(*i, xConnectionDisconnected.m_dwCallId, xConnectionDisconnected.m_sAddress, xConnectionDisconnected.m_eCause); 
	}
}

/**
 * Call has been disconnected from address.
 * Caller method will remove pCall from call-map and call-model.
 */
bool	ASTCstaProvider::handleConnectionDisconnected(AsteriskCall *pCall, call_cause cause, ConnectionDisconnected &xConnectionDisconnected) 
{
	const std::string	logicAddress = m_callMap.GetLogicAddress(pCall) ;

	m_log.LogTrace("ASTCstaProvider::handleConnectionDisconnected uniqueId=%s, logicaddress=%s, cause=%d", pCall->GetUniqueId().c_str(), logicAddress.c_str(), cause);

	//AsteriskCall	*pACSource = m_callMap.GetSourceCall(pCall);
	DWORD	dwCallId = m_callMap.GetLogicCallId(pCall);

	// TBD for DEBUG purposes!!!
	m_callMap.DumpMaps();

	// RULE-9: Do not send EVENTS to local channels
	if (pCall->IsLocalChannel()) {
		m_log.LogTrace("  event minimization RULE-10: Do not send DISCONNECTED to local channels. But send it if call has no destination.");
		return false;
	}

	//	RULE-11: Do not send EVENTS to AsyncGoto channels */
	if ((pCall->GetChannel()).find("AsyncGoto") != -1){
		m_log.LogTrace("  event minimization RULE-11: Do not send EVENTS to AsyncGoto channels");
		return false;
	}

	// fill xConnectionDisconnected
	xConnectionDisconnected.m_dwCallId = dwCallId;
	xConnectionDisconnected.m_sAddress = logicAddress;
	xConnectionDisconnected.m_eCause = cause;

	// can't RemoveConnectionFromCallModel() before fireOnConnectionClearedCallback() because nothing would get fired...	
	// m_callMap.RemoveConnectionFromCallModel(dwCallId, pCall->GetAddress());
	// m_callMap.DumpMaps();

	return true;
};

void	ASTCstaProvider::fireOnCallClearedCallback(CallDisconnected &xCallDisconnected)
{
	Set_xref set_Monitors;
	GetMonitorSet(xCallDisconnected.m_dwCallId, set_Monitors);
	if (set_Monitors.size() == 0){
		m_log.LogWarn("  do not send EVENTS because not exist any monitors in call %d", xCallDisconnected.m_dwCallId);
	}

	for (Set_xref::iterator i = set_Monitors.begin(); i!=set_Monitors.end(); i++) {
		// RULE-5: Do not send CONNCLEARED to SOURCE if source.address==this.address
		m_log.LogTrace("    sending event:CFTE:CALLCLEARED to xref=%d, callid=%d", *i, xCallDisconnected.m_dwCallId);
		m_cstaListener->OnCallCleared(*i, xCallDisconnected.m_dwCallId, "", xCallDisconnected.m_eCause);
	}
}

/**
 * Call has been disconnected.
 * Caller method will completely remove dwCallId rom call-map and call-model.
 */
bool	ASTCstaProvider::handleCallDisconnected(AsteriskCall *pCall, DWORD dwCallId, call_cause cause, CallDisconnected & xCallDisconnected) 
{
	m_log.LogTrace("ASTCstaProvider:: CallId=%d", dwCallId);

	if (!pCall)
		return false;

	//*	RULE-11: Do not send EVENTS to AsyncGoto channels */
	if ((pCall->GetChannel()).find("AsyncGoto") != -1){
		m_log.LogTrace("  event minimization RULE-11: Do not send EVENTS to AsyncGoto channels");
		return false;
	}

	//* RULE-5: Do not send CONNCLEARED to SOURCE if source.address==this.address

	// fill CallDisconnected
	xCallDisconnected.m_dwCallId = dwCallId;
	xCallDisconnected.m_eCause = cause;

	return true;
};


void	ASTCstaProvider::fireOnQueuedCallback(ConnectionQueued &xConnectionQueued)
{
	Set_xref set_Monitors;
	GetMonitorSet(xConnectionQueued.m_dwCallId, set_Monitors);
	if (set_Monitors.size() == 0){
		m_log.LogWarn("  do not send EVENTS because not exist any monitors in call %d", xConnectionQueued.m_dwCallId);
	}

	for (Set_xref::iterator i = set_Monitors.begin(); i!=set_Monitors.end(); i++) {
		m_log.LogTrace("    sending event:CFTE:QUEUED to xref=%d, address=%s, userData=%s", *i, xConnectionQueued.m_sAddress.c_str(), xConnectionQueued.m_sUserData.c_str());
		m_cstaListener->OnQueued(*i, xConnectionQueued.m_dwCallId, xConnectionQueued.m_sAddress, xConnectionQueued.m_sQueue, xConnectionQueued.m_nPosition, xConnectionQueued.m_nCount, xConnectionQueued.m_sCallerId, xConnectionQueued.m_sCallerIdName, xConnectionQueued.m_sCalled, xConnectionQueued.m_sUserData, xConnectionQueued.m_eCause);
	}
}

/**
 * Call has queued.
 */
bool	ASTCstaProvider::handleConnectionQueued( AsteriskCall *pCall, std::string queue, int position, int count, call_cause cause, ConnectionQueued & xConnectionQueued) 
{
	const std::string	address = pCall->GetLastRedirectionDevice() ;
	m_log.LogTrace("ASTCstaProvider::handleConnectionQueued uniqueId=%s, address=%s, queue=%s, position=%d, count=%d", pCall->GetUniqueId().c_str(), address.c_str(), queue.c_str(), position, count);

	AsteriskCall	*pACSource = m_callMap.GetSourceCall(pCall);
	if (!pACSource) {
		m_log.LogWarn("  pACSource is NULL");
		return false;
	}
	DWORD dwCallId = m_callMap.GetLogicCallId(pCall);

	// RULE-9: Do not send EVENTS to local channels
	if (pCall->IsLocalChannel() && (!pACSource->IsPredictiveCall() || m_callMap.IsLocalCall(pCall->GetCallId()))) {
		m_log.LogTrace("  event minimization RULE-9: Do not send EVENTS to local channels");
		return false;
	}

	std::string	called = pCall->GetLastRedirectionDevice();

	std::string	callerid, calleridname, userData;

	m_callMap.GetCIDInfoFromSourceCall(pCall, callerid, calleridname, userData);

	// RULE-7: Do not send ALERTING/ESTABLISHED/FAILED when callerid is unknown and calling=''
	if (pCall->GetCallerId()!="" && AsteriskCall::IsUnknown(pCall->GetCallerId())) {
		if (callerid.size()==0) {
			m_log.LogTrace("  event minimization RULE-7: callerid is unknown and calling==VOID");
			return false;
		}
	}

	// RULE-3: Do not send ALERTING/ESTABLISHED/FAILED if callerid is unknown and calleddevice==address
	if (pCall->GetCallerId()!="" && AsteriskCall::IsUnknown(pCall->GetCallerId()) && called==pCall->GetAddress()) {
		m_log.LogTrace("  event minimization RULE-3: callerid is unknown and calleddevice==address==%s", pCall->GetAddress().c_str());
		return false;
	}

	//*	RULE-11: Do not send EVENTS to AsyncGoto channels */
	if ((pCall->GetChannel()).find("AsyncGoto") != -1){
		m_log.LogTrace("  event minimization RULE-11: Do not send EVENTS to AsyncGoto channels");
		return false;
	}

	// TBD for DEBUG purposes!!!
	m_callMap.DumpMaps();
	m_callMap.AddConnectionToCallModel(dwCallId, queue, QUEUED /*"QUEUED"*/);	// add callid/address to m_mapCallId2DeviceId
	
	// RULE-4: Do not send ALERTING/ESTABLISHED/FAILED/QUEUED/DEQUEUED if caller==called==(address||numDevice)
	if (callerid==called && (called==pCall->GetAddress() ||called==pCall->GetNumDevice())) {
		m_log.LogTrace("  event minimization RULE-4: (callerid = %s) == (called = %s) == (address = %s)",callerid.c_str(), called.c_str(), pCall->GetAddress().c_str());
		return false;
	}

	// fill ConnectionQueued
	xConnectionQueued.m_dwCallId = dwCallId;
	xConnectionQueued.m_sAddress = address;
	xConnectionQueued.m_sQueue = queue;
	xConnectionQueued.m_nPosition = position;
	xConnectionQueued.m_nCount = count;
	xConnectionQueued.m_sCallerId = callerid;
	xConnectionQueued.m_sCallerIdName = calleridname;
	xConnectionQueued.m_sCalled = called;
	xConnectionQueued.m_sUserData = userData;
	xConnectionQueued.m_eCause = cause;

	return true;
};

void	ASTCstaProvider::fireOnDivertedCallback(ConnectionDiverted &xConnectionDiverted)
{
	Set_xref set_Monitors;
	GetMonitorSet(xConnectionDiverted.m_dwCallId, set_Monitors);
	if (set_Monitors.size() == 0){
		m_log.LogWarn("  do not send EVENTS because not exist any monitors in call %d", xConnectionDiverted.m_dwCallId);
	}

	for (Set_xref::iterator i = set_Monitors.begin(); i!=set_Monitors.end(); i++) {
		m_log.LogTrace("    sending event:CFTE:DIVERTED to xref=%d, address=%s", *i, xConnectionDiverted.m_sAddress.c_str());
		m_cstaListener->OnDiverted(*i, xConnectionDiverted.m_dwCallId, xConnectionDiverted.m_sAddress, xConnectionDiverted.m_sQueue, xConnectionDiverted.m_sCallerId, xConnectionDiverted.m_sCallerIdName, xConnectionDiverted.m_sCalled, "", xConnectionDiverted.m_eCause);
	}
}



/**
 * Call left a queue.
 */
bool	ASTCstaProvider::handleConnectionDiverted(AsteriskCall *pCall, std::string queue, call_cause cause, ConnectionDiverted &xConnectionDiverted) 
{
	const std::string	address = pCall->GetLastRedirectionDevice() ;
	m_log.LogTrace("ASTCstaProvider::handleConnectionDiverted uniqueId=%s, address=%s, queue=%s", pCall->GetUniqueId().c_str(), address.c_str(), queue.c_str());

	AsteriskCall	*pACSource = m_callMap.GetSourceCall(pCall);
	DWORD dwCallId = m_callMap.GetLogicCallId(pCall);

	// RULE-9: Do not send EVENTS to local channels except for localCalls with monitors
	if (pCall->IsLocalChannel() && (!pACSource->IsPredictiveCall() || m_callMap.IsLocalCall(pCall->GetCallId()))) {
		m_log.LogTrace("  event minimization RULE-9: Do not send EVENTS to local channels");
		return false;
	}

	std::string	callerid, calleridname, userData;
	std::string	called = pCall->GetLastRedirectionDevice();

	m_callMap.GetCIDInfoFromSourceCall(pCall, callerid, calleridname, userData);

	// RULE-4: Do not send ALERTING/ESTABLISHED/FAILED/QUEUED/DEQUEUED if caller==called==(address||numDevice)
	if (callerid==called && (called==pCall->GetAddress() ||called==pCall->GetNumDevice())) {
		m_log.LogTrace("  event minimization RULE-4: (callerid = %s) == (called = %s) == (address = %s)",callerid.c_str(), called.c_str(), pCall->GetAddress().c_str());
		m_callMap.RemoveConnectionFromCallModel(dwCallId, queue);	
		return false;
	}

	// RULE-7: Do not send ALERTING/ESTABLISHED/FAILED when callerid is unknown and calling=''
	if (pCall->GetCallerId()!="" && AsteriskCall::IsUnknown(pCall->GetCallerId())) {
		if (callerid.size()==0) {
			m_log.LogTrace("  event minimization RULE-7: callerid is unknown and calling==VOID");
			m_callMap.RemoveConnectionFromCallModel(dwCallId, queue);	
			return false;
		}
	}

	// RULE-3: Do not send ALERTING/ESTABLISHED/FAILED if callerid="" and calleddevice==address
	if (pCall->GetCallerId()!="" && AsteriskCall::IsUnknown(pCall->GetCallerId()) && called==pCall->GetAddress()) {
		m_log.LogTrace("  event minimization RULE-3: callerid is unknown and calleddevice==address==%s", pCall->GetAddress().c_str());
		m_callMap.RemoveConnectionFromCallModel(dwCallId, queue);
		return false;
	}

	// TBD for DEBUG purposes!!!
	m_callMap.DumpMaps();

	//*	RULE-11: Do not send EVENTS to AsyncGoto channels */
	if ((pCall->GetChannel()).find("AsyncGoto") != -1){
		m_log.LogTrace("  event minimization RULE-11: Do not send EVENTS to AsyncGoto channels");
		return false;
	}

	// fill ConnectionDiverted
	xConnectionDiverted.m_dwCallId = dwCallId;
	xConnectionDiverted.m_sAddress = address;
	xConnectionDiverted.m_sQueue = queue;
	xConnectionDiverted.m_sCallerId = callerid;
	xConnectionDiverted.m_sCallerIdName = calleridname ;
	xConnectionDiverted.m_sCalled = called;
	xConnectionDiverted.m_eCause = cause ;

	return true;
};


void	ASTCstaProvider::fireOnHeldCallback(ConnectionHeld &xConnectionHeld)
{
	Set_xref set_Monitors ;
	GetMonitorSet(xConnectionHeld.m_dwCallId, set_Monitors);	// get alerting device near-end set
	if (set_Monitors.size() == 0){
		m_log.LogWarn("  do not send EVENTS because not exist any monitors in call %d", xConnectionHeld.m_dwCallId);
	}
	
	for (Set_xref::iterator i = set_Monitors.begin(); i!=set_Monitors.end(); i++) {
		m_log.LogTrace("    sending event:CFTE:HOLDING to xref=%d, address=%s callid=%d", *i, xConnectionHeld.m_sAddress.c_str(), xConnectionHeld.m_dwCallId);
		m_cstaListener->OnHold(*i,  xConnectionHeld.m_dwCallId, xConnectionHeld.m_sAddress, xConnectionHeld.m_eCause);
	}
}

/**
*	Call has held
*	\param pCall
*	\param cause
*/
bool	ASTCstaProvider::handleConnectionHeld(AsteriskCall* pCall, call_cause cause, ConnectionHeld & xConnectionHeld)
{
	const std::string	address = pCall->GetAddress() ;
	m_log.LogTrace("ASTCstaProvider::handleConnectionHeld uniqueId=%s, address=%s", pCall->GetUniqueId().c_str(), address.c_str());

	DWORD	dwCallId = m_callMap.GetLogicCallId(pCall);

	// TBD for DEBUG purposes!!!
	m_callMap.DumpMaps();
	m_callMap.AddConnectionToCallModel(dwCallId, address, HELD /*"HELD"*/);

	// fill xConnectionHeld
	xConnectionHeld.m_dwCallId = dwCallId;
	xConnectionHeld.m_sAddress = address;
	xConnectionHeld.m_eCause = cause;

	return true;
}


void	ASTCstaProvider::fireOnRetrievedCallback(ConnectionHeld &xConnectionHeld)
{
	Set_xref set_Monitors ;
	GetMonitorSet(xConnectionHeld.m_dwCallId, set_Monitors);	// get alerting device near-end set
	if (set_Monitors.size() == 0){
		m_log.LogWarn("  do not send EVENTS because not exist any monitors in call %d", xConnectionHeld.m_dwCallId);
	}
	
	for (Set_xref::iterator i = set_Monitors.begin(); i!=set_Monitors.end(); i++) {
		m_log.LogTrace("    sending event:CFTE:UNHOLDING to xref=%d, address=%s callid=%d", *i, xConnectionHeld.m_sAddress.c_str(), xConnectionHeld.m_dwCallId);
		m_cstaListener->OnUnhold(*i, xConnectionHeld.m_dwCallId, xConnectionHeld.m_sAddress, xConnectionHeld.m_eCause);
	}
}

/**
*	Call has unheld
*	\param pCall
*	\param cause
*/
bool	ASTCstaProvider::handleConnectionUnhold(AsteriskCall *pCall, call_cause cause, ConnectionHeld & xConnectionHeld)
{
	const std::string	address = pCall->GetAddress() ;
	m_log.LogTrace("ASTCstaProvider::handleConnectionUnheld uniqueId=%s, address=%s", pCall->GetUniqueId().c_str(), address.c_str());

	DWORD	dwCallId = m_callMap.GetLogicCallId(pCall);

	// TBD for DEBUG purposes!!!
	m_callMap.DumpMaps();
	m_callMap.AddConnectionToCallModel(dwCallId, address, CONNECTED /*"CONNECTED"*/);

	// fill xConnectionHeld
	xConnectionHeld.m_dwCallId = dwCallId;
	xConnectionHeld.m_sAddress = address;
	xConnectionHeld.m_eCause = cause;

	return true;
}

