/*
	Name: invokeidmanager.cpp
	Copyright: Under the Mozilla Public License Version 1.1 or later
	Author: ACTIVA
	Date: MAY/06
*/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Activa ASTProvider.
 *
 * The Initial Developer of the Original Code is ACTIVA.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): none
 *
 * ***** END LICENSE BLOCK ***** */

#include <Windows.h>		
#include "invokeidmanager.h"
#include "astinterface.h"

#define INVOKEID_MAX	1000000

/*
** InvokeID
*/
InvokeID::InvokeID(std::string invoke, bool bLockReaderAfterResponse) 
{
	m_invoke = invoke;
	m_bReceived = false;
	m_hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);	// create non-signaled event
	m_bLockReaderAfterResponse = bLockReaderAfterResponse;
}

InvokeID::~InvokeID()
{
	CloseHandle(m_hEvent);
}

bool InvokeID::WaitReponse( int seconds ) 
{
	DWORD rc = WaitForSingleObject (m_hEvent, 1000*seconds);
	switch(rc) {
		case WAIT_TIMEOUT:
		case WAIT_FAILED:
			return false;
		default:
			break;
	}
	
	return m_bReceived ;
}
	
void InvokeID::SignalResponse(ManagerResponse result) {
	m_bReceived=true;
	m_result = result;
	SetEvent(m_hEvent);
}

ManagerResponse InvokeID::GetResult() {
	return m_result;
}


/**
 * InvokeIdManager
 */
InvokeIdManager::InvokeIdManager() 
{
    m_currentInvokeId = 0   ;
	m_hMutex = CreateMutex(NULL, FALSE, NULL); 
	m_hManagerEvent = CreateEvent(NULL, FALSE, FALSE, NULL);	// create non-signaled event, global to synchoronize readerThread and CompProc
}
	
/**
*	Wait for response of command
*/
bool InvokeIdManager::WaitResponse(std::string iInvoke, ManagerResponse &result, int responseTimeoutSeconds) 
{
	bool bFound = false;

	// protect map access with mutex
	WaitForSingleObject(m_hMutex, INFINITE);

	if (m_pendingInvokeID.find(iInvoke)!=m_pendingInvokeID.end()) { // found invokeid object
		InvokeID	* pInvokeID = m_pendingInvokeID[iInvoke];
		ReleaseMutex(m_hMutex);
		if (pInvokeID) { // just double check it is a real invokeId (not a callid)
			if ( pInvokeID->WaitReponse(responseTimeoutSeconds)) { // got response!
				result = pInvokeID->GetResult();
				bFound=true;
			}
			delete pInvokeID;		
		}

		WaitForSingleObject(m_hMutex, INFINITE);
		m_pendingInvokeID.erase(iInvoke);
		ReleaseMutex(m_hMutex);
	}

	return bFound;
}


ManagerResponse InvokeIdManager::GetResult(std::string iInvoke) 
{
	ManagerResponse response ;

	WaitForSingleObject(m_hMutex, INFINITE);
	if (m_pendingInvokeID.find(iInvoke)!=m_pendingInvokeID.end()) { // found invokeid object
		response = m_pendingInvokeID[iInvoke]->GetResult();
	}
	ReleaseMutex(m_hMutex);

	return response;
}

/**
*	Generate a new InvokeId
*	\return string with new InvokeId
*/
std::string InvokeIdManager::NewInvokeId( bool bLockReaderAfterResponse)
{
	WaitForSingleObject(m_hMutex, INFINITE);
	if (m_pendingInvokeID.size()==INVOKEID_MAX) { // map is full :-(
		ReleaseMutex(m_hMutex);
		return "";
	}

	std::string	sInvokeId ;
	do {
		m_currentInvokeId++ ;
		if (m_currentInvokeId > INVOKEID_MAX) {
			m_currentInvokeId=1;
		}
		sInvokeId = ToString(m_currentInvokeId);
	} while (m_pendingInvokeID.find(sInvokeId)!=m_pendingInvokeID.end());
	
	m_pendingInvokeID[sInvokeId] = new InvokeID(sInvokeId, bLockReaderAfterResponse); 
	ReleaseMutex(m_hMutex);
    
    return sInvokeId;
}

    
/**
*	Wait For response
*	\param result
*	\param invokeID
*/
bool InvokeIdManager::SignalResponse(ManagerResponse result, std::string invokeID) {
	bool bLockReaderAfterResponse=false;
	WaitForSingleObject(m_hMutex, INFINITE);
	if (m_pendingInvokeID.find(invokeID)==m_pendingInvokeID.end()) { // invoke not found
		ReleaseMutex(m_hMutex);
		return false;
	}

	bLockReaderAfterResponse = m_pendingInvokeID[invokeID]->GetFlagLockReaderAfterResponse();
	m_pendingInvokeID[invokeID]->SignalResponse(result);
	
	if(bLockReaderAfterResponse){
		ReleaseMutex(m_hMutex);
		WaitForSingleObject(m_hManagerEvent, INFINITE); //lock reader 
	}
	else{
		ReleaseMutex(m_hMutex);
	}
	
	return true;
}


/**
*	Signal all waiting threads with pending invokeid
*	\param result
*/
int InvokeIdManager::SignalAll(ManagerResponse result) {
	Map_invokeId::iterator it;
	int count=0;

	WaitForSingleObject(m_hMutex, INFINITE);

	for ( it=m_pendingInvokeID.begin() ; it != m_pendingInvokeID.end(); it++ )
	{
		(*it).second->SignalResponse(result);
		count++;
	}

	ReleaseMutex(m_hMutex);
	return (count);
}

/**
*	Unlock reader
*/
void InvokeIdManager::UnLockReader() 
{

	SetEvent(m_hManagerEvent);
}
