/*
	Name: originare.cpp
	Copyright: Under the Mozilla Public License Version 1.1 or later
	Author: ACTIVA
	Date: MAY/06
*/

/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is astprovider.
 *
 * The Initial Developer of the Original Code is ACTIVA.
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): none
 *
 * ***** END LICENSE BLOCK ***** */

#include "astprovider.h"

#include <list>
#include <map>
#include <algorithm>
#include <cctype>

/**
*	Action: originate
*	http://www.voip-info.org/wiki/view/Asterisk+Manager+API+Action+Originate
*/
/**
*	Originate a call by Dial
*	\param callingDevice identifier of who is calling
*	\param calledDevice identifier of who is called
*	\param callid identifier of call
*	\param outGoingChannel
*	\param timeout in seconds
*	\param invokeId identifier of action in manager API
*	\return AC_RESULT
*/
AC_RESULT	ASTProvider::OriginateByDial(std::string callingDevice, std::string calledDevice, DWORD callid, std::string outGoingChannel, DWORD timeout, std::string invokeId)
{
    std::string command;

	if (callid < 0){
		m_log.LogError(" callid has an invalid number");
		return invalid_callid;
	}

	command = "Action: Originate\r\n";
	command += "Channel: ";
	command += "local:" + callingDevice;
	command += "\r\n";
	command += "Application: Dial\r\n";

	//Ensure that we are sending nothing but numbers to Asterisk
	std::string::iterator it;
	std::string tempStr("");

	for ( it = calledDevice.begin(); it != calledDevice.end(); it++ ) {
		if ( *it >= '0' && *it <= '9' ) {
			tempStr += *it;
		}
	}

	if (timeout > 0){
		// Timeout must be in ms
		command += "Timeout: " + ToString(timeout*1000);
		command += "\r\n";
	}
	command += "Data: ";
	command += outGoingChannel;
	command += tempStr;
	command += "\r\n\r\n";
	SendCommand(command);

	return ok ;
}

/**
*	Originate a call
*	\param channel
*	\param callerId identifier of who is calling
*	\param callerIdName 
*	\param calledExtension identifier of who is called
*	\param callid identifier of call
*	\param context
*	\param timeout in seconds
*	\param invokeId identifier of action in manager API
*	\return AC_RESULT
*/
AC_RESULT	ASTProvider::OriginateByContext(std::string channel, std::string callerId, std::string callerIdName, std::string calledExtension, DWORD callid, std::string context, DWORD timeout, std::string invokeId, std::string variable)
{
    std::string command;
	std::string	extendedCallerId;
	std::string callIdStr;

	if (callid < 0){
		m_log.LogError(" callid has an invalid number");
		return invalid_callid;
	}

	if (callid > 0){
		callIdStr = "#Id="+ToString(callid);
	}

	if (callerIdName.size()>0) {
		extendedCallerId = callerIdName + callIdStr + " <" + callerId + ">";
	}else{
		extendedCallerId = callIdStr + " <" + callerId + ">";
	}

	/// channel ending with /n is only a feature for Local channels.
	/// http://www.voip-info.org/wiki/view/Asterisk+readme.channel
	/// Says: Adding "/n" at the end of the string will make the Local channel not
	/// do a native transfer (the "n" stands for "n"o release) upon the remote end
	///	answering the line. This is an esoteric, but important feature if you
	/// expect the Local channel to handle calls _exactly_ like a normal channel. 

	size_t npos = -1;
	if ((channel.find("Local") != npos) || (channel.find("local") != npos)){
		channel = channel + "/n";
	}

	if (variable!= "") variable = ","+variable;

	command = "Action: Originate\r\n";
	command += "Channel: " + channel;
	command += "\r\n";
	command += "Exten: " + calledExtension;
	command += "\r\n";
	command += "Priority: 1";
	command += "\r\n";
	command += "Callerid: " + extendedCallerId;
	command += "\r\n";
	command += "Context: " + context;
	command += "\r\n";
	command += "Variable: originating="+callerId+variable;
	command += "\r\n";
	command += "ActionID: " + ToString(invokeId);
	command += "\r\n";
	if (timeout > 0){
		// Timeout must be in ms
		command += "Timeout: " + ToString(timeout*1000);
		command += "\r\n";
	}
	command += "Async: 1";
	command += "\r\n\r\n";
	SendCommand(command);

	return ok;
}

